package net.hivecell.hive.agent;

import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

import net.hivecell.hive.agent.AgentInitializationException;
import net.hivecell.hive.cell.RemoteCell;
import net.hivecell.hive.event.ExternalEvent;
import net.hivecell.hive.event.HiveEvent;
import net.hivecell.hive.event.IncomingExternalEvent;
import net.hivecell.hive.event.OutgoingExternalEvent;
import net.hivecell.hive.shadow.EventPublisherShadow;
import net.hivecell.hive.shadow.ExternalInterfaceShadow;
import net.hivecell.hive.shadow.ExternalInterfaceShadowImpl;
import net.hivecell.hive.shadow.ShadowNotFoundException;
import net.hivecell.hive.shadow.Signalable;
import net.hivecell.hive.support.CellConnectException;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.support.MalformedHiveURLException;
import net.hivecell.hive.support.CellAddress;

import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Hashtable;


/**
 * this agent sits on top of the ExternalInterfaceShadow, and spits
 * the ExternalEvent to the specified agent who will presumably know
 * how to handle that event, or can "translate" that event into
 * something of more use to other agents 
 */
public class ExternalInterfaceAgentImpl
extends EventTransceivingAgentImpl
implements Signalable, ExternalInterfaceAgent {

    /**
     * our hook to our shadow
     */
    protected transient ExternalInterfaceShadow externalInterfaceShadow = null;


    /**
     * construct a new one of these agents
     *
     * @exception RemoteException if initalisation goes wrong
     */
    public ExternalInterfaceAgentImpl()
    throws RemoteException {
	super();
	
    }

    /**
     * set ourselves up to listen to events from the
     * ExternalInterfaceShadow, and then that will set our variable to
     * point to that shadow so we can use it
     *
     * @exception AgentInitializationException if something goes wrong 
     */
    public void doLocalSetup()
    throws AgentInitializationException {
	super.doLocalSetup();
	
	// get our hook to our external interface shadow -- there has to be abetter way to do this
	while( true ) {
	    try {
		Debug.notice( "getting a handle to the ExternalInterfaceShadow" );
		externalInterfaceShadow = (ExternalInterfaceShadow)myCell.getShadow( ( ExternalInterfaceShadowImpl.class ).getName() );
		externalInterfaceShadow.addListener( this );
		break;
		
	    } catch( ShadowNotFoundException error ) {
		Debug.error( "cannot find the ExternalInterfaceShadow on " + myCell.getAddress() );
		try {
		    Debug.notice( "waiting 5 seconds to query again" );
		    Thread.sleep( 5000 );

		} catch( InterruptedException internalError ) {

		}
		// throw new AgentInitializationException( error );
		
	    }

	}

	setIsReady( true ); 

    }

    /**
     * just sit here and wait
     */
    public void doBehavior() {
	waitUntilDeath();

    }

    /**
     * listen to these ExternalEvents which are coming up to me, and
     * then we need to figure out exactly who to deliver this event to
     *
     * @param source the source of this event
     */
    public void signal( EventPublisherShadow source ) {
	this.signal( source, null );

    }

    /**
     * listen to these ExternalEvents which are coming up to me, and
     * then we need to figure out exactly who to deliver this event to
     *
     * @param source the source of this event
     * @param information the event that we want to deal with 
     */
    public void signal( EventPublisherShadow source, Object information ) {
	// dump the event if it does not meet this following criteria
	if( ( source != externalInterfaceShadow ) ||
	    ( information == null ) ||
	    ( !( information instanceof OutgoingExternalEvent ) ) )
	    return;

	// parse out the XML document from the evnet
	ExternalEvent event = (ExternalEvent)information;
	Document document = event.getXMLDocument();

	// ok - at this point we have an external event that was given
	// to us by the external interface shadow.  let's figure out
	// where we should be delivering this event to and then send
	// the event off
	Node outgoingEventNode = document.getFirstChild();
	NodeList nodeList = outgoingEventNode.getChildNodes();
	Node destinationNode = null;
	for( int count=0;count<nodeList.getLength();count++ ) {
	    Node searchingNode = nodeList.item( count );
	    if( searchingNode.getNodeName().equals( "destination" ) ) {
		destinationNode = searchingNode;
		break;

	    }

	}
	if( destinationNode == null )
	    return;

	// ok - we have the destination node -- now let's pull the
	// information we are interested in out of this tag so we can
	// contact the appropriate agent
	Debug.notice( "pulling the destination information out of the external event" );
	String hostname = null;
	String className = null;
	Hashtable semantics = new Hashtable();
	NodeList destinationNodeList = destinationNode.getChildNodes();
	for( int count=0;count<destinationNodeList.getLength();count++ ) {
	    Node searchingNode = destinationNodeList.item( count );
	    String searchingNodeName = searchingNode.getNodeName();
	    if( searchingNodeName.equals( "hostname" ) )
		hostname = searchingNode.getFirstChild().getNodeValue();

	    else if( searchingNodeName.equals( "syntactic" ) )
		className = searchingNode.getFirstChild().getNodeValue();
	    
	    else if( searchingNodeName.equals( "semantic" ) ) {
		NodeList semanticNodeList = searchingNode.getChildNodes();
		for( int semanticNodeCount=0;semanticNodeCount<semanticNodeList.getLength();semanticNodeCount++ ) {
		    Node individualSemanticNode = semanticNodeList.item( semanticNodeCount );
		    if( individualSemanticNode.getNodeType() == Node.ELEMENT_NODE )
			semantics.put( individualSemanticNode.getNodeName(), individualSemanticNode.getFirstChild().getNodeValue() );
		    
		}

	    }

	}
	
	// and at this point, we are actually going to do the query to
	// get the agent we want to talk to
	RemoteCell remoteCell = null;
	if( hostname == null ) {
	    Debug.notice( "querying on the local cell" );
	    remoteCell = myCell;
	
	} else
	    try {
		Debug.notice( "attempting to query on " + hostname );
		remoteCell = ( new CellAddress( hostname ) ).getRemoteCell();
		Debug.notice( "have the handle to the remote cell for this event" );

	    } catch( MalformedHiveURLException error ) {
		Debug.error( "the XML specified a destination cell of " + hostname + " which is not valid" );
		return;
		
	    } catch( CellConnectException error ) {
		Debug.error( "could not connect to the cell at " + hostname, error );
		return;

	    }

	// construct the syntactic and the semantic arrays for the quries
	String[] syntactic = null;
	if( className != null ) {
	    syntactic = new String[1];
	    syntactic[0] = className;

	}

	String[] semantic = null;
	if( semantics.size() != 0 ) {
	    semantic = new String[semantics.size()];
	    int arrayPosition = 0;
	    Enumeration semanticTags = semantics.keys();
	    while( semanticTags.hasMoreElements() )
		semantic[arrayPosition++] = (String)semanticTags.nextElement();

	}
	
	// now let's do the query -- we should have all the information!
	try {
	    Debug.notice( "doing the query agents to find the agent handle" );
	    EventReceivingAgent eventReceivingAgent = (EventReceivingAgent)remoteCell.queryAgents( this, syntactic, semantic ).pickOne();
	    if( eventReceivingAgent != null ) {
		Debug.notice( "sending event out to " + eventReceivingAgent );
		eventSendingHelper.broadcastEventTo( event, eventReceivingAgent );		

	    } else 
		Debug.notice( "no matching agents to send event to" );

	} catch( ClassNotFoundException error ) {
	    Debug.error( "remote cell does not know anything about " + syntactic[0] );
	    return;

	} catch( RemoteException error ) {
	    Debug.error( "problems finding the handle to the agent", error );
	    return;

	} catch( ClassCastException error ) {
	    // this means the agent that we found is not an event
	    // receiving agent, so we couldn't send it an event anyway
	    Debug.error( "the agent specified is not an event receiving agent" );
	    return;

	}

    }

    /**
     * this is the notify event so we can get the events that we want
     * to pass onto the device.  this will only take external events
     * as they will be coded with the final device that the event is
     * going to
     *
     * @param theEvent the hive event that comes into 
     */
    public void notify( HiveEvent theEvent ) {
	// punt the evnet if it is not an external event
	if( !( theEvent instanceof IncomingExternalEvent ) )
	    return;
	
	// ok -- lets pass the event down to the shadow
	externalInterfaceShadow.sendEvent( (IncomingExternalEvent)theEvent );

    }


}
