// $Id: Spoke.java,v 1.1 2000/05/18 22:04:48 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.


package net.hivecell.hive.agent.desktop;

/** A spoke is a point in polar coordinates that is moving at a constant
 ** radial velocity. This class is intended for use with SpokeCanvas.
 **
 ** @author Nelson Minar <nelson@media.mit.edu>
 **/
public class Spoke extends Object implements java.io.Serializable {
    static final float minR = 0.2f, maxR = 0.8f;

    /** The length of the spoke, should be between 0 and 1. **/
    float r;
    /** The angle of the spoke, should be between 0 and 2*pi. **/
    float theta;
    /** The rate at which the spoke length is changing. **/
    float dr;
    /** The rate at which the spoke angle is changing. **/
    float dtheta;

    protected static final float TwoPI = (float)(2 * Math.PI);
    protected static final float maxDTheta = (float)(2 *Math.PI / 6);
    
    public Spoke(float r, float theta, float dr, float dtheta) {
        this.r = r;   this.theta = theta;
        this.dr = dr; this.dtheta = dtheta;
    }
             
    /** Update the position of the spoke by adding velocity.
     ** Spoke position is kept in range; if it gets too long or short,
     ** it "bounces".
     **/
    public synchronized void update() {
        r += dr;
        // if r is out of range, bring it back and change sign of dr.
        if (r > maxR) {                       
            r = maxR - (r - maxR);
            dr = -dr;
        } else if (r < minR) {
            r = minR + (minR - r);
            dr = -dr;
        }

        // if theta is out of range, bring it back in range.
        theta += dtheta;
        if (theta > TwoPI)
            theta -= TwoPI;
        else if (theta < 0)
            theta += TwoPI;
    }

    /** Make the spoke spin faster.
     ** @param magnitude The magnitude of the change. This value
     **   is multiplied by the current dtheta. As a special case,
     **   if dtheta is 0 and the magnitude is greater than 1, then
     **   dtheta is set to 0.05f.
     **/
    public synchronized void accelerateTheta(float magnitude) {
        // If we're too small and scaling up, set it up to a constant.
        if (dtheta == 0 && magnitude > 1)
            dtheta = 0.05f;
        // Scale dtheta
        dtheta *= Math.abs(magnitude);
        // If we're too small after scaling down, then stop.
        if (dtheta < 0.01f && magnitude < 1)
           dtheta = 0;
        // If we're too fast after scaling up, then max out.
        if (dtheta > (maxDTheta))
            dtheta = maxDTheta;
    }

    public synchronized String toString() {
        return "Spoke r=" + r + " theta= " + theta +
            " dr=" + dr + " dtheta=" + dtheta;
    }
}
