// $Id: SpokeCanvas.java,v 1.1 2000/05/18 22:04:48 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.


package net.hivecell.hive.agent.desktop;

import net.hivecell.hive.Global;
import net.hivecell.hive.support.BaseCanvas;

import java.awt.*;
import java.awt.event.*;
import java.awt.image.MemoryImageSource;
import java.lang.Math;
import java.util.Vector;
import java.util.Enumeration;

/** An animated canvas that shows whirling spokes.
 ** This was based on a design by Marius Watz, see
 ** http://www.notam.uio.no/~mariusw/ and
 ** http://www.notam.uio.no/~mariusw/form/java/rotline.html
 **
 ** @author Nelson Minar <nelson@media.mit.edu>
 **/
public class SpokeCanvas extends BaseCanvas {
    /** a list of all the spokes we are drawing. Contains SpokeData objs. **/
    protected Vector spokes = new Vector();

    /** An image used to represent a ball. Loaded from ballImageData array. **/
    protected Image ballImage = null;
        
    public SpokeCanvas() {
        setBackground(Color.white);
        setForeground(Color.black);
        setSize(200, 200);
        setDoubleBuffering(true);
        setVisible(true);
    }

    /** Create a new spoke with reasonable random values, including motion **/
    public Spoke createMovingSpoke() {
        float r = (Global.random.nextFloat() * (Spoke.maxR - Spoke.minR) + Spoke.minR);
        float dr = (Global.random.nextFloat() * 0.06f) - 0.03f;
        float theta = (float)((Global.random.nextFloat()) * Math.PI * 2);
        float dtheta = (Global.random.nextFloat() * 0.6f) - 0.3f;
        return createSpoke(r, theta, dr, dtheta);
    }

    /** Create a spoke with random position but no motion. **/
    public Spoke createStoppedSpoke() {
        float r = (Global.random.nextFloat() * (Spoke.maxR - Spoke.minR) + Spoke.minR);
        float theta = (float)((Global.random.nextFloat()) * Math.PI * 2);
        return createSpoke(r, theta, 0, 0);
    }

    /** Add a spoke with the specific values. **/
    public Spoke createSpoke(float r, float theta, float dr, float dtheta) {
        Spoke s = new Spoke(r, theta, dr, dtheta);
        synchronized(spokes) {
            spokes.addElement(s);
        }
        return s;
    }

    /** Remove the given spoke from the display. **/
    public void removeSpoke(Spoke s) {
        synchronized(spokes) {
            spokes.removeElement(s);
        }
    }
    
    /** Update all the spokes' positions and render. **/
    public void updateDisplay() {
        Enumeration e = spokes.elements();
        while (e.hasMoreElements())
            ((Spoke)e.nextElement()).update();
        render();
    }

    /** Draw all of the spokes. **/
    public void render(Graphics g) {
        final Dimension windowSize = getSize();

        // Make sure that we have an image object constructed for the ball.
        if (ballImage == null)
            createBallImage();
        
        // clear the window
        g.setColor(getBackground());
        g.fillRect(0, 0, windowSize.width, windowSize.height);

        // figure out where to draw things
        final int centerX = windowSize.width / 2;
        final int centerY = windowSize.height / 2;
        final int xScale = windowSize.width / 2;
        final int yScale = windowSize.height / 2;

        // Calculate the positions of all the spoke endpoints, store them
        // in xpts and ypts arrays. We also want to ensure that a spoke
        // isn't added or removed in the middle of this..
        final int numSpokes, xpts[], ypts[];
        synchronized(spokes) {
            numSpokes = spokes.size();
            xpts = new int[numSpokes];
            ypts = new int[numSpokes];
            int i = 0;
            
            Enumeration e = spokes.elements();
            while (e.hasMoreElements()) {
                final Spoke s = (Spoke)e.nextElement();
                xpts[i] = ((int)Math.round(s.r * Math.cos(s.theta) * xScale)) + centerX;
                ypts[i] = ((int)Math.round(s.r * Math.sin(s.theta) * yScale)) + centerY;
                i++;
            }
        }

        g.setColor(getForeground());
        // Now draw the rays
        for (int i = 0; i < numSpokes; i++)
            g.drawLine(centerX, centerY, xpts[i], ypts[i]);
        // And the circles
        for (int i = 0; i < numSpokes; i++)
            g.drawImage(ballImage, xpts[i] - 4, ypts[i] - 4, this);
    }

    /** Make the AWT Image for the ball at the end of a spoke. **/
    protected void createBallImage() {
        // An 8x8 ring, antialised and transparent on corners.
        // converted from GIF by ImageToArray
        int[] imagePixels = { 
            0x00fefefe,0x00fefefe,0xff707070,0xff1c1c1c,0xff1c1c1c,0xff707070,
            0x00fefefe,0x00fefefe,0x00fefefe,0xff000000,0xff1a1a1a,0xffadadad,
            0xffadadad,0xff1a1a1a,0xff000000,0x00fefefe,0xff707070,0xff1a1a1a,
            0xfffbfbfb,0xffffffff,0xffffffff,0xfffbfbfb,0xff1a1a1a,0xff707070,
            0xff1c1c1c,0xffadadad,0xffffffff,0xffffffff,0xffffffff,0xffffffff,
            0xffadadad,0xff1c1c1c,0xff1c1c1c,0xffadadad,0xffffffff,0xffffffff,
            0xffffffff,0xffffffff,0xffadadad,0xff1c1c1c,0xff707070,0xff1a1a1a,
            0xfffbfbfb,0xffffffff,0xffffffff,0xfffbfbfb,0xff1a1a1a,0xff707070,
            0x00fefefe,0xff000000,0xff1a1a1a,0xffadadad,0xffadadad,0xff1a1a1a,
            0xff000000,0x00fefefe,0x00fefefe,0x00fefefe,0xff707070,0xff1c1c1c,
            0xff1c1c1c,0xff707070,0x00fefefe,0x00fefefe
        };
        
        // create the image from the array.
        MemoryImageSource mis = new MemoryImageSource(8, 8, imagePixels, 0, 8);
        ballImage = Toolkit.getDefaultToolkit().createImage(mis);
    }
    
    /** A simple main function to create a spoke canvas and show it off. **/
    public static void main(String[] args) {
        int numSpokes = 15;
        try {
            numSpokes = Integer.parseInt(args[0]);
        } catch (Exception ex) { }

        SpokeCanvas sc = new SpokeCanvas();
        for (int i = 0; i < numSpokes; i++)
            sc.createMovingSpoke();

        Frame f = new Frame();
        f.add(sc);
        f.pack();
        f.show();
        f.addWindowListener(new WindowAdapter() {
             public void windowClosing(WindowEvent event) {
                 System.exit(0);
             }});

        while (true) {
            sc.updateDisplay();
            try {
                Thread.sleep(100);
            } catch (InterruptedException ex) { }
        }
    }
}
