// $Id: MovingAverageAgentImpl.java,v 1.2 2000/05/19 19:55:29 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.agent.filter;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.*;
import net.hivecell.hive.agent.AgentInitializationException;
import net.hivecell.hive.event.HiveEvent;
import net.hivecell.hive.cell.Cell;
import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.agent.EventReceivingAgentImpl;
import java.util.*;

/** Chaining agent - takes in event input from an agent, produces a
 ** time average. This has two parameters - how many milliseconds to
 ** remember, and how many milliseconds to average over. If
 ** historyLength is 300000 and averageWindow is 10000, then what we
 ** get is a 5 minute moving average of the number of events per 10
 ** seconds.
 **/
public class MovingAverageAgentImpl 
extends EventReceivingAgentImpl 
implements EventFloatTranslatingAgent {

    protected int eventCounts[], eventCountScan, lastValidNumber;
    protected float movingAverage;

    // parameters for the agent
    protected int historyLength = 300000, averageWindow = 10000;

    public MovingAverageAgentImpl() 
	throws java.rmi.RemoteException { 
	super(); 
    }

    public void doLocalSetup() throws AgentInitializationException  {
	super.doLocalSetup();
	buildEventCounts();
    }

    // Change how much history we record
    public void setAverageWindow(int aw) {
	averageWindow = aw;
	buildEventCounts();
    }

    public void setHistoryLength(int hl) {
	historyLength = hl;
	buildEventCounts();
    }

    protected void buildEventCounts() {
	eventCounts = new int[historyLength / averageWindow + 1]; // actual data
	eventCountScan = 0;                           // pointer in the data
	lastValidNumber = 0;                          // how far data we've written
    }

    public void doBehavior() {
	// Now main loop
	while (!timeToStop) {
	    try {
		Thread.sleep(averageWindow);
	    } catch (InterruptedException e) {}
	    updateAverage();
	}
    }

    // Refigure the moving average window - update the scan, calc the average
    // The average calculation is wasteful: you really only need to refigure
    //   based on the most recent number. But it's no big deal.
    protected synchronized void updateAverage() {
	int i, total = 0;

	// OK, we have one more valid number - update the count of data
	if (lastValidNumber < eventCounts.length)
	    lastValidNumber++;

	// Now calculate the total and average
	// Debug.print(Debug.NOTICE, "cnts ");
	for (i = 0; i < lastValidNumber; i++) {
	    total += eventCounts[i];
	    // if (Debug.getLevel() <= Debug.NOTICE)
	    //   System.err.print(eventCounts[i] + " ");
	}
	if (lastValidNumber > 0)
	    movingAverage = (float)total / lastValidNumber;
	else
	    movingAverage = 0;                          // special case
	// if (Debug.getLevel() <= Debug.NOTICE)
	//   System.err.println("avg: " + movingAverage);

	// And update the data pointer - zero out the new bin.
	eventCountScan = (eventCountScan + 1) % eventCounts.length;
	eventCounts[eventCountScan] = 0;
    }

    public float getFloat(Agent sender) {
	myCell.noteAgentMessage( this, "getFloat", sender, this ); // the message
	myCell.noteAgentMessage( this, "getFloat reply", this, sender ); // the reply
	return movingAverage;
    }

    // everytime an event is received, increment the appropriate bin
    public synchronized void notify(HiveEvent event) {
	super.notify(event);
	eventCounts[eventCountScan]++;
    }

    public int getAverageWindow() { return averageWindow; }
    public int getHistoryLength() { return historyLength; }
}


