// $Id: GenericFilterAgentImpl.java,v 1.2 2000/05/19 19:55:31 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.


package net.hivecell.hive.agent.image;

import java.awt.*;
import java.rmi.RemoteException;


import net.hivecell.hive.agent.AgentInitializationException;
import net.hivecell.hive.*;
import net.hivecell.hive.event.*;
import net.hivecell.hive.shadow.*;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.support.ImageEvent;
import net.hivecell.hive.support.*;

/** 
 * This abstract class is the superclass for all image
 * filters. Subclassers need merely to implement the abstract
 * <i>filter</i> method to have a working filter. It is very important
 * that filter return a NEW image, and not merely a modified copy of the
 * argument, or else there will be a lot of issues with multiple
 * agents sharing copies of the same image.
 */
public abstract class GenericFilterAgentImpl 
extends ImageSourceAgentImpl {

    /** This Queue handles the images waiting to be fetched.
     * 
     */
    protected Queue fetchQ;
    /** This threaded class fetches any images waiting in the fetchQ and puts them into the filterQ
     * 
     */
    protected transient ImageFetchingHelper fetcher;
    /** This Queue handles any images waiting to be filtered.
     * 
     */
    protected Queue filterQ;
    
    /** This is where the filtering takes place.
     * Subclassers will want to override this method.
     * IMPORTANT: The filter MUST return a NEW image, and not simply
     * modify the argument. If you do you will get wierd and unexpected
     * results because multiple ImageConsumers can have references to a
     * single image
     * 
     * @param si
     */
    protected abstract SerializableImage filter(SerializableImage si);

    /** Initialize the Filter. If you override this method be sure and call super();
     * 
     * @exception RemoteException 
     */
    public GenericFilterAgentImpl() 
    throws RemoteException {
	super();
	
    }

    /** Instantiates the ImageFetchingHelper and the fetchQ and filterQ queues.
     *
     */
    public void doLocalSetup() throws AgentInitializationException  {
        super.doLocalSetup();
        fetchQ = new Queue(); // Contains QueueEntries.
	filterQ = new Queue(); // Contains images.
        fetcher = new ImageFetchingHelper(this, fetchQ, filterQ, 10);
    }
    
    /** When an event is recieved, this method ensures that it is
     * an ImageEvent and that the sender is a ImageSourceAgent.
     * If the ImageEvent is a PENDING notification, that message is
     * rebroadcast. If it is a READY message, then if the filter is
     * busy, a IGNORED message is broadcast, otherwise, the image is
     * added to the fetchQ to be retrieved. If the fetchQ exceeds BUFSIZE
     * then the filter goes busy.
     * 
     * @param event 
     */
    public synchronized void translate(HiveEvent event) {
	ImageSourceAgent source = null;
	ImageEvent ce = null;
	try {
	    source = (ImageSourceAgent) event.getSource();
	    ce = (ImageEvent) event.getEventData();
	}
	catch (ClassCastException e) {
	    Debug.println(Debug.WARNING,
			  "Recieved an Event from a non ImageSourceAgent.");
	    return;
	}
	if (ce.getStatus() == ce.READY) {
	    if (isBusy()) {
		eventSendingHelper.broadcastEvent
		    (new ImageEvent(ImageEvent.IGNORED,
				  ce.getFrameNum()));
	    }
	    else {
		fetchQ.enQueue(new QueueEntry(source, ce));
                synchronized (fetcher) {
                    fetcher.notifyAll(); // this notifies the wait() in run in ImageFetchingHelper
                }
                
		if (fetchQ.getSize() >= BUFSIZE) {
		    setBusy(true);
		}
	    }
	}
	else if (ce.getStatus() == ce.PENDING) {
	    eventSendingHelper.broadcastEvent(ce);
	}
    }

    /** Waits until notified by the filterQ, grabs elements off the filterQ, filtering them and
     *  putting them into the ringbuffer.
     * 
     */
    public synchronized void doBehavior() {
            while(!timeToStop) {
                if (!(filterQ.isEmpty())) {
                    SerializableImage si = (SerializableImage) filterQ.deQueue();
                    SerializableImage modsi = filter(si);
                    if (modsi != null) {
                        rb.addElement(modsi);
                        eventSendingHelper.broadcastEvent(new ImageEvent
                            (ImageEvent.READY,
                             rb.getLatestIndex()));
                    }
                }
                else {
                    try {
                        wait(); // this waits for the notification from run in ImageFetchingHelper
                    }
                    catch(InterruptedException e) {
                        ;;
                    }
                }
            }
            fetcher.timeToStop();           
    }
}


class ImageFetchingHelper implements Runnable {

    /** The agent that owns this Helper class.
     * 
     */
    protected ImageSourceAgentImpl owner;
    /** This Queue holds images waiting to be fetched.
     * 
     */
    protected Queue fetchQ;
    
    /** This Queue holds images waiting to be filtered.
     * 
     */
    protected Queue filterQ;
    /** The local thread.
     * 
     */
    protected Thread thread;
    /** This is set to true when the filter Agent wants to die.
     * 
     */
    protected boolean timeToStop = false;

    /** If the fetchQueue goes is less than MAXQUEUE then owner is set to !BUSY.
     * 
     */
    protected final int MAXQUEUE;

    /**      * 
     * @param fetchQ The Queue from which to fetch images
     * @param owner The owner of this helper class.
     * @param maxQueue The maximum length to allow the fetchQ to grow 
     *     to.
     * @param filterQ The Queue onto which to put fetched images.
     */
    public ImageFetchingHelper(ImageSourceAgentImpl owner, 
			     Queue fetchQ, Queue filterQ,
			     int maxQueue) {
	this.owner = owner;
	this.fetchQ = fetchQ;
	this.filterQ = filterQ;
	MAXQUEUE = maxQueue;
	thread = new Thread(owner.getThreadGroup(), this, "ImageFetchingHelperThread");
	thread.start();
    }

    /** Called when the filter agent wants to die.
     * 
     */
    public void timeToStop() {
	timeToStop = true;
    }

    /** Loop continuously, grabbing images off the fetchQ, fetching them, and then putting them on the filterQ.
     * 
     */
    public synchronized void run() { 
        
        while(!timeToStop) {
            if (!(fetchQ.isEmpty())) {
                QueueEntry qe = (QueueEntry) fetchQ.deQueue();		
                if (fetchQ.getSize() < MAXQUEUE) {
                    owner.setBusy(false);
                }
                try {
                    SerializableImage si = qe.getSource().getImage(owner,
                                                                   qe.getEvent().getFrameNum());
                    filterQ.enQueue(si);
                    synchronized (owner) {
                        owner.notifyAll(); // this is the notification for the doBehavior in GenericFilterAgentImpl
                    }
                }
                catch(RemoteException e) {
                    Debug.println(Debug.ERROR, 
                                  "Failed to get Image from remote.", e);   
                }
            }
            else {
                try {
                    wait(); // this is waiting for translate in GenericFilterAgentImpl to notify it 
                    }
                catch(InterruptedException e) {
                        ;;
                }
            }            
        }    
    }
}

