// $Id: ImageDisplayCanvas.java,v 1.1 2000/05/18 22:04:58 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

// Canvas for ImageDisplay

package net.hivecell.hive.agent.image;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;

import net.hivecell.hive.support.Debug;
import net.hivecell.hive.support.BaseCanvas;

/** This is the class that does the work of displaying images for the ImageDisplayAgent.
 * 
 */
public class ImageDisplayCanvas extends BaseCanvas {
    
    /** The default size for the canvas.
     * 
     */
    protected static final Dimension defaultDim = new Dimension(320, 240);
    /** Whether or not to hatch the pending display. Setting this to false will provide a performance boost for JDKs with a slow AWT.
     * 
     */
    protected final static boolean HATCH = true;
    /** The font size for the pending notification.
     * 
     */
    protected static final int   FONTSIZE = 40;

    /** The image currently displayed.
     * 
     */
    protected Image image = null;
    /** The size of the current image.
     * 
     */
    protected Dimension imageDim = null;
    /** The font to display the pending notification.
     * 
     */
    protected Font font = null;
    /** Whether or not an image is pending currently.
     * 
     */
    protected boolean pending = false;
    
    /** Set up the canvas for instantiation.
     * 
     */
    public ImageDisplayCanvas() {
	super();
	setBackground(Color.black);
	setSize(defaultDim.width, defaultDim.height);
	setVisible(true);
    }
    
    /** Display a new image.
     * 
     * @param imageDim The dimension of the image
     * @param image The image to be displayed.
     */
    public void showImage(Image image, Dimension imageDim) {
	pending = false;
	this.image = image;
	this.imageDim = imageDim;
	render();
    }

    /** This method is called when a new image is pending.
     * 
     */
    public void imagePending() {
	pending = true;
	render();
    }

    /** Redisplay the canvas.
     * 
     * @param g 
     */
    public void render(Graphics g) {
	Dimension size = getSize();
	if (font == null) {
	    Font defFont = getFont();
	    font = new Font(defFont.getName(),
			    defFont.getStyle(),
			    FONTSIZE);
	    g.setFont(font);
	}
	g.setColor(Color.black);
	g.fillRect(0,0,size.width, size.height);
	if (image != null) {
	    if ((imageDim.width <= size.width / 2) || 
		(imageDim.height <= size.height / 2)) {
		int x = (size.width - (imageDim.width * 2)) / 2;
		int y = (size.height - (imageDim.height * 2)) / 2;
		g.drawImage(image,
			    x, y,
			    imageDim.width * 2, imageDim.height * 2,
			    this);
	    }
	    else {
		int x = (size.width - imageDim.width) / 2;
		int y = (size.height - imageDim.height) / 2;
		g.drawImage(image, 
			    x, y, 
			    imageDim.width, imageDim.height,
			    this);
	    }
	}
	if (pending) {
	    if (HATCH) {
		g.setColor(new Color(128, 128, 128));
		for (int i = 0; i < size.height + size.width; i+= 4) {
		    g.drawLine(i, 0, 0, i);
		}
	    }
	    g.setColor(new Color(192, 192, 192));
	    g.setFont(font);
	    drawCenterText(g, "Waiting", size, 1, 3);
	    drawCenterText(g, "For", size, 2, 3);
	    drawCenterText(g, "Image", size, 3, 3);
	}
    }
    
    /** Draw horizontally and vertically centered text on a graphics canvas.
     * 
     * @param string The string to display
     * @param lines The total number of lines that will be displayed.
     * @param dim The dimensions of the graphics canvas.
     * @param g The graphics context.
     * @param line The line we are drawing now.
     */
    private static void drawCenterText(Graphics g, String string,
				       Dimension dim, int line, int lines) {
	FontMetrics fm = g.getFontMetrics();
	int charHeight = (int) Math.round((float)
					  (g.getFontMetrics().getAscent() + 
					   g.getFontMetrics().getDescent()) *
					  1.25f);
	g.drawString(string, (dim.width / 2) -
		     (g.getFontMetrics().stringWidth(string) / 2), 
	     ((dim.height - (lines * charHeight)) / 2) +
		     (line * charHeight) - 
		     (g.getFontMetrics().getAscent() / 2));
    }
}
    



