// $Id: ImageSourceAgentImpl.java,v 1.1 2000/05/18 22:04:58 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.
package net.hivecell.hive.agent.image;

import java.awt.*;
import java.rmi.RemoteException;

import net.hivecell.hive.*;
import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.agent.EventTranslatingAgentImpl;
import net.hivecell.hive.agent.cell.CellControlAgent;
import net.hivecell.hive.agent.desktop.ImageTakingDisplayingAgentImpl;
import net.hivecell.hive.shadow.*;
import net.hivecell.hive.support.*;
import net.hivecell.hive.support.Debug;

/** This abstract agent is the superclass of all agents that have images
 * available for retrieval. It needs to be subclassed to implement some way to get
 * those images.
 */
public abstract class ImageSourceAgentImpl 
extends EventTranslatingAgentImpl
implements ImageSourceAgent {

    /** The size of the buffer.
     */
    protected static final int BUFSIZE = 10;

    /** The buffer where images are stored.
     */
    protected ImageRingBuffer rb;

    /** Is the image source busy? The rules for how this flag are set are implementation dependent.
     */
    private boolean busy = false;

    /** This constructor instantiates the Agent with a fixed size ringbuffer.
     * 
     * @exception java.rmi.RemoteException 
     */
    public ImageSourceAgentImpl() 
    throws RemoteException {
	super();
	rb = new ImageRingBuffer( BUFSIZE );
    }

    public Agent sendUIAgent( Agent requester, CellAddress destination ) {
	try {
	    ImageTakingDisplayingAgentImpl itdai = new ImageTakingDisplayingAgentImpl();
	    itdai.setImageSourceAgent( (ImageSourceAgent)this );

	    Debug.notice( "shipping the image taking and displaying agent over to " + destination );
	    return myCell.moveAgent( itdai, destination );

	} catch( Exception error ) {
	    Debug.error( "problem instantiating and serializing a remote interface to send over", error );

	}
	return null;

    }

    /** Checks if the Agent is busy retrieving an image. The rules for how this flag are set are implementation dependent.
     */
    public boolean isBusy() {
	return busy;
    }

    /** Set whether the agent is busy or not.
     * 
     * @param busy 
     */
    public void setBusy(boolean busy) {
	this.busy = busy;
    }

    /** returns the image of index i;
     */
    public SerializableImage getImage(Agent sender, int i) {
	myCell.noteAgentMessage( this, "getImage", sender, this ); // the message
	myCell.noteAgentMessage( this, "getImage reply", this, sender ); // the reply
	return rb.getElement(i);
    }

    /** returns the most recent image. Does not request a new one.
     */
    public SerializableImage getImage(Agent sender) {
	myCell.noteAgentMessage( this, "getImage", sender, this ); // the message
	myCell.noteAgentMessage( this, "getImage reply", this, sender ); // the reply
	return rb.getElement(rb.getLatestIndex());
    }

    /** Get the index of the most recent image.
     */
    public int getLatestIndex(Agent sender) {
	myCell.noteAgentMessage( this, "getHighVal", sender, this ); // the message
	myCell.noteAgentMessage( this, "getHighVal reply", this, sender ); // the reply
	return rb.getLatestIndex();
    }

    /** return the index of the earliest image still in the buffer.
     */
    public int getEarliestIndex(Agent sender) {
	myCell.noteAgentMessage( this, "getLowVal", sender, this ); // the message
	myCell.noteAgentMessage( this, "getLowVal reply", this, sender ); // the reply
	return rb.getEarliestIndex();
    }

    /** return the size of the ring buffer.
     */
    public int getBufSize(Agent sender) {
	myCell.noteAgentMessage( this, "getLowVal", sender, this ); // the message
	myCell.noteAgentMessage( this, "getLowVal reply", this, sender ); // the reply
	return rb.getSize();
    }

    /** return the busy flag
     */
    public boolean isBusy(Agent sender) {
	myCell.noteAgentMessage( this, "isCameraBusy", sender, this ); // the message
	myCell.noteAgentMessage( this, "isCameraBusy reply", sender, this ); // the reply.
	return busy;
    }

    ///// Convenience methods for manipulating pixels... ///

    /** Convenience class for pixel manipulation.
     * 
     * @param pixel The input pixel.
     * @return The alpha value of pixel.
     */
    protected static int getAlpha(int pixel) {
	return (pixel >> 24) & 255;
    }

    /** Convenience class for pixel manipulation.
     * 
     * @param alpha The desired alpha value.
     * @param pixel The input pixel
     * @return The resultant pixel.
     */
    protected static int setAlpha(int pixel, int alpha) {
	return ((alpha & 255) << 24) | pixel;
    }

    /** Convenience class for pixel manipulation.
     * 
     * @param alpha The desired alpha value.
     * @return The resultant pixel.
     */
    protected static int setAlpha(int alpha) {
	return setAlpha(0, alpha);
    }

    /** Convenience class for pixel manipulation.
     * 
     * @param pixel 
     */
    protected static int getRed(int pixel) {
	return (pixel >> 16) & 255;
    }

    /** Convenience class for pixel manipulation.
     * 
     * @param red 
     * @param pixel 
     */
    protected static int setRed(int pixel, int red) {
	return ((red & 255) << 16) | pixel;
    }

    /** Convenience class for pixel manipulation.
     * 
     * @param red 
     */
    protected static int setRed(int red) {
	return setRed(0, red);
    }

    /** Convenience class for pixel manipulation.
     * 
     * @param pixel 
     */
    protected static int getGreen(int pixel) {
	return (pixel >> 8) & 255;
    }

    /** Convenience class for pixel manipulation.
     * 
     * @param pixel 
     * @param green 
     */
    protected static int setGreen(int pixel, int green) {
	return ((green & 255) << 8) | pixel;
    }

    /** Convenience class for pixel manipulation.
     * 
     * @param green 
     */
    protected static int setGreen(int green) {
	return setGreen(0, green);
    }
    
    /** Convenience class for pixel manipulation.
     * 
     * @param pixel 
     */
    protected static int getBlue(int pixel) {
	return pixel & 255;
    }

    /** Convenience class for pixel manipulation.
     * 
     * @param blue 
     * @param pixel 
     */
    protected static int setBlue(int pixel, int blue) {
	return (blue & 255) | pixel; 
    }

    /** Convenience class for pixel manipulation.
     * 
     * @param blue 
     */
    protected static int setBlue(int blue) {
	return setBlue(0, blue);
    }

}





