// $Id: CricketToLocationTranslatingAgentImpl.java,v 1.1 2000/05/18 22:05:01 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.agent.wearable;

import net.hivecell.hive.support.Debug;
import net.hivecell.hive.event.*;
import net.hivecell.hive.agent.EventTransceivingAgentImpl;
import net.hivecell.hive.cell.*;
import net.hivecell.hive.Global;
import net.hivecell.hive.description.HiveXML;
import org.w3c.dom.*;
import net.hivecell.hive.agent.wearable.Location;
import java.util.Hashtable;
import java.util.Date;

/** Agent that turns Cricket ID numbers (integers) into  hive urls **/
public class CricketToLocationTranslatingAgentImpl extends EventTransceivingAgentImpl {
    public int valueToSend = 0;
    protected Integer valueObject = new Integer(valueToSend);
    public final static String nowhere = "nowhere";
    protected Hashtable idToURL = null;           // mapping
    Thread timer;
    /** Milliseconds since last event was received. **/
    private int lastEventReceivedTime = 0;
    /** Millseconds to wait testing for notifications. **/
    private final int timeoutInterval = 1500;
    /** Milliseconds to pass before we decide we're nowhere. **/
    private final int nowhereThreshold = 10000;
    /** What's the last thing we sent out? **/
    private String lastSent = null;
    /** What's the last thing we received? **/
    private String lastLocPerceived = null;
        
    public CricketToLocationTranslatingAgentImpl() throws java.rmi.RemoteException { 
        super();
        if (idToURL == null)
            this.configure(Global.getConfigAsXMLDocument("CricketLocationMap.xml"));
    }
    
    /** When an event from a cricket is received, send out a location **/
    public void notify(HiveEvent event) {
        super.notify(event);

        // First, let's update the time.
        lastEventReceivedTime = 0;

        // Now, grab the data and parse out the data in it.
        Object data = event.getEventData();
        String ID = null;
        if (data instanceof Number)               // number data from cricket
            ID = data.toString();
        else if (data instanceof String)          // data from a string sender
            ID = (String)data;                    // probably for debugging
        else {
            Debug.notice("Ignoring unknown event " + event);
            return;
        }

        // Finally, lookup the ID in the database
        String newLoc = (String)idToURL.get(ID);
        Debug.notice("ID " + ID + " NewLoc " + newLoc);
        maybeSendLocation(newLoc);
    }

    public void doBehavior() {
        // implement a timeout - if we haven't heard from anyone in awhile,
        // then send out a "nowhere" string.
        while (!timeToStop) {
            // sleep a bit
            try {
                Thread.sleep(timeoutInterval);
            } catch (InterruptedException ex) {
                continue;                         // probably time to stop.
            }

            // age the last event received info a bit
            lastEventReceivedTime += timeoutInterval;

            // Now check, has it been awhile? If so, send out a nowhere
            // event. Only send it once though!
            if (lastEventReceivedTime >= nowhereThreshold) {
                Debug.notice("No cricket data received in awhile.");
                maybeSendLocation(nowhere);
            }
        }
    }

    /** Send out an event for the location. But first check that we've
     ** been asked to same this send location twice in a row, for stability,
     ** and then insure we only send out each location once.
     **/
    private synchronized void maybeSendLocation(String s) {
        if (s == null)
            return;                               // don't send null
        if (s.equals(lastLocPerceived)) {         // we've already seen once
            if (!s.equals(lastSent)) {            // if we haven't sent it
                Debug.notice("Sending " + s);     // send it!
                eventSendingHelper.broadcastEvent(s);
                lastSent = s;
            }
        } else {                                  // only seen once
            lastLocPerceived = s;                 // remember we saw this
            Debug.notice("Only saw " + s + " once, not sending yet.");
        }
    }

    /** Configure this agent from an XML file. Main thing to load
     ** is the hashtable associating IDs with locations. Sample file:
     ** <?xml version="1.0"?>
     ** <config>
     **   <locationMap>
     **     <location id="16">hive://rosemary.media.mit.edu/</location>
     **     <location id="32">hive://basil.media.mit.edu/</location>
     **   </locationMap>
     ** </config>
     **/
    public void configure(Document doc) {
        NodeList locations;
        // Drill down to the <location> array.
        try {
            Node config = HiveXML.selectFirst(doc, "config");
            Node map = HiveXML.selectFirst(config, "locationMap");
            locations = HiveXML.select(map, "location");
        } catch (NullPointerException ex) {
            Debug.error("Problem parsing map file.");
            return;
        }

        // Build our hashtable
        idToURL = new Hashtable();

        // For each location, pull out ID and text.
        int numLocations = locations.getLength();
        for (int i = 0; i < numLocations; i++) {
            Node locItem = locations.item(i);

            String id, location;
            try {
                location = HiveXML.getFirstNonWhitespaceChild(locItem).getNodeValue();
                NamedNodeMap attrs = locItem.getAttributes();
                Node idNode = attrs.getNamedItem("id");
                id = idNode.getNodeValue();
            } catch (NullPointerException npe) {
                Debug.error("Problem parsing map file.");
                return;
            }

            // Phew, parsed it out! Store it away.
            idToURL.put(id, location);
        }
    }
}
