// $Id: ExternalInterfaceShadowImpl.java,v 1.1 2000/05/18 22:05:03 tucker Exp $
// Hive. Copyright (c) 1998-2000, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.shadow;

import net.hivecell.hive.agent.ExternalInterfaceAgentImpl;
import net.hivecell.hive.event.ExternalEvent;
import net.hivecell.hive.event.IncomingExternalEvent;
import net.hivecell.hive.event.OutgoingExternalEvent;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.description.Description;

import java.io.BufferedReader;
import java.io.IOException;
import java.io.InputStreamReader;
import java.io.OutputStreamWriter;
import java.io.PipedInputStream;
import java.io.PipedOutputStream;
import java.io.PrintWriter;
import java.io.StringBufferInputStream;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Hashtable;

import com.docuverse.dom.DOM;
import org.w3c.dom.Document;
import org.w3c.dom.Node;
import org.w3c.dom.NodeList;

/**
 * the external interface shadow listens for connections coming from
 * an external source.  the other source can connect to us and send
 * events to us, and we can spit events out through the socket.
 * because event sending is unreliable, these events may or may not
 * get to the remote agent.  eventually we will need to actually send
 * an exception down the stream, but right now i don't want to deal
 * with that
 *
 * to deal with events the device want to send out, we need XML which
 * looks like this:
 * <outgoingevent>
 *  <destination>
 *   <hostname>...</hostname>
 *   <syntatic>...</syntatic>
 *   <semantic>
 *     .
 *     tags
 *     .
 *   </semantic>
 *  </destinaton>
 *  <data>
 *   .
 *   .
 *  </data>
 * </outgoingevent>
 *
 * the destination tag needs to be there, but only one of the syntatic
 * or the semantic is necessary.  the hostname is optional, if left
 * out we will query the local cell for the appropriate agent.
 *
 * now the return, events can be sent to these devices and the stream
 * coming off this shadow will look like:
 * <incomingevent syntatic="the class coming from" hostname="the host coming from">
 *  <data>
 *   .
 *   .
 *  </data>
 * </incomingevent>
 *
 * the device has to keep the stream open if it ever wants the ability
 * to receive events.  if a device connects, disconnects, and
 * reconnects then it will not be "teh same" device here, and it will
 * not receive events meant for the previous connection of it (or
 * probably not) 
 */
public class ExternalInterfaceShadowImpl
extends EventPublisherShadowImpl
implements ExternalInterfaceShadow {
    /**
     * the socket we are going to listen to incoming connections from 
     */
    protected ServerSocket incomingServerSocket = null;

    /**
     * the port that it is listening on
     */
    protected int incomingServerSocketPort = -1;
    

    /**
     * the mapping from the DeviceInterface object to the UID string
     */
    private Hashtable deviceToUID;

    /**
     * the mapping from the UID string to the DeviceInterface
     */
    private Hashtable UIDToDevice;

    /**
     * the counter we are using to keep track of UIDs -- this will
     * always be increasing 
     */
    private int counter = 0;


    /**
     * initialise the shadow -- we are going to parse the XML so we
     * can figure out what the port is we want to connect on, then we
     * open the server socket and start spawning off threads 
     *
     * @exception ShadowInitializationException something went wrong
     */
    public void init()
    throws ShadowInitializationException {
	
	// try parsing the XML stuff so we can get our configuration information
	Description desc = this.getDescription();
	if( desc == null )
	    throw new ShadowInitializationException( "need a description to get configuration information" );
	desc = desc.configContext();
	if( !desc.selectFirst( "port" ) )
	    throw new ShadowInitializationException( "need port information" );
	
	try {
	    incomingServerSocketPort = Integer.parseInt( desc.getMatch().getText() );
	    try {
		// time to try opening the socket that we can listen over
		incomingServerSocket = new ServerSocket( incomingServerSocketPort );
		Debug.notice( "opened a socket on port " + incomingServerSocket.getLocalPort() );
		
	    } catch( IOException error ) {
		Debug.error( "could not open a socket on port " + incomingServerSocketPort, error );
		throw new ShadowInitializationException( error );

	    }

	} catch( NumberFormatException error ) {
	    Debug.error( desc.getMatch().getText() + " is not a valid port number" );
	    throw new ShadowInitializationException( error );

	}

	// set our datastructures up
	deviceToUID = new Hashtable();
	UIDToDevice = new Hashtable();

	setReady( true );

	// spawn off the ExternalInterfaceAgentImpl
	Debug.notice( "spawning off our external interface agent" );
	try {
	    myCell.createNewAgent( ExternalInterfaceAgentImpl.class );

	} catch( InstantiationException error ) {
	    throw new ShadowInitializationException( "something wrong with the creation of the external interface agent" );
	    
	}

	Debug.notice( "spawning off a thread to deal with incoming connections" );
	( new Thread( new Runnable() {
	    public void run() {
		// at this point we are going to need to wait for incoming
		// connections and deal with it
		while( true ) {
		    try {
			Socket incomingSocket = incomingServerSocket.accept();
			
			Debug.notice( "got an incoming connection from " + incomingSocket.getInetAddress().getHostName() );
			Debug.notice( "passing connection to the device interface abstraction" );
			( new Thread( new DeviceInterface( incomingSocket ) ) ).start();
			
		    } catch( IOException error ) {
			// this means something is going wrong with our acceptor
			Debug.error( "io exception on the incoming server socket -- restarting it" );
			try {
			    incomingServerSocket.close();
			    incomingServerSocket = new ServerSocket( incomingServerSocketPort );
			
			} catch( IOException internalError ) {
			    // whatever --  i'm not sure what to do here 

			}
			
		    }
		    
		}

	    }

	})).start();

    }

    /**
     * this will handle devices connecting in over a certain socket.
     * when this class is instantiated, it needs to register itself in
     * the external interface shadow's database so it can receive
     * things coming into it 
     */
    class DeviceInterface 
    implements Runnable {

	/**
	 * the socket that we are talking to the device over
	 */
	Socket socket;

	/**
	 * the writer object we are using to talk to the device to
	 */
	PrintWriter writer;

	/**
	 * the reader that we can read for stuff from
	 */
	BufferedReader reader;

	/**
	 * the UID
	 */
	String UID;

	
	/**
	 * create the connection we want.  this will attempt to get
	 * the streams rigged up to the socket, and then it will
	 * register itself in the database 
	 *
	 * @param incomingSocket the socket that we want to connect over
	 * @exception IOException if something went wrong with getting the streams
	 */
	public DeviceInterface( Socket incomingSocket )
	throws IOException {
	    reader = new BufferedReader( new InputStreamReader( incomingSocket.getInputStream() ) );
	    writer = new PrintWriter( incomingSocket.getOutputStream() );
	    registerInterface( this );
	    UID = getUIDInterface( this );
	    
	}

	/**
	 * this is the method which will be running.  it listens on
	 * the socket for something new coming in from it, and then
	 * changes that to an ExternalEvent and spits that out.  
	 */
	 public void run() {
	     Debug.notice( "creating DOM object for XML parsing" );
	     DOM dom = new DOM();
	     
	     try {
		 while( true ) {
		     // read the XML from the incoming stream
		     Debug.notice( "waiting for an event to come in from " + UID );
		     String incomingEvent = "", incoming = "";
		     while( !( incoming = reader.readLine() ).startsWith( "</outgoingevent>" ) ) {
			 Debug.notice( "reading: \"" + incoming + "\"" );
			 incomingEvent += incoming;

		     }
		     incomingEvent += incoming;
		     Debug.notice( "reading: \"" + incoming );
		     Debug.notice( "got a full event from " + UID );

		     // now convert the XML into a HiveEvent by
		     // putting the Document node as the event data
		     // and shipping the event out
		     Debug.notice( "about to parse XML stream" );
		     Document document = dom.openDocument( new StringBufferInputStream( incomingEvent ) );

		     Debug.notice( "creating an outgoing external event object" );
		     OutgoingExternalEvent event = new OutgoingExternalEvent( ExternalInterfaceShadowImpl.this, UID, eph.getCount(), document );

		     // spit the event off to our agent upstairs who
		     // will deal with sending it out via the event
		     // sending agent interface (i don't want to hook
		     // directly into the agent messaging agent here
		     // for compatibility reasons)
		     notifySubscribers( event );

		 }

	     } catch( IOException error ) {
		 Debug.error( "exception with the incoming stream", error );
		 removeInterface( this ); // remove this interface from the lists
		 
	     }

	 }

	/**
	 * this is the method that should be called to get the
	 * ExternalEvent to be delivered to this device
	 *
	 * @param event the External event 
	 */
	public synchronized void deliverEvent( IncomingExternalEvent event ) {
	    // convert the external event to XML, then spit that
	    // text down the stream to the device.  we just need
	    // to take the DocumentObject and convert it down to a
	    // string and then send that out over the stream
	    String XMLString = event.getXMLDocument().toString();
	    // pull out the <?xml version="1.0"?>
	    XMLString = XMLString.substring( XMLString.indexOf( ">" ) ).trim();
	    
	    Debug.notice( "sending the XML string to " + UID );
	    Debug.notice( "sending " + XMLString );
	    writer.println( XMLString );
	    
	    // check to see if there has been an error writing this
	    // string out
	    if( writer.checkError() )
		removeInterface( this ); // remove this interface from the lists

	}

    }

    /**
     * register the interface from the device in our database.  it
     * will come up with a unique identifier for this device in the
     * form of "external:cell address:UID" and use that as the key to
     * this device.
     *
     * @param device the device that we want to register 
     */
    protected synchronized void registerInterface( DeviceInterface device ) {
	String UID = "external:" + myCell.getAddress() + ":" + counter++;
	deviceToUID.put( device, UID );
	UIDToDevice.put( UID, device );
	
    }

    /**
     * remove the UID from the datastructures
     *
     * @param UID the UID that you want to remove
     * @return false if we didn't remove it from the datastructure
     */
    protected synchronized boolean removeInterface( String UID ) {
	Object toRemove = UIDToDevice.remove( UID );
	if( toRemove != null ) 
	    return ( deviceToUID.remove( toRemove ) != null );
	return false;

    }

    /**
     * remove the UID from the datastructures
     *
     * @param device the device that we want to remove
     * @return false if we didn't remove it from the datastructure
     */
    protected synchronized boolean removeInterface( DeviceInterface device ) {
	Object toRemove = deviceToUID.remove( device );
	if( toRemove != null ) 
	    return ( UIDToDevice.remove( toRemove ) != null );
	return false;

    }

    /**
     * get the UID of a connection
     *
     * @param device the device that we connect in
     * @return UID the uid of the connection (could possibly be null)
     */
    protected synchronized String getUIDInterface( DeviceInterface device ) {
	return (String)deviceToUID.get( device );

    }

    /**
     * get the device from teh UID
     *
     * @param UID the UID that we want to retrieve
     * @return teh device class (may return null)
     */
    protected synchronized DeviceInterface getDevice( String UID ) {
	return (DeviceInterface)UIDToDevice.get( UID );

    }

    /**
     * spit an external event off to the device specified by the UID
     *
     * @param UID the idnetiifer that we want to spit the event to
     * @param event the ExternalEvent that we want to spit down
     */
    public void sendEvent( IncomingExternalEvent event ) {
	DeviceInterface deviceInterface = (DeviceInterface)getDevice( event.getDestinationUID() );
	if( deviceInterface != null )
	    deviceInterface.deliverEvent( event );

    }

}
