// $Id: Global.java,v 1.1 2000/05/18 22:00:47 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive;

import net.hivecell.hive.support.async.*;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.description.HiveXML;

import java.awt.Font;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileOutputStream;
import java.io.InputStream;
import java.io.OutputStream;
import java.text.SimpleDateFormat;
import java.util.Date;
import java.util.Vector;
import java.util.Properties;
import java.util.Random;
import org.w3c.dom.Document;

/** Provide some useful global methods for all of Hive.
 ** No important state should be here, only some conveniences.
 **
 ** @version $Revision: 1.1 $
 ** @author Nelson Minar <nelson@media.mit.edu>
 **/
public class Global 
extends Object {

    static {
	// this code block will be loaded the first time this class is
	// loaded.  since this is loaded by Hive.java, this will be
	// loaded immediately at the beginning.  and since system
	// properties can be thought to be global, i'm going to set
	// some of the global variables here
	Properties props = System.getProperties();
	props.put( "hive.UI", "true" );
	props.put( "hive.GUIAgent", "true" );
	props.put( "hive.JoinCellList", "true" );
	props.put( "hive.LazyShadowInstantiation", "false" );
	props.put( "hive.VisualShadowConfig", "false" );

    }

    /** Version number of system, only bumped up occasionally. */
    public static final String version = "0.9.1";

    /** Port that Hive uses if not told otherwise. **/
    public static final int defaultPort = 23235;

    /** Date formatter used by dateString(). **/
    private static final SimpleDateFormat timestampFormatter = new SimpleDateFormat( "dMMMyyyy-HH:mm:ss.SSS" );

    /** URL to use to get the list of servers.
     **/
    public static final String defaultCellListURL =
	"http://nelson.www.media.mit.edu/people/nelson/tmp/hiveMasterCell";

    /** Definitions for property names used to pass data into the Cell
     * at construction..
     */
    public static final String UI = "hive.UI";
    public static final String GUIAgent = "hive.GUIAgent";
    public static final String ConfigDir = "hive.ConfigDir";
    public static final String CellDescription = "HiveCellDescription";
    //    public static final String ShadowConfig = "hive.ShadowConfig";
    //    public static final String AgentConfig = "hive.AgentConfig";
    //    public static final String HiveStartup = "hive.HiveStartup";
    public static final String JoinCellList = "hive.JoinCellList";
    public static final String MasterCellListCell = "hive.MasterCellListCell";
    public static final String LazyShadowInstantiation = "hive.LazyShadowInstantiation";
    public static final String VisualShadowConfig = "hive.VisualShadowConfig";
    public static final String AgentClassdir = "hive.AgentClassdir";  // NOTE: Kill this as soon as finding the classpath is done.


    /** List of those options which were set on the command line at startup, so
     *  the cell won't override the values if the same option is in its description.
     */
    public static final Vector argList = new Vector();

    
    /** Flag to know when we're on Windows, used in a few contexts. **/
    public static boolean windowsOS = false;

    /** Default font all windows should use **/
    public static Font defaultFont = null;

    /** Shared random source for any agents that need a random value. (Like
     ** the CameraAgent)
     **/
    public static Random random = new Random();

    /** The default implementation for the async calling **/
    public static final AsyncCallerClass asyncCallerClass = AsyncOneMethodCaller.getInstance();

    /** Default Configuration directory. Can be changed by 
     ** setConfigDirectory()
     **/
    private static String configDir = System.getProperty( "user.home" ) + File.separator + ".hive";

    /** Set the configDirectory */
    public static void setConfigDir(String string) {
	configDir = string;
    }

    /** Get the configDirectory */
    public static String getConfigDir() {
	return configDir;
    }

    /** Return a handle to a config file. */
    public static File getConfigAsFile(String name) {
	return new File(configDir, name);
    }

    /** Return an input stream from a config file */
    public static InputStream getConfigAsInputStream(String name) {
	try {
	    return new FileInputStream(getConfigAsFile(name));
	}
	catch (Exception e) {
	    Debug.notice("Unable to open config: " + name);
	    return null;
	}
    }

    /** Return an output stream from a config file. */
    public static OutputStream getConfigAsOutputStream(String name) {
	try {
	    return new FileOutputStream(getConfigAsFile(name));
	}
	catch (Exception e) {
	    Debug.warning("Unable to open config for writing: " + name);
	    return null;	    
	}
    }

    /** Get a configuration file as an XML object. **/
    public static Document getConfigAsXMLDocument(String name) {
        InputStream is = Global.getConfigAsInputStream(name);
        if (is == null)
            return null;
        else {
            HiveXML hiveXML = new HiveXML(is);
            return hiveXML.getRoot();
        }
    }
    
    /** Returns the directory where the user's menu rdf file are. */
    public static String getMenuDir() { 
	String fs = System.getProperty("file.separator");
	return (configDir + fs + "menu" + fs);
    }

    /** Returns the directory where the user's agent rdf file are. */
    public static String getAgentDir() { 
	String fs = System.getProperty("file.separator");
	return (configDir + fs + "agent" + fs);
    }

    /** Returns the directory where the user's shadow rdf file are. */
    public static String getShadowDir() { 
	String fs = System.getProperty("file.separator");
	return (configDir + fs + "shadow" + fs);
    }

    /** Given an object, return a string that's the typename.
     ** Shortens Class.toString() - remove "class " and well known packages
     **/
    public static String shortClassName(Object o) {
	if (o == null)
	    return "(null)";
    
	String s = o.getClass().toString().substring(6);   // cut off "class "
	return stripName(s);
    }

    /** Given an agent, return a short string describing it.
     ** We do this by parsing the RMI string representation.
     ** TODO - it'd be good if we could ask the object where it's
     ** server is, but we can't get that without a method call.
     **
     ** @param obj The object to print a short representation of.
     ** @return If local, just the shortest type name. If remote,
     **         then typename_stub@host
     **/
    public static String shortString(Object obj) {
	if (!(obj instanceof java.rmi.server.RemoteObject))
	    return obj.toString();

	String rs = obj.toString();

	// First, get the type name out of the front.
	int typeNameLocation = rs.indexOf('[');
	String typeName;
    
	if (typeNameLocation < 0) {
	    Debug.warning("Couldn't parse RMI string " + rs);
	    typeName = "???" + rs + "???";
	} else {
	    typeName = stripName(rs.substring(0, typeNameLocation));
	}

	// Now, find the location.
	String location;
	int locationIndicator = rs.indexOf("(remote)");
	if (locationIndicator >= 0) {                 // it's a remote!
	    String firstHalf = rs.substring(0, locationIndicator);
	    try {
		// Pull out the hostname, ignore the port.
		location = firstHalf.substring(firstHalf.lastIndexOf('[') + 1,
					       firstHalf.lastIndexOf(':'));
	    } catch (StringIndexOutOfBoundsException ex) {
		location = "???";
		// ignore it
	    }
	} else {
	    locationIndicator = rs.indexOf("(local)");
	    if (locationIndicator < 0) {                // what is this???
		//Debug.warning("Warning, we couldn't parse RMI string " + r);
		location = "???";
	    } else {
		location = "";
	    }
	}
	if (location.length() > 0)
	    return typeName + "@" + location;
	else
	    return typeName;
    }

    /** Strip out the package portion of the name. **/
    private static String stripName(String s) {
	return s.substring(s.lastIndexOf('.') + 1);
    }
      
    /** Format a time into a standard string format.
     ** @param ms The time to format, in milliseconds since the epoch.
     **/
    public static String dateString(long ms) {
	return timestampFormatter.format(new Date(ms));
    }

    /** Format the current time into a standard string format.
     **/
    public static String dateString() {
        return dateString(System.currentTimeMillis());
    }

    /**
     * test to see if these two arrays contain the same pointers.
     * does not require for them to be in the same order
     *
     * @param first the first array
     * @param second the second array
     * @return true if they are equal 
     */
    public static boolean arrayEquals( Object[] first, Object[] second ) {
	if( ( first == null ) && ( second == null ) )
	    return true;

	if( first.length != second.length )
	    return false;

	for( int count=0;count<first.length;count++ ) {
	    boolean found = false;
	    for( int goThrough=0;goThrough<second.length;goThrough++ )
		if( first[count].equals( second[goThrough] ) ) {
		    found = true;
		    break;

		}
	    if( !found )
		return false;

	}

	return true;

    }

    /** 
     * Munge a server address to be suitable for inclusion in a
     * properties file. 
     *
     * @param address the address that we want to munge
     * @return the munged address
     */
    public static String mungeAddress( String address ) {
	return address.replace( ':', '-' );

    }

}

