// $Id: AgentImpl.java,v 1.2 2000/05/19 19:55:53 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.agent;
import java.rmi.RemoteException;
import java.rmi.server.UnicastRemoteObject;
import java.io.Serializable;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.util.Vector;
import java.util.Hashtable;
import java.awt.event.ActionEvent;

import net.hivecell.hive.Global;
import net.hivecell.hive.description.Description;
import net.hivecell.hive.description.RemoteDescribable;
import net.hivecell.hive.description.Describable;
import net.hivecell.hive.cell.RemoteCell;
import net.hivecell.hive.cell.Cell;
import net.hivecell.hive.cell.AgentDeliveryException;
import net.hivecell.hive.shadow.cell.ComponentManagerShadow;
import net.hivecell.hive.shadow.cell.ComponentManagerShadowImpl;
import net.hivecell.hive.shadow.ShadowNotFoundException;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.support.CellAddress;
import net.hivecell.hive.support.IconLoader;
import net.hivecell.hive.support.PPM;
import net.hivecell.hive.support.SerializableImage;

import org.w3c.dom.Document;

/** 
 * Base class for Hive agent implementations.
 * 
 * An agent has a life cycle.<ol>
 * 
 * <li>Once in its lifetime, the agent is constructed by the no
 * argument constructor.
 *
 * <li>The agent arrives on a server; its <code>doLocalSetup()</code>
 * method is called. This happens whenever it first comes to a server,
 * whether created by <code>Cell.createNewAgent(Class)</code> or when
 * it moves to the server.
 *
 * <li>The agent's <code>doBehavior()</code> method is called. This
 * method is the agent's main loop, where it's behavior is
 * implemented. The agent is free to do what it wants, but should
 * respect <code>timeToStop</code>.  The agent sould never exit this
 * method, a convienience method <code>waitUntilDeath</code> is
 * provided for this purpose.
 *
 * <li>When an agent is being killed, the notification
 * <code>onDying</code> method is called.  The <code>onMoving</code>
 * is called for the notification for moving
 *
 * <li><code>onLocalCleanup</code> is called to ask the agent to free
 * all its local resources.  The reason for agent death can be
 * discovered by looking at the value of <code>stopCode</code> if
 * cleanup is dependant on the "type" of agent death </ol>
 *
 * An agent might travel to another server, either by moving itself
 * or being moved by someone authorized. The <code>moveTo()</code>
 * method initiates the move: when the agent arrives at the remote
 * server, <code>arriveAt()</code> and <code>doBehaviour()</code>
 * will be called again.
 * 
 * All agents have a notion of a list of "connections", which agents
 * they talk to. It is up to the agent to define the semantics of
 * what a connect means. Agents also have an icon that represents
 * them: by default this icon is loaded from
 * icons/AgentClassName.ppm.
 *
 * Agents contain something similar to "popup menus" called
 * ActionCommands.  These can be rendered by a UI if so wished, but
 * this is also so things can issue a string comamnd to the agent and
 * it will perform the command.  this is just another way to ask the
 * agent to perform an action
 * 
 * @version $Revision: 1.2 $ 
 */
public abstract class AgentImpl 
extends UnicastRemoteObject 
implements Agent, Serializable, Describable {

    /** 
     * Pointer to the server I am currently on.
     */
    protected transient Cell myCell = null;

    /**
     * the address this agent is currently sitting at
     */
    protected transient CellAddress myAddress = null;

    /**
     * Pointer to the thread group for this agent -- ell behaved
     * agents, if they need to spawn separate threads, will make those
     * threads members of this agents thread group.  There is no way,
     * unfortunately, for Hive to enforce this though 
     */
    protected transient ThreadGroup agentThreadGroup = null;

    /**
     *  Pointer to my XML description.
     */
    protected Description description = null;
  

    /**
     * Just describing whether or not this agent is ready for
     * operation or not 
     */
    protected transient boolean readyFlag = false;

    /** 
     * Variable indicating this agent should die soon. This is set
     * asynchronously, typically by the server. Agents should inspect
     * this in their main loop.
     * 
     * @see doBehavior
     */
    protected transient boolean timeToStop = false;
    
    /**
     * Variable describing the reason why an agent was asked to stop
     */
    protected transient int stopCode;

    /**
     * the stop code which means the agent is being killed
     */
    public static final int AGENTKILLED = 1;

    /**
     * the stop code which means the agent is being moved
     */
    public static final int AGENTMOVED = 2;


    /** 
     * The icon that represents this agent. Should be loaded at create time.
     */
    protected SerializableImage icon = null;

    /**
     * the name of the icon file
     */
    protected String iconName = "";

    /** 
     * Hashtable and list for the popup menu items or their
     * non-graphical equivalents.
     */
    protected Hashtable commands = new Hashtable();  /** Hashtable of CommandObjects */
    protected Vector commandList = new Vector();     /** List of the CommandObject's names */


    /** 
     * Basic constructor for agents. This only gets called once in an
     * agent's lifetime, when it is first created. It is <b>not</b>
     * called when the agent arrives at a new host.
     * 
     * @see doLocalSetup  Note - because of Java syntax rules, all subclasses must explicitly declare a constructor, even if all it does is call  the superclass constructor.  
     */
    public AgentImpl() 
    throws RemoteException {
        super();
	loadIcon();
	description = new Description( this, this.getName() );

    }

    /** 
     * Return the semantic description of this agent.
     *
     * @return the description of the agent
     */
    public Description getDescription() {
	return description;
    }

    /** 
     * Set the semantic deescription of this agent.  we take the xml
     * description, rip it apart and make sure the description of the
     * xml description is the right thing
     *
     * @param desc the desc that we want to add to this agent 
     */
    public void setDescription( Description desc ) { 
	desc.noContext();
	this.description = new Description( this, desc.getNode() );

    }

    /**
     * Get the value of iconName.
     * 
     * @return Value of iconName.
     */
    public String getIconName() { 
	return iconName; 

    }
    
    /**
     * Set the value of iconName.
     *
     * @param v  Value to assign to iconName.
     */
    public void setIconName( String v ) {
	this.iconName = v; 
	loadIcon();

    }
    
    /**
     * get a "name" for this agent -- if the description contains a
     * nickname, then we should return that, else we return the short
     * string of the name 
     *
     * @return the "name" of this agent
     */
    public String getName() {
	if( description != null ) {
	    Description desc = description.semanticContext();
	    if( desc != null )
		if( description.selectFirst( "nickname" ) )  // NOTE: OK for now, but should probably make this "name" later.
		    return description.getMatch().getText();

	}
	return Global.shortClassName(this);

    }

    /** 
     * Tell this agent to connect to some other agent.  It is up to
     * the agent to define the semantics of what this means.  By
     * default, AgentImpl just prints an error message and returns
     * false.
     * 
     * @param otherAgent reference to the agent to connect to
     * @return whether the connection was successful.  
     */
    public boolean connectTo( Agent otherAgent ) {
	Debug.notice( "there is no default connection for an agent." );
	return false;

    }
  
    /** 
     * Disconnects this agent from all agents.  It is up to the agent
     * to define the semantics of what this means.  By default,
     * AgentImpl just prints an error message and returns false.
     */
    public void disconnectFromAll() {
	Debug.notice( "there is no default disconnection for an agent." );

    }

    /** 
     * Disconnects this agent from the specified other agent. It is up
     * to the agent to define the semantics of what this means.  By
     * default, AgentImpl just prints an error message and returns
     * false.  
     */
    public void disconnectFrom( Agent otherAgent ) {
	Debug.notice( "there is no default disconnection for an agent." );

    }

    /**
     * Return a vector of all the agents that are connected to me.  It
     * is up to the agent to define the semantics of what this means.
     * By default, AgentImpl just prints an error message and returns
     * false.  
     *
     * @return a vector containing the incoming connections -- in this case it is empty
     */
    public Vector listAllIncomingConnections() {
	Debug.warning( "there is no default connection list for an agent." );
	return new Vector();

    }

    /**
     * Return a vector of all that agents that i am connected to.  It is up
     * to the agent to define the semantics of what this means.  By
     * default, AgentImpl just prints an error message and returns
     * false.  
     *
     * @return a vector containing the outgoing connections -- in this case it is empty
     */
    public Vector listAllOutgoingConnections() {
	Debug.warning( "there is no default connection list for an agent." );
	return new Vector();

    }


    /**
     * Handle the setup when an agent arrives on a host -- it is
     * appropriate to allocate system resources (file descriptors,
     * etc) here and even to locate other agents that this agent needs
     * to communicate with.
     *
     * @exception AgentInitializationException can be thrown by the agent if for some reason it cannot fufil its setup stuff
     */
    public void doLocalSetup() 
    throws AgentInitializationException { 

    }

    /** 
     * Set the cell flag on the agent. Normally user code should not
     * call this - creation code in the cell should handle this 
     *
     * @param c the cell that we want to set this agent to
     */
    public void setCell( Cell c ) {
	if( myCell != null )
	    Debug.warning( "setCell() was called on " + Global.shortString( this ) + " when it had a cell set!" );
	myCell = c;
	myAddress = myCell.getAddress();

    }

    /**
     * Set the thread group of this agent.  Normally user code should
     * not call this -- the creation methods in the server will set
     * this properly 
     *
     * @param tg the thread group for this agent
     */
    public void setThreadGroup( ThreadGroup tg ) {
	this.agentThreadGroup = tg;

    }
    
    /** 
     * Do the basic behavior for an agent, the agent's main loop.
     * Override this to provide your particular agent's behavior.
     * <B>Note</b>: the doBehavior loop should not exit.  Please call
     * waitUntilDeath() at the end of this method, this will prevent
     * the Cell from believing this agent prematurely exited.
     * 
     * @see timeToStop
     * @see doLocalSetup 
     */
    public abstract void doBehavior();

    /**
     * Used at the end of the doBehavior method.  Simply wait until
     * the timeToStop flag has been set and the notify method has been
     * called so we can exit properly, This should be done, otherwise
     * the agnet will fall out of its doBehavior loop, and the Cell
     * will "clean" the agnet up
     *
     * @see doBehavior 
     */
    protected void waitUntilDeath() {
	while( !timeToStop ) 
	    synchronized( this ) {
		try {
		    wait();
		} catch( InterruptedException error ) { }
	    }

    }

    /**
     * Used to notify the agent that it is about to be moved 
     */
    public void onMoving() { }

    /** 
     * Move this agent to a new host.  this basically moves the call
     * up to the cell who handles it
     * 
     * @param address the address that we want to move this agent to 
     * @exception AgentDeliveryException if something went wrong with the movement
     */
    public void moveTo( CellAddress address ) 
    throws AgentDeliveryException {
	//	Debug.notice( Global.shortString( this ) + ".moveTo(" + address + ")" );
        myCell.moveAgent( this, address );

    }

    /**
     * Command the agent to free its local resources here.  It may or
     * may not equal what happens when the agent is supposed to die --
     * but here would be the appropriate place to free File
     * descriptors, and stop threads, but do not do anything related
     * to the agent community -- do not notify them that this agent is
     * about to die.  When this method terminates, this agent should
     * be ready to be killed 
     */
    public void doLocalCleanup() { }
    
    /**
     * Used to notify the agent that it is about to be killed 
     */
    public void onDying() { }

    /**
     * Ask the agent to politely die -- by default, the agent just
     * relays the call to the server and asks it tkill itself 
     */
    public void diePlease() {
	myCell.killAgent( (AgentImpl)this );

    }

    /** 
     * Sets the time to stop flag 
     *
     * @param flag the code to whcih we will ask the agent to stop.  either AgentImpl.AGENTKILLED, or AgentImpl.AGENTMOVED
     */
    public final void setTimeToStop( int flag ) {
	timeToStop = true;
	stopCode = flag;
	synchronized( this ) {
	    notifyAll();

	}

    }

    /** 
     * Load the icon for an agent off of disk.  If the agent is class
     * <code>net.hivecell.hive.FooAgent</code> then this looks for
     * <code>net.hivecell.hive/icons/FooAgent.ppm</code> in the class
     * path or Jar file.  If that doesn't exist, then a default is
     * substituted.  
     */
    protected void loadIcon() {
 	InputStream is = null;
	if( iconName.equals( "" ) ) {
	    icon = IconLoader.loadIcon( this.getClass() );
	    return;

	} else {
	    is = Global.getConfigAsInputStream( iconName );    // Tries to load it from the user's home directory
	    if( is == null ) {
		Debug.notice( "trying to get user-specified icon from directory in CLASSPATH" );
		is = getClass().getResourceAsStream( iconName ); // Tries to load it from the current directory

	    }

	}

	if( is == null ) {
	    // Note: if these warnings are ever thrown, something has
	    // gone very wrong... (UnknownAgent.ppm not in
	    // hive/agent/icons)
	    Debug.warning( "no icon found for " + this.getClass().toString() + ", substituting " + IconLoader.defaultIconName );
	    is = getClass().getResourceAsStream( IconLoader.defaultIconName );
	    if( is == null ) {
		Debug.critical( IconLoader.defaultIconName + " doesn't exist therefore agent icons are broken -- using a blank icon." );
		
		icon = new SerializableImage( new int[]{ 1 }, new java.awt.Dimension( 1, 1 ) );
		return;

	    }

	}

	// OK, got the file as a stream, let's parse the PPM out and cast it into SerializableImage.
	try {
	    icon = new PPM( is );

	} catch( Exception ex ) { 
	    Debug.critical( "error loading agent's icon: " + ex ); 

	}

    }

    /** 
     * Return the SerializableImage object that is the agent's icon.
     *
     * @return the serializable image which is the icon
     */
    public SerializableImage getIcon() {
	return icon;

    }

    /** 
     * Return the cell I'm living on.
     *
     * @return the remote cell that this agent is on
     */
    public RemoteCell getCell() {
	return (RemoteCell)myCell;

    }

    /**
     * return the address that i'm at
     *
     * @return the cell address that this agnet is on
     */
    public CellAddress getAddress() {
	return myAddress;

    }

    /**
     * Return the thread group that we are in
     *
     * @return the thread group that the agent has
     */
    public ThreadGroup getThreadGroup() {
	return agentThreadGroup;

    }

    /**
     * Override this function to let an agent configure itself from a Description.
     * This method is called by Hive if the agent was created through
     * The user's Global.CellDescription file. If you want to load a
     * configuration by hand, see Global.getConfigAsXMLDocument().
     *
     * @param o the object that asked to reconfigure the agent
     * @param desc A description which the agent should use to configure itself.
     */
    public void configure( Object o, Description desc ) {

    }


    /**
     * Add a popup menu item or non-graphical equivalent to the 
     * agent's behavior.
     *
     * @param com the action command which is the "menu item" for the agent
     */
    public void addActionCommand( ActionCommand com ) {
	commandList.addElement( com.getName() );
	commands.put( com.getName(), com );

    }

    /**
     * Invoke an action command by name.  Currently, this will only be
     * called in a baseUICanvas or its non-graphical equivalent.
     * 
     * @param com the name of the command that we want to execute
     */
    public boolean invokeActionCommand( String com ) {
	if( commands.containsKey( com ) ) {
	    ((ActionCommand)commands.get( com )).invoke();
	    return true;

	}
	return false;

    }

    /**
     * Get the names of the ActionCommands this agent uses in the
     * order in which they were entered.
     *
     * @return a vector of the commands 
     */
    public Vector getActionCommands() { 
	return commandList; 
    
    }

    /** 
     * A convenience method to return the ComponentManagerShadow or 
     * throws an exception if one is not present. 
     *
     * @return the ComponentManagerShadow if one exists on this cell
     * @exception ShadowNotFoundException if we can't find the ComponentManagerShadow
     */
    public ComponentManagerShadow getComponentManagerShadow() 
    throws ShadowNotFoundException {
	return (ComponentManagerShadow)myCell.getShadow( (ComponentManagerShadowImpl.class).getName() );

    }

    /** 
     * Handle serialization requests.
     *
     * @param out the stream that we want to serialize to
     * @exception IOException if something goes wrong with the serialization
     */
    private void writeObject( ObjectOutputStream out ) 
    throws IOException {
	out.defaultWriteObject();

    }

    /**
     * Handle deserialization
     *
     * @param in the stream that we want to deserialize from
     * @exception IOException if something goes wrong with deserialization
     * @exception ClassNotFoundException if we don't know what class to deserialize to
     */
    private void readObject( ObjectInputStream in ) 
    throws IOException, ClassNotFoundException {
	in.defaultReadObject();

    }

    /**
     * This method tells the world that this agent is ready for
     * operation.  Deciding when an agent is ready is up to the 
     * agent's author, however the agent should -not- report that it 
     * is ready until it can be connected to, receive and send 
     * events, and perform any other special duties it needs to.
     * setIsReady(true) is called automatically when the agent's 
     * doLocalSetup is done in Cell.java runAgentThread().
     *
     * @param flag true if this agent is ready 
     */
    public final void setIsReady( boolean flag ) {
	readyFlag = flag;
	synchronized( this ) {
	    notifyAll();
	}

    }

    /**
     * this flag tells another object whether this agent is ready to
     * do its stuff.  the semantics of being ready are left up to the
     * agent's author, however the only restriction is that the agent
     * should -not- report that it is ready until the server calls
     * setIsReady( true ) on this agent
     *
     * @return if the agent is "ready" 
     */
    public boolean isReady() {
	return readyFlag;

    }

    /**
     * this method should block until the agent is ready -- it just
     * sits in the wait method until it gets notified from the
     * setIsReady method 
     */
    public void blockUntilReady() {
	while( !isReady() ) {

	    try {
		synchronized( this ) {
		    wait();
		} 
	    } catch( InterruptedException error ) { }

	}

    }



    /** 
     * Tell if the agent been told to stop. 
     *
     * @return true if the agent has been asked to stop
     */
    public boolean isTimeToStop() {
        return timeToStop;

    }
    
}










