// $Id: ConnectAgentHelper.java,v 1.1 2000/05/18 22:00:49 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.agent;

import java.util.*;
import java.rmi.RemoteException;

import net.hivecell.hive.event.*;
import net.hivecell.hive.Global;
import net.hivecell.hive.cell.*;
import net.hivecell.hive.description.AgentSet;
import net.hivecell.hive.support.Debug;


/** Helper class for ConnectAgentHelper interface;
 ** agent implementations can delegate methods to this class.
 **
 ** This class handles automatic wiring of an agent to and from 
 ** other agents.
 **
 ** Biggest thing still to do: add semantic criteria as well as the syntactic
 **/
public class ConnectAgentHelper extends Object {
    /** Agent we're helping **/
    private AgentImpl agent;
    /** Cell we're on **/
    private Cell agentCell;
    /** Can we receive events? **/
    private boolean weReceive;
    /** Can we send events? **/
    private boolean weSend;

    /** Construct a helper for an event sending or receiving agent to delegate to.
     * @param agentHelped The agent to be helped.
     * @param server The Cell we're on.
     **/
    public ConnectAgentHelper(AgentImpl agentHelped, Cell server) {
        this.agentCell = server;
        this.agent = agentHelped;

        this.weReceive = (agent instanceof EventReceivingAgent);
        this.weSend = (agent instanceof EventSendingAgent);

        //        if (this.weReceive) Debug.notice("We receive"); else Debug.notice("We do NOT receive");
        //        if (this.weSend) Debug.notice("We send"); else Debug.notice("We do NOT send");
    }



    // NOTE: 2/27, tucker: This is the same thing (essentially) as queryAgents, and very similar to querySet.
    // Select agents that match syntactic and semantic descriptions
    // This should probably be moved to the RDF package somewhere (Lookup?)
    public AgentSet lookupAgentMatches(Vector agentVec, String[] syntactic, String[] semantic) {

        if (agentVec == null) return(null);
	AgentSet ret = new AgentSet(agentVec);

	return (ret.querySet(syntactic, semantic));
    }



    // Are we listening to an agent with this syntactic yet?
    public boolean listeningTo (String[] syntactic, String[] semantic) throws ClassNotFoundException {
        // If we can't receive or it can't send, then we're clearly not listening to it
        if (!this.weReceive || !doesThisSend(syntactic)) {
            Debug.notice("Our agent doesn't receive, or "+syntactic+" doesn't send, so not listening.");
            return(false);
        }
        else {
            Vector subscriptions = ((EventReceivingAgentImpl)agent).agentSubscriptions;
            if (subscriptions == null) return(false);
            else {
                AgentSet matches = lookupAgentMatches(subscriptions, syntactic, semantic);
                if (matches.size() > 0) {
                    //                    Debug.notice("We're already listening to "+syntactic);
                    return(true);
                }
                else return(false);
            }
        }
    }

    // Are we sending to an agent with this syntactic yet?
    public boolean sendingTo (String[] syntactic, String[] semantic) throws ClassNotFoundException {
        // If we can't send or it can't receive, then we're clearly not sending to it
        if (!this.weSend || !doesThisReceive(syntactic)) {
            Debug.notice("Our agent doesn't send, or "+syntactic+" doesn't receive, so not sending to it.");
            return(false);
        }
        else {
            EventSendingHelper esh = null;
            if (agent instanceof EventTransceivingAgentImpl) 
                esh = ((EventTransceivingAgentImpl)agent).eventSendingHelper;
            else if (agent instanceof EventSendingAgentImpl)
                esh = ((EventSendingAgentImpl)agent).eventSendingHelper;

            if (esh == null) {
                Debug.warning("Can't get an eventSendingHelper for "+agent+", even though we think it sends.");
                return(false);
            }
            Vector listeners = esh.getListeners();
            if (listeners == null) return(false);
            else {
                AgentSet matches = lookupAgentMatches(listeners, syntactic, semantic);
                if (matches.size() > 0) {
                    //                    Debug.notice("We're already sending to "+syntactic);
                    return(true);
                }
                else return(false);
            }
        }
    }

    // Return true if an agent implementing these syntactic description can send events
    // If syntactic == null, also return true
    public boolean doesThisSend (String[] syntactic) {
        if (syntactic == null) return(true);
        Class eventSender = null;
        Class requestedClass = null;
        try { eventSender = Class.forName("net.hivecell.hive.agent.EventSendingAgent"); }
        catch (Exception e) { 
            Debug.warning("Can't find the class net.hivecell.hive.agent.EventSendingAgent"); 
            return(false);
        }
        // Do any of the syntactic classes mentioned implement EventSendingAgent?
        nextclass: 
        for (int i=0; i < syntactic.length; i++) {
            try { requestedClass = Class.forName(syntactic[i]); }
            catch (Exception e) { 
                Debug.warning("Can't find the class "+syntactic[i]);
                continue nextclass;
            }
            if (eventSender.isAssignableFrom(requestedClass)) {
                //                Debug.notice("The class "+requestedClass+" can send");
                return(true);
            }
        }
        //        Debug.notice("The class "+requestedClass+" can NOT send");
        return(false);
    }


    // Return true if an agent implementing these syntactic description can receive events
    // If syntactic == null, also return true
    public boolean doesThisReceive (String[] syntactic) {
        if (syntactic == null) return(true);
        Class eventReceiver = null;
        Class requestedClass = null;
        try { eventReceiver = Class.forName("net.hivecell.hive.agent.EventReceivingAgent"); }
        catch (Exception e) { 
            Debug.warning("Can't find the class net.hivecell.hive.agent.EventReceivingAgent"); 
            return(false);
        }
        // Do any of the syntactic classes mentioned implement EventReceivingAgent?
        nextclass: 
        for (int i=0; i < syntactic.length; i++) {
            try { requestedClass = Class.forName(syntactic[i]); }
            catch (Exception e) { 
                Debug.warning("Can't find the class "+syntactic[i]);
                continue nextclass;
            }
            if (eventReceiver.isAssignableFrom(requestedClass)) {
                //                Debug.notice("The class "+requestedClass+" can receive");
                return(true);
            }
        }
        //        Debug.notice("The class "+requestedClass+" can NOT receive");
        return(false);
    }


    public boolean connectMeTo (String syntactic, String semantic) {
        String[] s1 = { syntactic };
        String[] s2 = { semantic };
        if(syntactic == null)
            s1 = null;
        if(semantic == null)
            s2 = null;
        return(connectMeTo(s1, s2));
    }

    public boolean connectToReceive (String syntactic, String semantic) {
        String[] s1 = { syntactic };
        String[] s2 = { semantic };
        if(syntactic == null)
            s1 = null;
        if(semantic == null)
            s2 = null;
        return(connectToReceive(s1, s2));
    }

    public boolean connectToSend (String syntactic, String semantic) {
        String[] s1 = { syntactic };
        String[] s2 = { semantic };
        if(syntactic == null)
            s1 = null;
        if(semantic == null)
            s2 = null;
        return(connectToSend(s1, s2));
    }

    // If we're not already connected to at least one agent of this
    // syntactic type, connect to one.  This makes connections from me
    // to agents (sending).  Return true if we are either already
    // connected to at least one of the type of agent given or if we
    // cant send or they can't receive, false we're not connected and
    // can't find one to connect to.

    public boolean connectToSend (String[] syntactic, String[] semantic) {
        boolean connectedToAll = true;
        Agent matchingagent = null;
        AgentSet ds = null;
        
        // If we can send to this sort of agent, and aren't already
        try {
            if (this.weSend && doesThisReceive(syntactic) && !sendingTo(syntactic, semantic)) {
                
                // OK, we haven't subscribed to one of these yet
                ds = agentCell.queryAgents(agent, syntactic, semantic);

                if (ds.size() != 0) {
                    matchingagent = (Agent)ds.pickOne();
                }
                if (matchingagent == null) {
                    Debug.notice("Couldn't find a receiver for us, but we want one");
                    connectedToAll = false;    // Couldn't find one
                }
                else {    // Got one, let's hook up
                    try { Debug.notice("Hooking "+agent.getName()+" up to "+matchingagent.getName()+" to send");} catch (Exception e) {}
                    ((EventReceivingAgentImpl)matchingagent).subscribeTo(agent);
                }
            }
        }        
        catch (Exception e) { 
            Debug.warning("Error hooking up agent: "+e);
            return(false);
        }
        
        return(connectedToAll);
    }


    // If we're not already connected to at least one agent of this
    // syntactic type, connect to one.  This makes connections to me
    // from the agents (receiving).  Return true if we are either
    // already connected to at least one of the type of agent given or
    // if we cant receive or they can't send, false we're not
    // connected and can't find one to connect to.

    public boolean connectToReceive (String[] syntactic, String[] semantic) {
        boolean connectedToAll = true;
        Agent matchingagent = null;
        AgentSet ds = null;
        
           // If we can receive from this sort of agent, and aren't connected already
        try {
            if (this.weReceive && doesThisSend(syntactic) && !listeningTo(syntactic, semantic)) {
                // OK, we haven't subscribed to one of these yet
                matchingagent = null;
                ds = agentCell.queryAgents(agent, syntactic, semantic);

                if (ds.size() != 0) {
                    matchingagent = (Agent)ds.pickOne();
                }
                if (matchingagent == null) {
                    Debug.notice("Couldn't find a sender for us, but we want one");
                    connectedToAll = false;    // Couldn't find one
                }
                else {    // Got one, let's hook up
                    try { Debug.notice("Hooking "+agent.getName()+" up to "+matchingagent.getName()+" to receive");} catch (Exception e) {}
                    if (matchingagent instanceof EventSendingAgent) {
                        try {
                        agent.connectTo((EventSendingAgent)matchingagent);
                        }
                        catch (Exception e) {
                            try {Debug.warning("Caught exception "+e+" while hooking up "+matchingagent.getName());}
                            catch (Exception e2) {}
                        }
                    }
                    else {
                        Debug.warning("We thought "+matchingagent+" could send, but it can't!");
                    }
                }
            }
        }
        catch (Exception e) { 
            Debug.warning("Error hooking up agent: "+e);
            return(false);
        }
        return(connectedToAll);
    }


    // If we're not already connected to at least one agent of this
    // syntactic type, connect to one.  This makes connections in one
    // or both directions, depending on what's possible given the
    // agent types.  Return true if we are either already connected to
    // at least one of the type of agent given (or if the connection
    // is impossible because one of us can't send or receive), false
    // we're not connected and can't find one to connect to.

    // Things to add: option for non-local server

    public boolean connectMeTo (String[] syntactic, String[] semantic) {
        boolean connectedSend, connectedReceive;
        connectedSend = connectToSend(syntactic, semantic);
        connectedReceive = connectToReceive(syntactic, semantic);
        return (connectedSend && connectedReceive);
    }
}
