// $Id: EventReceivingAgentImpl.java,v 1.2 2000/05/19 19:55:53 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.agent;

import net.hivecell.hive.Global;
import net.hivecell.hive.event.HiveEvent;
import net.hivecell.hive.event.RemoteEvent;
import net.hivecell.hive.cell.RemoteCell;
import net.hivecell.hive.cell.Cell;
import net.hivecell.hive.support.Debug;

import java.util.*;
import java.rmi.RemoteException;

/** Base class for agents that listen to events on one host.
 ** Handles finding the first event sending agent on the remote host,
 ** notifying the server of event() messages.
 **
 ** @version $Revision: 1.2 $
 ** @author Nelson Minar <nelson@media.mit.edu>
 **/

public abstract class EventReceivingAgentImpl 
extends AgentImpl 
implements EventReceivingAgent {

    /** Vector of EventSending agents we're subscribed to. We keep track of
     * these so we can politely desubscribe when we die.
     *
     * This is marked as transient so when we deserialize them from
     * session management, nothing will "goto wrong" due to remote
     * handles.  during session management, the shadow keeps track of
     * the connections so it will reconnect them -- the question is when
     * the agent is moved to another server
     **/
    protected transient Vector agentSubscriptions;

    public EventReceivingAgentImpl() 
    throws RemoteException {
	super();
    }
  
    public void doLocalSetup() throws AgentInitializationException {
        super.doLocalSetup();
	agentSubscriptions = new Vector();
    }

    /** Called whenever the agent receives an event. By default, just
     * notifies the server of the message. Override with agent's behaviour,
     * but remember to call this superclass method!
     **/
    public void notify( HiveEvent event ) {
	myCell.noteAgentMessage( event.getSource(), "notify", (Agent)event.getSource(), this );
    }

    /** This method is here for compatibility with Distributed Events, but
     * should never be invoked in Hive.
     **/
    public void notify( RemoteEvent ev ) {
	Debug.error( "Huh? notify(RemoteEvent) was called!" );
    }
  
    /** Override diePlease to also unsubscribe from any agents we were
     ** subscribed to.
     **/
    public void diePlease() {
	super.diePlease();
	this.disconnectFromAll();
    }

    /** Basic implementation of connectTo for event receivers.
     ** Just subscribe to events on the remote agent.
     **/
    public boolean connectTo( Agent otherAgent ) {
	boolean connected = this.subscribeTo( otherAgent );
	if( connected == true )
	    myCell.noteAgentConnect( this, this, otherAgent );
	return connected;
    }

    /** Disconnect from a particular agent. Checks that we are
     * currently subscribing to it, then sends it a removeListener
     * message and removes it from the subscription list.
     **/
    public void disconnectFrom( Agent otherAgent ) {
	// Check that we subscribe.
	if( agentSubscriptions.contains( otherAgent ) != true ) {
	    Debug.warning( "Ignoring request to unsubscribe from an agent we don't subscribe to, " + Global.shortString( otherAgent ) );
	    return;
	}

	// Actually unsubscribe from the agent.
	EventSendingAgent infoAgent = (EventSendingAgent)otherAgent;
	Debug.notice( "unsubscribing from " + Global.shortString( infoAgent ) );
	try { 
	    infoAgent.removeListener( this );
	} catch( RemoteException ex ) {
	    Debug.warning( "Attempt to unsubscribe failed, message was " + ex.getMessage() + ". If the other host is down, it's OK." );
	}

	// And remove it from our list.
	agentSubscriptions.removeElement( otherAgent );

	// And tell the server what we did.
	myCell.noteAgentMessage( this, "removeListener", this, otherAgent );
	myCell.noteAgentDisconnect( this, this, otherAgent );
    }

    /** Walk the list of subscriptions. **/
    public synchronized void disconnectFromAll() {
	if( agentSubscriptions != null ) {

	    // Clone to prevent the remove in disconnectFrom from affecting us.
	    Vector list = (Vector)agentSubscriptions.clone();     

	    // tell the server what is happening
	    myCell.noteAgentDisconnectFromAll( this, this );

	    Enumeration e = list.elements();
	    while (e.hasMoreElements())
		disconnectFrom( (Agent)e.nextElement() );

	}
    }

    /** Subscribe to an event source.
     * This is the nuts and bolts of the subscribing. Checks the type
     * of the remote agent.
     * @param vba The agent to subscribe to (should be EventSending)
     **/
    public boolean subscribeTo( Agent otherAgent ) {
	EventSendingAgent infoAgent = null;
	// Check the type of the agent we're subscribing to.
	if ( ! ( otherAgent instanceof EventSendingAgent ) ) {
	    Debug.error( "EventReceivingAgent was told to subscribe to a non-EventSending." );
	    return false;
	}
    
	infoAgent = (EventSendingAgent)otherAgent;

	// Make the connection.
	myCell.noteAgentMessage( this, "addListener", this, infoAgent );
	try {
	    infoAgent.addListener( this );
	} catch( RemoteException re ) {
	    Debug.critical( "Subscription failed ", re );
	    return false;
	}
    
	// And track the subscription.
	agentSubscriptions.addElement( infoAgent );

	return true;
    }

    /** Return all the agents I'm connected to
     */
    public Vector listAllIncomingConnections() {
	return agentSubscriptions;

    }

}










