package net.hivecell.hive.agent.cell;

import net.hivecell.hive.Global;
import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.agent.AgentImpl;
import net.hivecell.hive.agent.EventReceivingAgent;
import net.hivecell.hive.cell.RemoteCell;
import net.hivecell.hive.event.HiveEvent;
import net.hivecell.hive.event.UnknownEventException;
import net.hivecell.hive.shadow.EventPublisherShadow;
import net.hivecell.hive.shadow.ShadowNotFoundException;
import net.hivecell.hive.shadow.Signalable;
import net.hivecell.hive.shadow.cell.AgentMessagingShadow;
import net.hivecell.hive.shadow.cell.AgentMessagingShadowImpl;
import net.hivecell.hive.support.CellAddress;
import net.hivecell.hive.support.CellConnectException;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.support.async.AsyncCall; 
import net.hivecell.hive.support.async.AsyncResult; 

import java.lang.reflect.Method;
import java.rmi.RemoteException;

/**
 * this is the agent that will be doing the negotiation for the
 * agentmonitoringshadows for each cell.  this agent is simply out
 * here to find out what port the other cells are operating on, and to
 * advertise which port this cell is working on 
 *
 * @see net.hivecell.hive.shadow.cell.AgentMonitoringShadow
 */
public class AgentMessagingShadowAgentImpl
extends AgentImpl
implements AgentMessagingShadowAgent, Signalable {

    // the internal variables
    private transient AgentMessagingShadow agentMessagingShadow = null; // our hook back to the shadow

    private transient Method notifyMethod = null; // this is the notify method that we are "caching"

    
    // the initialization stuff

    /**
     * create the agent messaging shadow
     */
    public AgentMessagingShadowAgentImpl()
    throws RemoteException { 
	super();

    }

    /**
     * there is nothing to do in the setup -- it all has to happen in
     * the doBehavior since that will be called after the
     * AgentMessagingShadow is in the database 
     */
    public void doLocalSetup() { 
	try {
	    Class[] parameters = { (HiveEvent.class) };
	    notifyMethod = (EventReceivingAgent.class).getMethod( "notify", parameters );
	    Debug.notice( "caching the java reflection stuff for the event receiving agent" );

	} catch( Exception error ) {
	    // we should never have an exception here

	}

    }

    /**
     * first we get the hook down to its shadow, then this agent has
     * nothing really to do while it is alive except respond to the
     * one RMI call, getAgentMessagingPort.  besides that, we just sit
     * around and wait until we are killed.  what an existance. 
     *
     * this agent is usually only spawned from the
     * AgentMessagingShadow, so if we can't find one, then we should
     * sit around and wait for a bit.  perhaps Cell.java should be
     * modified so we can recieve notification when certain shadows
     * come alive on the cell, for timing issues such as this.
     */
    public void doBehavior() {
	try {
	    agentMessagingShadow = (AgentMessagingShadow)myCell.getShadow( (AgentMessagingShadowImpl.class).getName() );
	    agentMessagingShadow.addListener( this );
	    waitUntilDeath();
	    
	} catch( ShadowNotFoundException error ) {
	    Debug.error( "cannot find an agent messaging shadow to bind to" );
	
	}

    }

    /**
     * ignore this method as we only have it since we implement
     * signalable -- maybe sometime later we will do soemthing useful
     * with the shadow's signaling 
     *
     * @param source the agent messaging shadow
     */
    public void signal( EventPublisherShadow source ) { }

    /**
     * ignore this method as we only have it since we implement
     * signalable -- maybe sometime later we will do soemthing useful
     * with the shadow's signaling 
     *
     * @param source the agent messaging shadow
     * @param information the information being passed us
     */
    public void signal( EventPublisherShadow source, Object information ) { }


    // the remote methods
    
    /**
     * this is the method that other AgentMessagingShadowAgents will
     * be calling to figure out what port to use to send HiveEvents to
     * this cell.
     *
     * @param a the agent that is requesting this information 
     */
    public int getAgentMessagingPort( Agent a ) {
	Debug.notice( Global.shortString( a ) + " requested event port number" );
	return agentMessagingShadow.getEventListeningPort();
	
    }

    
    // methods for the AgentMessagingShadow
    
    /**
     * the AgentMessagingShadow can call this method with a new cell
     * address, and this agent will do its best to try to tell us what
     * port that cell is listening on for its cell events
     *
     * @param address the cell address of the cell we are trying to find the port for
     * @return the port number that we can connect to that cell on 
     * @exception CellConnectException this is thrown if we cannot even connect to the cell address specified
     * @exception UnknownAgentException if there is no AgentMessagingShadowAgent on the cell address to connect to
     */
    public int getCellAgentMessagingPort( CellAddress address ) 
    throws CellConnectException, UnknownAgentException {
	
	RemoteCell rc = address.getRemoteCell();
	try {
	    String[] agentMessagingShadowAgentString = { (AgentMessagingShadowAgent.class).getName() };
	    AgentMessagingShadowAgent agentMessagingShadowAgent = (AgentMessagingShadowAgent)rc.queryAgents( this, agentMessagingShadowAgentString, null ).pickOne();
	    if( agentMessagingShadowAgent == null )
		throw new UnknownAgentException( "there is no agent messaging shadow agent on " + address );
	    return agentMessagingShadowAgent.getAgentMessagingPort( this );

	} catch( ClassNotFoundException error ) {
	    // this just won't happen

	} catch( RemoteException error ) {
	    // if we get a problem talking to the remote cell to query
	    // agents, then we might as well have not just connected
	    // to the cell at all
	    Debug.error( "remote exception while talking to the remote cell at " + address );
	    throw new CellConnectException( error );

	}

	// we'll never get here
	return -1;

    }

    /**
     * deliver this event via RMI for us because we are having
     * problems doing it any other way -- there is no guarantee that
     * this will work.  this is a synchronous call as it is designed
     * only to be able to be called from the AgentMessaingShadow which
     * will force it to asynchonous
     *
     * @param agent the EventReceivingAgent that we are delivering the event to
     * @param event the event that we want to send 
     * @exception RemoteException if something goes wrong
     * @exception UnknownEventException is covered by the jini spec
     */
    public void deliverEvent( EventReceivingAgent agent, HiveEvent event ) 
    throws RemoteException, UnknownEventException {
	agent.notify( event );
	
    }


}
