// $Id: AgentMonitoringAgent.java,v 1.1 2000/05/18 22:00:51 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.agent.cell;

import java.rmi.RemoteException;
import java.util.Vector;

import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.agent.EventReceivingAgent;
import net.hivecell.hive.agent.EventSendingAgent;
import net.hivecell.hive.event.cell.CellEvent;
//import net.hivecell.hive.rdf.Description;
//import net.hivecell.hive.rdf.DescSet;
import net.hivecell.hive.description.Description;
import net.hivecell.hive.description.AgentSet;
import net.hivecell.hive.cell.RemoteCell;
import net.hivecell.hive.support.AgentGraph;
import net.hivecell.hive.support.CellAddress;
import net.hivecell.hive.support.SerializableImage;

/** 
 * Agent to watch for interesting events on the server.
 *
 * @see CellEvent
 * @author Nelson Minar <nelson@media.mit.edu>
 */
public interface AgentMonitoringAgent 
extends EventSendingAgent, CellAgent {

    /**
     * Return which cells that this agent is currently monitoring
     *
     * @return an array of the cell addresses that this agent is monitoring
     */
    public CellAddress[] getMonitoredCells( Object requester )
    throws RemoteException;

    /** 
     * Ask this agent if we are already monitoring this cell.
     *
     * @param requester the object requesting this information
     * @param isMonitoring the address that we want to know about
     * @return true if this agent is monitoring the cell, false otherwise 
     */
    public boolean isMonitoringCell( Object requester, CellAddress isMonitoring )
    throws RemoteException;

    /**
     * Ask this agent to start monitoring this cell address.  It is a
     * safe thing to do to ask this agent just to monitor this cell
     * whenever we start.  This agent will be keeping a data structure
     * around recording who asked for what cells.
     *
     * @param requester the object requesting this agent monitor another address
     * @param toMonitor the address of the cell this agent should monitor
     * @exception UnableToMonitorCellException when something goes wrong in the cell monitoring 
     */
    public void monitorCell( Object requester, CellAddress toMonitor )
    throws RemoteException, UnableToMonitorCellException;

    /**
     * Ask this agent to stop monitoring a cell.  This may or may not
     * cause the ama to actually unsubscribe itself from the event
     * stream -- it is just that if nobody at all is interested in
     * this stream, then it will unsubscribe itself.  For politeness,
     * an agent should ask the ama to start monitoring a cell when it
     * needs to, and to stop monitoring a cell when it needs to
     *
     * @param requester the object requesting this agent stop monitoring another address
     * @param toStopMonitoring the address to stop monitoring
     */
    public void stopMonitoringCell( Object requester, CellAddress toStopMonitoring )
    throws RemoteException;

    /**
     * Request that the event log be changed in length
     *
     * @param length the new length that we wish the log to be
     * @return the new length of the log -- not guaranteed to be the length requested
     */
    public int setEventLogLength( Object requester, int length )
    throws RemoteException;

    public int getEventLogLength( Object requester )
    throws RemoteException;


    /**
     * Get the n latest events from the event log.  It is undefined
     * what will be returned if the requested amount is greater than
     * the length of the event log 
     * 
     * @param requester the object requesting the information
     * @param n the amount of events that we want from the log
     * @return an array of length n with the newest events toward the 0 end
     */
    public CellEvent[] getEventLog( Object requester, int n )
    throws RemoteException;

    /**
     * A method similar to the query agents in the RemoteCell
     * interface, except this also takes an array of cell addresses to
     * make the query across.  A null in any of the array fields is
     * treated as a wildcard.
     *
     * @param requester the object that is requesting this information
     * @param addresses the array of addresses to do the query across
     * @param syntatic the syntatic description to query on
     * @param semantic the semantic description to query on
     * @return a desc set of all the agents that match this query
     * @exception ClassNotFoundException when the syntatic class we are querying doesn't exist */
    public AgentSet queryAgents( Object requester, CellAddress[] addresses, String[] syntatic, String[] semantic )
    throws RemoteException, ClassNotFoundException;

    public void addListener( EventReceivingAgent subscriber )
    throws RemoteException;

    public void addListener( EventReceivingAgent subscriber, CellAddress[] address )
    throws RemoteException;

    public void addListener( EventReceivingAgent subscriber, CellAddress[] address, long eventMask )
    throws RemoteException;

//      /**
//       * Get the agent graph of the agents on these servers.
//       *
//       * @param requester the object requesting this information
//       * @param addresses the addersses to get this information from
//       * @return a graph of the agents on these servers
//       */
//      public AgentGraph listAllConnections( Object requester, CellAddress[] addresses ) 
//      throws RemoteException;
    
    /**
     * List all the agents that are connected on the "outgoing" connections from these agents
     * 
     * @param requester the object that is requesting this information
     * @param agent the agent that we want this information about
     * @return a desc set of the agents that are connected on the outgoing connections
     */
    public AgentSet listOutgoingAgentConnections( Object requester, Agent agent ) 
    throws UnknownAgentException, RemoteException;
  
    /** 
     * List all the agents that are connected on the "incoming" connections from these agents
     * 
     * @param requester the object that is requesting this information
     * @param agent the agent that we want this information about
     * @return a desc set of the agents that are connected on the incoming connections
     */
    public AgentSet listIncomingAgentConnections( Object requester, Agent agent )
    throws UnknownAgentException, RemoteException;

    /** 
     * Get the cached information as to where this agent is
     * 
     * @param requester the object asking for this information
     * @param agent we want the information for
     * @return the cell address of the agent
     */
    public CellAddress getCellAddress( Object requester, Agent agent ) 
    throws UnknownAgentException, RemoteException;

    /** 
     * Get the cached information as to the agent's server's remote server handle
     * 
     * @param requester the object asking for this information
     * @param agent the agent we want the information for
     * @return the remote server handle for this agent's server
     */
    public RemoteCell getRemoteCell( Object requester, Agent agent )
    throws UnknownAgentException, RemoteException;

    /**
     * Get the cached description of this agent
     *
     * @param requester the object asking for this information
     * @param agent the agent we want the information for
     * @return the description of the agent 
     */
    public Description getDescription( Object requester, Agent agent )
    throws UnknownAgentException, RemoteException;

}


