// $Id: AgentMonitoringAgentImpl.java,v 1.2 2000/05/19 19:55:55 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.agent.cell;

import net.hivecell.hive.Global;
import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.agent.AgentImpl;
import net.hivecell.hive.agent.EventReceivingAgent;
import net.hivecell.hive.agent.EventSendingAgent;
import net.hivecell.hive.agent.EventSendingHelper;
import net.hivecell.hive.agent.GenericSignalPassthroughAgentImpl;
import net.hivecell.hive.agent.AgentInitializationException;
import net.hivecell.hive.event.HiveEvent;
import net.hivecell.hive.event.RemoteEvent;
import net.hivecell.hive.event.cell.*;
import net.hivecell.hive.description.Description;
import net.hivecell.hive.description.AgentSet;
import net.hivecell.hive.cell.AgentDescriptionPair;
import net.hivecell.hive.cell.QueryAgentsHelper;
import net.hivecell.hive.cell.RemoteCell;
import net.hivecell.hive.cell.Cell;
import net.hivecell.hive.shadow.EventPublisherShadow;
import net.hivecell.hive.shadow.cell.CellEventMonitor;
import net.hivecell.hive.support.AgentGraph;
import net.hivecell.hive.support.CellAddress;
import net.hivecell.hive.support.CellConnectException;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.support.Null;
import net.hivecell.hive.support.SerializableImage;
import net.hivecell.hive.support.graph.NotInGraphException;

import net.hivecell.hive.agent.EventMaskWrapper;

import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

public class AgentMonitoringAgentImpl 
extends GenericSignalPassthroughAgentImpl 
implements AgentMonitoringAgent, EventReceivingAgent {

    /**
     * this is just the object we are synchronzing on whenever trying
     * to read or change the state 
     */
    protected Object dataLockingObject = new Integer( -1 );


    /** 
     * vector of eventsending agents we're subscribed to. we keep
     * track of these so we can politely desubscribe when we die.  
     *
     * this needs to be looked at more carefully in case we are moved,
     * we need to deal witht his better 
     */
    private transient Vector agentSubscriptions;

    
    /**
     * these are the cells that we are monitoring -- these are
     * actually to be a mapping between cell addresses to vectors of
     * agents which are asking to monitor 
     */
    private Hashtable cellsMonitored;

    /**
     * these are cells that we are in the process of monitoring
     */
    private Vector beginningMonitor;

    /**
     * a mapping from cell addresses to remote cells so we don't have
     * to make the remote call every single time
     */
    private Hashtable cellAddressToRemoteCell;

    /**
     * a mappging from cell addresses to the agent monitoring agent on
     * that cell 
     */
    private Hashtable cellAddressToAgentMonitoringAgent;

    /**
     * these are agents that we "know" about, these agents may or may
     * not be in the other cache information (but if they are in one
     * of the cache information, they are in all of them).  if an
     * agent is in here, but not in the rest of the data objects, then
     * that means that we got a remote exception or something when
     * trying to query for more information.  
     */
    private Vector knownAgents;

    /**
     * a hastable mapping agents to cell addresses 
     */
    private Hashtable agentsToCellAddresses;

    /**
     * a hashtable mapping cell addresses to vectors containsing
     * agents 
     */
    private Hashtable cellAddressesToAgents;

    /**
     * a hashtable mapping agents to descriptions
     */
    private Hashtable agentsToDescriptions;

    /**
     * a hashtable mapping agents to vectors of agents which are
     * outgoing 
     */
    private Hashtable agentsToOutgoing;

    /**
     * a hashtable mapping agents to vectors of agents which are
     * incoming 
     */
    private Hashtable agentsToIncoming;

    /**
     * a hashtable mapping event mask wrappers which are subscribing
     * to an array of cell addresses that they are subscribing to.  an
     * agent can be in here, but map to a null array which means that
     * all events to go to it.  its different from being mapped to an
     * array with 0 elements -- that means that it is subscribed, but
     * receives nothing (don't know why you would want to do thta 
     */
    private Hashtable subscriptions;

    /**
     * mapping between agents and their event mask wrappers
     */
    private Hashtable agentsToEventMaskWrapper;


    /**
     * the log of events that we are sending through the system -- by
     * default we are specifing an eventLog of 64.  this can be set
     * and changed */
    private transient CellEvent[] eventLog; 
    int eventLogFront = 0, eventLogEnd = 0;

    /**
     * get the current length of the event log
     *
     * @return the current length of the event log
     */
    public int getEventLogLength( Object requester ) {
	synchronized( dataLockingObject ) {
	    return eventLog.length;

	}

    }

    /**
     * set the length of the event log to whatever is specified --
     * thsi will attempt to keep as much information as we had before,
     * punt the oldest information.  it will also ignore any request
     * to set the cell log to a negative amount, but setting the
     * length to 0 turns off all logging
     *
     * @param requester the object requesting this agent stop monitoring another address
     * @param length the length that we are going to set the log to 
     * @return the previous length of the log
     */
    public int setEventLogLength( Object requester, int length ) {
	synchronized( dataLockingObject ) {
	    int oldLength = getEventLogLength( this );
	    if( length < 0 )
		return oldLength;
	    
	    CellEvent[] newLog = new CellEvent[length];
	    
	    int startPosition = eventLogFront;
	    if( length < oldLength )
		startPosition = ( eventLogFront + oldLength - length ) % oldLength;
	    if( startPosition > eventLogEnd ) {
		System.arraycopy( eventLog, startPosition, newLog, 0, oldLength - startPosition );
		System.arraycopy( eventLog, 0, newLog, oldLength - startPosition + 1, eventLogEnd );
		eventLogEnd = oldLength - startPosition + eventLogEnd;
		
	    } else {
		System.arraycopy( eventLog, startPosition, newLog, 0, eventLogEnd - startPosition );
		eventLogEnd = eventLogEnd - startPosition;
		
	    }
	    eventLog = newLog;
	    eventLogFront = 0;
	    
	    return oldLength;

	}

    }


    public AgentMonitoringAgentImpl() 
    throws RemoteException {
	super();

	// these are created here because they are not transient
	synchronized( dataLockingObject ) {
	    cellsMonitored = new Hashtable();
	    beginningMonitor = new Vector();
	    cellAddressToRemoteCell = new Hashtable();
	    cellAddressToAgentMonitoringAgent = new Hashtable();
	    knownAgents = new Vector();
	    agentsToCellAddresses = new Hashtable();
	    cellAddressesToAgents = new Hashtable();
	    agentsToDescriptions = new Hashtable();
	    agentsToOutgoing = new Hashtable();
	    agentsToIncoming = new Hashtable();
	    subscriptions = new Hashtable();
	    agentsToEventMaskWrapper = new Hashtable();

	}

    }

    /**
     * setup this agent to be ready to go.  setup all our data
     * structures, and set ourselves into a position to repeat the
     * evnts from our CellEventMonitor 
     */
    public void doLocalSetup() throws AgentInitializationException  {
	// setup our datastructures
	synchronized( dataLockingObject ) {
	    agentSubscriptions = new Vector();

	    eventLog = new CellEvent[64];
	    eventLogFront = 0; eventLogEnd = 0;

	    // get the population of our cell so we can get teh cache
	    // right from the beginning
	    CellAddress localAddress = myCell.getAddress();
	    cellAddressToRemoteCell.put( localAddress, myCell );
	    cellAddressToAgentMonitoringAgent.put( localAddress, this );
	    
	    cacheAgentsFrom( getCell() );
	    Vector monitors = new Vector();
	    monitors.addElement( this );
	    cellsMonitored.put( myCell.getAddress(), monitors ); // put our cell in the monitored list

	    try {
		Vector localAgents = new Vector();
		cellAddressesToAgents.put( localAddress, localAgents );
		AgentSet agents = myCell.queryAgents( this, null, null );
		Enumeration e = agents.elements();
		while( e.hasMoreElements() ) {
		    Agent a = (Agent)e.nextElement();
		    if( !cacheAgent( a ) )
			Debug.error( "problem caching " + a );
		    
		}

	    } catch( ClassNotFoundException error ) {
		// this won't get thrown

	    }
	    
	}

	// set ourselves up to repeat events from the cell event monitor
  	setShadowType( (CellEventMonitor.class).getName() ); 

	super.doLocalSetup();

    }

    /**
     * add an event to the event log -- this moves the front and back
     * pointers appropriately
     * 
     * @param event the event that we want to add to the log 
     */
    private void addToEventLog( CellEvent event ) {
	synchronized( dataLockingObject ) {
	    eventLog[eventLogEnd++] = event;
	    eventLogEnd = eventLogEnd % eventLog.length;
	    if( eventLogEnd == eventLogFront )
		eventLogFront++;

	}
	
    }

    /**
     * retrieve a portion of the event log
     *
     * @param requester the object requesting the event log
     * @param amount signify that you wish the last amount of events
     * @return an array of CellEvents that are amount or shorter
     */
    public CellEvent[] getEventLog( Object requester, int amount ) {
	CellEvent[] returning = null;

	synchronized( dataLockingObject ) {
	    // calculate the pointer positions inside the eventLog array
	    int length = amount;
	    int logSize = eventLogEnd - eventLogFront;
	    if( logSize < 0 )
		logSize = eventLogEnd + eventLog.length - eventLogFront;
	    if( logSize < amount )
		length = logSize;
	    returning = new CellEvent[length];
	    
	    // copy the eventLog into the new array for returning
	    if( eventLogFront - length < 0 ) {
		System.arraycopy( eventLog, eventLog.length + eventLogFront - length, returning, 0, Math.abs( eventLogFront - length ) );
		System.arraycopy( eventLog, 0, returning, Math.abs( eventLogFront - length ) + 1, eventLogFront );
		
	    } else
		System.arraycopy( eventLog, eventLogFront - length, returning, 0, length );

	}

	return returning;

    }

    /**
     * get a list of all the cells we are monitoring
     *
     * @param requester the object requesting the information
     * @return an array of cell addresses
     */
    public CellAddress[] getMonitoredCells( Object requester ) {
	CellAddress[] toReturn = null;
	synchronized( dataLockingObject ) {
	    toReturn = new CellAddress[cellsMonitored.size()];
	    Enumeration e = cellsMonitored.keys();
	    int count = 0;
	    while( e.hasMoreElements() ) 
		toReturn[count++] = (CellAddress)e.nextElement();

	}
	return toReturn;

    }

    /**
     * return true if we are monitoring this cell
     *
     * @param requester the object requesting the informatino
     * @param isMonitoring the cell address we want to test
     */
    public boolean isMonitoringCell( Object requester, CellAddress isMonitoring ) {
	synchronized( dataLockingObject ) {
	    return cellsMonitored.containsKey( isMonitoring );

	}

    }

    /**
     * request this agent start monitoring another cell
     *
     * @param requester the object requesting that we start monitoring
     * @param toMonitor the address that we want to start monitoring
     * @exception UnableToMonitorCellException something went wrong, and we are -not- monitoring the cell
     */
    public void monitorCell( Object requester, CellAddress toMonitor )
    throws UnableToMonitorCellException {
	// ok - we want to monitor another cell - if we are already
	// attempting to monitor this cell, then we need ot just wait,
	// if we are not attempting to monitor, then just add ourself
	// to a list saying that we are going to
	synchronized( dataLockingObject ) {
	    if( cellsMonitored.containsKey( toMonitor ) ) {
		((Vector)cellsMonitored.get( toMonitor )).addElement( requester );
		return;

	    }

	    if( !beginningMonitor.contains( toMonitor ) )
		beginningMonitor.addElement( toMonitor );

	    else {
		while( beginningMonitor.contains( toMonitor ) ) {
		       try {
			   dataLockingObject.wait();
			   
		       } catch( InterruptedException error ) {
			   
		       }

		}
		return;

	    }

	}

	// ok - now lets attempt to make the connection
	try {
	    RemoteCell remoteCell = toMonitor.getRemoteCell();
	    cellAddressToRemoteCell.put( toMonitor, remoteCell );

	    // now lets find the AgentMonitoringAgentImpl, subscribe
	    // to it, lock the data structure, then add all the agents
	    // in the AgentSet
	    AgentSet agents = null;
	    try {
		agents = remoteCell.queryAgents( this, null, null );
	    } catch( ClassNotFoundException error ) {
		// this is never going ot be thrown

	    }
	    AgentSet otherAgentMonitoringAgentSet = agents.querySet( (AgentMonitoringAgent.class).getName(), null );
	    AgentMonitoringAgent otherAgentMonitoringAgent = (AgentMonitoringAgent)otherAgentMonitoringAgentSet.pickOne();
	    cellAddressToAgentMonitoringAgent.put( toMonitor, otherAgentMonitoringAgent );
	    if( otherAgentMonitoringAgent != null ) {
		if( !subscribeTo( otherAgentMonitoringAgent ) )
		    throw new RemoteException();

	    } else
		throw new UnableToMonitorCellException( "no agent monitoring agent on " + toMonitor );

	    Enumeration e = agents.elements();
	    synchronized( dataLockingObject ) {
		while( e.hasMoreElements() ) {
		    Agent agent = (Agent)e.nextElement();
		    if( !cacheAgent( agent ) )
			Debug.error( "problem caching " + agent );

		}
		beginningMonitor.removeElement( toMonitor );
	      
		Vector monitors = new Vector();
		monitors.addElement( toMonitor );
		cellsMonitored.put( toMonitor, monitors );

		dataLockingObject.notifyAll();

	    }
		
	} catch( CellConnectException error ) {
	    throw new UnableToMonitorCellException( "problems contacting " + toMonitor, error );

	} catch( RemoteException error ) {
	    synchronized( dataLockingObject ) {
		cellAddressToRemoteCell.remove( toMonitor );
		cellAddressToAgentMonitoringAgent.remove( toMonitor );
		
	    }
	    throw new UnableToMonitorCellException( "problems making calls to " + toMonitor, error );

	} catch( Exception error ) {
	    Debug.error( error );

	}

    }

    /**
     * stop monitoring a cell
     *
     * @param requester the object that is asking to stop requesting a cell
     * @param toStopMonitoring the cell address that we want to stop monitoring
     */
    public void stopMonitoringCell( Object requester, CellAddress toStopMonitoring ) {
	AgentMonitoringAgent otherAgentMonitoringAgent = null;
	synchronized( dataLockingObject ) {
	    if( cellsMonitored.containsKey( toStopMonitoring ) ) {
		((Vector)cellsMonitored.get( toStopMonitoring )).removeElement( requester );
		if( ((Vector)cellsMonitored.get( toStopMonitoring )).size() != 0 )
		    return; // there are other people monitoring this cell

	    } else
		return; // we were never monitoring this cell

	    // get the handle to the agent monitoring agent on the cell that we want to stop monitoring
	    otherAgentMonitoringAgent = (AgentMonitoringAgent)cellAddressToAgentMonitoringAgent.get( toStopMonitoring );
	    cellAddressToAgentMonitoringAgent.remove( toStopMonitoring );
	    
	    // ok - remove all references to this cell from our datastructures
	    cellsMonitored.remove( toStopMonitoring );
	    cellAddressToRemoteCell.remove( toStopMonitoring );
	    Vector agents = (Vector)cellAddressesToAgents.get( toStopMonitoring );
	    cellAddressesToAgents.remove( toStopMonitoring );
	    Enumeration e = agents.elements();
	    while( e.hasMoreElements() ) {
		Agent agent = (Agent)e.nextElement();
		knownAgents.removeElement( agent );
		agentsToCellAddresses.remove( agent );
		agentsToDescriptions.remove( agent );
		agentsToOutgoing.remove( agent );
		agentsToIncoming.remove( agent );
		
	    }

	}
	
	// now actually do the unsubscribing
	disconnectFrom( otherAgentMonitoringAgent );
	
    }

    /**
     * A method similar to the query agents in the RemoteCell
     * interface, except this also takes an array of cell addresses to
     * make the query across.  A null in any of the array fields is
     * treated as a wildcard.
     *
     * @param requester the object that is requesting this information
     * @param addresses the array of addresses to do the query across
     * @param syntatic the syntatic description to query on
     * @param semantic the semantic description to query on
     * @return an agent set of all the agents that match this query
     * @exception ClassNotFoundException when the syntatic class we are querying doesn't exist */
    public AgentSet queryAgents( Object requester, CellAddress[] addresses, String[] syntatic, String[] semantic )
    throws RemoteException, ClassNotFoundException {
	Vector agentsToAdd = new Vector();
	synchronized( dataLockingObject ) {
	    if( addresses == null ) {
		Enumeration cells = cellAddressesToAgents.keys();
		while( cells.hasMoreElements() ) {
		    Vector agentList = (Vector)cellAddressesToAgents.get( (CellAddress)cells.nextElement() );
		    Enumeration agents = agentList.elements();
		    while( agents.hasMoreElements() ) {
			Agent agent = (Agent)agents.nextElement();
			if( !agentsToAdd.contains( agent ) )
			    agentsToAdd.addElement( agent );
			
		    }
		    
		}
		
	    } else
		for( int count=0;count<addresses.length;count++ ) {
		    Vector agentList = (Vector)cellAddressesToAgents.get( addresses[count] );
		    Enumeration agents = agentList.elements();
		    while( agents.hasMoreElements() ) {
			Agent agent = (Agent)agents.nextElement();
			if( !agentsToAdd.contains( agent ) )
			    agentsToAdd.addElement( agent );

		    }

		}
		     
	}

	AgentSet set = new AgentSet( agentsToAdd );
	return set.querySet( syntatic, semantic );
    }


    /**
     * List all the agents that are connected on the "outgoing" connections from these agents
     * 
     * @param requester the object that is requesting this information
     * @param agent the agent that we want this information about
     * @return a desc set of the agents that are connected on the outgoing connections
     */
    public AgentSet listOutgoingAgentConnections( Object requester, Agent agent ) 
    throws UnknownAgentException {
	synchronized( dataLockingObject ) {
	    if( knownAgents.contains( agent ) ) {
		Vector outgoingEdges = (Vector)agentsToOutgoing.get( agent );
		if( outgoingEdges == null )
		    return null;

		return new AgentSet( outgoingEdges );

	    } else
		throw new UnknownAgentException();

	}

    }
  
    /** 
     * List all the agents that are connected on the "incoming" connections from these agents
     * 
     * @param requester the object that is requesting this information
     * @param agent the agent that we want this information about
     * @return a desc set of the agents that are connected on the incoming connections
     */
    public AgentSet listIncomingAgentConnections( Object requester, Agent agent )
    throws UnknownAgentException {
	synchronized( dataLockingObject ) {
	    if( knownAgents.contains( agent ) ) {
		Vector incomingEdges = (Vector)agentsToIncoming.get( agent );
		if( incomingEdges == null )
		    return null;

		return new AgentSet( incomingEdges );

	    } else
		throw new UnknownAgentException();

	}

    }

    /** 
     * Get the cached information as to where this agent is
     * 
     * @param agent we want the information for
     * @return the cell address of the agent -- this is possible to be
     * null!  it just means that we don't know where the agent is
     * from, but we are awknowledging its existnace 
     * @exception if we don't even know about this agent 
     */
    public CellAddress getCellAddress( Object requester, Agent agent ) 
    throws UnknownAgentException {
	synchronized( dataLockingObject ) {
	    if( knownAgents.contains( agent ) )
		return (CellAddress)agentsToCellAddresses.get( agent );
	    throw new UnknownAgentException();

	}

    }

    /** 
     * Get the cached information as to the agent's server's remote server handle
     * 
     * @param agent the agent we want the information for
     * @return the remote server handle for this agent's server
     */
    public RemoteCell getRemoteCell( Object requester, Agent agent )
    throws UnknownAgentException {
	synchronized( dataLockingObject ) {
	    CellAddress address = getCellAddress( requester, agent );
	    if( address == null )
		return null;
	    return (RemoteCell)cellAddressToRemoteCell.get( address );

	}

    }

    /**
     * Get the cached description of this agent
     *
     * @param agent the agent we want the information for
     * @return the description of the agent 
     */
    public Description getDescription( Object requester, Agent agent )
    throws UnknownAgentException {
	synchronized( dataLockingObject ) {
	    if( knownAgents.contains( agent ) )
		return (Description)agentsToDescriptions.get( agent );
	    throw new UnknownAgentException();

	}

    }
    
    /**
     * add listener to all the events. period.
     *
     * @param subscriber the agent subscribing
     */
    public void addListener( EventReceivingAgent subscriber ) {
	addListener( subscriber, (CellAddress[])null );

    }

    /**
     * add ourselves as a listener to this agent monitoring agent --
     * this registers only for cell address about which we are
     * putting.  however, if you ask for a cell address taht we are
     * not monitoring, then nothing happens -- we will be subscribed,
     * but we don't get events
     *
     * @param subscriber the person subscribing
     * @param address the addresses that we want to be listening to -- null is a wildcard
     */
    public void addListener( EventReceivingAgent subscriber, CellAddress[] address ) {
	addListener( subscriber, address, HiveEvent.ALL_EVENTS );

    }

    /**
     * add ourselves as a listener to this agent montiroing agent with a specified event mask
     *
     * @param subscriber the agent subscribing
     * @param address the addresses thatw ewant to listen to
     * @param eventMask the event mask that we are registering for
     */
    public void addListener( EventReceivingAgent subscriber, CellAddress[] address, long eventMask ) {
	while( eventSendingHelper == null )
	    try {
		Debug.notice( "Waiting on EventSendingHelper creation" );
		synchronized( this ) {
		    wait();
		}

	    } catch( InterruptedException error ) { }

	// we have to maintina a different data structure here
	synchronized( dataLockingObject ) {
	    // perhaps i should just define .equals on the emw to
	    // mathc if we give the appropriate agen?
	    EventMaskWrapper emw = new EventMaskWrapper( subscriber, eventMask );
	    agentsToEventMaskWrapper.put( subscriber, emw );
	    if( address == null )
		subscriptions.put( emw, new Null() );

	    else
		subscriptions.put( emw, address );

	}

    }

    /**
     * remove the listener from all our data structures
     *
     * @param sender the sender who wants to be removed
     */
    public void removeListener( EventReceivingAgent sender ) {
	synchronized( dataLockingObject ) {
	    EventMaskWrapper emw = (EventMaskWrapper)agentsToEventMaskWrapper.remove( sender );
	    if( emw != null )
		subscriptions.remove( emw );

	}

    }

    /**
     * the signal method that is going to be called by the
     * cellEventMontior -- we just need to hack this so we can use it
     * to keep track of which agents are in the system, and for the
     * evnet log
     *
     * @param source the source of the evnets -- its going to be the CellEventMonitor
     * @param information the cell event that we are going to be asked to broadcast out 
     */
    public void signal( EventPublisherShadow source, Object information ) {
	if( !( information instanceof CellEvent ) )
	    return;

	// need to reconstruct the cell event so this agent is the
	// source of it, and not the cell event monitor.  we also have
	// to set the addresses of the from and to agents.  this was
	// not set in the cell event monitor, because the cell event
	// montior can't do anything with the network
	CellEvent cellEvent = null;
	if( information instanceof TwoAgentCellEvent ) {
	    TwoAgentCellEvent info = (TwoAgentCellEvent)information;
	    CellAddress sourceAddress = null;
	    CellAddress destinationAddress = null;
	    try {
		sourceAddress = getCellAddress( this, info.getSourceAgent() );
		destinationAddress = getCellAddress( this, info.getDestinationAgent() );
		
	    } catch( UnknownAgentException error ) {
		try {
		    if( sourceAddress == null )
			sourceAddress = info.getSourceAgent().getCell().getAddress();

		    if( destinationAddress == null )
			destinationAddress = info.getDestinationAgent().getCell().getAddress();

		} catch( RemoteException internalError ) {
		    // i'm not sure what to do if this happens!!!!!!!!!!!!!!!!!!!!!!!!
		    Debug.error( internalError );

		}

	    } finally {
		cellEvent = (CellEvent)info.createNewEvent( this, eventSendingHelper.getEventCount( this ), sourceAddress, destinationAddress );

	    }
	    		
	} else
	    cellEvent = ((CellEvent)information).createNewEvent( this, eventSendingHelper.getEventCount( this ) );

	handleCellEvent( cellEvent );

	// broadcast the information
	cellEvent.addPassThrough( myCell.getAddress() );
	broadcast( cellEvent );

    }

    public void notify( RemoteEvent theEvent ) {
	Debug.error( "Huh? notify(RemoteEvent) was called!" );

    }

    /**
     * called when we recieve events from another
     * AgentMonitoringAgent.  we have to make sure are updating the
     * cache if this happens.  
     *
     * the first thing we have to do is to make sure that the event is
     * coming from an agent that we are supposed to be getting events
     * from.  if it is, then we need to further examine the event and
     * update our cache.
     *
     * we do not repeat these events - they die one away from their
     * source
     *
     * @param event the hive event that is sent to us 
     */
    public void notify( HiveEvent event ) {
	// first make sure we have a cell event, and then make sure it
	// hasn't come thorugh this cell already
	if( !( event instanceof CellEvent ) || ((CellEvent)event).passedThrough( myCell.getAddress() ) )
	    return;

	// oh baby -- this is the gate for the cell event to get
	// through.  ok - first we have to check to see if the
	// celladdress this came from is part of the montoring cells
	// list.  if it is, let it fall through.  if it is not, then
	// we go through the list of the beginning to monitor.  if we
	// are there, we go into wait()
	CellEvent cellEvent = (CellEvent)event;
	synchronized( dataLockingObject ) {
	    boolean found = false;
	    Enumeration cells = cellsMonitored.keys();
	    while( cells.hasMoreElements() )
		if( cellEvent.matchesCellAddress( (CellAddress)cells.nextElement() ) ) {
		    found = true;
		    break;

		}

	    if( !found ) {
		CellAddress waitingOn = null;
		cells = beginningMonitor.elements();
		while( cells.hasMoreElements() ) {
		    waitingOn = (CellAddress)cells.nextElement();
		    if( cellEvent.matchesCellAddress( waitingOn ) ) {
			found = true;
			break;

		    }
		    
		}
		 
		if( !found ) {
		    Debug.warning( "got an event from a host we are not monitoring" );
		    return;

		}
		    
		while( found && beginningMonitor.contains( waitingOn ) )
		    try {
			dataLockingObject.wait();

		    } catch( InterruptedException error ) {

		    }

	    }

	}

	handleCellEvent( cellEvent );

	cellEvent.addPassThrough( myCell.getAddress() );
	broadcast( cellEvent );

    }

    /**
     * take a cell event, and broadcasts it out to the appropriate
     * people -- it checks to get a list of agent s which are
     * interested for a cell event from this particular address, and
     * then it narrows that list soem more using the netmask, and then
     * it broadcasts to those particular agents
     *
     * @param cellEvent the cell event that we want to broadcast out 
     */
    public void broadcast( CellEvent cellEvent ) {
	Vector sendingTo = new Vector();
  	synchronized( dataLockingObject ) {
	    Enumeration e = subscriptions.keys();
	scanningEventMaskWrapper:
	    while( e.hasMoreElements() ) {
		EventMaskWrapper emw = (EventMaskWrapper)e.nextElement();
		Object addressesObject = subscriptions.get( emw );
		if( addressesObject instanceof Null )
		    sendingTo.addElement( emw );

		else {
		    CellAddress[] addresses = (CellAddress[])subscriptions.get( emw );
		    for( int count=0;count<addresses.length;count++ ) 
			if( cellEvent.matchesCellAddress( addresses[count] ) ) {
			    sendingTo.addElement( emw );
			    continue scanningEventMaskWrapper;
			    
			}

		}

	    }

	}
	
	EventMaskWrapper[] array = new EventMaskWrapper[sendingTo.size()];
	sendingTo.copyInto( array );
	eventSendingHelper.broadcastEventTo( cellEvent, (EventReceivingAgent[])null, array );

    }

    /**
     * contacts a cell, queries for all its agents, and then attempts
     * to put them all into the cache
     *
     * @param address the address of the cell that we are quering
     * @return the number of agents we got into the cache
     * @exception CellConnectException if we cannot even contact the cell
     * @exception RemoteException if anything went wrong with contacting the cell
     */
    protected int cacheAgentsFrom( CellAddress address ) 
    throws CellConnectException, RemoteException {
	return cacheAgentsFrom( address.getRemoteCell() );

    }

    /**
     * contacts a cell, queries for all its agents, and then attempts
     * to put them all into the cache
     *
     * @param remoteCell the cell that we are contacting for the information
     * @return the number of agents we got into the cache
     */
    protected int cacheAgentsFrom( RemoteCell remoteCell ) {
	AgentSet agents = null;
	try {
	    agents = remoteCell.queryAgents( this, null, null );

	} catch( ClassNotFoundException error ) {
	    // this won't happen

	} catch( RemoteException error ) {
	    Debug.error( "remote exception while trying to query agents from " + remoteCell );
	    return 0;

	}

	synchronized( dataLockingObject ) {
	    Enumeration e = agents.elements();
	    int count = 0;
	    while( e.hasMoreElements() ) {
		Agent a = (Agent)e.nextElement();
		if( !cacheAgent( a ) )
		    Debug.warning( "problem putting " + a + " in the cache" );
		
		else
		    count++;
		
	    }
	    return count;

	}

    }

    /**
     * does a best effort at caching an agent into our data structures.
     *
     * @param a the agent that we want to cache
     * @return true if we could cache it, false otherwise 
     */
    protected boolean cacheAgent( Agent a ) {
	try {
	    return cacheAgent( a, a.getCell().getAddress() );

	} catch( RemoteException error ) {
	    Debug.error( "remote exception while trying to get the cell address of " + a + ".  not caching." );
	    return false;

	}

    }

    /**
     * does a best effort at caching an agent into our data
     * structures.  this is exactly the same as the other method by
     * the same name, except in this one we alreay have the cell
     * address that we want to use.
     *
     * this assumes that cellAddressesToAgents has been prepped for
     * this cell address by calling monitor cell with that address
     * already.
     *
     * @param a the agent that we want to cache
     * @param address the address this agent is at
     * @return true if we could cache it, false otherwise 
     */
    protected boolean cacheAgent( Agent a, CellAddress address ) {
	try {
	    Vector incoming = a.listAllIncomingConnections();
	    if( incoming == null )
		incoming = new Vector();
	    Vector outgoing = a.listAllOutgoingConnections();
	    if( outgoing == null )
		outgoing = new Vector();
	    Description description = a.getDescription();
	    
	    // if we are here, then we have the appropriate
	    // data structures set up
	    agentsToCellAddresses.put( a, address );

	    // we go through this juggling here just to make sure an
	    // agent doesn't appear twice in the list
	    Vector agentList = (Vector)cellAddressesToAgents.get( address );
	    if( agentList == null ) {
		agentList = new Vector();
		cellAddressesToAgents.put( address, agentList );

	    }
	    if( !agentList.contains( a ) )
		agentList.addElement( a );

	    agentsToDescriptions.put( a, description );
	    agentsToOutgoing.put( a, outgoing );
	    agentsToIncoming.put( a, incoming );
	    knownAgents.addElement(a);

	    return true;

	} catch( RemoteException error ) {
	    Debug.error( "remote exception while trying to get information needed to cache " + a + ".  not caching." );
	    return false;

	}

    }

    /**
     * remove an agent from the cache
     *
     * @param a the agent that we want to remove from all caching structures
     * @return true if the agent was removed, false otherwise
     */
    protected boolean decacheAgent( Agent a ) {
	CellAddress cellOn = (CellAddress)agentsToCellAddresses.get( a );
	if( cellOn == null )
	    return false;

	// remove the stuff from all the data structures
	return knownAgents.removeElement( a ) &&
	    ( agentsToCellAddresses.remove( a ) != null ) &&
	    ((Vector)cellAddressesToAgents.get( cellOn )).removeElement( a ) &&
	    ( agentsToDescriptions.remove( a ) != null ) &&
	    ( agentsToOutgoing.remove( a ) != null ) &&
	    ( agentsToIncoming.remove( a ) != null );

    }

    /**
     * deal with the cell event that we either received via the notify
     * or the signal methods.
     *
     * @param event the event that we are going to deal with 
     */
    protected void handleCellEvent( CellEvent incomingEvent ) {
	// we are assuming that we only have the remote reference to
	// the agent by the time we get here!  this means make sure
	// the cell event monitor is working on things right.
	synchronized( dataLockingObject ) {
	    CellEvent event = (CellEvent)incomingEvent;
	    addToEventLog( event );
	    
	    if( event instanceof AgentBornCellEvent ) {
		
		Agent agent = ((AgentBornCellEvent)event).getAgent();
		if( !knownAgents.contains( agent ) )
		    // this should never contain it, but in case it
		    // does
		    knownAgents.addElement( agent );

		else
		    Debug.warning( "the known agents list already seems to contain " + agent + " which is supposed to be just born" );

		cacheAgent( agent );

	    } else if( event instanceof AgentDeathCellEvent ) {
		Agent agent = ((AgentDeathCellEvent)event).getAgent();
		decacheAgent( agent );

	    } else if( event instanceof AgentConnectCellEvent ) {
		Agent originAgent = ((AgentConnectCellEvent)event).getSourceAgent();
		Agent destinationAgent = ((AgentConnectCellEvent)event).getDestinationAgent();

		// ok - we would not have gotten this event unless one
		// of these agents is in our cache -- so let's start
		// with the origin agent.  if it is in the cache, then
		// update its information.  if it is not forget about
		// it.  then examine the destinationAgent
		if( knownAgents.contains( originAgent ) )
		    ((Vector)agentsToOutgoing.get( originAgent )).addElement( destinationAgent );
		
		if( knownAgents.contains( destinationAgent ) )
		    ((Vector)agentsToIncoming.get( destinationAgent )).addElement( originAgent );

	    } else if( event instanceof AgentDisconnectCellEvent ) {
		Agent originAgent = ((AgentDisconnectCellEvent)event).getSourceAgent();
		Agent destinationAgent = ((AgentDisconnectCellEvent)event).getDestinationAgent();

		if( knownAgents.contains( originAgent ) )
		    ((Vector)agentsToOutgoing.get( originAgent )).removeElement( destinationAgent );

		if( knownAgents.contains( destinationAgent ) )
		    ((Vector)agentsToIncoming.get( destinationAgent )).removeElement( originAgent );

	    } else if( event instanceof AgentMovingToCellEvent ) {
		// when we get a moving to event, we just remove the
		// agent from the data structures -- we'll requery it
		// when it arrives at the other cell just as though it
		// were born on the other cell

		Agent agent = ((AgentMovingToCellEvent)event).getAgent();
		decacheAgent( agent );

	    } else if( event instanceof AgentMovingFromCellEvent ) {
		// when we get this moving from, then we just deal as
		// though the agent was birthed on this cell and then
		// we requery it to get all the appropriate
		// information
		
		Agent agent = ((AgentMovingFromCellEvent)event).getAgent();
		CellAddress cellOn = ((AgentMovingFromCellEvent)event).getDestinationAddress();

		if( !knownAgents.contains( agent ) )
		    // this should never contain it, but in case it
		    // does
		    knownAgents.addElement( agent );

		else
		    Debug.warning( "the known agents list already seems to contain " + agent + " which is supposed to be just born" );

		cacheAgent( agent, cellOn );

	    }

	}

    }
    

    /*****************************************************************
     *          CUT AND PASTED FROM EVENTRECEIVINGAGENTIMPL          *
     *****************************************************************/

    /**
     * basic implementation of connectto for event receivers.  just
     * subscribe to events on the remote agent.  
     *
     * @param otherAgent the other agent we are supposed to connect to
     * @return true if the connection was made, false otherwise
     */
    public boolean connectTo( Agent otherAgent ) {
	boolean connected = this.subscribeTo( otherAgent );
	if( connected == true )
	    myCell.noteAgentConnect( this, this, otherAgent );
	return connected;

    }

    /**
     * disconnect from a particular agent. checks that we are
     * currently subscribing to it, then sends it a removelistener
     * message and removes it from the subscription list.  
     * 
     * @param otherAgent the agent we are supposed to disconnect from
     */
    public void disconnectFrom( Agent otherAgent ) {
	// Check that we subscribe.
	if( agentSubscriptions.contains( otherAgent ) != true ) {
	    Debug.warning( "Ignoring request to unsubscribe from an agent we don't subscribe to, " + Global.shortString( otherAgent ) );
	    return;
	}

	// Actually unsubscribe from the agent.
	EventSendingAgent infoAgent = (EventSendingAgent)otherAgent;
	Debug.notice( "unsubscribing from " + Global.shortString( infoAgent ) );
	try { 
	    infoAgent.removeListener( this );
	} catch( RemoteException ex ) {
	    Debug.warning( "Attempt to unsubscribe failed, message was " + ex.getMessage() + ". If the other host is down, it's OK." );
	}

	// And remove it from our list.
	agentSubscriptions.removeElement( otherAgent );

	// And tell the server what we did.
	myCell.noteAgentMessage( this, "removeListener", this, otherAgent );
	myCell.noteAgentDisconnect( this, this, otherAgent );

    }

    /**
     * disconnect ourself from all agents
     */
    public synchronized void disconnectFromAll() {
	if( agentSubscriptions != null ) {

	    // Clone to prevent the remove in disconnectFrom from affecting us.
	    Vector list = (Vector)agentSubscriptions.clone();     

	    // tell the server what is happening
	    myCell.noteAgentDisconnectFromAll( this, this );

	    Enumeration e = list.elements();
	    while( e.hasMoreElements() )
		disconnectFrom( (Agent)e.nextElement() );

	}

    }

    /** 
     * subscribe to an event source.  this is the nuts and bolts of
     * the subscribing. checks the type of the remote agent.
     *
     * @param otherAgent the agent that we are supposed to be subscribing to
     * @return true if the agent was subscribed to
     */
    public boolean subscribeTo( Agent otherAgent ) {
	EventSendingAgent infoAgent = null;
	// Check the type of the agent we're subscribing to.
	if ( ! ( otherAgent instanceof EventSendingAgent ) ) {
	    Debug.error( "EventReceivingAgent was told to subscribe to a non-EventSending." );
	    return false;
	}
    
	infoAgent = (EventSendingAgent)otherAgent;

	// Make the connection.
	myCell.noteAgentMessage( this, "addListener", this, infoAgent );
	try {
	    infoAgent.addListener( this );
	} catch( RemoteException re ) {
	    Debug.critical( "Subscription failed ", re );
	    return false;
	}
    
	// And track the subscription.
	agentSubscriptions.addElement( infoAgent );

	return true;
    }

    /** 
     * return all the agents i'm connected to
     *
     * @return a vector of all the incoming connections
     */
    public Vector listAllIncomingConnections() {
	return agentSubscriptions;

    }

}

