// $Id: CellListAgentImpl.java,v 1.2 2000/05/19 19:55:55 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.


package net.hivecell.hive.agent.cell;

import net.hivecell.hive.Global;
import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.agent.AgentInitializationException;
import net.hivecell.hive.agent.EventTransceivingAgentImpl;
import net.hivecell.hive.agent.cell.AgentMonitoringAgent;
import net.hivecell.hive.agent.hiveui.iconicui.CellListCanvas;
import net.hivecell.hive.event.HiveEvent;
import net.hivecell.hive.event.cell.*;
import net.hivecell.hive.description.Description;
import net.hivecell.hive.description.AgentSet;
import net.hivecell.hive.cell.RemoteCell;
import net.hivecell.hive.shadow.ShadowNotFoundException;
import net.hivecell.hive.support.async.*;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.support.CellAddress;
import net.hivecell.hive.support.CellConnectException;

import java.awt.Panel;
import java.awt.TextArea;
import java.io.BufferedReader;
import java.io.InputStreamReader;
import java.io.IOException;
import java.net.URL;
import java.net.MalformedURLException;
import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;

/** An agent that keeps a list of servers. Agents are in two flavours:
 ** a "master" agent that broadcasts events to whomever cares to listen,
 ** and "client" agents that subscribe to masters to track who is up or down.
 ** Eventually, the distinction between masters and clients might disappear,
 ** to make a more decentralized system.
 **
 ** The agent sends events telling subscribers when servers are coming
 ** up or going down. Typically, only the master agent will be
 ** subscribed to.  However, the slave agents can also rebroadcast
 ** these events, but -only- to other local agents
 **
 ** The agent receives events from AgentMonitoringAgents to learn about
 ** when a server is going down, and from the master CellListAgent to
 ** hear both when servers are coming up and going down.
 **
 ** The agent defines three new methods; listCells(),
 ** announceCell(), and withdrawCell(). See their documentation
 ** for details.
 **
 ** The agent's default behaviour is to ping other servers to see if
 ** they are still up. It maintains an on-screen display of who is up
 ** or down.
 **/
public class CellListAgentImpl
extends EventTransceivingAgentImpl
implements CellListAgent, AsyncResultListener {

    /** Is this agent a master? Controls its behaviour.. **/
    public boolean master = false;
    /** Pointer to the "master" agent. **/
    protected transient CellListAgent masterCellListAgent = null;
    /** List of CellListInfo objects, data about each other server. **/
    protected transient Vector otherCells;
    /** Hashtable associating CellAddress to CellListInfo. **/
    protected transient Hashtable addressToInfo;
    /** How often to update the display. **/
    protected int updateInterval = 3000;
    /** How often to ping servers, multiply by updateInterval **/
    protected int pingFrequency = 20;
    //protected transient TextArea serverListDisplay;

    /** How long to wait for ping, in milliseconds. Negative means forever. **/
    protected int pingTimeout = 5000;
    /** The object we use when making ping calls. **/
    private AsyncCall pingCall;

    public CellListAgentImpl() 
    throws RemoteException {

        super();
        final Object[] pingCallParams = {};
        try {
            // We use "getAddress" for now, because the method was there.
            // ideally, it should be just a "ping()" method?.
            pingCall = new AsyncCall( "ping", RemoteCell.class, pingCallParams );

        } catch( NoSuchMethodException ex ) {
            Debug.critical( "Couldn't find RemoteCell.ping()?!", ex );
        }

    }


    
    public void doLocalSetup() throws AgentInitializationException {
	super.doLocalSetup();

        otherCells = new Vector();
        addressToInfo = new Hashtable();

        // Always track ourselves.
        if ( !trackCell( myCell.getAddress(), myCell, true ) )
	    Debug.critical( "Something is really funky -- we can't track ourselves" );

        // Now see if we should connect to the master list agent.
        if( Boolean.getBoolean( Global.JoinCellList ) )
            connectToMasterListAgent();
        else
            Debug.notice( "Not joining Hive server federation as requested" );

    }
    
    public void doBehavior() {

        // Now enter the main loop.
        int pingCounter = 0;
        while( !timeToStop ) {
            if( master ) {
                // ping other servers. Currently, only master does this to
                // cut down on the impact of this system.
                pingCounter++;
                if( pingCounter >= pingFrequency ) {
                    pingCounter = 0;
                    Enumeration e = otherCells.elements();
                    while( e.hasMoreElements() ) {
                        CellListInfo sli = (CellListInfo)e.nextElement();
                        if( !sli.getAddress().equals( myCell.getAddress() ) && sli.isUp() ) {
                            RemoteCell rs = sli.getCell();
                            AsyncResult ar = pingCall.callOn( rs, pingTimeout );
                            ar.addExceptionOnlyListener( this );
                            // Note, this isn't strictly right...
                            sli.updateLastInfo();
                        }
                    }
                }
            }

            try {
                Thread.sleep( updateInterval );
            } catch( InterruptedException ex ) { }
        }
    }

    /** Return a list of all of the servers this agent knows about.
     ** Elements in the Vector are of type CellListInfo.
     **/
    public Vector listCells( Agent sender ) {

        myCell.noteAgentMessage( this, "listCells", sender, this );

	// wait until the crucial part of doBehavior is done
	while( !isReady() )
	    synchronized( this ) {
		try {
		    wait();
		} catch( InterruptedException error ) { }
	    }


        myCell.noteAgentMessage( this, "listCells reply", this, sender );

	return otherCells;

    }

    /**
     * return a list of all the servers that match this semantic
     * description that are up and running -- it will ignore those
     * cells which are not up
     *
     * @param requester the agent asking for this information
     * @param semantic the semantic array to make the query with -- a query will null here will just return all servers which are up
     * @return a desc set of remote servers and descriptions that match this query 
     */
    public AgentSet queryCells( Agent requester, String[] semantic ) {

        myCell.noteAgentMessage( this, "queryCells", requester, this );
	// wait until the crucial part of doBehavior is done
	while( !isReady() )
	    synchronized( this ) {
		try {
		    wait();
		} catch( InterruptedException error ) { }
	    }
        
	// get a vector of all those which are up -- remote servers
	Vector upCells = new Vector();
	synchronized( this ) {
	    Enumeration e = otherCells.elements();
	    while( e.hasMoreElements() ) {
		CellListInfo sli = (CellListInfo)e.nextElement();
		if( sli.isUp() ) upCells.addElement( sli.getCell() );
	    }
	}
	AgentSet ds = new AgentSet( upCells );
	ds = ds.querySet(null, semantic);
	/*	if( semantic != null )
	    for( int i=0; i < semantic.length; i++ )
		ds = Lookup.requireRole( ds, semantic[i] );
	*/
        myCell.noteAgentMessage( this, "queryCells reply", this, requester );
	return ds;

    }

    /** Ask the server list agent to announce that a new server is up.
     ** @return Whether the announcement was sent out.
     **/
    public boolean announceCell( Agent sender, CellAddress ca, RemoteCell rs ) {
	
        myCell.noteAgentMessage( this, "announceCell", sender, this );

	if( !trackCell( ca, rs, true ) )
	    return false;

	
	if( !master ) {
	    try {
		return masterCellListAgent.announceCell( sender, ca, rs );
	    } catch( RemoteException error ) {
		Debug.error( "Remote exception while trying to announce server to master server list agent." );
		return false;
	    }

	} else {
            CellBornCellEvent sbse = new CellBornCellEvent( this, eventSendingHelper.getEventCount( this ), ca );
            eventSendingHelper.broadcastEvent( sbse );
            return true;
        }

    }

    /** Ask the server list agent to annouce that a server is withdrawing
     ** from the network.
     ** @return Whether the announcement was sent out.
     **/
    public boolean withdrawCell( Agent sender, CellAddress ca, RemoteCell rs ) {

        myCell.noteAgentMessage( this, "withdrawCell", sender, this );
        stopTrackingCell( ca, rs );
        Debug.warning("Told to withdraw a server, that's not well tested.");

        // Now, if we're a master, broadcast the information about withdrawing
        if( master ) {
            CellWithdrawnCellEvent swse = new CellWithdrawnCellEvent( this, eventSendingHelper.getEventCount( this ), ca );
            eventSendingHelper.broadcastEvent( swse );
            return true;
        } else {
	    try {
		return masterCellListAgent.withdrawCell( sender, ca, rs );
	    } catch( RemoteException error ) {
		Debug.error( "Remote exception while trying to withdraw server from master server list agent." );
		return false;
	    }
        }

    }

    /** When an event is received, see if it's an event the server list
     ** agent cares about, and act on it. Update the data structures about
     ** what servers are up, and if this is a master then possibly repropagate
     ** the event.
     **/
    public void notify( HiveEvent he ) {
        super.notify( he );
        if ( !( he instanceof CellEvent ) ) {
            Debug.notice( "Cell list agent received a hive event that wasn't a server event. Ignoring." );
            return;
        }
        CellEvent se = (CellEvent)he;

	// please please please let the event code for server events
	// not become bigger than an int!!!!
        switch( (int)se.eventCode() ) {
            // if it's SERVERBORN, add it to our list
            case (int)CellEvent.SERVERBORN:
		try {
		    trackCell( ((CellBornCellEvent)se).getAddress(), ((CellBornCellEvent)se).getAddress().getRemoteCell(), true );
		} catch( CellConnectException error ) {
		    Debug.error( "Problem converting a cell address to a remote server reference" );
		}
                break;
            // if it's SERVERGOINGDOWN, take it down in our list
            case (int)CellEvent.SERVERGOINGDOWN:
		try {
		    noteGoingDown( ((CellGoingDownCellEvent)se).getAddress(), ((CellGoingDownCellEvent)se).getAddress().getRemoteCell() );
		} catch( CellConnectException error ) {
		    Debug.error( "Problem converting a cell address to a remote server reference" );
		}
                break;
            // if it's SERVERWITHDRAWN, remove it from our list
            case (int)CellEvent.SERVERWITHDRAWN:
		try {
		    stopTrackingCell( ((CellWithdrawnCellEvent)se).getAddress(), ((CellWithdrawnCellEvent)se).getAddress().getRemoteCell() );
		} catch( CellConnectException error ) {
		    Debug.error( "Problem converting a cell address to a remote server reference" );
		}
                break;
            default:
                return;                           // exit function
        }

        // Now, repropagate the event if appropriate. Note, code above
        // has already fitered the event type, this only gets executed
        // if it's an event the server list agent cares about.  If
        // this agent is the master server list agent, then we should
        // tell everybody of this new update.  If we are not, then
        // only notify our local agents of this fact
        if( master ) {
            if ( he.getSource() instanceof CellListAgent ) {
                Debug.notice( "Not repropagating server event received from another server list agent." );
                return;
            } else {
                eventSendingHelper.broadcastEvent( he );
            }
        } else {
	    eventSendingHelper.broadcastEventTo( he, eventSendingHelper.getLocalListeners(), null );

	}

    }



    /** Find the master list agent and connect to it. **/
    protected void connectToMasterListAgent() {
        RemoteCell masterRS = findMasterCell();

        // Test, are we destined to be the master server list agent?
        if( myCell.equals( masterRS ) ) {
            Debug.notice( "I am on the master server! I will act as a master." );
            master = true;
        } else {
            master = false;
        }

        // OK, now let's find the master server list agent itself. Note,
        // this could just be ourselves; that's OK.
        if( masterRS != null )
            masterCellListAgent = findMasterAgent( masterRS );
        if( masterCellListAgent == null ) {
            Debug.error( "Couldn't find the remote master server list agent, we cannot join the federation." );
        } else {
            // Alright, now are we the master?
            if( master ) {
                // Yes, let's skip the rest of this, just do a sanity check.
                Debug.assert( masterCellListAgent.equals( this ) );
            } else {
                // OK, we've found the master, let's talk to it.
                try {
                    // First, subscribe ourselves as a listener to the master.
                    this.connectTo( masterCellListAgent );

                    // Second, update our list based on the serve.r
                    updateListFrom( masterCellListAgent );

                    // Third, ask the master list to announce us.
                    myCell.noteAgentMessage( this, "announceCell", this, masterCellListAgent );
                    masterCellListAgent.announceCell( this, myCell.getAddress(), myCell );
                    myCell.noteAgentMessage( this, "announceCellReply", this, masterCellListAgent );
                } catch( RemoteException re ) {
                    Debug.error( "Couldn't subscribe to the remote server list agent." );
                }
            }
        }
    }

    
    /** Look for the server that hosts the master server list agent. **/
    protected RemoteCell findMasterCell() {
        String cellName = System.getProperty( Global.MasterCellListCell );
        if( cellName == null ) {
            try {
                URL url = new URL( Global.defaultCellListURL );
                BufferedReader br = new BufferedReader( new InputStreamReader( url.openStream() ) );
                cellName = br.readLine().trim();
            } catch( MalformedURLException ex ) {
                Debug.error( "URL malformed " + Global.defaultCellListURL );
                return null;
            } catch( IOException ex ) {
                Debug.error( "Error looking up data in " + Global.defaultCellListURL + " use " + Global.MasterCellListCell + " to override." );
                return null;
            }
        }

        try {
            CellAddress masterCA;
            RemoteCell masterRS;
            masterCA = new CellAddress( cellName );
            return masterCA.getRemoteCell();
        } catch( CellConnectException cce ) {
            Debug.error( "Couldn't connect to master server list cell." );
            return null;
        } catch( net.hivecell.hive.support.MalformedHiveURLException mhue ) {
            Debug.error( "Bad Hive URL " + cellName );
            return null;
        }
    }

    protected CellListAgent findMasterAgent( RemoteCell rs ) {
        String[] dsName = { "net.hivecell.hive.agent.cell.CellListAgent" };
        AgentSet ds = null;
        try {
            ds = rs.queryAgents( this, dsName, null );
        } catch( ClassNotFoundException cnfe ) {
            Debug.error( "Major problem on remote host, didn't have CellListAgent class." );
            return null;
        } catch( RemoteException re ) {
            Debug.error( "Couldn't connect to remote host to find master CellListAgent." );
            return null;
        }
        
        if( ds.size() == 0 ) {
            Debug.error( "Couldn't find a CellListAgent on master cell." );
            return null;
        } else if( ds.size() > 1 ) {
            Debug.warning( "Found more than one CellListAgent on master cell?? Picking one.." );
        }
        return (CellListAgent)ds.pickOne();
    }

    /** Make sure we're tracking a server. If we haven't seen it before,
     ** we sanity check it and add it to our lists.
     ** @return whether this is a new server
     **/
    protected synchronized boolean trackCell( CellAddress ca, RemoteCell rs, boolean isUpNow ) {
        Debug.notice( "Tracking " + ca + " " + Global.dateString() );
        if( isUpNow ) {
            if( checkSanity( ca, rs ) == false ) {
                Debug.warning( "Cell list agent not tracking cell " + ca + " because the sanity check on it failed." );
                return false;
            }
        }
        
        // First, test to see if we already know this one.
        CellListInfo storedSLI = (CellListInfo)addressToInfo.get( ca );

        if( storedSLI == null ) {
            // Nope, this is new, time to create a data structure for it.
            CellListInfo newSLI = new CellListInfo( ca, rs, isUpNow );
            addressToInfo.put( ca, newSLI );
            otherCells.addElement( newSLI );
            // If we're the master, subscribe to the server event info agent.
            // (slaves could, but they don't to avoid overwhelming everyone.
            // Currently they rely on the master to rebroadcast information.)
            if( master && isUpNow )
                subscribeToInfoAgent( newSLI );
            return true;
        } else {
            // We've seen this one before. Just update our information.
            Debug.assert( otherCells.contains( storedSLI ) == true );
            storedSLI.setUp( isUpNow );
	    //            return false;
	    return true;
        }
    }

    /** Remove a server from our lists.
     ** @return whether we were tracking this in the first place.
     **/
    protected synchronized boolean stopTrackingCell( CellAddress ca, RemoteCell rs ) {
        CellListInfo storedSLI = (CellListInfo)addressToInfo.get( ca );
        if( storedSLI == null ) {
            Debug.notice( "Told to stop tracking " + ca + " but we were not watching it." );
            return false;
        } else {
            Debug.notice( "No longer tracking " + ca + " "+Global.dateString() );
            
            Debug.assert( otherCells.contains( storedSLI ) == true );
            addressToInfo.remove( ca );
            otherCells.removeElement( storedSLI );
            return true;
        }
    }

    /** Update the data structure noting that a server is going down. **/
    protected synchronized void noteGoingDown( CellAddress ca, RemoteCell rs ) {
        CellListInfo storedSLI = (CellListInfo)addressToInfo.get( ca );
        if( storedSLI == null ) {
            Debug.notice( "We were told " + ca + " was going down, but we didn't know about it." );
            return;
        } else {
            Debug.notice( "Noting that " + ca + " is down " + Global.dateString() );
            // note the down status.
            storedSLI.setUp( false );
        }
    }

    /** Subscribe myself to the server list info agent on the given host. **/
    protected void subscribeToInfoAgent( CellListInfo sli ) {
        String[] amaName = { "net.hivecell.hive.agent.cell.AgentMonitoringAgent" };
        AgentSet ds = null;
        AgentMonitoringAgent ama = null;
        try {
            ds = sli.getCell().queryAgents( this, amaName, null );
        } catch( ClassNotFoundException cnfe ) {
            Debug.error( "Major problem on remote host, didn't have server event info class." );
        } catch( RemoteException re ) {
            Debug.error( "Couldn't connect to remote host to find master server event info agent." );
        }
        if( ds.size() == 0 ) {
            Debug.error( "Couldn't find a server event info agent on master cell." );
        } else {
            if( ds.size() > 1 )
                Debug.warning( "Found more than one server event info agent on master cell?? Picking one.." );
            ama = (AgentMonitoringAgent) ds.pickOne();
        }
        if( ama == null ) {
            Debug.warning( "Couldn't subscribe to server event info agent, if it goes down we might not know." );
            return;
        }
        this.connectTo( ama );
    }
    
    /** Update our list with the info from another agent. **/
    protected void updateListFrom( CellListAgent sla ) 
    throws RemoteException {
        // Subscribe to 
        Vector newList = null;
        myCell.noteAgentMessage( this, "listCells", this, sla );
        newList = sla.listCells( this );
        myCell.noteAgentMessage( this, "listCells reply", sla, this );
        
        // And assimilate all the servers there.
        if( newList != null ) {
            Enumeration e = newList.elements();
            while( e.hasMoreElements() ) {
                CellListInfo sli = (CellListInfo)e.nextElement();
                trackCell( sli.getAddress(), sli.getCell(), sli.isUp() );
            }
        }
    }

    /** A little sanity check on cell addresses, remote servers. Do they
     ** resolve correctly to each other? This involves network traffic.
     **/
    protected boolean checkSanity( CellAddress ca, RemoteCell rs ) {
        CellAddress lookedupCA;
        RemoteCell lookedupRS;
        try {
            lookedupCA = rs.getAddress();
        } catch( RemoteException re ) {
	    Debug.warning( "Couldn't get cell address from remote server." );
            return false;
        }

        try {
            lookedupRS = ca.getRemoteCell();
        } catch( CellConnectException cce ) {
            Debug.warning( "Couldn't get remote server from cell address." );
            return false;
        }
        
        if ( !( lookedupCA.equals( ca ) && lookedupRS.equals( rs ) ) ) {
            Debug.warning( "Sanity check failed; inconsistency between cell address and remote server." );
            Debug.warning( lookedupCA + " " + ca + " " + lookedupRS + " " + rs );
            return false;
        } else
            return true;
    }
    

    /** Debugging method - just dump out all the servers we know. **/
    private void dumpCellList() {
        Debug.assert( addressToInfo.size() == otherCells.size() );
        Debug.notice( "Dumping list of " + addressToInfo.size() + " servers." );
        Enumeration e = otherCells.elements();
        while( e.hasMoreElements() ) {
            CellListInfo sli = (CellListInfo)e.nextElement();
            Debug.assert( addressToInfo.get( sli.getAddress() ) == sli );
            Debug.notice( sli );
        }
    }

    /** Handle asynchronous replies Currently, we only get these when
     ** a ping fails - we then note that the server is down.
     ** TODO - this should get all ping messages, and update the time
     ** on the ones that are still up. But that requires an efficient
     ** data structure to map RemoteCell to CellListInfo.
     **/
    public void asyncFinished( AsyncResult ar ) {
        // First check, was this a ping call?
        if( !ar.getCall().equals( pingCall ) )
            return;

        if( ar.wasException() == false ) {
            Debug.warning( "Got a result that was not an exception, ignoring." );
            return;
        }

        RemoteCell rs = (RemoteCell)ar.getReceiver();
        Debug.assert( rs != null );

        // Scan our data structures for the CellListInfo corresponding
        // to this server. TODO - make this more efficient by changing
        // our data structures.
        Enumeration e = otherCells.elements();
        CellListInfo sliDown = null;
        while( e.hasMoreElements() ) {
            CellListInfo sli = (CellListInfo)e.nextElement();
            if( rs.equals( sli.getCell() ) ) {
                Debug.assert( sliDown == null );
                sliDown = sli;
            }
        }

        if( sliDown == null ) {
            // Ooops, didn't find it!
            Debug.error( "Had a bad ping for host, but it's not on our list!" );
            return;
        } else {
            // Alright, note the failure.
            Debug.notice( "Ping for " + sliDown.getAddress() + " failed." );
            noteGoingDown( sliDown.getAddress(), sliDown.getCell() );
            if( master ) {
                CellGoingDownCellEvent sgdse = new CellGoingDownCellEvent( this, eventSendingHelper.getEventCount( this ), sliDown.getAddress() );
                eventSendingHelper.broadcastEvent( sgdse );
            }
        }
    }
    
    public void setUpdateInterval( int i ) { updateInterval = i; }
    public int getUpdateInterval() { return updateInterval; }
    public void setPingTimeout( int i ) { pingTimeout = i; }
    public int getPingTimeout() { return pingTimeout; }
    public void setPingFrequency( int i ) { pingFrequency = i; }
    public int getPingFrequency() { return pingFrequency; }

}

