// $Id: BaseAgentIcon.java,v 1.1 2000/05/18 22:00:53 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

/**
 ** Base class for icons in a "Boxes and lines" UI. In bad haiku:
 ** 
 ** icons of agents
 ** colored pictures; images
 ** of life in the hive
 **/

package net.hivecell.hive.agent.hiveui.iconicui;

import java.io.*;
import java.awt.*;
import java.awt.image.*;
import java.util.*;

import net.hivecell.hive.Global;
import net.hivecell.hive.agent.*;
import net.hivecell.hive.support.*;


/**
 ** Terminology: 
 ** Image - the picture to be displayed
 ** Icon - the image and the border together
 **/
public class BaseAgentIcon {
    
    public static final Dimension defaultIconSize = new Dimension(32, 32);
    public static final Color defaultBorderColor = Color.cyan;
    public static final int defaultBorderWidth = 3;

    protected Dimension currentIconSize;

    protected Color borderColor;
    protected int borderWidth;
    
    protected Image image;
    protected Rectangle rect;  // Bounding rectangle of the agent.  Makes it easy to find out if it is at point p.

    protected Agent agent;
    protected Vector connections;   // BaseConnection to the agents this agent is connected to.


    /** Constructor should be called in the class responsible for drawing the icon, eg an instance of BaseUIAgent */
    public BaseAgentIcon(Agent agent, Image image) {
	this.agent = agent;

	connections = new Vector();

	currentIconSize = new Dimension( defaultIconSize.width + (2 * defaultBorderWidth), 
					 defaultIconSize.height + (2 * defaultBorderWidth) );
	
	this.image = image.getScaledInstance(defaultIconSize.width, defaultIconSize.height, Image.SCALE_SMOOTH);
	/*  Cool icons (that raffi doesn't like)
	this.image = image.getScaledInstance(currentIconSize.width, currentIconSize.height, Image.SCALE_SMOOTH);
	currentIconSize.width += defaultBorderWidth;
	currentIconSize.height += defaultBorderWidth;
	*/
	rect = new Rectangle(currentIconSize);
	borderWidth = defaultBorderWidth;
	borderColor = defaultBorderColor;
    }

    /** Constructor for specifying an icon size other than the default */
    public BaseAgentIcon(Agent agent, Image image, Dimension size) {
	this.agent = agent;
	currentIconSize = size;

	connections = new Vector();

	this.image = image.getScaledInstance(defaultIconSize.width, defaultIconSize.height, Image.SCALE_SMOOTH);

	rect = new Rectangle(currentIconSize);
	borderWidth = defaultBorderWidth;
	borderColor = defaultBorderColor;
    }



    public boolean equals( Object o ) {
        if (o == null)
            return false;

	if( !( o instanceof BaseAgentIcon ) )
	    return false;

	BaseAgentIcon bai = (BaseAgentIcon)o;

        if (bai.borderColor == null || bai.agent == null)
            return false;

	if( !bai.borderColor.equals( borderColor ) )
	    return false;

	if( !bai.agent.equals( agent ) )
	    return false;
	
	return true;

    }

    /** Set a new image for the icon */
    public void setImage(Image image) { this.image = image; }

    /** Returns the icon image. */
    public Image getImage() { return (image); }


    /**
     ** Returns the icon image with dimension size.
     ** @param size the dimension of the image to be returned.
     **/
    public Image getImage(Dimension size) { 
	currentIconSize = size;
	rect.setSize(size);
	return (image.getScaledInstance(size.width, size.height, Image.SCALE_FAST));
    }

    /** Returns the current icon's size. */
    public Dimension getCurrentSize() { return (currentIconSize); }


    /**
     ** Move the icon to location Point p on the UICanvas, or to location int x, int y.
     ** Bounds checking should be done in the UICanvas
     **/
    public void move(Point p) { move(p.x, p.y); }
    public void move(int x, int y) { rect.move(x, y); }


    /**
     ** Sets the size (in pixels) of the border around the image.
     ** @param width Width of the border
     **/
    public void setBorderWidth(int width) { borderWidth = width; }
    /**
     ** Returns the color of the border around the image.
     ** @param color Color of the border
     **/
    public void setBorderColor(Color col) { borderColor = col; }

    /** Returns the size (in pixels) of the border around the image. */
    public int getBorderWidth() { return(borderWidth); }
    /** Returns the color of the border around the image. */
    public Color getBorderColor() { return(borderColor); }
    
    
    /** Returns the x location of the upper left corner of the icon. */
    public int getX() { return(rect.getLocation().x); }
    /** Returns the y location of the upper left corner of the icon. */
    public int getY() { return(rect.getLocation().y); }
    /** Returns the location of the upper left corner of the icon. */
    public Point getLocation() { return(rect.getLocation()); }

    /** Returns the center point of the icon. */
    public Point getCenter() { 
	return ( new Point(rect.getLocation().x + (currentIconSize.width / 2), 
			   rect.getLocation().y + (currentIconSize.height / 2)) );
    }


    /** Returns the agent represented by the icon. */
    public Agent getAgent() { return(agent); }



    /** 
     * adds a new BaseConnection to this agent -- it first attempts to
     * make sure that we are not adding a duplicate conection
     *
     * @param c the base connection we wish to add
     * @return true if we have added it, false otherwise 
     */
    public boolean addConnection( BaseConnection c ) { 
	if( !connections.contains( c ) ) {
	    connections.addElement( c ); 
	    return true;
	}
	return false;
    }

    /**
     * remove a connection from this agent
     *
     * @param c the base connectino object taht we wish to remove
     * @return true if the connection was removed, false otherwise
     */
    public boolean removeConnection( BaseConnection c ) { 
	if( connections.contains( c ) ) {
	    connections.removeElement( c );
	    return true;
	}
	return false;
    }

    /**
     * return a list of all the connections for this icon
     *
     * @return a vector of all the BaseConnection objects
     */
    public Vector getConnections() { return connections; }


    /**
     ** Returns true if pt is contained in the icon.
     ** @param pt Does the icon contain pt?
     **/
    public boolean contains(Point pt) { return (rect.contains(pt)); }
    

    /** Update/(re)draw the icon.  This is called whenever the canvas is rendered. */
    public void draw(Graphics g, ImageObserver io) {
	if (g == null) return;         // Always make sure you have a valid Graphics object.
	g.setColor( borderColor );
	// Standard draw:
	g.fill3DRect( this.getX(), this.getY(), currentIconSize.width, currentIconSize.height, true );
	// 3d draw
	//g.fill3DRect( this.getX(), this.getY(), currentIconSize.width - defaultBorderWidth, currentIconSize.height - defaultBorderWidth, true );

	g.drawImage( this.image, this.getX() + borderWidth, this.getY() + borderWidth, io );
    }
    
    /** 
     ** Update the icon's animation.  This should be overridden in a class that implements AnimatedIcon.
     ** It is called in the uiCanvas's animation loop. 
     **/
    public void update(Graphics g, ImageObserver io) { }
}
