// $Id: BaseGraphicalArrow.java,v 1.1 2000/05/18 22:00:53 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

/** 
    BaseGraphicalArrow is more complicated - it has to deal with rendering
    Briefly, an arrow is several things:
    A coloured line between two AgentIcons to show a connection exists
    Two coloured arrowheads to show the age of that half of the link
    Dots to show magic packets travelling along.
    Colors of these things are dependent on how long ago that half of
    the arrow was used. lastAtoB, lastBtoA, and aToBLast track that.
*/

package net.hivecell.hive.agent.hiveui.iconicui;

import java.awt.*;

import net.hivecell.hive.support.Debug;

// note the equality semantics below
public class BaseGraphicalArrow {
    // basic state
    //AgentGraphCanvas canvas;              // where are we drawing?
    //IconicUIAgentImpl uiAgent;              // who are we drawing for?
    BaseAgentIcon aData, bData;             // what icons are used?

    // tracking last use of each half of the link
    private int lastAtoB, lastBtoA;         // when were these dirs last used?
    private boolean aToBLast;               // if the same time, which dir last?
    private boolean connectedAToB=false, connectedBToA=false;
    private boolean needsErasing = false;          // set in disconnectDir()

    // Save this to help erase old parts of the arrow
    private Point lastDot;                  // save to erase it later

    // Constants to change the look
    static final int dotSize = 6;               // should be even on Unix
    static final int numDots = 5;               // needs to be odd
    static final int msToFade = 10000;          // how long to colour blue
    static final int msToDisappear = 60000;     // after this, no more draw **
    // ** currently disabled!
    static final int blueDivisions = msToFade / BaseUICanvas.updateInterval;
    static final int blackAge = msToDisappear / BaseUICanvas.updateInterval;
  
    public BaseGraphicalArrow( BaseAgentIcon a, BaseAgentIcon b ) {
	aData = a;
	bData = b;
	lastAtoB = BaseUICanvas.timer;
	lastBtoA = BaseUICanvas.timer;
    }

    public void connectDirection( BaseAgentIcon from, BaseAgentIcon to ){
	if( ( this.aData == from ) && ( this.bData == to ) ){
	    connectedAToB = true;
	}
	if( ( this.aData == to ) && ( this.bData == from ) ) {
	    connectedBToA = true;
	}
    }

    public void disconnectDirection( BaseAgentIcon from, BaseAgentIcon to ){
	if( ( this.aData == from ) && ( this.bData == to ) )
	    connectedAToB = false;
	else if( ( this.aData == to ) && ( this.bData == from ) )
	    connectedBToA = false;
	else
	    Debug.println( Debug.ERROR, "Invalid arrow data given!" );
	if ( ( connectedAToB == false ) || ( connectedBToA == false ) )
	    needsErasing = true;
    }

    // Equality is just are we talking about the same A and B?
    public int hashCode() {
	int code = aData.hashCode() * bData.hashCode();
	return code;
    }

    // Equality is just are we talking about the same A and B?
    public boolean equals( Object obj ) {
	if( obj instanceof BaseGraphicalArrow ) {
	    BaseGraphicalArrow that = (BaseGraphicalArrow)obj;
	    return ( ( this.aData == that.aData ) && ( this.bData == that.bData ) ||
		     ( this.aData == that.bData ) && ( this.bData == that.aData ) );
	} else
	    return false;
    }

    // Call this everytime a link is used - update the right timer
    //   aToBLast is used as a tiebreaker to show msg/reply pairs
    void noteLinkUsed( BaseAgentIcon from, BaseAgentIcon to ) {
	if( ( aData == from ) && ( bData == to ) ) {
	    lastAtoB = BaseUICanvas.timer;
	    aToBLast = true;                            // aToB was seen last
	} else if( ( aData == to ) && ( bData == from ) ) {
	    lastBtoA = BaseUICanvas.timer;
	    aToBLast = false;                           // bToA was seen last
	} else
	    Debug.println( Debug.ERROR, "noteLinkUsed() called with bad BaseAgentIcon" );
    }

    // render the arrow on the given graphics
    synchronized void drawSelf( Graphics g ) {
	if (g == null) return;         // Always make sure you have a valid Graphics object.
	Color bg;

	// First, let's erase any old dot that might be around
	if( lastDot != null ) {
	    g.setColor( IconicUIAgentImpl.bgColor );
	    g.fillOval( lastDot.x-dotSize/2, lastDot.y-dotSize/2, dotSize, dotSize );
	    lastDot = null;
	}

	// Second, let's calculate the colours for things
	Color lineColor, aColor, bColor;
	int ageAtoB = (BaseUICanvas.timer - lastAtoB < 0 ? 0 : BaseUICanvas.timer - lastAtoB);
	int ageBtoA = (BaseUICanvas.timer - lastBtoA < 0 ? 0 : BaseUICanvas.timer - lastBtoA);
	int ageEither = Math.min( ageBtoA, ageAtoB );

	// The arrow is only drawn if the agents are connected, or if they
	// were just disconnected. If they are connected, then the colour
	// depends on the age of last communication. If they aren't connected,
	// then we draw the arrow with the background colour.
	if( connectedAToB || connectedBToA ) {
	    // Arrowhead colours depend on the age of last communication
	    aColor = calculateColor( ageBtoA );
	    bColor = calculateColor( ageAtoB );
	    lineColor = calculateColor( ageEither );
	    // Ok, let's draw the arrow!
	    drawArrow( g, aData.getCenter(), bData.getCenter(), lineColor, aColor, bColor );
	} else if( needsErasing ) {
	    aColor = IconicUIAgentImpl.bgColor;
	    bColor = aColor;
	    lineColor = aColor;
	    // Ok, let's draw the arrow!
	    drawArrow( g, aData.getCenter(), bData.getCenter(), lineColor, aColor, bColor );
	    needsErasing = false;                       // we've erased it now.
	} else {
	    // Not connected, not just disconnected, don't drawn an arrow.
	}
	
	// Now then, is this link fresh enough to draw dots?
	if( ageEither < numDots ) {
	    Point from = null, to = null;
	    int dotDistance;

	    g.setColor( Color.red );                      // dots always red

	    // Which direction is last used on the link?
	    if( lastAtoB > lastBtoA )
		lastDot = drawDot( g, aData.getCenter(), bData.getCenter(), ageAtoB );
	    else if ( lastBtoA > lastAtoB )
		lastDot = drawDot( g, bData.getCenter(), aData.getCenter(), ageBtoA );
	    else {
		// Ugh, both are the same age. Must be a msg/reply, use aToBLast
		// and animate the dot twice as fast, back and forth
		int age = ageEither;
		int age2;
		if ( age * 2 <= numDots - 1 )
		    age2 = age * 2;
		else
		    age2 = ( numDots - 1 - age ) * 2;
		if( aToBLast )
		    lastDot = drawDot( g, bData.getCenter(), aData.getCenter(), age2 );
		else
		    lastDot = drawDot( g, aData.getCenter(), bData.getCenter(), age2 );
	    }
	}
    }

    // calculate a colour based on age
    //   red when very young
    //   blue fade to black until msToFade
    //   black until msToDisappear
    //   then null, means draw nothing (currently turned off)
    Color calculateColor( int age ) {
	if( age < numDots ) 
	    return Color.red;
	else if( age < blueDivisions ) {
	    float blue = 1.0f - ( age / (float)blueDivisions );
	    return ( new Color( blue / 2, blue / 2, blue ) );       // should cache
	} else if( age < blackAge ) {
	    return Color.black;
	} else {
	    // return null;
	    return Color.black;
	}
    }

    // Actual dot drawing code. Takes a logical distance [0..numDots)
    // and turn it into a point. Uses the same halo code the arrows do.
    public Point drawDot( Graphics g, Point from, Point to, int dotDistance ) {
	if (g == null) return null;         // Always make sure you have a valid Graphics object.
	int halo = BaseUICanvas.normalIconWidth * 5 / 4; // space around centers

	double angle = Math.atan2( to.y - from.y, to.x - from.x );
	int x1, x2, y1, y2;
	x1 = from.x + (int)Math.round( halo * Math.cos( angle ) );
	y1 = from.y + (int)Math.round( halo * Math.sin( angle ) );
	x2 = to.x - (int)Math.round( halo * Math.cos( angle ) );
	y2 = to.y - (int)Math.round( halo * Math.sin( angle ) );

	if( from.x < to.x && x1 > x2 || from.x > to.x && x1 < x2 ||
	    from.y < to.y && y1 > y2 || from.y > to.y && y1 < y2 ) {
	    // these two boxes are too close, let's just skip them.
	    return null;
	}

	// ok, put the dot the right distance along the line
	Point dot = new Point( x1 + ( x2 - x1 ) * dotDistance / ( numDots-1 ),
			       y1 + ( y2 - y1 ) * dotDistance / ( numDots-1 ) );
	g.fillOval( dot.x - dotSize/2, dot.y - dotSize/2, dotSize, dotSize );
	return dot;
    }
      
    // Draw an arrow with given colors. Color null means don't draw that part.
    protected void drawArrow( Graphics g, Point a, Point b, Color lc, Color ac, Color bc ) {
	if (g == null) return;         // Always make sure you have a valid Graphics object.
	if( a.equals( b ) )            // don't draw a length 0 arrow.
	    return;
	int halo = ( BaseUICanvas.normalIconWidth + BaseUICanvas.iconBorderWidth * 2 ) * 6 / 8;  // space around centers

	double angleBtoA = Math.atan2( b.y - a.y, b.x - a.x );
	double angleAtoB = Math.atan2( a.y - b.y, a.x - b.x );

	int x1, x2, y1, y2;
	x1 = a.x - (int)Math.round( halo * Math.cos( angleAtoB ) );
	y1 = a.y - (int)Math.round( halo * Math.sin( angleAtoB ) );
	x2 = b.x - (int)Math.round( halo * Math.cos( angleBtoA ) );
	y2 = b.y - (int)Math.round( halo * Math.sin( angleBtoA ) );

	if( lc != null ) {
	    g.setColor( lc );
	    g.drawLine( x1, y1, x2, y2 );
	}

	if( ac != null ) {
	    g.setColor( ac );
	    if( connectedAToB || needsErasing )
		drawArrowHead( g, angleAtoB, x1, y1, 10 );
	    else
		drawArrowHead( g, angleAtoB, x1, y1, 6 );
	}
	if( bc != null ) {
	    g.setColor( bc );
	    if( connectedBToA || needsErasing )
		drawArrowHead( g, angleBtoA, x2, y2, 10 );
	    else
		drawArrowHead( g, angleBtoA, x2, y2, 6 );
	}
    }

    // Draw the arrow head at the specified location
    void drawArrowHead( Graphics g, double angle, int x1, int y1 ) {
	drawArrowHead( g, angle, x1, y1, 8 );
    }

    void drawArrowHead( Graphics g, double angle, int x1, int y1, int size ) {
	if (g == null) return;         // Always make sure you have a valid Graphics object.
	int arrowSize = size;
	int ax1, ay1, ax2, ay2;
	ax1 = x1  - (int)Math.round( arrowSize * Math.cos( angle - Math.PI/6 ) );
	ay1 = y1  - (int)Math.round( arrowSize * Math.sin( angle - Math.PI/6 ) );
	ax2 = x1  - (int)Math.round( arrowSize * Math.cos( angle + Math.PI/6 ) );
	ay2 = y1  - (int)Math.round( arrowSize * Math.sin( angle + Math.PI/6 ) );
	int xp[] = { x1, ax1, ax2 };
	int yp[] = { y1, ay1, ay2 };
	g.fillPolygon( xp, yp, 3 );
    }
}
