// $Id: BaseLayoutManager.java,v 1.1 2000/05/18 22:00:53 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

/**
 ** Base class for Layout manager in a "Boxes and Lines" UI.
 ** This class is responsible for updating the graphics whenever 
 ** any of the following actions takes place:
 **
 ** an agent is added/removed
 ** two agents are connected/disconnected
 ** an agent (icon) is moved
 **
 ** In general: whenever a "box" or "line" needs to be rearranged 
 ** for any reason, this is the class that determines where it
 ** goes and how it looks.  The BaseUICanvas then queries this class
 ** for the placement and appearence data it needs to render the UI.
 **
 ** It also manipulates agents internally through their BaseAgentIcons,
 ** only pulling them out when it needs to perform some sort of operation
 ** on them.
 **/

package net.hivecell.hive.agent.hiveui.iconicui;

import java.awt.*;
import java.util.*;

import net.hivecell.hive.agent.*;


public abstract class BaseLayoutManager {
    
    protected Vector agents;            // BaseAgentIcons being layed out by the layout manager
    protected Hashtable agentIcons;     // Hash Agents -> BaseAgentIcons
    
    public BaseLayoutManager() {
	agents = new Vector();
	agentIcons = new Hashtable();
    }

    /** If using a fully animated UI, these should call update, otherwise only add single agent. */
    public abstract BaseAgentIcon arrangeAddAgent(Agent agent, Image image);
    public abstract BaseAgentIcon arrangeRemoveAgent(Agent agent);

    /** Handles graphics of (dis)connecting agents, not server side. */
    public abstract BaseConnection arrangeConnectAgents(Agent from, Agent to);
    public abstract BaseConnection arrangeDisconnectAgents(Agent from, Agent to);

    /** Handle a message between agents */
    public abstract void arrangeAgentMessage(Agent from, Agent to);
    
    /** Handle an agent moving around on the screen. */
    public abstract void arrangeMoveAgentIcon(Agent a, int nx, int ny);
    
    /**
     ** To be called only when the entire UI needs to be updated.
     ** Ordinarily should only be called on refresh, but if a fully
     ** animated UI is being used (eq. relaxation layout), then it
     ** will be called every timing interval.
     **/
    public abstract void update();


    /**
     ** Returns the location of Agent which's icon.
     ** Called in the BaseUICanvas when it needs to render the screen.
     **
     ** @param which the agent whose position is needed.
     **/
    public Point getLocation(Agent which) { return ( ((BaseAgentIcon)agentIcons.get(which)).getLocation() ); }


    /**
     ** Returns the icon at Point where.
     ** Mostly using mouse coordinates to find an icon.
     **
     ** @param where The point at which to check for an agent's existence
     **/
    public BaseAgentIcon getIconAtLocation(int x, int y) { return ( getIconAtLocation(new Point(x, y)) ); }
    public BaseAgentIcon getIconAtLocation(Point where) { 
	
	// Go backwards, because newest Agents are rendered on top of older ones.
	for (int i = agents.size() - 1; i >= 0; i--)
	    {
		if ( ((BaseAgentIcon)agents.elementAt(i)).contains(where) )
		    return ((BaseAgentIcon)agents.elementAt(i));
	    }
	return (null);
    }

    /** 
     ** Returns agent's icon as an AWT Image.
     **
     ** @param agent The agent whose icon image is needed.
     **/
    public BaseAgentIcon getIconForAgent(Agent agent) { return ( (BaseAgentIcon)agentIcons.get(agent) ); }

    
    /**
     ** Returns a vector of the agent icons for the agents being monitored by this layout manager.
     **/
    public Vector getAgentIcons() { return agents; }
}
