// $Id: BaseUICanvas.java,v 1.1 2000/05/18 22:00:53 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

/**
 ** Base class for UICanvas for a "Boxes and Lines" UI.
 ** This class is responsible for rendering the graphics 
 ** as dictated by the Layout manager.
 **
 ** The layout manager determines where the icons and lines go, 
 ** the UICanvas decides how to render that information and 
 ** how frequently to render it.
 **
 ** This is also the class that handles the mouse listening.
 **
 ** An instance of a sub-class of this should be in an agent 
 ** that is a sub-class of a GUIAgent, eg. newAWTUIAgentImpl (whenever it gets written).
 **/

package net.hivecell.hive.agent.hiveui.iconicui;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;

import net.hivecell.hive.Global;
import net.hivecell.hive.agent.*;
import net.hivecell.hive.agent.hiveui.BaseUIAgentImpl;
import net.hivecell.hive.support.*;


public abstract class BaseUICanvas 
extends BaseCanvas
implements MouseListener, MouseMotionListener, ImageObserver {
 
    /**
     * some constants
     */
    static final int updateInterval = 150;
    static final int normalIconHeight = 32, normalIconWidth = 32;
    static final int iconBorderWidth = 3;


    protected BaseLayoutManager layout;
    protected BaseUIAgentImpl uiAgent;


    // menu state command used for connecting and disconnecting agents
    int menuState = 0; // 0 when nothing is happening, 1 for connections, 2 for dis
    BaseAgentIcon connectFromIcon = null; // used for the clicking action of connect and disconnect

    // state used to handle mouse dragging
    int mouseButtonForDrag;                         // 0, 1, or 2.
    BaseAgentIcon selectedIcon = null;
    Point iconDragOffset, mousePressLocation, lastDragLocation;  // locations
    Graphics dragGraphics;                          // graphics for drawing

    // Animation stuff for arrows and/or icons.
    static int timer = 0;


    public BaseUICanvas(BaseUIAgentImpl uiAgent, BaseLayoutManager layout) {
	super();
	this.uiAgent = uiAgent;
	this.layout = layout;
    }

    /** 
     ** Put the UICanvas into connect mode.
     **/
    public void setConnectMode() {
	menuState = 1; // setting the agent graph canvas for a connection
	connectFromIcon = selectedIcon; // this is where the connection is coming from
    }
    /** 
     ** Put the UICanvas into disconnect mode.
     **/
    public void setDisconnectMode() {
	menuState = 2; // setting the agent graph canvas for a connection
	connectFromIcon = selectedIcon; // this is where the connection is coming from
    }


    /** Get the currently selected icon */
    public BaseAgentIcon getSelectedIcon() { return (selectedIcon); }

    /**
     ** This method queries layout and draws the
     ** agent icons and arrows as directed by the layout
     **/
    public abstract void render(Graphics g);



    /** Mouse handling code */
    // Mouse pressed - start a drag here. (This is where we should put in
    // the click-an-agent-to-edit stuff)
    public void mousePressed(MouseEvent event) {
	// Figure out which mouse button it was from the meta/alt flags.
	if (event.isAltDown()) {
	    mouseButtonForDrag = 1;
	} else if (event.isMetaDown()) {
	    mouseButtonForDrag = 2;
	} else {
	    mouseButtonForDrag = 0;
	}

	// Figure out which icon was pressed.
	selectedIcon = layout.getIconAtLocation(event.getX(), event.getY());

	if ( selectedIcon == null ) {
	    // Whoops, mouse missed.
	    Debug.println(Debug.NOTICE, "There's nothing at (" + event.getX() + ", " + event.getY() + ")" );
	    iconDragOffset = null;
	    dragGraphics = null;
	    return;

	} else if ( selectedIcon != null ) {

	    // OK, we got a click and it was an agent. Now then.. stores some positions
	    mousePressLocation = new Point( event.getX(), event.getY() );
	    lastDragLocation = new Point( event.getX(), event.getY() );
	    // Figure out where the user clicked inside the icon
	    iconDragOffset = new Point( event.getX() - selectedIcon.getX(),
					event.getY() - selectedIcon.getY() );

	    // Do something based on which mouse button was pressed.
	    if (mouseButtonForDrag == 0) {                       
		// left button -- ok -- if menuState is non-zero, then we
		// need to do something, else we need to move.
		if( menuState == 0 ) {
		    // Set up XOR drawing for frame
		    dragGraphics = this.getGraphics();
		    dragGraphics.setXORMode(this.getBackground());
		
		    // draw an XORed rectangle in the right spot
		    dragGraphics.drawRect(lastDragLocation.x - iconDragOffset.x,
					  lastDragLocation.y - iconDragOffset.y,
					  selectedIcon.getCurrentSize().width,
					  selectedIcon.getCurrentSize().height);


		} else if( menuState == 1 ) {
		    // we need to do a connect
		    uiAgent.connectAgentsCommand( connectFromIcon.getAgent(), selectedIcon.getAgent() );

		    menuState = 0;
		    connectFromIcon = null;

		} else if( menuState == 2 ) {
		    // we need to disconnect
		    uiAgent.disconnectAgentsCommand( connectFromIcon.getAgent(), selectedIcon.getAgent() );

		    menuState = 0;
		    connectFromIcon = null;

		}

	    } else if (mouseButtonForDrag == 2) { // Handle Popup menu stuff.
		
		String cname = selectedIcon.getAgent().getClass().getName();
		String aname = cname.substring(cname.lastIndexOf('.') + 1);
		PopupMenu pm = new PopupMenu( aname );
		
		if( Global.windowsOS == true ) {     // hack stupid Windows
		    pm.add( aname );                 // JDK bug in popups
		    pm.add( "-" );
		}
		
		pm.add( uiAgent.getAgentAddress(selectedIcon.getAgent()).toString() );
		pm.addSeparator();
		
		String cmds[] = uiAgent.getMenuNames( selectedIcon.getAgent() );
		for (int i = 0; i < cmds.length; i++)
		    pm.add(cmds[i]);
		pm.addActionListener( (ActionListener)uiAgent );
		add(pm);

		
		if (cmds.length > BaseUIAgentImpl.defaultCmds.length)  // More commands than default commands, put in a separator.
		    pm.insertSeparator(cmds.length + 1);

		pm.show(this, event.getX(), event.getY());


	    } else if (mouseButtonForDrag == 1) {                // right button, draw lines
		dragGraphics = this.getGraphics();
		dragGraphics.setColor(Color.red);
		dragGraphics.setXORMode(this.getBackground());
	    } else {
		Debug.error("Unknown mouse button.  What the hell is going on?");     // shouldn't reach here.
	    }

	}

    }

    // When the user lets go of the mouse, then we handle the drag
    public void mouseReleased(MouseEvent event) {
	if (selectedIcon == null)
	    return;

	// Find the coordinates of the upper left hand corner
	int xpt = event.getX() - iconDragOffset.x;
	int ypt = event.getY() - iconDragOffset.y;
	Dimension dim = new Dimension(selectedIcon.getCurrentSize());

	if (xpt + dim.width < 4 || xpt > this.getSize().width - 4 ||
	    ypt + dim.height < 4 || ypt > this.getSize().height - 4 )
	    Debug.println(Debug.NOTICE, "Dragged off screen, ignored.");
	else {

	    if (mouseButtonForDrag == 0) {              // button 0 - move icon
		selectedIcon.move(event.getX() - iconDragOffset.x,
				    event.getY() - iconDragOffset.y);
	    } else if (mouseButtonForDrag == 2) {       // button 2 - popup did this
	    } else if (mouseButtonForDrag == 1) {       // button 1 - draw link
		BaseAgentIcon destIcon = layout.getIconAtLocation(event.getX(), event.getY());
		if (destIcon == null) {
		    Debug.println(Debug.NOTICE, "There's nothing at " + event.getX() + "," + event.getY());
		    // erase the last line
		    dragGraphics.drawLine(mousePressLocation.x, mousePressLocation.y,
					  lastDragLocation.x, lastDragLocation.y);
		} else {
		    uiAgent.connectAgentsCommand(selectedIcon.getAgent(), destIcon.getAgent());
		}
	    } else {
		Debug.assert(false);
	    }
	}

	// valid drag - let's take action
	if (dragGraphics != null) {
	    dragGraphics.dispose();
	    dragGraphics = null;
	}
	this.render();                                // always redraw
    }


    // as the mouse is dragged, draw whatever is appropriate (uses XOR)
    // this interacts badly with double buffering
    public void mouseDragged(MouseEvent e) {
	if (selectedIcon == null)
	    return;
	if (mouseButtonForDrag == 0) {
	    // erase the last rectangle
	    dragGraphics.drawRect(lastDragLocation.x - iconDragOffset.x,
				  lastDragLocation.y - iconDragOffset.y,
				  selectedIcon.getCurrentSize().width,
				  selectedIcon.getCurrentSize().height);

	    // update the point
	    lastDragLocation.x = e.getX();
	    lastDragLocation.y = e.getY();
	    // and draw the new rectangle
	    dragGraphics.drawRect(lastDragLocation.x - iconDragOffset.x,
				  lastDragLocation.y - iconDragOffset.y,
				  selectedIcon.getCurrentSize().width,
				  selectedIcon.getCurrentSize().height);

	} else if (mouseButtonForDrag == 2) {
	    // do nothing (popup)
	} else if (mouseButtonForDrag == 1) {
	    // erase the last line
	    dragGraphics.drawLine(mousePressLocation.x, mousePressLocation.y,
				  lastDragLocation.x, lastDragLocation.y);
	    lastDragLocation.x = e.getX();
	    lastDragLocation.y = e.getY();
	    // and draw the new line
	    dragGraphics.drawLine(mousePressLocation.x, mousePressLocation.y,
				  lastDragLocation.x, lastDragLocation.y);
	} else {
	    Debug.assert(false);
	}
    }

    /** We don't care about these events */
    public void mouseClicked(MouseEvent event) {}
    public void mouseEntered(MouseEvent event) {}
    public void mouseExited(MouseEvent event) {}
    public void mouseMoved(MouseEvent event) {}
    
}
