// $Id: IconicUICanvas.java,v 1.1 2000/05/18 22:00:53 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

/**
 ** UICanvas for a "Boxes and Lines" UI.
 ** This class is responsible for rendering the graphics 
 ** as dictated by the Layout manager.
 **
 ** The layout manager determines where the icons and lines go, 
 ** the UICanvas decides how and when to render that information.
 **
 ** This is also the class that handles the mouse listening.
 **
 ** An instance of a sub-class of this should be in an agent 
 ** that is a sub-class of a GUIAgent, eg. newAWTUIAgentImpl (whenever it gets written).
 **/


package net.hivecell.hive.agent.hiveui.iconicui;

import java.awt.*;
import java.awt.image.*;
import java.awt.event.*;
import java.util.*;

import net.hivecell.hive.agent.*;
import net.hivecell.hive.agent.hiveui.BaseUIAgentImpl;
import net.hivecell.hive.support.*;
import net.hivecell.hive.agent.cell.CellAgent;

public class IconicUICanvas extends BaseUICanvas {

    private Thread updateThread = null;

    // Let the Canvas know it shouldn't start any new renders, as its agent is being killed.
    protected boolean dontRender = false;
    // Animater the icons, or not?
    protected boolean animIcons = true;

    // For the animated UIIcon
    protected Hashtable smallIcons;         // Cache of the small icons    
    protected Dimension smallSize;          // size of the small icons


    public IconicUICanvas( BaseUIAgentImpl uiAgent, BaseLayoutManager aLayout ) {
	super(uiAgent, aLayout);
 
	setSize(getPreferredSize());                  // set up our size
	this.setDoubleBuffering(false);               // double buffering is buggy; doesn't play thread-safe...
	setVisible(true);                             // map, please

	addMouseListener(this);                       // arrange for mouse events
	addMouseMotionListener(this);

	smallIcons = new Hashtable();
	smallSize  = new Dimension( (int)(Math.pow(BaseAgentIcon.defaultIconSize.width, 2)  / getPreferredSize().width) + 1,
				    (int)(Math.pow(BaseAgentIcon.defaultIconSize.height, 2) / getPreferredSize().height) + 1);
	
	updateThread = new Thread( new Runnable() {
	    public void run() {
		while( true ) {
		    // First, increment the timer
		    timer++;
		    updateArrows();       // Update the arrow animations
		    if (animIcons)
		        updateIcons();    // Update the icon animations
		    layout.update();      // Update the layout animations (ex. Agent relaxation placement)

		    // Sleep a bit
		    try {
			Thread.sleep( updateInterval );
		    } catch( InterruptedException ex ) { 
			Debug.println( Debug.ERROR, ex ); 
		    }
		}
	    }
	});

	updateThread.start();
    }

    public void setAnimateIcons(boolean val) { animIcons = val; }

    /**
     ** This method queries layout and draws the
     ** agent icons and arrows as directed by the layout.
     **/
    public void render(Graphics g) {
	if (g == null) return;  // Has the uiAgent been stopped, or has the frame been killed (g == null)?
	//System.out.print("Rendering...\n");

	// Clear the buffer
	Color oldColor = g.getColor();
	g.setColor(getBackground());
	g.fillRect(0, 0, getSize().width, getSize().height);
	g.setColor(oldColor);

	// Re-draw the icons
	Enumeration e = layout.getAgentIcons().elements();
	while (e.hasMoreElements()) {
	    BaseAgentIcon b = (BaseAgentIcon)e.nextElement();

	    b.draw(g, this);

	    // Re-draw the arrows
	    Enumeration c = b.getConnections().elements();
	    while (c.hasMoreElements()) {
		GraphicalConnectionArrow gca = (GraphicalConnectionArrow)c.nextElement();
		
		gca.draw(g);
	    }
	}

	if (animIcons) // If not being called from the MirrorIcon, and animations are on, render the IcUIAg icon.
	    renderUIIcon();
    }



    /** Tell the canvas to stop the animation thread. */
    public void stopAnim() { updateThread.stop(); }//tGroup.stop(); }
    /** Tell the canvas not to render anything else. */
    public void stopRenders() { dontRender = true; }

    /** Updates the arrows in the animation/color fading thread (See constructor) */
    private void updateArrows() {

	Enumeration e = layout.getAgentIcons().elements();
	while (e.hasMoreElements()) {
	    Enumeration c = ((BaseAgentIcon)e.nextElement()).getConnections().elements();
	    while (c.hasMoreElements()) {
		((GraphicalConnectionArrow)c.nextElement()).update(getGraphics());
	    }
	}
    }

    private void updateIcons() {

	Enumeration e = layout.getAgentIcons().elements();
	while (e.hasMoreElements()) {
	    BaseAgentIcon b = (BaseAgentIcon)e.nextElement();
	    b.update(getGraphics(), this);
	}
    }



    // Overriden abstract methods.
    /** Render a single icon */
    public void renderAgentIcon(BaseAgentIcon i) {
	i.draw(getGraphics(), this);
    }
    /** Render a single arrow */
    public void renderArrow(GraphicalConnectionArrow a) {
	a.draw(getGraphics());
    }


    // Get the default size of the canvas.
    public Dimension getMinimumSize() {
	//return new Dimension(12 * BaseAgentIcon.defaultIconSize.width, 
	//		     7 * BaseAgentIcon.defaultIconSize.height);
	//15
	return new Dimension(14 * BaseAgentIcon.defaultIconSize.width, 
			     10 * BaseAgentIcon.defaultIconSize.height);
    }

    public Dimension getPreferredSize() {
	return getMinimumSize();
    }


    // Implement hideous hack to get IconicUIIcon to display the canvas:
    // This is the thread that does the drawing in the background
    public void renderUIIcon() {
	BaseAgentIcon icUIAgIc = layout.getIconForAgent(uiAgent);

	// The agent hasn't finished loading yet, or the canvas hasn't initialized yet. This should be removed once timing is fixed
	// We need to have some sort of sequence of locks in place so that the UIAgent doesn't try to render anything until
	// the canvas has been initialized.
	if (icUIAgIc == null || this.getSize().width == 0) return;

	// buffer image to prevent flickering
	Image image = createImage(BaseAgentIcon.defaultIconSize.width,
				  BaseAgentIcon.defaultIconSize.height );
	Graphics imageG = image.getGraphics();     // Graphics for the buffer image


	// Clear the buffer image
	imageG.setColor(Color.white);
	imageG.fillRect(0, 0, BaseAgentIcon.defaultIconSize.width, BaseAgentIcon.defaultIconSize.height);


	// Add any new icons that have appeared since last render
	Enumeration e = layout.getAgentIcons().elements();
	while (e.hasMoreElements()) {

	    BaseAgentIcon b = (BaseAgentIcon)e.nextElement();
	    SmallAgentIcon smalli = null;

	    if (!smallIcons.containsKey(b)) {       // Need to create a new small icon
		smalli = new SmallAgentIcon(b.getBorderColor(), smallSize);
		smallIcons.put(b, smalli);

	    } else smalli = (SmallAgentIcon)smallIcons.get(b);

	    // Scale the position of the small agent
	    smalli.move( (b.getX() * BaseAgentIcon.defaultIconSize.width) / this.getSize().width,
			 (b.getY() * BaseAgentIcon.defaultIconSize.height) / this.getSize().height );
	    
	}


	// Remove any old icons that have disappeared since last render
	if (smallIcons.size() != layout.getAgentIcons().size()) {
	    e = smallIcons.keys();
	    while (e.hasMoreElements()) {
		BaseAgentIcon sic = (BaseAgentIcon)e.nextElement();
		if ( !layout.getAgentIcons().contains(sic) )
		    smallIcons.remove(sic);
	    }
	}


	// Go through and draw the lines before the arrows so that the icons are on top
	e = layout.getAgentIcons().elements();
	while (e.hasMoreElements()) {
	    BaseAgentIcon b = (BaseAgentIcon)e.nextElement();

	    // Draw the arrows first so the icons will be on top
	    Enumeration c = b.getConnections().elements();
	    while (c.hasMoreElements()) {
		GraphicalConnectionArrow gca = (GraphicalConnectionArrow)c.nextElement();
		BaseAgentIcon to, from;
		to = gca.getToIcon();
		from = gca.getFromIcon();

		SmallAgentIcon toS, fromS;
		toS = (SmallAgentIcon)smallIcons.get(to);
		fromS = (SmallAgentIcon)smallIcons.get(from);
		if( ( toS != null ) && ( fromS != null ) ) {
		    imageG.setColor(Color.lightGray);
		    imageG.drawLine(fromS.getCenter().x, fromS.getCenter().y, toS.getCenter().x, toS.getCenter().y);
		}
		
	    }
	}
	
	// Go through and draw the shrunken icons
	e = smallIcons.keys();
	while( e.hasMoreElements() ) {
	    SmallAgentIcon smi = (SmallAgentIcon)smallIcons.get( e.nextElement() );
	    if( smi != null )
		smi.draw( imageG );
	}
	
	icUIAgIc.setImage(image);
	icUIAgIc.draw(getGraphics(), this);
    }
}


class SmallAgentIcon {
    
    protected Color col;
    protected int x, y;
    protected Dimension size;

    public SmallAgentIcon(Color c, Dimension s) {
	col = c;
	size = s;
    }

    public void move(int xp, int yp) {
	x = xp;
	y = yp;
    }

    public Point getCenter() { return ( new Point(x + (size.width / 2), y + (size.height / 2)) ); }

    public void draw(Graphics g) {
	g.setColor(col);
	g.fill3DRect(x, y, size.width, size.height, true );
    }

}
