// $Id: IconicUILayoutManager.java,v 1.1 2000/05/18 22:00:53 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

/**
 ** The default layout manager for IconicUIAgent.
 ** This class is responsible for updating the graphics whenever 
 ** any of the following actions takes place:
 **
 ** an agent is added/removed
 ** two agents are connected/disconnected
 ** an agent (icon) is moved
 **
 ** In general: whenever a "box" or "line" needs to be rearranged 
 ** for any reason, this is the class that determines where it
 ** goes and how it looks.  The BaseUICanvas then queries this class
 ** for the placement and appearence data it needs to render the UI.
 **
 ** It also manipulates agents internally through their BaseAgentIcons,
 ** only pulling them out when it needs to perform some sort of operation
 ** on them.
 **/
package net.hivecell.hive.agent.hiveui.iconicui;

import java.io.*;
import java.awt.*;
import java.util.*;
import java.lang.reflect.*;

import net.hivecell.hive.support.*;
import net.hivecell.hive.agent.*;

public class IconicUILayoutManager 
extends BaseLayoutManager {

    /**
     * this is a hash of all the icons that were ever rendered to this
     * and a Point object reflecting where we placed them -- this is
     * just so if the icon disappears and reappears, it comes back in
     * the same place 
     */
    private Hashtable icons;

    protected Dimension grid;
    protected int xspacing = 25, yspacing = 25;
    
    protected int lx = 0, ly = 0;  // Location last Icon was dropped.


    public IconicUILayoutManager() { 

	super(); 

	grid = new Dimension( 12 * BaseAgentIcon.defaultIconSize.width, 7 * BaseAgentIcon.defaultIconSize.height ); 

	icons = new Hashtable();

    }

    /**
     * put a new agent into the system -- this method figures out its
     * placement and wheterh it is animated and stuff like that 
     *
     * @param agent the agent that we are adding 
     * @param image the image that we are displaying for the agent
     * @return the base agent icon for the agent
     */
    public BaseAgentIcon arrangeAddAgent( Agent agent, Image image ) { 

	BaseAgentIcon a = null;

	if( icons.containsKey( agent ) )
	    // if we have this one cached, then just use it
	    a = (BaseAgentIcon)icons.get( agent );

	else {
	    // create a new agent icon
	    a = new BaseAgentIcon( agent, image );
	    
	    if( lx + a.getCurrentSize().width + 10 < grid.width ) {
		a.move( lx + xspacing, ly + yspacing );  // Do better layout later.
		lx += a.getCurrentSize().width + xspacing;
	    } else {
		lx = 0;
		ly += a.getCurrentSize().height + yspacing;
		a.move( lx + xspacing, ly + yspacing );  // Do better layout later.
		lx += a.getCurrentSize().width + xspacing;
	    }
	    icons.put( agent, a );
	}	    

	// Now, see if it has an animated icon...
	BaseAgentIcon aic = hasAnimatedIcon( agent, image );
	if( aic == null ) {
	    agents.addElement( a );
	    agentIcons.put( agent, a );
	    return a;

	} else {
	    aic.move( a.getLocation() );
	    agents.addElement( aic );
	    agentIcons.put( agent, aic );
	    return aic;

	}
	
    }

    /**
     * remove an icon from the display system
     *
     * @param agent the agent that we wish to remove
     * @return the base agent icon that we removed or null if we don't have this agent
     */
    public BaseAgentIcon arrangeRemoveAgent( Agent agent ) { 

	if( agentIcons.containsKey( agent ) ) {
	    BaseAgentIcon a = (BaseAgentIcon)agentIcons.get( agent );
	    agents.removeElement( a );
	    return ((BaseAgentIcon)agentIcons.remove( agent ));

	} else {
	    Debug.warning( "Asked to remove agent '" + agent + "', which I'm not in charge of!" );
	    return null;
	}

    }

    /** 
     * handles the connection of two agents
     *
     * @param from the agent that we want to remove the connection from
     * @param to the agent that the connection goes to
     * @return the base connection object that represents the connection
     */
    public BaseConnection arrangeConnectAgents( Agent from, Agent to ) { 

	GraphicalConnectionArrow a = new GraphicalConnectionArrow( (BaseAgentIcon)agentIcons.get( from ),
								   (BaseAgentIcon)agentIcons.get( to ) );
	((BaseAgentIcon)agentIcons.get( from )).addConnection( a );
	return a;

    }

    /**
     * handles the disconnection of two agents
     *
     * @param from the agent that the connection is originating from
     * @param to the agent that the connection is going to
     * @return the base connection object that we are removing, or null if we don't have the pointer
     */
    public BaseConnection arrangeDisconnectAgents( Agent from, Agent to ) { 

	if (agentIcons.containsKey(from)) {
	    BaseAgentIcon fIcon = (BaseAgentIcon)agentIcons.get( from );
	    Enumeration e = fIcon.getConnections().elements();
	    while( e.hasMoreElements() ) {
		BaseConnection b = (BaseConnection)e.nextElement();
		if ( b.getToIcon().getAgent().equals( to ) ) {
		    fIcon.removeConnection( b );
		    return b;
		}
	    }
	    return null;  // Not connected to anything.
	}
	else {
	    Debug.warning( "Asked to disconnect '" + from.getClass().getName() + "', which I'm not in charge of, and '" + to.getClass().getName() + "'." );
	    return null;
	}

    }
    
    /**
     * arrange an agent message to be rendered going between two icons
     *
     * @param from the agent that the message is coming from
     * @param to the agent that the message is going to
     */
    public void arrangeAgentMessage( Agent from, Agent to ) {
	BaseAgentIcon toIcon = (BaseAgentIcon)agentIcons.get( to );
	if (toIcon == null) {
  	    Debug.warning("Icon for agent '" + from + "' not found. Not going to render message.  (Probably just hasn't been noticed by the ui yet.)");
  	    return;
  	}

	Enumeration e = toIcon.getConnections().elements();
	while( e.hasMoreElements() ) {
	    GraphicalConnectionArrow b = (GraphicalConnectionArrow)e.nextElement();
	    if ( b.getToIcon().getAgent().equals( from ) ) {
		b.message( b.getToIcon(), toIcon );
		return;
	    }
	}
    }

    /**
     * handle moving an agent around the screen. since this layout
     * manager won't be refreshing their locations, i'm going to hash
     * away the new location for this icon
     *
     * @param a the agent that we are moving
     * @param nx the new x location
     * @param ny the new y location 
     */
    public void arrangeMoveAgentIcon( Agent a, int nx, int ny ) {

	((BaseAgentIcon)agentIcons.get( a )).move( nx, ny );
    }
    
    /**
     * to be called only when the entire ui needs to be updated.
     * ordinarily should only be called on refresh, but if a fully
     * animated ui is being used (eq. relaxation layout), then it
     * will be called every timing interval.
     */
    public void update() { }


    private BaseAgentIcon hasAnimatedIcon(Agent a, Image i) {
	// Get the name of the agent's class:
	String fName = a.getClass().getName();
	String jName = fName.substring(fName.lastIndexOf('.') + 1) + "Icon.java";
	String cName = fName + "Icon";

	// See if it has an Icon .java file (implying a class)
	InputStream is = a.getClass().getResourceAsStream(jName);

	// if it does, return a new icon of that type.
	if (is != null) {
	    Debug.failure(cName + " Does exist.");

	    // Use some reflection.
	    Class aicType = null, argTypes[] = null;
	    Constructor aicConst;
	    Object args[] = new Object[] { a, i };

	    try {

		aicType = Class.forName(cName);
		argTypes = new Class[] { Agent.class, Image.class };
		aicConst = aicType.getConstructor(argTypes);

		return ( (BaseAgentIcon)makeObject(aicConst, args) );

	    } catch(ClassNotFoundException e) {
		Debug.error("AnimatedIcon class '" + cName + "' not found.", e);
	    } catch (NoSuchMethodException e) {
		Debug.error("Need to have a constructor accepting an Agent and an Image in the AnimatedIcon subclass '" + cName + "'.",e);
	    }
	}

	return (null);
    }
    private Object makeObject(Constructor constructor, Object[] arguments) {
	
	System.out.println ("Constructor: " + constructor.toString());
	Object object = null;
	
	try {
	    object = constructor.newInstance(arguments);
	    System.out.println ("Object: " + object.toString());
	    return object;
	} catch (InstantiationException e) {
	    Debug.error("Exception while creating instance of type AnimatedIcon.", e);
	} catch (IllegalAccessException e) {
	    Debug.error("Exception while creating instance of type AnimatedIcon.", e);
	} catch (IllegalArgumentException e) {
	    Debug.error("Exception while creating instance of type AnimatedIcon.", e);
	} catch (InvocationTargetException e) {
	    Debug.error("Exception while creating instance of type AnimatedIcon.", e);
	}
	return object;
    }

}

