// $Id: GraphTreeFilterer.java,v 1.1 2000/05/18 22:00:57 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.agent.hiveui.treeui;

import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.agent.hiveui.BaseUIAgentImpl;
import net.hivecell.hive.agent.hiveui.BaseGUIAgentImpl;
import net.hivecell.hive.agent.hiveui.BaseSwingGUIAgentImpl;
import net.hivecell.hive.support.AgentGraph;
import net.hivecell.hive.support.CellAddress;
import net.hivecell.hive.support.Debug;

import java.awt.Component;
import java.awt.IllegalComponentStateException;
import java.awt.Graphics;
import java.awt.Color;
import java.awt.Image;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import java.awt.event.MouseEvent;
import java.awt.event.MouseListener;
import java.io.InputStream;
import java.io.InputStreamReader;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;
import javax.swing.ImageIcon; 
import javax.swing.JPopupMenu; 
import javax.swing.JTree;
import javax.swing.event.TreeExpansionEvent;
import javax.swing.event.TreeExpansionListener;
import javax.swing.event.TreeSelectionEvent;
import javax.swing.event.TreeSelectionListener;
import javax.swing.tree.DefaultMutableTreeNode;
import javax.swing.tree.DefaultTreeCellRenderer;
import javax.swing.tree.TreePath;
import javax.swing.tree.TreeSelectionModel;


/**
 * this interface is needed so a "pluggable" object can be created to
 * query the agent monitoring agent for information and convert it to
 * a DefaultMutableTreeNode for display on the tree 
 */
abstract class GraphTreeFilterer {

    /**
     * our hook back to our base gui agent so we can make use of its
     * serivces to get information on the agents 
     */
    protected BaseGUIAgentImpl bga;

    /**
     * the jtree object that anything that extends this should be
     * updating.  this is usefful because if somebody doesn't care
     * about certain server events and does not care to render them,
     * then this JTree will be returned by default 
     */
    protected JTree jtree;

    /**
     * our mapping between agents and their popup menus
     */
    protected Hashtable agentMenus;

    /**
     * our list of nodes that are expanded
     */
    private Vector expandedNodes;

    public GraphTreeFilterer( BaseGUIAgentImpl bga ) {

	this.bga = bga;	

	jtree = new JTree( new DefaultMutableTreeNode( "Cell Addresses" ) );
	agentMenus = new Hashtable();
	expandedNodes = new Vector();

    }

    public JTree renderAgentGraph( AgentGraph ag ) { return jtree; }
    public JTree renderNewAgent( Agent a ) { return jtree; }
    public JTree renderRemoveAgent( Agent a, CellAddress ca, Agent[] incoming, Agent[] outgoing ) { return jtree; }
    public JTree renderConnectAgents( Agent s, Agent d ) { return jtree; }
    public JTree renderDisconnectAgents( Agent s, Agent d ) { return jtree; }
    public JTree renderDisconnectFromAll( Agent a, Agent[] incoming, Agent[] outgoing ) { return jtree; }
    public JTree renderAgentMessage( Agent s, Agent d ) { return jtree; }
    public JTree renderCellReport( String message ) { return jtree; }

    /**
     * this class needs to be created to provide a mapping between a
     * tree node and the agent that it is supposed to be representing 
     */
    public abstract Agent getAgentNode( DefaultMutableTreeNode dmtn );

    /**
     * creates the popup menu associated with the agent
     * 
     * @param a the agent that the menu is created for
     * @return the PopupMenu for that agent 
     */
    protected JPopupMenu getAgentPopupMenu( final Agent a ) {

	if( agentMenus.containsKey( a ) )
	    return (JPopupMenu)agentMenus.get( a );

	String popupTitle = bga.getAgentName( a );
	JPopupMenu pm = new JPopupMenu( popupTitle );

	String[] menuItemNames = bga.getMenuNames( a );
	for( int count=0;count < menuItemNames.length;count++ ) {
	    final String menuItemName = menuItemNames[count];
	    pm.add( menuItemName ).addActionListener( new ActionListener() {
		public void actionPerformed( ActionEvent ae ) {
		    bga.invokeMenuCommand( a, menuItemName );
		}
	    });
	}

	agentMenus.put( a, pm );

	return pm;
    }

    /**
     * make sure the appropriate nodes are expanded
     *
     * @param tree the tree to expand and collapse nodes on
     */
    protected synchronized void expandTree( JTree tree ) {

	Vector checkedNodes = new Vector();

	// first filter out which paths are actually in this tree
	Enumeration e = expandedNodes.elements();
	while( e.hasMoreElements() ) {
	    TreePath tp = (TreePath)e.nextElement();
	    Object[] o = tp.getPath();
	    boolean inTree = false; 
	    DefaultMutableTreeNode position = (DefaultMutableTreeNode)tree.getModel().getRoot();
	    for( int count=1;count < o.length;count++ ) {
		if( count == o.length - 1 ) {
		    inTree = true;
		    break;
		}
		if( position.isNodeChild( (DefaultMutableTreeNode)o[count] ) )
		    position = (DefaultMutableTreeNode)o[count];
	    }
	    if( inTree )
		checkedNodes.addElement( tp );
	}

	expandedNodes = checkedNodes;
	e = expandedNodes.elements();
	while( e.hasMoreElements() )
	    tree.expandPath( (TreePath)e.nextElement() );

    }

    /**
     * add the listeners for the mouse following and for the popup menus
     *
     * @param tree teh JTree to add teh listenres to
     */
    protected void addListeners( final JTree jtree ) {


	// watch the tree for expansions and collapses -- this way we
	// know what to expand and contract later
	jtree.addTreeExpansionListener( new TreeExpansionListener() {
	    public void treeExpanded( TreeExpansionEvent event ) {
		expandedNodes.addElement( event.getPath() );
	    }

	    public void treeCollapsed( TreeExpansionEvent event ) {
		expandedNodes.removeElement( event.getPath() );
	    }
	});

	// set up the ability to select on the agents
	jtree.addMouseListener( new MouseListener() {
	    public void mouseClicked( MouseEvent me ) {
		int selectedRow = jtree.getRowForLocation( me.getX(), me.getY() );
		TreePath selectedPath = jtree.getPathForLocation( me.getX(), me.getY() );
		jtree.setSelectionPath( selectedPath );
		if( me.isMetaDown() && ( selectedRow != -1 ) && ((DefaultMutableTreeNode)selectedPath.getLastPathComponent()).isLeaf() ) {
		    JPopupMenu jmenu = getAgentPopupMenu( getAgentNode( (DefaultMutableTreeNode)selectedPath.getLastPathComponent() ) );
		    try {
			if( jmenu != null ) jmenu.show( jtree, me.getX(), me.getY() );
		    } catch( IllegalComponentStateException error ) {
			// this is just to safeguard in the case that
			// we are attempting to show a popup menu for
			// something that we just removed from the
			// system
		    }
		}

	    }

	    public void mouseEntered( MouseEvent me ) { }
	    public void mouseExited( MouseEvent me ) { }
	    public void mousePressed( MouseEvent me ) { }
	    public void mouseReleased( MouseEvent me ) { }
	});


	class TreeUIRenderer 
	extends DefaultTreeCellRenderer {
	    public Component getTreeCellRendererComponent( JTree tree, Object value, boolean sel, boolean expanded, boolean leaf, int row, boolean hasFocus) {
		
		super.getTreeCellRendererComponent( tree, value, sel, expanded, leaf, row, hasFocus ); 
		Agent a = getAgentNode( (DefaultMutableTreeNode)value );
		if( a != null ) {
		    Image img = jtree.createImage(16, 16);
		    Graphics g = img.getGraphics();
		    g.setColor(Color.white);
		    g.drawRect(0, 0, 16, 16);
		    g.drawImage( bga.getAgentIcon( a ).getImage().getScaledInstance( 14, 14, Image.SCALE_FAST ),
				 1, 1, jtree );
		    
		    g.dispose();
		    setIcon( new ImageIcon( img ) );
		}
		return this;

	    }
	}
	jtree.setCellRenderer( new TreeUIRenderer() );
 
    }

}
