// $Id: AutoMonitoringAgentImpl.java,v 1.2 2000/05/19 19:55:58 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.agent.service;

import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.agent.AgentImpl;
import net.hivecell.hive.agent.AgentInitializationException;
import net.hivecell.hive.agent.EventReceivingAgentImpl;
import net.hivecell.hive.agent.cell.AgentMonitoringAgent;
import net.hivecell.hive.agent.cell.CellListAgent;
import net.hivecell.hive.agent.cell.CellListInfo;
import net.hivecell.hive.agent.cell.UnableToMonitorCellException;
import net.hivecell.hive.event.HiveEvent;
import net.hivecell.hive.event.cell.CellEvent;
import net.hivecell.hive.event.cell.CellBornCellEvent;
import net.hivecell.hive.event.cell.CellGoingDownCellEvent;
import net.hivecell.hive.event.cell.CellWithdrawnCellEvent;
import net.hivecell.hive.description.AgentSet;
import net.hivecell.hive.support.CellAddress;
import net.hivecell.hive.support.Debug;

import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Vector;
 

/**
 * this agent's sole purpose is to ask the agent monitoring agent to
 * monitor any servers which have appeared on the scene (as according
 * to the server list agent).  this works by having this agent
 * subscribe to the local server list agent and whenever it gets a
 * message that a server goes up, it will ask the agent monitoring
 * agent to monitor it.  if a cell goes down or withdraws, then it
 * will ask the agent monitoring agent to stop monitoring it 
 */
public class AutoMonitoringAgentImpl
extends EventReceivingAgentImpl
implements net.hivecell.hive.agent.cell.CellAgent {

    /**
     * our hook to our local agent monitoring agent to send our commands to
     */ 
    protected AgentMonitoringAgent ama;

    /**
     * the server list agent we need to subscribe to to get our server events
     */
    protected CellListAgent sla;


    public AutoMonitoringAgentImpl() 
    throws RemoteException {

	super();

    }

    public void doLocalSetup() throws AgentInitializationException {

	super.doLocalSetup();
	try {
	    String[] amaName = { "net.hivecell.hive.agent.cell.AgentMonitoringAgent" };
	    AgentSet ds = myCell.queryAgents( this, amaName, null );
	    ama = (AgentMonitoringAgent)ds.pickOne(); 

	    String[] slaName = { "net.hivecell.hive.agent.cell.CellListAgent" };
	    ds = myCell.queryAgents( this, slaName, null );
	    sla = (CellListAgent)ds.pickOne();
	    if( !connectTo( sla ) ) {  // subscribe to the event stream
		Debug.warning( "Could not subscribe to local server list agent" );
		myCell.killAgent( (AgentImpl)this );
		
	    }
	    
	} catch( ClassNotFoundException error ) {
	    Debug.critical( "Huh?  This should -never- happen!!", error );

	}

	try {
	    Enumeration cells = sla.listCells( this ).elements();
	    while( cells.hasMoreElements() ) {
		CellListInfo cli = (CellListInfo)cells.nextElement();
		if( cli.isUp() ) {
		    try {
			ama.monitorCell( this, cli.getAddress() );
		    } catch( RemoteException error ) {
			Debug.warning( "Remote exception while trying to ask the agent monitoring agent to monitor " + cli.getAddress(), error ); 

		    } catch( UnableToMonitorCellException error ) {
			Debug.warning( "Agent monitoring agent unable to monitor cell at " + cli.getAddress(), error );

		    }
		}

	    }
	} catch( RemoteException error ) {
	    Debug.warning( "Could not retrieve the list of cells which are up from the server list agent", error );

	}

    }

    public void doBehavior() { waitUntilDeath(); }

    public void notify( HiveEvent event ) {

	super.notify( event );
	if( !event.getSource().equals( sla ) ) {
	    Debug.warning( "Got an event from a source other than the server list agent (" + event.getSource() + ")...  Ignoring it." );
	    return;
	}

	if( !( event instanceof CellEvent ) ) {
	    Debug.warning( "The server list agent sent us a non server event...  Ignoring it." );
	    return;
	}

	CellEvent se = (CellEvent)event;
	if( ( se.eventCode() & ( CellEvent.SERVERBORN | CellEvent.SERVERGOINGDOWN | CellEvent.SERVERWITHDRAWN ) ) != 0 ) {
	     
	    if( se instanceof CellBornCellEvent ) {
		CellAddress ca = ((CellBornCellEvent)se).getAddress();
		try {
		    ama.monitorCell( this, ca );

		} catch( RemoteException error ) {
		    Debug.error( "Remote exception while trying to ask the agent monitoring agent to monitor " + ca );

		} catch( UnableToMonitorCellException error ) {
		    Debug.error( "Agent monitoring agent is unable to monitor " + ca );

		}

	    } else if( ( se instanceof CellGoingDownCellEvent ) || 
		       ( se instanceof CellWithdrawnCellEvent ) ) {
		CellAddress ca = null;
		if( se instanceof CellGoingDownCellEvent ) 
		    ca = ((CellGoingDownCellEvent)se).getAddress();
		else
		    ca = ((CellWithdrawnCellEvent)se).getAddress();
		try {
		    ama.stopMonitoringCell( this, ca );
		    
		} catch( RemoteException error ) {
		    Debug.error( "Remote exception while trying to ask the agent monitoring agent to stop monitoring " + ca );

		}

	    }

	}

    }

    public void doLocalCleanup() {
	
	super.doLocalCleanup();
	disconnectFrom( sla );

	try {
	    Enumeration cells = sla.listCells( this ).elements();
	    while( cells.hasMoreElements() ) {
		CellListInfo cli = (CellListInfo)cells.nextElement();
		if( cli.isUp() ) {
		    Debug.notice( "Asking to stop monitoring " + cli.getAddress() );
		    try {
			ama.stopMonitoringCell( this, cli.getAddress() );
		    } catch( RemoteException error ) {
			Debug.warning( "Remote exception while trying to ask the agent monitoring agent to stop monitoring " + cli.getAddress(), error ); 

		    }
		}

	    }
	} catch( RemoteException error ) {
	    Debug.warning( "Could not retrieve the list of cells which are up from the server list agent", error );

	}
	
    }

}
