// $Id: CellCommand.java,v 1.1 2000/05/18 22:01:03 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.cell;

import java.io.Serializable;
import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Random;
import java.util.StringTokenizer;
import java.util.Vector;

import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.agent.cell.CellControlAgent;
import net.hivecell.hive.agent.cell.CellControlAgentImpl;
//import net.hivecell.hive.rdf.DescSet;
import net.hivecell.hive.description.AgentSet;
import net.hivecell.hive.support.CellAddress;
import net.hivecell.hive.support.CellConnectException;
import net.hivecell.hive.support.MalformedHiveURLException;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.Global;

import gnu.getopt.*;


public class CellCommand 
implements Serializable {

    private static void usage() {

	System.out.println( "Usage: java net.hivecell.hive.cell.CellCommand" );
	System.out.println( "        [-l <cell>] [-s <cell>] [-k <agent> [<agent>]] [-d <agent>] [-c <agent> <agent>]" );
	System.out.println( "        [-m <cell> <classname>]" );
	System.out.println();
	System.out.println( " [-l <cell>]         : List agents on a cell" );
	System.out.println( " [-s <cell>]         : Shutdown a cell" );
	System.out.println( " [-k <agent>]          : Kill an agent" );
	System.out.println( " [-d <agent>]          : Disconnect an agent from everybody" );
	System.out.println( " [-d <agent> <agent>]  : Disconnect two agents from each other" );
	System.out.println( " [-c <agent> <agent>]  : Connect two agents to each other" );
	System.out.println( " [-m <cell> <agent>]   : Create a specfic agent on a cell" );

    }
    
    private static CellControlAgent getCellControlAgent( CellAddress ca ) 
    throws CellConnectException, RemoteException {

	Debug.notice( "Getting the remote server handle" );
	RemoteCell rs = ca.getRemoteCell();
	
	Debug.notice( "Attempting to get a handle to the server control agent on " + ca.getHost() );
	CellControlAgent sca = null;
	try {
	    String[] serverCommandAgentClassName = { (CellControlAgentImpl.class).getName() };
	    sca = (CellControlAgent)rs.queryAgents( new CellCommand(), serverCommandAgentClassName, null ).pickOne();

	} catch( ClassNotFoundException error ) {
	    Debug.error( "Huh?  This should never be thrown!", error );
	    System.exit( 1 );
	    
	} 
	return sca;

    }

    private static Agent getAgent( CellAddress ca ) 
    throws CellConnectException, RemoteException {

	Debug.notice( "Getting the handle to the remote server" );
	RemoteCell rs = ca.getRemoteCell();

	Debug.notice( "Getting a list of all the agents" );
	AgentSet ds = null;
	try {
	    ds = rs.queryAgents( new CellCommand(), null, null );
	} catch( ClassNotFoundException error ) {
	    Debug.error( "Huh?  This should never be thrown!", error );
	    System.exit( 1 );
	}

	Debug.notice( "Looking for the " + ca.getFile() + " agent" );
	Vector possibleAgents = new Vector();
	Enumeration e = ds.elements();
	while( e.hasMoreElements() ) {
	    Agent a = (Agent)e.nextElement();
	    if( a.getName().equals( ca.getFile().substring( 1 ) ) )
		possibleAgents.addElement( a );
	}
	
	if( possibleAgents.size() == 0 ) {
	    Debug.error( "No " + ca.getFile() + " found on " + ca.getHost() + "!" );
	    System.exit( 1 );
	    return null;
	} else if( possibleAgents.size() == 1 ) {
	    Debug.notice( "Found " + ca.getFile() + " on " + ca.getHost() );
	    return ((Agent)possibleAgents.elementAt( 0 ));
	} else {
	    Debug.notice( "Found multiple agents by " + ca.getFile() + " on " + ca.getHost() );
	    Debug.notice( "Choosing one at random" );
	    Random r = new Random();
	    int agentNumber = r.nextInt() % possibleAgents.size();
	    Debug.notice( "Returning candidate number " + agentNumber );
	    return ((Agent)possibleAgents.elementAt( agentNumber ));
	}

    }

    public static void main( String[] args ) {
	// the stuff we need to setup for the parsing
	CellAddress serverShutdownCellAddress = null;
	CellAddress serverListAgentsCellAddress = null;
	CellAddress serverCreateAgentCellAddress = null;
	String agentClassesToCreate = null;
	CellAddress disconnectAgent1Address = null;
	CellAddress disconnectAgent2Address = null;
	CellAddress connectAgent1Address = null;
	CellAddress connectAgent2Address = null;
	CellAddress agentToKill = null;

	// the parsing
	int c, lastFlag = -1;
	String arg;
	Vector arguments = new Vector(); // try to keep the order of the arguments

	LongOpt[] longopts = new LongOpt[7];
	longopts[0] = new LongOpt( "shutdown", LongOpt.REQUIRED_ARGUMENT, null, 's' );
	longopts[1] = new LongOpt( "list", LongOpt.REQUIRED_ARGUMENT, null, 'l' );
	longopts[2] = new LongOpt( "make", LongOpt.REQUIRED_ARGUMENT, null, 'm' );
	longopts[3] = new LongOpt( "disconnect", LongOpt.REQUIRED_ARGUMENT, null, 'd' );
	longopts[4] = new LongOpt( "connect", LongOpt.REQUIRED_ARGUMENT, null, 'c' );
	longopts[5] = new LongOpt( "kill", LongOpt.REQUIRED_ARGUMENT, null, 'k' );
	longopts[6] = new LongOpt( "help", LongOpt.REQUIRED_ARGUMENT, null, 'h' );
	Getopt g = new Getopt( "CellCommand", args, "-s:l:m:d:c:k:h", longopts );
	g.setOpterr( false );
	while( ( c = g.getopt() ) != -1 ) {
	    if( arguments.contains( new Character( (char)c ) ) ) {
		Debug.error( "This version only allows one of each flag on the command line at a time" );
		System.exit( 1 );
	    }
	    arguments.addElement( new Character( (char)c ) ); // store the option here
	    switch( c ) {
	    case 's':
		arg = g.getOptarg();
		lastFlag = 's';
		if( arg == null ) {
		    Debug.error( "The shutdown server command requires a Hive cell address" );
		    System.exit( 1 );
		}
		try {
		    serverShutdownCellAddress = new CellAddress( arg );
		} catch( MalformedHiveURLException error ) {
		    Debug.error( arg + " is not a valid Hive cell address" );
		    System.exit( 1 );
		}
		break;

	    case 'l':
		arg = g.getOptarg();
		lastFlag = 'l';
		if( serverListAgentsCellAddress != null ) {
		    Debug.error( "This version only allows one of each flag on the command line at a time" );
		    System.exit( 1 );
		}
		if( arg == null ) {
		    Debug.error( "The list agents on server command requires a Hive cell address" );
		    System.exit( 1 );
		}
		try {
		    serverListAgentsCellAddress = new CellAddress( arg );
		} catch( MalformedHiveURLException error ) {
		    Debug.error( arg + " is not a valid Hive cell address" );
		    System.exit( 1 );
		}
		break;
		
	    case 'm':
		arg = g.getOptarg();
		lastFlag = 'm';
		if( arg == null ) {
		    Debug.error( "The create agent command requires a Hive cell address and a class name" );
		    System.exit( 1 );
		} 
		try {
		    serverCreateAgentCellAddress = new CellAddress( arg );
		} catch( MalformedHiveURLException error ) {
		    Debug.error( arg + " is not a valid Hive cell address" );
		    System.exit( 1 );
		}
		break;

	    case 'd':
		arg = g.getOptarg();
		lastFlag = 'd';
		if( arg == null ) {
		    Debug.error( "The disconnect agent command requires one or two Hive agent address" );
		    System.exit( 1 );
		} 
		try {
		    disconnectAgent1Address = new CellAddress( arg );
		    if( disconnectAgent1Address.getFile().equals( "" ) || disconnectAgent1Address.getFile().substring( 1 ).equals( "" ) ) {
			Debug.error( arg + " does not have an agent address attached to it" );
			System.exit( 1 );
		    }
		} catch( MalformedHiveURLException error ) {
		    Debug.error( arg + " is not a valid Hive agent address" );
		    System.exit( 1 );
		}
		break;

	    case 'c':
		arg = g.getOptarg();
		lastFlag = 'c';
		if( arg == null ) {
		    Debug.error( "The connect agent command requires one or two Hive agent address" );
		    System.exit( 1 );
		} 
		try {
		    connectAgent1Address = new CellAddress( arg );
		    if( connectAgent1Address.getFile().equals( "" ) || connectAgent1Address.getFile().substring( 1 ).equals( "" ) ) {
			Debug.error( arg + " does not have an agent address attached to it" );
			System.exit( 1 );
		    }
		} catch( MalformedHiveURLException error ) {
		    Debug.error( arg + " is not a valid Hive agent address" );
		    System.exit( 1 );
		}
		break;

	    case 'k':
		arg = g.getOptarg();
		lastFlag = 'k';
		if( arg == null ) {
		    Debug.error( "The kill agent command requires a Hive agent address" );
		    System.exit( 1 );
		}
		try {
		    agentToKill = new CellAddress( arg );
		    if( agentToKill.getFile().equals( "" ) || agentToKill.getFile().substring( 1 ).equals( "" ) ) {
			Debug.error( arg + " does not have an agent address attached to it" );
			System.exit( 1 );
		    }
		} catch( MalformedHiveURLException error ) {
		    Debug.error( arg + " is not a valid Hive agent address" );
		    System.exit( 1 );
		}
		break;

	    case 1:
		if( arguments.size() != 0 )
		    arguments.removeElementAt( arguments.size() - 1 ); // remove this invalid flag from the list
		String errorarg = g.getOptarg();
		if( lastFlag == 'm' ) {
		    // trying to get the class name to create
		    Class toCreate = null;
		    try {
			toCreate = Class.forName( errorarg );
			agentClassesToCreate = errorarg;
		    } catch( ClassNotFoundException error ) {
			Debug.error( errorarg + " is not a valid class name" );
			System.exit( 1 );
		    }

		} else if( lastFlag == 'd' ) {
		    // trying to get the second agent address to disconnect from
		    try {
			disconnectAgent2Address = new CellAddress( errorarg );
			if( disconnectAgent2Address.getFile().equals( "" ) || disconnectAgent2Address.getFile().substring( 1 ).equals( "" ) ) {
			    Debug.error( errorarg + " does not have an agent address attached to it" );
			    System.exit( 1 );
			}
		    } catch( MalformedHiveURLException error ) {
			Debug.error( errorarg + " is not a valid Hive agent address" );
			System.exit( 1 );
		    }

		} else if( lastFlag == 'c' ) {
		    // trying to get the second agent address to connect to
		    try {
			connectAgent2Address = new CellAddress( errorarg );
			if( connectAgent2Address.getFile().equals( "" ) || connectAgent2Address.getFile().substring( 1 ).equals( "" ) ) {
			    Debug.error( errorarg + " does not have an agent address attached to it" );
			    System.exit( 1 );
			}
		    } catch( MalformedHiveURLException error ) {
			Debug.error( errorarg + " is not a valid Hive agent address" );
			System.exit( 1 );
		    }

		} else {
		    // hmm - something happened
		    Debug.error( errorarg + " is not a valid flag" );
		    System.exit( 1 );

		}
		lastFlag = -1; // reset this so we won't catch it the next time around
		break;
		
	    case 'h':
		usage();
		System.exit( 1 );
		break;

	    default:
		if( arguments.size() != 0 )
		    arguments.removeElementAt( arguments.size() - 1 ); // remove this invalid flag from the list
		Debug.error( g.getOptarg() + " is not a valid option" );

		usage();

		System.exit( 1 );
		break;

	    }
	}

	// do some dependancy testing to make sure all the proper things are set
	if( ( serverCreateAgentCellAddress != null ) && ( agentClassesToCreate == null ) ) {
	    Debug.error( "The create agent command requires a hive cell address and a class name" );
	    System.exit( 1 );
	}
	if( ( connectAgent1Address != null ) && ( connectAgent2Address == null ) ) {
	    Debug.error( "The connect agent command requires two hive agent addresses" );
	    System.exit( 1 );
	}


	// ok -- now let's do stuff in order
	Enumeration flags = arguments.elements();
	while( flags.hasMoreElements() ) {
	    char flagChar = ((Character)flags.nextElement()).charValue();
	    System.out.println( flagChar );
	    
	    if( flagChar == 's' ) {
		Debug.notice( "Processing the shutdown command" );
		
		Debug.notice( "Getting the handle to the server control agent on " + serverShutdownCellAddress.getHost() );
		CellControlAgent sca = null;
		try {
		    sca = getCellControlAgent( serverShutdownCellAddress );

		    Debug.notice( "Sending the shutdown command" );
		    sca.shutDownCell();
		    Debug.notice( "Cell shutting down" );

		} catch( CellConnectException error ) {
		    Debug.error( "Cannot get a handle to the server control agent on " + serverShutdownCellAddress.getHost() + "!" );
		    System.exit( 1 );

		} catch( RemoteException error ) {
		    Debug.error( "Remote exception while trying to talk to the remote server", error );
		    System.exit( 1 );

		}

	    } else if( flagChar == 'l' ) {
		Debug.notice( "Processing the list agents command" );
		
		Debug.notice( "Getting the handle to the remote server" );
		try {
		    RemoteCell rs = serverListAgentsCellAddress.getRemoteCell();
		    
		    Debug.notice( "Asking for a list of all the agents" );
		    AgentSet ds = rs.queryAgents( new CellCommand(), null, null );

		    String agents = "";
		    Enumeration a = ds.elements();
		    while( a.hasMoreElements() ) {
			Agent agent = (Agent)a.nextElement();
			agents += "  " + agent.getName() + " (" + agent.getClass().getName() + ")\n";
		    }
		    agents = "Agents alive on " + serverListAgentsCellAddress.getHost() + ":\n" + agents;
		    Debug.notice( agents );

		} catch( ClassNotFoundException error ) {
		    Debug.notice( "Huh?  This should never be thrown!", error );
		    System.exit( 1 );
		    
		} catch( CellConnectException error ) {
		    Debug.notice( "Problem getting the handle to the remote server at " + serverListAgentsCellAddress.getHost() );
		    System.exit( 1 );

		} catch( RemoteException error ) {
		    Debug.notice( "Remote exception while trying to talk to remote host!", error );
		    System.exit( 1 );
		}

	    } else if( flagChar == 'm' ) {
		Debug.notice( "Processing request to create agent" );
		
		CellControlAgent sca = null;
		try {
		    Debug.notice( "Getting a handle to the server control agent on " + serverCreateAgentCellAddress.getHost() );
		    sca = getCellControlAgent( serverCreateAgentCellAddress );

		    Debug.notice( "Asking the server control agent to create a " + agentClassesToCreate );
		    sca.createAgent( agentClassesToCreate );

		} catch( CellConnectException error ) {
		    Debug.error( "Problem connecting to cell at " + serverCreateAgentCellAddress.getHost() );
		    System.exit( 1 );

		} catch( ClassNotFoundException error ) {
		    Debug.error( serverCreateAgentCellAddress.getHost() + " reports that it does not know how to create a " + agentClassesToCreate );
		    System.exit( 1 );

		} catch( InstantiationException error ) {
		    Debug.error( serverCreateAgentCellAddress.getHost() + " says problem creating " + agentClassesToCreate );
		    System.exit( 1 );

		} catch( RemoteException error ) {
		    Debug.error( "Remote exception while trying to connect to remote host!", error );
		    System.exit( 1 );

		}

	    } else if( flagChar == 'd' ) {
		if( disconnectAgent2Address == null ) 
		    Debug.notice( "Processing disconnect from all command" );
		else
		    Debug.notice( "Processing disconnect agents command" );

		try {
		    Agent agent1 = null, agent2 = null;
		    Debug.notice( "Getting the agent handle to " + disconnectAgent1Address.getFile().substring( 1 ) );
		    agent1 = getAgent( disconnectAgent1Address );
		    if( agent1 == null ) {
			Debug.error( "Cannot get a handle to " + disconnectAgent1Address.getFile().substring( 1 ) + "!" );
			System.exit( 1 );
		    }

		    if( disconnectAgent2Address != null ) {
			Debug.notice( "Getting the agent handle to " + disconnectAgent2Address.getFile().substring( 1 ) );
			agent2 = getAgent( disconnectAgent2Address );
			if( agent2 == null ) {
			    Debug.error( "Cannot get a handle to " + disconnectAgent2Address.getFile().substring( 1 ) + "!" );
			    System.exit( 1 );
			}
		    }

		    Debug.notice( "Getting a handle to the server control agent on " + disconnectAgent1Address.getHost() );
		    CellControlAgent sca = getCellControlAgent( disconnectAgent1Address );

		    if( disconnectAgent2Address == null ) {
			Debug.notice( "Asking the server control agent to disconnect " + disconnectAgent1Address.getFile().substring( 1 ) + " on " + disconnectAgent1Address.getHost() + " from everybody" );
			sca.disconnectAgentAll( agent1 );

		    } else {
			Debug.notice( "Asking the server control agent to disconnect " + disconnectAgent1Address.getFile().substring( 1 ) + " on " + disconnectAgent1Address.getHost() + " from " + disconnectAgent2Address.getFile().substring( 1 ) + " on " + disconnectAgent2Address.getHost() );
			sca.disconnectAgents( agent1, agent2 );
		    }
		    Debug.notice( "Sent command to server" );

		} catch( CellConnectException error ) {
		    Debug.error( "Problem connecting to cell at " + disconnectAgent1Address.getHost() );
		    System.exit( 1 );

		} catch( RemoteException error ) {
		    Debug.error( "Remote exception while trying to connect to remote host!", error );
		    System.exit( 1 );

		}
		
	    } else if( flagChar == 'c' ) {
		Debug.notice( "Processing request to connect agents" );

		try {
		    Debug.notice( "Getting the agent handle to " + connectAgent1Address.getFile().substring( 1 ) );
		    Agent agent1 = getAgent( connectAgent1Address );
		    if( agent1 == null ) {
			Debug.error( "Cannot get a handle to " + connectAgent1Address.getFile().substring( 1 ) + "!" );
			System.exit( 1 );
		    }

		    Debug.notice( "Getting the agent handle to " + connectAgent2Address.getFile().substring( 1 ) );
		    Agent agent2 = getAgent( connectAgent2Address );
		    if( agent2 == null ) {
			Debug.error( "Cannot get a handle to " + connectAgent2Address.getFile().substring( 1 ) + "!" );
			System.exit( 1 );
		    }

		    Debug.notice( "Getting a handle to the server control agent on " + connectAgent1Address.getHost() );
		    CellControlAgent sca = getCellControlAgent( connectAgent1Address );

		    Debug.notice( "Asking the server control agent to connect " + connectAgent1Address.getFile().substring( 1 ) + " on " + connectAgent1Address.getHost() + " to " + connectAgent2Address.getFile().substring( 1 ) + " on " + connectAgent2Address.getHost() );
		    sca.connectAgents( agent1, agent2 );
		    Debug.notice( "Sent connect request to server" );

		} catch( CellConnectException error ) {
		    Debug.error( "Problem connecting to cell at " + connectAgent1Address.getHost() );
		    System.exit( 1 );

		} catch( RemoteException error ) {
		    Debug.error( "Remote exception while trying to connect to remote host!", error );
		    System.exit( 1 );

		}

	    } else if( flagChar == 'k' ) {
		Debug.notice( "Processing kill request" );

		try {
		    Debug.notice( "Getting the agent handle to " + agentToKill.getFile().substring( 1 ) );
		    Agent agent = getAgent( agentToKill );
		    if( agent == null ) {
			Debug.error( "Cannot get a handle to " + agentToKill.getFile().substring( 1 ) + "!" );
			System.exit( 1 );
		    }

		    Debug.notice( "Getting a handle to the server control agent on " + agentToKill.getHost() );
		    CellControlAgent sca = getCellControlAgent( agentToKill );
		    
		    Debug.notice( "Asking agent to die" );
		    sca.killAgent( agent );

		    Debug.notice( "Agent asked" );

		} catch( CellConnectException error ) {
		    Debug.error( "Problem connecting to cell at " + agentToKill.getHost() );
		    System.exit( 1 );

		} catch( RemoteException error ) {
		    Debug.error( "Remote exception while trying to connect to remote host!", error );
		    System.exit( 1 );

		}
		

	    }

	}
    }

}





