// $Id: FirewallingSecurityManager.java,v 1.1 2000/05/18 22:01:04 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.cell;

import net.hivecell.hive.support.Debug;
//import net.hivecell.hive.rdf.Description;
//import net.hivecell.hive.rdf.DescSet;
import net.hivecell.hive.description.Description;

import java.io.FileDescriptor;
import java.net.InetAddress;
import java.net.UnknownHostException;
import java.util.Enumeration;
import java.util.StringTokenizer;
import java.util.Vector;

public class FirewallingSecurityManager
extends HiveSecurityManager {

    /**
     * these will hold our filtering rules
     */
    Vector filteringRules = new Vector();

    public FirewallingSecurityManager( Description ds ) {

	int count = ds.select( "filteringRule" );  // NOTE: This will need to be set up to the security manager tag first.

	for( int i = 0; i < count; i++ ) {
	    Description filteringRuleDesc = ds.getMatch(i);
	    try {
		FirewallingFilteringRule ffr = new FirewallingFilteringRule( filteringRuleDesc.getText() );
		filteringRules.addElement( ffr );
	    } catch( FilteringRuleConstructionException error ) {
		Debug.error( "problem creating a filtering rule for \"" + filteringRuleDesc.getText() + "\"", error );
	    }
	}
    }


    /**
     * this method tests to see if a piece of remote code is allowing
     * for an incoming connection, if it is then we need to throw our
     * security exception
     *
     * @param host the host that the connection is comming from
     * @param port the port the connection is coming in 
     */
    public void checkAccept( String host, int port ) {

	String error = "code is attempting to accept a connection from " + host + " on port " + port;
	if( !testFilteringMatch( host ) )
	    throw new SecurityException( error );
	testRemoteExecuting( "remote " + error );

    }


    /**
     * we are probably going to have to let this one go since the
     * mobile code will need to be in contact with its implementation.
     * if we go bezerk, we might be able to do some test to make sure
     * the mobile code is only talking to its implementation and not
     * something else
     *
     * @param host the host that the code is connecting to
     * @param port the port that it is connecting to 
     */
    public void checkConnect( String host, int port ) { 

	String error = "code is attempting to connect to " + host + " on port " + port;
	if( !testFilteringMatch( host ) )
	    throw new SecurityException( error );

    }

    /**
     * same deal as the other checkConnect method
     */
    public void checkConnect( String host, int port, Object context ) { 

	String error = "code is attempting to connect to " + host + " on port " + port;
	if( !testFilteringMatch( host ) )
	    throw new SecurityException( error );

    }
    
    /**
     * test to see if this hostname matches any of our filtering rules
     *
     * @param hostname the test to see if the hostname matches our filtering rule
     * @return true if it does matches the filtering rule, false otherwise
     */
    protected boolean testFilteringMatch( String hostname ) {
	Enumeration e = filteringRules.elements();
	while( e.hasMoreElements() )
	    if( ((FirewallingFilteringRule)e.nextElement()).filteringTest( hostname ) )
		return true;
	return false;
    }

}

class FirewallingFilteringRule {

    int[] filteringNumbers = new int[4];

    public FirewallingFilteringRule( String filteringRule ) 
    throws FilteringRuleConstructionException {

	StringTokenizer st = new StringTokenizer( filteringRule, "." );
	if( st.countTokens() > 4 ) throw new FilteringRuleConstructionException( "invalid filtering rule specification" );
	int count;
	int totalTokens = st.countTokens();
	for( count=0;count < totalTokens;count++ ) {
	    String filteringValue = st.nextToken();
	    if( filteringValue.equals( "*" ) )
		filteringNumbers[count] = -1;
	    else {
		try {
		    int filteringNumber = Integer.valueOf( filteringValue ).intValue();
		    if( ( filteringNumber > 255 ) || ( filteringNumber < 0 ) ) throw new FilteringRuleConstructionException( "invalid number for filtering -- must be between 0 to 255" );
		    filteringNumbers[count] = filteringNumber;
		} catch( NumberFormatException error ) {
		    throw new FilteringRuleConstructionException( filteringValue + " is not a valid ip number to filter on" );
		}
	    }
	}
	for( count=count;count < 4;count++ )
	    filteringNumbers[count] = -1;
    }


    /**
     * see if a host matches this filtering rule
     *
     * @param host the host to test
     * @return true if the filtering test matches this host
     */
    public boolean filteringTest( String host ) {
	try {
	    InetAddress address = InetAddress.getByName( host );
	    byte[] addressBytes = address.getAddress();
	    for( int count=0;count < 4;count++ )
		if( ( filteringNumbers[count] != -1 ) && ( (byte)filteringNumbers[count] != addressBytes[count] ) )
		    return false;
	} catch( UnknownHostException error ) {
	    Debug.error( host + " is not a valid computer name" );
	    return false;
	}
	return true;
    }

}

class FilteringRuleConstructionException 
extends Exception {
    public FilteringRuleConstructionException() { super(); }
    public FilteringRuleConstructionException( String message ) { super( message ); }
}

