package net.hivecell.hive.cell;

import java.rmi.AccessException;
import java.rmi.AlreadyBoundException;
import java.rmi.NotBoundException;
import java.rmi.Remote;
import java.rmi.RemoteException;
import java.rmi.registry.LocateRegistry;
import java.rmi.registry.Registry;

public class HiveRMIRegistry 
implements Registry {

    /**
     * the port that the registry sis on
     */
    protected int port;

    /**
     * the real registry.  we won't let the get to teh real registry
     * so we can not allow people to unbind the cell from it 
     */
    protected Registry registry;


    /**
     * create the registry at a certain port
     *
     * @param port the port that you want to connect to
     * @exception RemoteException if we can't create the port that you want
     */
    public HiveRMIRegistry( int port ) 
    throws RemoteException {
	registry = LocateRegistry.createRegistry( port );

    }

    /**
     * get the port that this registry is running on
     *
     * @return the port 
     */
    public int getPort() {
	return port;
	
    }

    /**
     * Returns the remote object associated with the specified name in the registry. 
     *
     * @param name the name that we want to lookup
     * @return the object that is bound
     * @exception RemoteException something went wrong with this call
     * @exception NotBoundException there is nothing bound to this object
     * @exception AccessException not allowed to do this
     */
    public Remote lookup( String name ) 
    throws RemoteException, NotBoundException, AccessException {
	return registry.lookup( name );

    }

    /**
     * Binds the name to the specified remote object. 
     *
     * @param name the name that we want to bind with
     * @param obj the object that we want to bind to the name
     * @exception RemoteException when things go wrong
     * @exception AlreadyBoundException if something is bound to that name
     * @exception AccessException this must be the secuiryt problem
     */
    public void bind( String name, Remote obj ) 
    throws RemoteException, AlreadyBoundException, AccessException {
	registry.bind( name, obj );

    }
    
    /**
     * Unbind the name. 
     *
     * @param name the object that we want to unbind
     * @exception RemoteException if something went wrong trying to perform this action
     * @exception NotBoundException if this object is not already bound
     * @exception AccessException if we are not allowed to unbind this exception
     */
    public void unbind( String name ) 
    throws RemoteException, NotBoundException, AccessException {
	if( name.equals( "cell" ) )
	    throw new AccessException( "cannot unbind the cell" );
	registry.unbind( name );

    }

    /**
     * Rebind the name to a new object, replacing any existing binding. 
     *
     * @param name the name we want to rebind
     * @param obj the object we want to bind it to
     * @exception RemoteException somethign went wrong when trying to do this
     * @exception AccessException if you are not allowed to do this action
     */
    public void rebind( String name, Remote obj ) 
    throws RemoteException, AccessException {
	if( name.equals( "cell" ) )
	    throw new AccessException( "cannot rebind over the cell" );
	registry.rebind( name, obj );
	
    }

    /**
     * Returns an array of the names in the registry. 
     *
     * @exception RemoteException if something went wrong while trying to do this
     * @exception AccessException if we are not allowed to perform this action
     */
    public String[] list() 
    throws RemoteException, AccessException {
	return registry.list();

    }



}
