// $Id: Description.java,v 1.1 2000/05/18 22:01:06 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

/**
 ** This class handles browsing an XML file like a DescSet browsing an RDF file.
 ** Definintion of an interface to XML that fully replaces RDF and is better designed than HiveXML
 ** 
 ** Thesis:  You have a DOM tree.  This is used to search the tree in different ways.
 ** 	 Once you find the node you are looking for, pull it out with getMatch(int) and query
 ** 	 it with the wrapper functions to the DOM commands.
 **/

package net.hivecell.hive.description;

import java.util.Vector;
import java.util.Hashtable;
import java.util.Enumeration;
import java.util.Properties;
import java.io.*;

import org.xml.sax.HandlerBase;
import org.xml.sax.AttributeList;
import org.xml.sax.SAXParseException;
import org.xml.sax.SAXException;
import org.xml.sax.InputSource;

import org.w3c.dom.*;

import com.docuverse.dom.util.SAXReader;
import com.docuverse.dom.util.*;
import com.docuverse.dom.*;

import net.hivecell.hive.support.Debug;
import net.hivecell.hive.Global;


// NOTE: Although require might be worthless, requireProperty might be very useful...  I think the menu
//       stuff could really benefit from it.
// NOTE// NOTE: Still todo: really want to be able to search the tree for a specific property, or a tag that has a specific property.
// NOTE: However, this can wait until after everything has been converted to XML (I think)
// 5-1-00:  Why did I use a linked list instead of a Vector?
// NOTE 5-1-00:  Check to make sure null values are handled consistently (ex. selecting a null value should
// return either a full set or an empty set depending on the logical context).  I belive they are, but I
// want to make sure.

public class Description implements Serializable {

    public static final int ANYWHERE = -1;
    public static final String RootTag = "description";
    public static final String SemanticTag = "semantic";
    public static final String ConfigTag = "config";


    // How matches now work: linked list of all Descriptions matching query.
    protected Description next = null;
    protected Node self = null;
    protected String name = null;
    protected Description matches = null;   // Stores the result of the most recent query.
    //    Document doc;
    Object thing;   // Reference to Agent or whatever that is being described.  NOTE: SHOULD THIS BE A DESCRIBABLE INTEAD?


    /* CONSTRUCTORS */
    public Description(Object what, Node me) {
	thing = what;
	self = me;
	name = Global.shortClassName(what);
    };

    public Description(Object what, String Name) {
	thing = what;
	name = Name;

	// NOTE: is there a difference between the two methods below?
	//	DOM dom = new DOM();
	//	Document doc = dom.createDocument("XML");
	Document doc = new BasicDocument();
	self = doc.createElement( RootTag );
    }


    /* These are internal linked list methods */
    private void setNext(Description node) {
	next = node;
    }
    /*
     * Returns the next in the linked list.  Useful for walking the matches.
     */
    public Description getNext() {
	return (next);
    }
    public Node getNode() {
	return (self);
    }


    /** Returns the reference to the Object this Description is describing.
     **/
    public Object getObject() {
	return (thing);
    }


    /**
     ** Return the name for the agent/cell/shadow this instance is describing.
     **/
    public String getName() {
	return (name);
    }


    
    /** Return all Descriptions that matched the last select or find query. **/
    public Description getMatches() {
	return (matches);
    }
    /** Return the first Description that matched the last select or find query. **/
    public Description getMatch() {
	return (getMatch(0));
    }
    /** Return a certain Descriptions that matched the last select or find query. **/
    public Description getMatch(int which) {
	if (matches == null) return null;
	    
	Description match = matches;
	while (which > 0 && match != null) {
	    match = match.getNext();
	    which--;
	}

	return match;
    }


    /** Jump to the previous level in the tree. **/
    public Description upContext() {
	if (self == null) return null;
	return (new Description (thing, self.getParentNode()));
    }
    /** Jump to the root node of the tree. **/
    public Description noContext() {
	if (self == null) return null;
	Node n = self;
	while (n.getParentNode() != null)
	    n = n.getParentNode();

	return (new Description(thing, n));
    }

    /** Jump to the branch of the tree which contains the agent/cell/shadow's seantic description. **/
    public Description semanticContext() {
	Description desc = this.noContext();
	if (desc != null && desc.findFirst( SemanticTag )) {
	    return (desc.getMatch());
	}
	return null;
    }
    /** Jump to the branch of the tree which contains the agent/cell/shadow's configuration information. **/
    public Description configContext() {
	Description desc = this.noContext();
	if (desc != null && desc.findFirst( ConfigTag )) {
	    return (desc.getMatch());
	}
	return null;
    }


    /**
     ** Adds a tag under <semantic> dynamically to a description.
     ** There will always be a semantic tag when a blank Description
     ** is created, this will also create one if the description is from 
     ** an xml file without a semantic section.
     **/
    public void addSemantic(String tagName, String tagVal, String[] attrTags, String[] attrVals) {

	try {
	    Description sem = this.semanticContext();
	    Document doc;

	    if (sem == null) {
		Description d = this.noContext();
		if ( !d.findFirst(RootTag) ) {
		    Debug.warning("This description has no root tag (" + RootTag + ")! Not adding tag '" + tagName + "'.");
		    return;
		}
		d = d.getMatch();
		doc = d.getNode().getOwnerDocument();
		
		Element a = doc.createElement( SemanticTag );
		d.getNode().appendChild(a);
		addSemantic(tagName, tagVal, attrTags, attrVals);
	    }


	    if (sem != null) {
		doc = sem.getNode().getOwnerDocument();
		Element a = doc.createElement(tagName);
		Node n = doc.createTextNode(tagVal);
		
		if (attrTags != null && attrVals != null) {
		    if (attrTags.length != attrVals.length)
			Debug.warning("Attribute tags array and values array not the same length!  Not adding attributes.");
		    else {
			for (int i = 0; i < attrTags.length; i++)
			    a.setAttribute(attrTags[i], attrVals[i]);
		    }
		}
		
		sem.getNode().appendChild(a);
		a.appendChild(n);
	    }

	} catch (Exception f) {
	    Debug.error("An error occured while trying to add nodes.", f);
	}
    }


    /**
     ** Adds a tag under the current node dynamically to a description.
     ** This allows the building of descriptions on the fly, which 
     ** can then be written out with toString().
     **/
    public void addTag(String tagName, String tagVal, String[] attrTags, String[] attrVals) {

	try {
	    Node cur = this.getNode();

	    if (cur != null) {
		Document doc = cur.getOwnerDocument();
		Element a = doc.createElement(tagName);
		Node n = doc.createTextNode(tagVal);
		
		if (attrTags != null && attrVals != null) {
		    if (attrTags.length != attrVals.length)
			Debug.warning("Attribute tags array and values array not the same length!  Not adding attributes.");
		    else {
			for (int i = 0; i < attrTags.length; i++)
			    a.setAttribute(attrTags[i], attrVals[i]);
		    }
		}
		
		cur.appendChild(a);
		a.appendChild(n);
	    }
	} catch (Exception f) {
	    Debug.error("An error occured while trying to add nodes.", f);
	}
    }




    /**
     ** Determines whether or not the Description contains a tag/value combo
     **/
    public boolean containsTagValue(String tag, String value) {

	int num = find(tag);
	if (num == 0) return false;
	if (value == null) return true;   // Was found, and don't care about the value
	
	for (int i = 0; i < num; i++) {
	    Description current = this.getMatch(i);
	    if ( current.getText() != null && value != null ) {
		if ( current.getText().equals(value) || value == null ) return true;
	    } 
	}

	return false;
    }



    /**
     ** Find a list of nodes matching 'tname' in current's children.  Returns number of matches.
     ** Current node and list of matches does not change if there wasn't a successful match.
     **
     ** @param tname The name of the tag to search for
     **/
    public int select(String tname) {
	if (self == null) return 0;

	NodeList nl = self.getChildNodes();
	Description prevNode = null;
	Description first = null;
	int num = 0;

	for (int i = 0; i < nl.getLength(); i++) {
	    Node c = nl.item(i);
	    // Only bother with the string compare if it is of ELEMENT_NODE type (the tags)
	    if (c.getNodeType() == Node.ELEMENT_NODE && c.getNodeName().equals(tname)) {
		if (prevNode == null) {
		    prevNode = new Description(thing, c);
		    first = prevNode;
		}
		else {
		    prevNode.setNext(new Description(thing, c));  // next pointer of last match points to this match
		    prevNode = prevNode.getNext();                   // last match becomes pointer to this node
		}
		num++;
	    }
	}

	matches = first;
	return (num);
    }


    /** 
     ** Finds first node matching 'tname' in current's children.  Returns true if a node is found.
     ** Current node and list of matches does not change if there wasn't a successful match.
     **
     ** @param tname The name of the tag to search for
     **/
    public boolean selectFirst(String tname) {
	if (self == null) return false;

	NodeList nl = self.getChildNodes();

	for (int i = 0; i < nl.getLength(); i++) {
	    Node c = nl.item(i);
	    // Only bother with the string compare if it is of ELEMENT_NODE type (the tags)
	    if (c.getNodeType() == Node.ELEMENT_NODE && c.getNodeName().equals(tname)) {
		matches = new Description(thing, c);
		return (true);
	    }
	}

	return (false);
    }




    /** 
     ** FIND
     **
     ** Finds all nodes matching 'tname' anywhere in the hierarchy of this's childNodes.
     ** Returns number of matches.  Does a breadth-first search.
     ** Current node does not change after a find, only the list of matches.  
     ** If no successful matches were found, matches doesn't change.
     **
     ** @param tname The name of the tag to search for
     ** @param depth max depth to search, Description.ANYWHERE 
     **              for the entire tree.
     **/
    public int find(String tname) {
	return ( find(tname, Description.ANYWHERE) );
    }

    public int find(String tname, int depth) {
	if (self == null) return (0);

	NodeListImpl n = new NodeListImpl();
	n.appendNode(self);   // Set up the queue

	return ( find(tname, depth, false, 0, null, n, false) );
    }


    /** 
     ** FIND FIRST
     **
     ** Find the first node matching 'tname' anywhere in the hierarchy with current as root.
     ** Returns true if a match was found.  Does a breadth-first search.
     ** Current node does not change after a find, only the list of matches.  
     ** If no successful matches were found, matches doesn't change.
     **
     ** @param tname The name of the tag to search for
     ** @param depth max depth to search, Description.ANYWHERE 
     **              for the entire tree.
     **/
    public boolean findFirst(String tname) {
	return ( findFirst(tname, Description.ANYWHERE) );
    }

    public boolean findFirst(String tname, int depth) {
	if (self == null) return (false);

	// Set up the queue
	NodeListImpl n = new NodeListImpl();
	n.appendNode(self);
	
	if ( find(tname, depth, true, 0, null, n, false) == 0 )
	    return (false);
	else
	    return (true);
    }



    /**
     ** Workhorse function for the find procedures, does a breadth-first search for the item.
     ** You should never need to use this function, use the front-end find procedures instead.
     **
     ** @param tname Name of the tag or attribute that is being searched for
     ** @param depth How far down in the tree to search
     ** @param first Return the first match, or find them all
     ** @param numSoFar How many matches have currently been found (this is a recursive function.  This is updated internally)
     ** @param prevMatch Another internal variable.  Newest node in the linked list.
     ** @param queue Queue of the nodes still left to search (Starts out as the self node of the Description being queried)
     ** @param attrib Search on the attributes of a tag, or the tag itself?  If true, tname refers to the Tag name, otherwise the attrib.
     ** 
     **/
    public int find(String tname, int depth, boolean first, int numSoFar, Description prevMatch, NodeList queue, boolean attrib) {
	if (queue.getLength() == 0 || depth == 0)
	    return (numSoFar);

	Node c = null;
	NodeListImpl newQueue = new NodeListImpl();

	for (int i = 0; i < queue.getLength(); i++) {

	    c = queue.item(i);

	    if (attrib) { // Search on the node's attributes:
		NamedNodeMap nnm = c.getAttributes();
		if (nnm == null) continue;

		for (int j = 0; j < nnm.getLength(); j++) {
		    Node p = nnm.item(j);
		    
		    if (p.getNodeName().equals(tname)) {
			if (prevMatch == null) {
			    matches = new Description(thing, c);
			    prevMatch = matches;
			}
			else {
			    prevMatch.setNext(new Description(thing, c));  // next pointer of last match points to this match
			    prevMatch = prevMatch.getNext();                  // last match becomes pointer to this node
			}

			numSoFar++;
			if (first) return (1);
		    }
		}
	    }

	    else {  // Search on the name of the tag:
		if (c.getNodeName().equals(tname)) {
		    if (prevMatch == null) {
			matches = new Description(thing, c);
			prevMatch = matches;
		    }
		    else {
			prevMatch.setNext(new Description(thing, c));  // next pointer of last match points to this match
			prevMatch = prevMatch.getNext();                  // last match becomes pointer to this node
		    }

		    numSoFar++;
		    if (first) return (1);
		}
	    }

	    if (depth > 1 || depth == ANYWHERE) {  // Don't bother filling the next queue if you're not going to search it.

		NodeList nql = c.getChildNodes();
		for (int j = 0; j < nql.getLength(); j++) {
		    newQueue.appendNode(nql.item(j));
		}
	    }
	}

	if (depth == ANYWHERE)
	    return ( find(tname, ANYWHERE, first, numSoFar, prevMatch, newQueue, attrib) );
	else return ( find(tname, depth - 1, first, numSoFar, prevMatch, newQueue, attrib) );
    }
    


    // ATTRIBUTE QUERYING METHODS:
    
    /**
     ** These methods perform queries on the attributes of an agent's XML description.
     **
     ** @param attribute The name of the attribute, null to search for any attributes.
     ** @param val The value of the attribute we're querying for, or null for any value.
     **/
    public boolean containsAttributeValue(String attribute, String val) {

	int num = findAttribute(attribute);
	if (num == 0) return false;     // Not found
	if (val == null) return true;   // Was found, and don't care about the attribute value 

	for (int i = 0; i < num; i++) { // Do they match?
	    Description current = this.getMatch(i);
	    String pval = current.getAttribute(attribute);

	    if ( pval != null && pval.equals(val) ) 
		return true;
	}

	return false;    
    }
    
    public boolean selectFirstAttribute(String attribute) {
	if (self == null) return false;

	NodeList nl = self.getChildNodes();

	for (int i = 0; i < nl.getLength(); i++) {
	    Node c = nl.item(i);
	    
	    if (c.getNodeType() == Node.ELEMENT_NODE) {
		Description d = new Description(thing, c);

		if (d.getAttribute(attribute) != null) {
		    matches = d;
		    return (true);
		}
	    }
	}

	return (false);
    }

    public int selectAttribute(String attribute) {
	if (self == null) return 0;

	NodeList nl = self.getChildNodes();
	Description prevNode = null;
	Description first = null;
	int num = 0;

	for (int i = 0; i < nl.getLength(); i++) {
	    Node c = nl.item(i);
	    // Only bother with the string compare if it is of ELEMENT_NODE type (the tags)
	    if (c.getNodeType() == Node.ELEMENT_NODE) {
		Description d = new Description(thing, c);

		if (d.getAttribute(attribute) != null) {

		    if (prevNode == null) {
			prevNode = d;
			first = prevNode;
		    }
		    else {
			prevNode.setNext(d);      // next pointer of last match points to this match
			prevNode = prevNode.getNext();    // last match becomes pointer to this node
		    }
		    num++;
		}
	    }
	}
	matches = first;
	return (num);
    }



    public int findAttribute(String attribute) {
	return ( findAttribute(attribute, Description.ANYWHERE) );
    }
    public int findAttribute(String attribute, int depth) {
	if (self == null) return (0);

	NodeListImpl n = new NodeListImpl();
	n.appendNode(self);   // Set up the queue
	
	return ( find(attribute, depth, false, 0, null, n, true) );
    }


    public boolean findFirstAttribute(String attribute) {
	return ( findFirstAttribute(attribute, Description.ANYWHERE) );
    }
    public boolean findFirstAttribute(String attribute, int depth) {
	if (self == null) return (false);

	NodeListImpl n = new NodeListImpl();
	n.appendNode(self);   // Set up the queue
	
	if ( find(attribute, depth, true, 0, null, n, true) == 0 )
	    return (false);
	else
	    return (true);
    }

    


    // NODE QUERY METHODS:
    
    /** 
     **	This is simply a wrapper for Node.getNodeName().  
     **	Just have it to keep a consistent interface.  
     **/
    public String getNodeName() { 
	if (self == null) return null;
	return (self.getNodeName()); 
    }

    
    /**
     ** This is an interface to the Node.getAttributes, 
     ** which returns a NamedNodeMap and is a mess.
     **
     ** @param propName The name of the property to get the value for
     ** @return The value of the property with tag propName or null
     **/
    public String getAttribute(String propName) { 
	if (self == null) return (null);

	NamedNodeMap nnm = self.getAttributes();

	for (int i = 0; i < nnm.getLength(); i++) {
	    Node p = nnm.item(i);
	    if (p.getNodeName().equals(propName)) {
		return (p.getNodeValue());
	    }
	}

	return (null);
    }

    /**
     ** If the first child of this node is a '#text' node (type TEXT_NODE),
     ** <tag>
     **     This part returned.
     ** </tag>
     ** then this returns the text: 'This part returned.'
     ** Otherwise, returns null.
     **/
    public String getText() {
	String raw = getUnstrippedText();

	if (raw != null) {  // Strip the whitespace from the front of the text.  String.trim() doesn't do it for us...

	    int firstWord = 0;
	    while (firstWord < raw.length() && 
		   raw.charAt(firstWord) == ' ' || 
		   raw.charAt(firstWord) == '\n' || 
		   raw.charAt(firstWord) == '\t') 
		firstWord++;

	    return (raw.substring(firstWord).trim());
	}
	return (null);
    }
    
    /**
     ** If the first child of this node is a '#text' node (type TEXT_NODE), 
     ** <tag>
     **    This part returned.
     ** </tag>
     ** 
     ** then this returns the text with all trailing and leading whitespace intact:
     '
         This part returned.'
     **
     ** Otherwise, returns null.
     **/
    public String getUnstrippedText() {
	if (self == null) return null;
	NodeList nl = self.getChildNodes();
	DOMUtil dutil = new DOMUtil();

	for (int i = 0; i < nl.getLength(); i++) {
	    Node n = nl.item(i);
	    if (dutil.isWhitespace(n)) continue;

	    if (n.getNodeType() == Node.TEXT_NODE)
		return (n.getNodeValue());
	}

	return (null);
    }



    /**
     ** Utility debugging methods.
     **/

    public void printTree() { printTree(getNode(), 0, false); }
    public void printTree(boolean showWS) { printTree(getNode(), 0, showWS); }
    public static void printTree(Node node) { printTree(node, 0, false); }
    public static void printTree(Node node, boolean showWS) { printTree(node, 0, showWS); }
    public static void printTree(Node node, int depth, boolean showWS) {

	DOMUtil dutil = new DOMUtil();

	if (!showWS && dutil.isWhitespace(node))
	    return;
	
	/* USING A FILTER:
	WhitespaceFilter whitespace = new WhitespaceFilter();
	if (whitespace.filterNode(node) == NodeFilter.ACCEPT)
	    return;
	*/
	
	for (int j = 0; j < depth; j++)
	    System.out.print("    ");
	System.out.println();


	if (depth != 0)
	    System.out.print(depth + "---");
	for (int i = 1; i < depth; i++)
	    System.out.print("----");

	// Print out the values for this node
	System.out.println("Name: " + node.getNodeName() + 
			   ", Type: " + getNameforType(node.getNodeType()) +
			   ", Value: " + node.getNodeValue());
	


	// Print out the properties (extra tags) for this node
	NamedNodeMap nnm = node.getAttributes();

	if (nnm != null && nnm.getLength() > 0) {
	    for (int i = 0; i < depth; i++)
		System.out.print("    ");
	    
	    System.out.println("        Properties (" + nnm.getLength() + "): ");
	    
	    
	    for (int i = 0; i < nnm.getLength(); i++) {
		for (int j = 0; j < depth; j++)
		    System.out.print("    ");
		System.out.print("            ");
		
		Node p = nnm.item(i);
		System.out.println(i + ") Name: " + p.getNodeName() +
				   ", Type: " + getNameforType(p.getNodeType()) +
				   ", Value: " + p.getNodeValue());
	    }
	}


	NodeList kids = node.getChildNodes();
	for (int i = 0; i < kids.getLength(); i++)
	    printTree(kids.item(i), depth + 1, showWS);
    }


    public static String getNameforType(short type) {
	if (type == Node.TEXT_NODE)
	    return ("TEXT_NODE");
	else if (type == Node.DOCUMENT_NODE)
	    return ("DOCUMENT_NODE");
	else if (type == Node.ATTRIBUTE_NODE)
	    return ("ATTRIBUTE_NODE");
	else if (type == Node.ENTITY_NODE)
	    return ("ENTITY_NODE");
	else if (type == Node.ELEMENT_NODE)
	    return ("ELEMENT_NODE");
	else if (type == Node.CDATA_SECTION_NODE)
	    return ("CDATA_SECTION_NODE");

	else return (Integer.toString(type));
    }


    /**
     ** This prints out the XML that makes up the Description.
     ** It prints out the entire XML document, no matter where 
     ** in the tree you are.
     **/
    public String toString() {
	Description t = this.noContext();
	return(t.getNode().toString());
    }

    /**
     ** This prints out the XML that makes up the Description.
     ** It prints out the XML document rooted at desc.
     **/
    public static String toString(Description desc) {
	return(desc.getNode().toString());
    }
}
