// $Id: HiveEvent.java,v 1.1 2000/05/18 22:01:07 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.event;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.Global;

/** Event classed used by Hive.
 ** This class inherits from the Jini RemoteEvent object, adding fields
 ** for event data and timestamps. Handbacks are not currently supported.
 **
 ** @version $Revision: 1.1 $
 ** @author Nelson Minar <nelson@media.mit.edu>
 **/

public class HiveEvent 
extends RemoteEvent {

    /** Extra event-specific data. **/
    private Object eventData;
    private long eventTime;

    /** What ID we use if the event creator didn't specify one. **/
    public static final long ALL_EVENTS = -1L;

    /** Construct a Hive event with all fields specified
     ** @param source The object (agent or shadow) that sent the event.
     ** @param count The count for this event; starts at 0.
     ** @param eventData Arbitrary event data to pass along.
     ** @param timeMillis The time in milliseconds of the event
     ** @param eventID The event ID code
     **/
    public HiveEvent(Object source, long count, Object eventData, long timeMillis, long eventID) {
        super(source, eventID, count, null);
        this.eventData = eventData;
        this.eventTime = timeMillis;
    }

    /** Construct a Hive event with a specified time but no ID **/
    public HiveEvent(Object source, long count, Object eventData, long timeMillis) {
	this(source, count, eventData, timeMillis, ALL_EVENTS);
    }

        
    /** Construct a Hive event with the current time and unspecified ID **/
    public HiveEvent(Object source, long count, Object eventData) {
	this(source, count, eventData, System.currentTimeMillis(), ALL_EVENTS);
    }

    /** Returns the sender of this event.
     ** In Hive, this field will always be type Agent.
     **/
    public Object getSource() {
	return super.getSource();
    }

    /** Returns the event count. Every event source maintains a count of
     ** all events sent out, starting at zero. In Hive, this field is
     ** equivalent to the sequence number.
     **/
    public long getCount() {
	return getSequenceNumber();
    }

    /** Returns any event-specific data. **/
    public Object getEventData() {
	return eventData;
    }

    /** Returns the time the event was sent (in milliseconds).
     ** Note: there are no guarantees about distributed time synchronization,
     ** believe this field at your own risk. Typically, this will just be
     ** the local time at the moment the sender generated the event object.
     **/
    public long getEventTime() {
	return eventTime;
    }

    /** This method is currently not used in Hive. **/
    public MarshalledObject getRegistrationObject() {
	Debug.println(Debug.WARNING, "HiveEvents do not use handbacks.");
	return null;
    }

    /** Format the event in a nice way. **/
    public String toString() {
	Object source = getSource();
	String sourceString;
    
	if (source instanceof Agent)
	    sourceString = Global.shortString(source);
	else
	    sourceString = source.toString();
    
	return "HiveEvent(#" + getCount() +
	    " id=" + Long.toHexString(eventID) + " " +
	    Global.dateString(getEventTime()) +
	    " src=" + sourceString +
	    " data=" + getEventData() + ")";
    }
}



