// $Id: ShadowImpl.java,v 1.1 2000/05/18 22:01:12 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.shadow;
import net.hivecell.hive.Global;
import net.hivecell.hive.agent.AgentImpl;
import net.hivecell.hive.cell.Cell;
import net.hivecell.hive.description.Describable;
import net.hivecell.hive.description.Description;
import net.hivecell.hive.support.Debug;

import java.io.*;
import java.util.*;
import java.lang.reflect.Constructor;
import java.lang.reflect.Method;

import com.docuverse.dom.DOM;

import org.w3c.dom.*;


/**
 * A shadow is a class that exposes a local resource to agents.
 *
 * @author Matthew Gray
 */
public abstract class ShadowImpl
implements Describable, Shadow {

    /** The cell this shadow exists on. */
    protected Cell myCell;

    /** Semantic description for shadow (parsed). **/
    protected Description description = null;

    /**
     * Just a simple flag to use and set if the shadow requires some
     * hardware startup that cannot be done instantaneously -- like
     * the LazyFishShadow 
     */
    private boolean ready = false;
    
    /**
     * the stub for this shadow
     */
    protected ShadowStub stub;


    /**
     * create this shadow and create the stub for it -- this
     * constructor might throw a lot of problems if we can't load our
     * stub.  but those will all be run time problems just like java
     * rmi will have a fit if we can't load the stub properly
     *
     * @exception ShadowStubNotFoundException if we can't load the stub class
     * @exception ShadowStubInstantiationException if we can't construct the stub class or if the class is not "right" 
     */
    public ShadowImpl() {

	// the stub will be net.hivecell.hive.shadow.TestShadow_Stub
	String stubClassName = this.getClass().getName() + "_Stub";
	Debug.notice( "attempting to load " + stubClassName + " as the stub for " + this.getClass().getName() );
	Class stubClass = null;
	try {
	    stubClass = Class.forName( stubClassName );

	} catch( ClassNotFoundException error ) {
	    throw new ShadowStubNotFoundException( stubClassName + " not found as the stub for " + this.getClass().getName() );

	}

	// ok -- we have the class loaded, let's make sure that we
	// have something which implements (directly, or indirectly)
	// the ShadowInterface
	Debug.notice( "testing " + stubClassName + " to make sure it is a valid stub" );
	Vector allInterfaces = getAllInterfaces( stubClass );
	if( !getAllInterfaces( stubClass ).contains( Shadow.class ) )
	    throw new ShadowStubInstantiationException( "stub class does not implement the ShadowInterface" );

	// if we have gotten this far, then we are good to go, let's
	// instantiate this stub and hold onto it.  but before we do
	// that, we need to find the right constructor
	Constructor stubConstructor = null;
	try {
	    Class[] constructorParameters = { ShadowImpl.class };
	    stubConstructor = stubClass.getConstructor( constructorParameters );

	} catch( Exception error ) {
	    throw new ShadowStubInstantiationException( stubClassName + " does not have the proper constructor" );

	}

	try {
	    Object[] parameters = { this };
	    stub = (ShadowStub)stubConstructor.newInstance( parameters );
	    Debug.notice( "constructed the stub for " + this.getClass().getName() );

	} catch( Exception error ) {
	    throw new ShadowStubInstantiationException( stubClassName + " has problems during construction" );

	}

    }

    /**
     * start with a single class and get all the interfaces
     *
     * @param startingClass the class to start everything with
     * @return a vector of interfaces
     */
    public static Vector getAllInterfaces( Class startingClass ) {
	Vector allInterfaces = new Vector();
	Class[] interfaces = startingClass.getInterfaces();
	for( int count=0;count<interfaces.length;count++ ) {
	    if( !allInterfaces.contains( interfaces[count] ) )
	        allInterfaces.addElement( interfaces[count] );
	    Vector localInterfaces = getAllInterfaces( interfaces[count] );
	    Enumeration e = localInterfaces.elements();
	    while( e.hasMoreElements() ) {
		Class singleInterface = (Class)e.nextElement();
		if( !allInterfaces.contains( singleInterface ) )
		    allInterfaces.addElement( singleInterface );
		
	    }
	    
	}
	return allInterfaces;

    }

    /**
     * this gets all the methods which are in interfaces which extend
     * ShadowInterface.  necessary so we can confirm that the stub we
     * are loading is the stub that we need
     *
     * @param interfaceClass the interfaces that we are going to search through
     * @return a vector of methods which are the ones that have to be in the stub 
     */
    public static Vector getAllShadowMethods( Class interfaceClass ) {
	 return getAllShadowMethods( interfaceClass, new Vector() );

     }

     /**
      * this gets all the methods which are in interfaces which extend
      * ShadowInterface.  necessary so we can confirm that the stub we
      * are loading is the stub that we need.  this needs to only grab
      * the public methods.  make sure of this!
      *
      * @param interfaceClass the interfaces that we are going to search through
      * @param path this is the current path through the interfaces that we are working with -- if calling this from outside, start empty
      * @return a vector of methods which are the ones that have to be in the stub 
      */
     private static Vector getAllShadowMethods( Class interfaceClass, Vector path ) {
	 Vector methods = new Vector();
	 if( interfaceClass.equals( Shadow.class ) ) {
	     // this means that we have found something that is a
	     // shadow interface, so we should traverse up the path
	     // and add all those things to the vector of methods
	     Enumeration e = path.elements();
	     while( e.hasMoreElements() ) {
		 try {
		     Class singleClass = (Class)e.nextElement();
		     Method[] classMethods = singleClass.getMethods();
		     for( int methodCount=0;methodCount<classMethods.length;methodCount++ )
			 if( !methods.contains( classMethods[methodCount] ) )
			     methods.addElement( classMethods[methodCount] );

		 } catch( SecurityException error ) {
		     // just ignore this class then

		 }

	     }

	 }

	 Class[] interfaces = interfaceClass.getInterfaces();
	 for( int count=0;count<interfaces.length;count++ ) {
	     // now we have to still search through all the other interfaces this extends
	     Vector clonedPath = (Vector)path.clone();
	     clonedPath.addElement( interfaces[count] );
	     Enumeration e = getAllShadowMethods( interfaces[count], clonedPath ).elements();
	     while( e.hasMoreElements() ) {
		 Method method = (Method)e.nextElement();

		 // we cannot do a simple test to see if the methods
		 // are .equals().  we need to test if everything
		 // -except- the return is the same.  then we have to
		 // see if one of the methods return is a superclass of
		 // the other, then we return the superclass
		 boolean add = true;
		 boolean replace = false;
		 int goThrough = 0;
		 for( goThrough=0;goThrough<methods.size();goThrough++ ) {
		     Method testing = (Method)methods.elementAt( goThrough );

		     // first make sure we don't have the same thing
		     if( method.equals( testing ) ) {
			 add = false;
			 break;

		     }
		     
//  		     // first -- let's see if the method we want
//  		     // to add is from a superclass of the method
//  		     // already in there
//  		     if( !testing.getDeclaringClass().isAssignableFrom( method.getDeclaringClass() ) )
//  			 continue;
		     
		     // test to see if the name is the same
		     String methodName = method.getName().substring( method.getName().lastIndexOf( "." ) + 1 );
		     String testingName = testing.getName().substring( testing.getName().lastIndexOf( "." ) + 1 );
		     if( !methodName.equals( testingName ) )
			 continue;
		     
		     // test to see if the same parameters are given
		     Class[] methodParams = method.getParameterTypes();
		     Class[] testingParams = testing.getParameterTypes();
		     if( !Global.arrayEquals( methodParams, testingParams ) )
			 continue;
		     
		     // test to see if the exceptions are teh same
		     Class[] methodExceptions = method.getExceptionTypes();
		     Class[] testingExceptions = testing.getExceptionTypes();
		     if( !Global.arrayEquals( methodExceptions, testingExceptions ) )
			 continue;
		     
		     // now see if the return type is a super
		     // class -- if it is, then we need to replcae
		     // it
		     if( testing.getReturnType().isAssignableFrom( method.getReturnType() ) ||
			 ( testing.getReturnType().isPrimitive() && method.getReturnType().isPrimitive() ) ) {
			 add = false;
			 replace = true;
			 break;
			 
		     }
		     
		 }
		 if( replace ) {
		     methods.removeElementAt( goThrough );
		     methods.insertElementAt( method, goThrough );
		     
		 }
		 if( add ) 
		     methods.addElement( method );

	     }

	 }
	 return methods;
	
     }

    /**
     * get all the "relavent" interfaces for creating the stub.  this
     * is similar to going through to find the appropriate methods,
     * but instead of recording the methods, let's record the
     * interfaces.  this is needed for the stub, so the stub can be
     * cast into the appropriate thing
     *
     * @param startingClass the class that we want to get the interfaces from
     * @return a vector of teh appropriate interfaces -- if calling this from outside, pass a new empty vector
     */
    public static Vector getShadowInterfaces( Class startingClass ) {
	return getShadowInterfaces( startingClass, new Vector() );
	
    }

    /**
     * get all the "relavent" interfaces for creating the stub.  this
     * is similar to going through to find the appropriate methods,
     * but instead of recording the methods, let's record the
     * interfaces.  this is needed for the stub, so the stub can be
     * cast into the appropriate thing
     *
     * @param startingClass the class that we want to get the interfaces from
     * @param path the path that we took so far to get to this point in the search
     * @return a vector of teh appropriate interfaces
     */
    private static Vector getShadowInterfaces( Class startingClass, Vector path ) {
	Vector interfaces = new Vector();
	if( startingClass.equals( Shadow.class ) ) {
	    // this means that we have found something that is a
	    // shadow interface, so we should traverse up the path
	    // and add all those things to the vector of methods
	    Enumeration e = path.elements();
	    while( e.hasMoreElements() ) {
		Class singleClass = (Class)e.nextElement();
		if( !interfaces.contains( singleClass ) )
		    interfaces.addElement( singleClass );
		
	    }

	}

	Class[] interfaceSearching = startingClass.getInterfaces();
	for( int count=0;count<interfaceSearching.length;count++ ) {
	    // now we have to still search through all the other interfaces this extends
	    Vector clonedPath = (Vector)path.clone();
	    clonedPath.addElement( interfaceSearching[count] );
	    Enumeration e = getShadowInterfaces( interfaceSearching[count], clonedPath ).elements();
	    while( e.hasMoreElements() ) {
		Class interfaceFound = (Class)e.nextElement();
		if( !interfaces.contains( interfaceFound ) )
		    interfaces.addElement( interfaceFound );

	    }

	}
	return interfaces;
	
    }
    

    /**
     * get the stub for the shadow
     *
     * @return the shadow's stub
     */
    public ShadowStub getStub() {
	return stub;

    }

    public void setCell(Cell cell) {
        if (myCell == null)
            myCell = cell;
        else
            Debug.error("someone tried to set my cell, but it's been set!");
    }
    
    /**
     * Returns true if the shadow reports it is ready for action
     *
     * @return true if ready, false otherwise
     */
    public boolean isReady() { return ready; }

    /**
     * Set the value of ready in the shadow -- by default it is ready,
     * but you can use this in the constructor to say its not ready by
     * deafult.  once a shadow is set as ready, it cannot be unset
     *
     * @param ready set the value of the readiness of this shadow 
     */
    protected void setReady( boolean ready ) { 
	if( ready ) return;
	this.ready = ready; 
	if( ready )
	    synchronized( this ) {
		notifyAll();
	    }
    }

    /**
     * Something that needs to wait until a shadow becomes ready can
     * call this.  This method will block until the shadow calls
     * setReady( true )
     *
     * @see setReady 
     */
    public void blockUntilReady() {
	while( !this.ready ) 
	    try {
		synchronized( this ) {
		    wait();
		}
	    } catch( InterruptedException error ) { }
    }

    /**
     * Get the value of desc.
     * @return Value of desc.
     */
    public Description getDescription() { return description; }
    
    /**
     * Set the value of desc.
     * @param v  Value to assign to desc.
     */
    public void setDescription( Description v ) { this.description = v; }
    

    /**
     * a shadow may specify a default agent, such as a basic export
     * agent.  the server on which it is being run may choose not to
     * request or instantiate the default agent, however.
     *
     * @return in theory it would return an agent impl to use, but in this case it reutnrs null
     */
    public AgentImpl getDefaultAgent() { 
	Debug.notice( "there is no default agent for this shadow" );
	return null; 
    }

    /**
     * Override this to do any initialization
     */
    public abstract void init() 
    throws ShadowInitializationException;

    /**
     * load the description file
     */
    public void loadDescription() { loadDescription( null ); }

    /**
     * load the description from a file
     *
     * @param semanticFile the name of the file in which to load the description from -- it can be null, then this class will attempt to find a file to load from
     */
    public void loadDescription(String semanticFile) {
	//	Description.initParser();
	DOM dom = new DOM();

	String className = this.getClass().toString();
	if( semanticFile == null) {
	    // turn "foo.bar.FooShadow" into "xml/FooShadow.xml"
	    int nameStart = className.lastIndexOf('.') + 1;
	    String faceName = new String("xml/" + className.substring(nameStart) + ".xml");

	    // Open the icon file as a resource (uses classpath)
	    InputStream is = Global.getConfigAsInputStream(faceName);
	    if(is != null) {

		Document doc = dom.openDocument(is);
		Description d = new Description(this, doc);
		
		setDescription(d);

		/*
		Description d = null;
		try {
		    setDescription( d = new Description(this, "hive:" + faceName,
							new InputStreamReader(is)) );
		} catch (MalformedRDFFileException e) {
		    Debug.warning("RDF File errors in " + faceName + ", " + e, e);
		}
		*/

		InputStream ris = getClass().getResourceAsStream(faceName);
		if (ris != null && d != null) {
		    Debug.warning("Two descriptions available for shadow! I still have to write the XML merging code tho... - Tucker");
		    //d.addDoc(this, className, new InputStreamReader(ris));
		}
	    }
	}

	else {
	    try {
		FileReader fr = new FileReader(Global.getConfigAsFile(semanticFile));
		Document doc = dom.openDocument(fr);
		Debug.notice("Loading semantic description file from " + semanticFile + ", " + fr);

		setDescription(new Description(this, doc));
	    } catch (Exception e){
		Debug.warning("Failed to load Description " + semanticFile + " as specified in config", e);
	    }
	}
    }
}
