package net.hivecell.hive.shadow.cell;

import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.agent.AgentImpl;
import net.hivecell.hive.agent.EventReceivingAgent;
import net.hivecell.hive.agent.cell.AgentMessagingShadowAgentImpl;
import net.hivecell.hive.agent.cell.UnknownAgentException;
import net.hivecell.hive.cell.MobileAgentInputStream;
import net.hivecell.hive.cell.MobileAgentOutputStream;
import net.hivecell.hive.event.HiveEvent;
import net.hivecell.hive.event.UnknownEventException;
import net.hivecell.hive.shadow.EventPublisherShadow;
import net.hivecell.hive.shadow.EventPublisherShadowImpl;
import net.hivecell.hive.shadow.ShadowInitializationException;
import net.hivecell.hive.shadow.ShadowNotFoundException;
import net.hivecell.hive.shadow.Signalable;
import net.hivecell.hive.shadow.cell.CellEventMonitor;
import net.hivecell.hive.support.CellAddress;
import net.hivecell.hive.support.CellConnectException;
import net.hivecell.hive.support.Debug;
import net.hivecell.hive.support.Null;
import net.hivecell.hive.support.Queue;
import net.hivecell.hive.support.async.AsyncCall;
import net.hivecell.hive.support.async.AsyncResult;

import java.io.BufferedInputStream;
import java.io.BufferedOutputStream;
import java.io.ByteArrayInputStream;
import java.io.ByteArrayOutputStream;
import java.io.ObjectInputStream;
import java.io.ObjectOutputStream;
import java.io.InputStream;
import java.io.IOException;
import java.io.NotSerializableException;
import java.io.Serializable;
import java.lang.reflect.Method;
import java.net.ServerSocket;
import java.net.Socket;
import java.rmi.RemoteException;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.Vector;


/**
 * this class will allow us to remove the need for each individual
 * agent to do its own event sending and receiving through RMI -- this
 * will allow us to remove the need to spawn off so many asynchronous
 * messaging threads, and will also allow us to cut down on the number
 * of sockets hive needs especuially when there is a significant
 * amount of messaging occuring.
 *
 * by default, the event shadow will open a random port there will be
 * an agent which will sit on the cell to help shadows negotiate what
 * ports they are sending to.  right now (and therefore probably
 * never) the port on which the shadow is operating can ont be changed
 * on the fly.  that just seems like too much of a headache to
 * negotiate that one.  
 */
public class AgentMessagingShadowImpl
extends EventPublisherShadowImpl
implements AgentMessagingShadow {
    // the internal variables
    
    public final static int defaultPort = 23234; // this is the default messaging port

    private ServerSocket eventReceivingSocket = null; // the socket everything is living on
    private int eventListeningPort = -1; // the port on which the socket is on
	

    /**
     * get the port number we are listening on
     *
     * @return the port number of the eventReceivingSocket
     */
    public int getEventListeningPort() { return eventListeningPort; }

    /**
     * set the port number of the event receiving port.  right now
     * this does nothing -- it is just here for java beans type stuff
     *
     * @param portNumber the port number we want to set as 
     */
    public void setEventListeningPort( int portNumber ) { 
	Debug.warning( "changing of event port is not supported (still on port " + getEventListeningPort() + ")" );

    }


    protected Method notifyMethod; // use this to "cache" the notify method object
    protected Method sendEventMethod; // use this to "cache"

    // these are the three queues that are all "interlinked" -- if one
    // of them is being updated, the other two should not be being
    // touched at that time.  for this semantics, use the
    // outgoingQueues object as the object to synchronize on whenever
    // touching any of these hashtables
    protected Hashtable outgoingQueues; // these are the outgoing queues.  celladdress->outgoingQueue
    protected Hashtable incomingQueues; // these are the incoming queues.  celladdress->incomingQueue
    protected Hashtable openSockets; // these are all the open sockets we are using.  celladdress->socket

    // these are the queues that are being created.  this is a
    // synchronization nightmare.  so this hashtable will map the
    // celladdress to the net.hivecell.hive.support.Null object.  the
    // first person to start a queue will create this mapping, all
    // others will simply grab the object.  when the first object has
    // created the queue, it will remove the mapping and then send a
    // notify through the Null
    protected Hashtable intermediateQueues;


    // our hook to the agent that we need to do our work with 
    private AgentMessagingShadowAgentImpl agentMessagingShadowAgent;

    // our hook to the cell event monitor so we can notify whenever an
    // event is sent -- we are no longer (march 30, 2000) guaranteeing
    // that all the messages will be notated -- we will do our best.
    // in order to do it right, we will need to redo RMI's socket
    // stuff
    private CellEventMonitor cellEventMonitor;


    // the initialization stuff

    /**
     * create ourselves an event shadow
     */
    public AgentMessagingShadowImpl() { super(); }

    /**
     * clients can call this to add themselves as a listener to this
     * type of event.  we will also be looking for the agent messaging
     * shadow agent impl through this
     *
     * @param o the client 
     */
    public void addListener( Signalable o ) { 
	super.addListener( o );
	if( ( agentMessagingShadowAgent == null ) && ( o instanceof AgentMessagingShadowAgentImpl ) ) {
	    Debug.notice( "got a connection from our agent messaging shadow agent" );
	    agentMessagingShadowAgent = (AgentMessagingShadowAgentImpl)o;

	}

    }

    /**
     * initialize the event shadow -- this creates the server socket
     * that we need to receive the vents, sets up the
     * incoming/outgoing queues, and create the agent that we will use
     * to query other cells to see what port we need to talk to them
     * on
     *
     * @exception ShadowInitalizationException is thrown when we
     *            cannot create the server socket to listen on 
     */
    public void init() 
    throws ShadowInitializationException {
	try {
	    cellEventMonitor = (CellEventMonitor)myCell.getShadow( (CellEventMonitorImpl.class).getName() );

	} catch( ShadowNotFoundException error ) {
	    // this shouldn't happen

	}

	try {
	    try {
		eventReceivingSocket = new ServerSocket( defaultPort );

	    } catch( IOException internalError ) {
		Debug.notice( "port " + defaultPort + " already used" );

	    }
	    if( eventReceivingSocket == null )
		eventReceivingSocket = new ServerSocket( 0 );
	    eventListeningPort = eventReceivingSocket.getLocalPort();
	    Debug.notice( "created event receiving port on " + eventListeningPort );

	} catch( IOException error ) {
	    Debug.critical( "having trouble opening an event receiving port.  shutting down." );
	    throw new ShadowInitializationException( error );
	}

	try {
	    Class[] notifyParameterTypes = { (HiveEvent.class) };
	    notifyMethod = (EventReceivingAgent.class).getMethod( "notify", notifyParameterTypes );
	    Debug.notice( "got the notify method : " + notifyMethod );

	    Class[] sendEventParameterTypes = { (EventReceivingAgent[].class), (CellAddress[].class), (HiveEvent.class) };
	    sendEventMethod = (AgentMessagingShadowImpl.class).getMethod( "sendEventHelper", sendEventParameterTypes );
	    Debug.notice( "got the sendEventHelper method : " + sendEventMethod );

	} catch( Exception error ) {
	    // this exception should not happen

	}

	outgoingQueues = new Hashtable();
	incomingQueues = new Hashtable();
	intermediateQueues = new Hashtable();
	openSockets = new Hashtable();
	Debug.notice( "finished setting up the data structures needed to maintain the queues." );

	// this is the little complicated part -- we are first going
	// to spawn of a thread just to handle the accept() methods on
	// the server socket so this way we can have an incoming queue
	// from everything we are connected to.  whenever we get a
	// connection to teh server socket, it spawns a thread to
	// handle that.  now that thread first opens a
	// ObjectInputStream (this class should be renamed) to
	// the host -- this way in case we need any mobility, the
	// "black magic" will just happen.  the first thing that
	// should be sent through that stream is a serialized version
	// of the cell address of the cell that has made the
	// connection, then everything after that will just be events.
	// if for some reason we get an exception on the stream, then
	// we will protect ourselves and just punt the connection
	(new Thread( new Runnable() {
		public void run() {
		    while( true ) {
			try {
			    final Socket incoming = eventReceivingSocket.accept();
			    Debug.notice( "receiving a connection from an incoming event stream." );
			    
			    // start monitoring for an incoming connection
			    IncomingRunningQueue incomingEventQueue = new IncomingRunningQueue( incoming );
			    (new Thread( incomingEventQueue ) ).start();
			    
			    // wait on the incoming connection to see
			    // if it is safe to test to see if the
			    // socket is in the appropriate data
			    // structures
			    while( !incomingEventQueue.readyToTest() )
				synchronized( incomingEventQueue ) {
				    try {
					incomingEventQueue.wait();
				    } catch( InterruptedException error ) { }
				}
			    
			    // test to see if the socket is there, if
			    // it is, then let's create a queue to it
			    CellAddress incomingAddress = incomingEventQueue.getIncomingAddress();
			    if( incomingAddress != null ) {
				Debug.notice( "made a connection from " + incomingAddress );
				synchronized( outgoingQueues ) {
				    incomingQueues.put( incomingAddress, incomingEventQueue );

				}

				Null intermediateFlag = new Null();
				synchronized( outgoingQueues ) {
				    intermediateQueues.put( incomingAddress, intermediateFlag );
				}
				try {
				    OutgoingRunningQueue outgoingQueue = new OutgoingRunningQueue( incomingAddress, incoming );
				    (new Thread( outgoingQueue ) ).start(); // launch the thread
				    synchronized( outgoingQueues ) {
					// ok!  we have the queue!  let's sync up our
					// data structures, and then lets send the
					// notoify through the intermediate creation
					// flag
					outgoingQueues.put( incomingAddress, outgoingQueue );
					intermediateQueues.remove( incomingAddress );
					synchronized( intermediateFlag ) {
					    intermediateFlag.notifyAll();
					    
					}
				    }
				} catch( IOException error ) {
				    Debug.warning( "had trouble creating the outgoing queue to " + incomingAddress );
				    // ok -- at this point, it just
				    // isn't working.  everything --
				    // incoming and outgoing.  better
				    // to be a bastard than to
				    // introduce some other
				    // complications

				    // this should also manage the
				    // data structures
				    incomingEventQueue.stop();

				}
			    }
			    
			} catch( IOException error ) {
			    // for some reason we could not accept an
			    // incoming stream -- oh well, just ignore
			    // it then
			    Debug.warning( "exception while trying to accept an incoming event stream.", error );
			    
			}
			
		    }
		}
	})).start();
	Debug.notice( "started listening on port " + eventListeningPort + " for events" );
	
	setReady( true );

    }

    /**
     * this class will handle an incoming connection from another host
     * and deal with the negotiation and with the method calls that
     * need to be made 
     */
    class IncomingRunningQueue
    implements Runnable {
	private Socket incoming; // this is the incoming socket
	private boolean readyToTest;  // this is just a flag that signals whether we can safely test the data structures to see if the socket has been added
	private CellAddress incomingAddress = null;
	
	public IncomingRunningQueue( Socket incomingSocket ) {
	    this.incoming = incomingSocket;

	}

	/**
	 * shut this queue down
	 */
	public void stop() {
	    try {
		incoming.close();

	    } catch( IOException error ) {
		// oh well -- whatever.  just forget it.

	    }

	}

	/**
	 * a test to see whether we can check the data structures to
	 * see if the socket has been added to it.  if it has, then we
	 * can tell what is the cell address
	 *
	 * @return true if we can test this, false otherwise
	 */
	public boolean readyToTest() {
	    return readyToTest;

	}
	
	/**
	 * return what cell address we are talking to 
	 *
	 * @return the cell address if we have it, false otherwise
	 */
	public CellAddress getIncomingAddress() {
	    return incomingAddress;

	}
	
	public void run() {
	    try {
		BufferedInputStream inputStream = new BufferedInputStream( incoming.getInputStream() );

		// read the address
		byte[] lengthBuffer = readStream( inputStream, 4 );
		int destinationAddressLength = byteArrayToInt( lengthBuffer );
		byte[] readingBuffer = new byte[destinationAddressLength];
		inputStream.read( readingBuffer, 0, destinationAddressLength );
		ObjectInputStream deserializingStream = new ObjectInputStream( new ByteArrayInputStream( readingBuffer ) );
		incomingAddress = (CellAddress)deserializingStream.readObject();
		
		boolean containsSocket = false;
		synchronized( outgoingQueues ) {
		    if( openSockets.containsKey( incomingAddress ) )
			containsSocket = true;

		}
		if( containsSocket )
		    Debug.warning( "just received a connection from " + incomingAddress + ", but we already have an open socket" );

		else {
		    Debug.notice( "have an incoming event stream from " + incomingAddress );
		    synchronized( outgoingQueues ) {
			openSockets.put( incomingAddress, incoming );

		    }
		    synchronized( this ) {
			readyToTest = true;
			notifyAll(); // signal that the socket is ok and that we have put it in the data structures

		    }
		}

		EventReceivingAgent[] destinationAgents = null;
		Object hiveEventObject = null;
	    incomingRead:
		while( true ) {
		    // get the array of event receiving agents
		    lengthBuffer = readStream( inputStream, 4 );
		    int destinationAgentsArrayLength = byteArrayToInt( lengthBuffer );
		    Debug.notice( "reading a destination array of " + destinationAgentsArrayLength + " bytes" );
		    readingBuffer = readStream( inputStream, destinationAgentsArrayLength );
		    deserializingStream = new ObjectInputStream( new ByteArrayInputStream( readingBuffer ) );
		    destinationAgents = (EventReceivingAgent[])deserializingStream.readObject();
		    Debug.notice( "deserialized an event destination array from " + incomingAddress );

		    // get the hive event
		    lengthBuffer = readStream( inputStream, 4 );
		    int hiveEventLength = byteArrayToInt( lengthBuffer );
		    if( hiveEventLength != 0 ) {
			Debug.notice( "reading a hive event of " + hiveEventLength + " bytes from " + incomingAddress );
			readingBuffer = readStream( inputStream, hiveEventLength );
			deserializingStream = new ObjectInputStream( new ByteArrayInputStream( readingBuffer ) );
			hiveEventObject = deserializingStream.readObject();

		    } else {
			Debug.notice( "reading a null hive event from " + incomingAddress );
			hiveEventObject = null;

		    }

		    // we just need to protect ourselves in case we don't actually get
		    // an AgentMessagingShadowEvent down the stream
		    if( !( hiveEventObject instanceof HiveEvent ) ) {
			Debug.notice( "received a non-event from " + incomingAddress );
			continue incomingRead;

		    }
		    HiveEvent event = (HiveEvent)hiveEventObject;
		    Debug.notice( "received an event from " + incomingAddress );
						
		    // call the notify method on the local reference of the
		    // agent asynchronously and deliver the event
		    for( int count=0;count < destinationAgents.length;count++ ) {
			AgentImpl localReference = AgentMessagingShadowImpl.this.myCell.getLocalReference( destinationAgents[count] );
			if( ( localReference != null ) &&
			    ( localReference instanceof EventReceivingAgent ) ) {
			    // ok!  we have a local reference to the event receiving 
			    // agent -- call notify on him asynchronously.  we don't 
			    // care about the result of the call, so don't even bother 
			    // keeping it around
			    Object[] notifyParameters = { event };
			    AsyncCall asyncNotify = new AsyncCall( AgentMessagingShadowImpl.this.notifyMethod, notifyParameters );
			    asyncNotify.callOn( (EventReceivingAgent)localReference );
			    
			}

		    }
		}			
		    
	    } catch( Exception error ) {
		Debug.warning( "exception on the incoming event stream from " + incomingAddress + ".  disconnecting." );
		synchronized( this ) {
		    readyToTest = true;
		    incomingAddress = null;
		    notifyAll(); // signal that we have failed.  it probably menas that we have a bad socket

		}
					    
	    } finally {
		// if we ever get here, this means that we are closing down the connection
		try {
		    incoming.close();
		    
		} catch( IOException error ) {
		    // whatever
		    
		}
		if( incomingAddress != null )
		    synchronized( outgoingQueues ) {
			openSockets.remove( incomingAddress );

			// this is something that i need 
			outgoingQueues.remove( incomingAddress );
			incomingQueues.remove( incomingAddress );
			
		    }
	    }

	}
    }

    
    // methods meant for the event sending helpers to call

    /**
     * call this method to send an event.  this method will wrap the
     * HiveEvent in another object so that the AgentMessagingShadow
     * will be able to decode the destination of the event.  this
     * needs to throw exceptions just so in case it doesn't work
     *
     * @param destinationAgents the agents which is meant to receive the HiveEvent
     * @param destinationAddresses the cell addresses of teh agent that we are going to
     * @param he the HiveEvent which is going to be sent to the other cell 
     * @return an async result object which has the result of the event being sent
     */
    public AsyncResult sendEvent( EventReceivingAgent[] destinationAgents, CellAddress[] destinationAddresses, HiveEvent he ) {
	Object[] eventParameters = { destinationAgents, destinationAddresses, he };
	AsyncCall asyncCall = new AsyncCall( sendEventMethod, eventParameters );
	return asyncCall.callOn( this );

    }


    /**
     * this is the helper function for the send event.  this will be
     * called using the async package so we can return exceptions
     * asynchronously and stuff.  do not call this method directly, it
     * is only made to be called from the sendEvent function.
     *
     * @param destinationAgents the agents which is meant to recieve the HiveEvent
     * @param destinationAddresses the cell addresses of the agent that we are going to
     * @param he the HiveEvent which is going to be sent to the other cell 
     * @exception CellConnectException is thrown if we have problems talking to the other cell
     * @exception UnknownEventException is thrown if the event receiver doesn't understand the event
     * @exception RemoteException is thrown if we have to defer to our agent to send the event and it has a problem
     * @exception IOException is thrown whenever we have problems with the socket we are connecting over
     */
    public void sendEventHelper( EventReceivingAgent[] destinationAgents, CellAddress[] destinationAddresses, HiveEvent he ) 
    throws CellConnectException, UnknownEventException, RemoteException, IOException {

	// first -- let's arrange this information in a manner which
	// is a lot more easy to deal with.  take the two arrays, and
	// compact them into a hashtable which will be mapping cell
	// address to a vector of all the agents (eww -- there has to
	// a better way to form a data structure like this)
	Hashtable recipients = new Hashtable();
	for( int count=0;count < destinationAddresses.length;count++ ) {
	    Vector destinations = (Vector)recipients.get( destinationAddresses[count] );
	    if( destinations == null ) {
		destinations = new Vector();
		recipients.put( destinationAddresses[count], destinations );

	    }
	    destinations.addElement( destinationAgents[count] );

	}

	// if the destination address is our address, then let's just
	// call the right notify method! -- so lets get all the local ones out of the way
	Vector localRecipients = (Vector)recipients.remove( myCell.getAddress() );
	if( localRecipients != null ) {
	    Enumeration e = localRecipients.elements();
	    while( e.hasMoreElements() ) {
		EventReceivingAgent era = (EventReceivingAgent)e.nextElement();
		if( !(era instanceof java.rmi.server.RemoteStub) )
		    // weird -- we somehow got the real local reference to the agent
		    era.notify( he );
		
		else if( myCell.isAgentLocal( era ) )
		    // ok -- this agent is local
		    ((EventReceivingAgent)myCell.getLocalReference( era )).notify( he );

		else {
		    // this should -never- happen -- this is a failsafe in
		    // case it does.  but if it does, then we did
		    // something wrong in Cell.java
		    
		    Debug.warning( era + " is local, but we cannot get the local reference" );
		    // dammit, ok -- ship ourselves upstairs to our agent
		    // and ask him to just deliver the notify via RMI or
		    // something -- blech
		    agentMessagingShadowAgent.deliverEvent( era, he );
		    
		}
	    }
	}

	// if we don't have anybody else to talk to, then we just punt
	// off at this point
	if( recipients.size() == 0 )
	    return;

	// first lets see if we even have any other agents to send to
	// via the socket protocol, if we don't, punt this method
	// right here -- if we do, let's serialize the event right
	// here since that is the slow part.  also, if it turns out
	// that the event cannot be serialized, then we don't want
	// each of the individual queues figuring this out for
	// themselves -- let's catch it right here
	byte[] serializedHiveEvent = null;
	ByteArrayOutputStream serializedHiveEventStream = new ByteArrayOutputStream();
	ObjectOutputStream objectOutputStream = new ObjectOutputStream( serializedHiveEventStream );
	try {
	    objectOutputStream.writeObject( myCell.getRemoteObject( he ) ); 
	    serializedHiveEvent = serializedHiveEventStream.toByteArray();
	    
	} catch( Exception error ) {
	    Debug.warning( he + " from " + he.getSource() + " is not serializable.  punting event." );
	    return;
	    
	}

	// now go through the rest of the addresses by enumerating through
	Enumeration e = recipients.keys();
	while( e.hasMoreElements() ) {
	    CellAddress destinationAddress = (CellAddress)e.nextElement();
	    Vector destinationAddressAgents = (Vector)recipients.get( destinationAddress );

	    // this needs to happen quickly because we don't want to have
	    // to hold up other events because one is taking a long time
	    // to go through.  the proposed method for this happening is
	    // going to be as follows:
	    //
	    // the shadow will synchronously check its hash table to see
	    // if it already has a queue, and if it doesn't then we check
	    // to see if one is being made to the appropriate cell
	    // address, if it does exist it simply enqueues this event for
	    // that queue.  if we don't, then this gets a little more
	    // difficult -- we will need to spawn off a thread to handle
	    // this one.  the thread will first register that it is
	    // attempting to make a hive event connection to that cell
	    // address, then attempt to negotiate a connection to that
	    // cell.  once it has the connection, it will spew the event.
	    OutgoingRunningQueue outgoingQueue = null;
	    Socket outgoingSocket = null;
	    Null intermediateCreationFlag = null;
	    boolean creation = false;
	    synchronized( outgoingQueues ) {
		// first see if we already have an outgoing queue
		outgoingQueue = (OutgoingRunningQueue)outgoingQueues.get( destinationAddress );

		// ok - so we don't have a queue already, do we have a socket to that machine
		if( outgoingQueue == null ) {
		    outgoingSocket = (Socket)openSockets.get( destinationAddress );

		    if( outgoingSocket == null ) {
			// if we don't have a socket, then we need to register
			// ourselves as beginning to make a connection
			intermediateCreationFlag = new Null();
			intermediateQueues.put( destinationAddress, intermediateCreationFlag );
			creation = true;

		    } else {
			// ok - we do have a socket to that machine, but
			// we can't use that socket to do anything yet.
			// the question is, is anybody working on trying
			// to make a queue to that destination?
			intermediateCreationFlag = (Null)intermediateQueues.get( destinationAddress );
			if( intermediateCreationFlag == null ) {
			    // nobody is trying to make a queue to this
			    // socket, ok -- let's grab it so we can
			    // create it
			    intermediateCreationFlag = new Null();
			    intermediateQueues.put( destinationAddress, intermediateCreationFlag );
			    creation = true;

			}
		    }
		}

		// ok - we are out of the synchronized zone, what should
		// we do?  let's test the flags.  go backwards
	    
		// the outgoingQueue = null, the socket = null, the
		// intermediateCreationFlag != null, and the creationflag
		// = false
		//
		// or
		//
		// the outgoingQueue = null, the socket != null,
		// intermediateCreationFlag != null, and the creationflag
		// = false.  let's just hold off and wait for the notify
		// through the intermediateCreationFlag
		if( ( outgoingQueue == null ) &&
		    ( creation == false ) &&
		    ( intermediateCreationFlag != null ) ) {

		    // wait on this object -- once we are done waiting,
		    // this means that we can grab the queue
		    boolean waiting = true;
		    while( waiting ) {
			synchronized( intermediateCreationFlag ) {
			    try {
				intermediateCreationFlag.wait();
				waiting = false;

			    } catch( InterruptedException error ) { }
			}
		    }

		    // grab the queue -- need to make sure we actually
		    // have a queue!  there is a chance the queue does not
		    // exist
		    synchronized( outgoingQueues ) {
			outgoingQueue = (OutgoingRunningQueue)outgoingQueues.get( destinationAddress );

		    }

		    // reset the rest of our stuff
		    outgoingSocket = null;
		    intermediateCreationFlag = null;
		    creation = false;

		}

		// the outgoingQueue = null, the socket = null, the
		// intermediateCreationFlag != null and the creationflag =
		// true -- this means we need to create the socket, set
		// the socket to != null, then fall through
		if( ( outgoingQueue == null ) &&
		    ( outgoingSocket == null ) &&
		    ( intermediateCreationFlag != null ) &&
		    ( creation == true ) ) {
		    Debug.notice( "creating a socket to " + destinationAddress );
		    try {
			outgoingSocket = new Socket( destinationAddress.getHost(), agentMessagingShadowAgent.getCellAgentMessagingPort( destinationAddress ) );
		    
		    } catch( CellConnectException error ) {
			Debug.error( "we cannot connect to " + destinationAddress + " to send an event" );
			// at this point we need to notify the sender that this is not going to work
			throw error;

		    } catch( UnknownAgentException error ) {
			Debug.notice( "cell at " + destinationAddress + " does not have an AgentMessagingShadowAgent" );
			// if they don't have the agent messaging shadow
			// agent, then let's try on the default agent
			// messaging port before giving up
			try {
			    outgoingSocket = new Socket( destinationAddress.getHost(), defaultPort );
			
			} catch( IOException internalError ) {
			    Debug.error( "cannot create a messaging socket to " + destinationAddress );
			    // now notify the sender that this is not going to work
			    throw internalError;
			
			}

		    } catch( IOException error ) {
			Debug.error( "cannot create a messaging socket to " + destinationAddress );
			// tell sender that this is not going to work
			throw error;

		    }

		    // if we got this far, then we have the socket, so let's update the tables
		    synchronized( outgoingQueues ) {
			openSockets.put( destinationAddress, outgoingSocket );

		    }
		}

		// the outgoingQueue = null, the socket != null,
		// intermediateCreationFlag != null, and the creationflag
		// = true -- this means take the socket and make the
		// queue.  we are also going to take the socket and
		// prepare it to listen to incoming messages
		if( ( outgoingQueue == null ) &&
		    ( outgoingSocket != null ) &&
		    ( intermediateCreationFlag != null ) &&
		    ( creation == true ) ) {
		    Debug.notice( "taking the socket to " + destinationAddress + " and creating an outgoing queue" );
		    try {
			outgoingQueue = new OutgoingRunningQueue( destinationAddress, outgoingSocket );
			(new Thread( outgoingQueue ) ).start(); // launch the thread
			synchronized( outgoingQueues ) {
			    // ok!  we have the queue!  let's sync up our
			    // data structures, and then lets send the
			    // notoify through the intermediate creation
			    // flag
			    outgoingQueues.put( destinationAddress, outgoingQueue );
			    intermediateQueues.remove( destinationAddress );
			    synchronized( intermediateCreationFlag ) {
				intermediateCreationFlag.notifyAll();

			    }
			}
			Debug.notice( "created an outgoing stream to " + destinationAddress );

			// manage the incoming end of the connection also
			(new Thread( new IncomingRunningQueue( outgoingSocket ) ) ).start();
		    
		    } catch( IOException error ) {
			Debug.error( "problem creating the outgoing queue with the socket to " + destinationAddress );
			// notify sender that we are screwed at this point
			throw error;

		    }

		}

		// the outgoingQueue != null.  woo-hoo!
		Debug.notice( "sending the event(s) out to queue for " + destinationAddress );	
		EventReceivingAgent[] receivingAgents = new EventReceivingAgent[destinationAddressAgents.size()];
		destinationAddressAgents.copyInto( receivingAgents );
		outgoingQueue.sendEvent( receivingAgents, serializedHiveEvent );

	    }

	}

    }


    // internal classes for AgentMessagingShadow

    /**
     * this is the queue used to send events to the certain cell
     * addresses.  this queue takes an AgentMessagingShadowEvent through
     * the enqueueEvent method, and then it will just go through the list
     * and serialize the events and send them through.  the sender of the
     * event will not get any notification that the event went through, or
     * didn't go through as the case may be.  
     */
    class OutgoingRunningQueue
    implements Runnable {
    
	// internal variables
	private boolean timeToStop = false;
	
	private CellAddress destinationAddress; // the address that we are talking to

	private Queue queue; // the queue holding the events to be sent

	private BufferedOutputStream outputStream; // the stream we will be using to communicate
	private Socket socket; // the socket to the other agent messaging shadow

	private ObjectOutputStream serializingStream; // t he stream we will use to serialize stuff
	private ByteArrayOutputStream serializedStream; // the stream we grab our serialized objects from
    
	// initialization stuff
	/**
	 * create a new agent messaging shadow -- this takes the
	 * socket, and the first thing it does is ship a serialized
	 * version of our cell's cell address through so the other
	 * cell knows where this connection is coming from
	 *
	 * @param destinationAddress the cell address that we are talking to 
	 * @param socket the socket that we will use to communicate to the other agent messaging shadow
	 * @exception IOException if we can't make the connection to the other cell or something like that 
	 */
	public OutgoingRunningQueue( CellAddress destinationAddress, Socket socket )
	throws IOException {
	    this.destinationAddress = destinationAddress;
	    this.socket = socket;
	    queue = new Queue();
	    
	    serializedStream = new ByteArrayOutputStream();
	    serializingStream = new ObjectOutputStream( serializedStream );
	    
	    try {
		serializingStream.writeObject( AgentMessagingShadowImpl.this.myCell.getAddress() );

	    } catch( Exception error ) {
		// this will never happen unless somehting is massively fucked
		Debug.critical( error );

	    }

	    outputStream = new BufferedOutputStream( socket.getOutputStream() );
	    byte[] addressSerialized = serializedStream.toByteArray();
	    byte[] byteLength = intToByteArray( addressSerialized.length );
	    outputStream.write( byteLength, 0, 4 );
	    outputStream.write( addressSerialized, 0, addressSerialized.length );
	    outputStream.flush();
	    Debug.notice( "sent address to " + destinationAddress + " in creation of outgoing queue" );

	}

	/**
	 * this is the main loop -- it just sits and waits for
	 * something to appear on the queue, then it dequeues it,
	 * serializes it, and sends it across.  if any problems occur
	 * in the shipping of the event (an IOException), then the
	 * connection is terminated and this queue disappears.  if a
	 * problem happens in serialization, the event to be sent is
	 * simply punted 
	 */
	public void run() {
	    while( !timeToStop ) {
		while( queue.isEmpty() && !timeToStop ) {
		    try {
			synchronized( this ) {
			    wait();
			}
		    } catch( InterruptedException error ) { }
		}
		if( timeToStop ) return;

		AgentMessagingShadowEvent amsEvent = (AgentMessagingShadowEvent)queue.deQueue();
 		try {
		    // this is the fun part -- serialize the array of
		    // agents which are the destination.  then our
		    // protocol is as follows -- two bytes signifies
		    // the size of the array of agents that are coming
		    // down, then send the array, then two bytes for
		    // the length of the hive event stuff, then send
		    // that array.
		    
		    // curious -- what i want to do is to reset the
		    // stream we are outputting to, but unfortunatelly
		    // by simply doing a reset(), then we don't get
		    // teh deserialized object coming back out.
 		    // serializedStream.reset();
		    // since we are doing this -- we should keep track of how many times and maybe just reset it every once in a while so it won't grow out of control
		    serializedStream = new ByteArrayOutputStream();
		    serializingStream = new ObjectOutputStream( serializedStream );
		    serializingStream.writeObject( amsEvent.getDestinationAgents() );
		    byte[] serializedDestinationAgents = serializedStream.toByteArray();
		    Debug.notice( "sending an destination event array of " + serializedDestinationAgents.length + " bytes to " + destinationAddress );

		    // write the array of destination agents
		    byte[] byteLength = intToByteArray( serializedDestinationAgents.length );
		    outputStream.write( byteLength, 0, 4 );
		    outputStream.write( serializedDestinationAgents, 0, serializedDestinationAgents.length );

		    // write the hive event -- need to take into account for null hive events
		    int runningValue;
		    if( amsEvent.getHiveEvent() != null ) {
			runningValue = amsEvent.getHiveEvent().length;
			Debug.notice( "sending a hive event of " + runningValue + " bytes to " + destinationAddress );

		    } else {
			runningValue = 0;
			Debug.notice( "sending a null hive event to " + destinationAddress );

		    }
		    byteLength = intToByteArray( runningValue );
		    outputStream.write( byteLength, 0, 4 );
		    if( amsEvent.getHiveEvent() != null )
			outputStream.write( amsEvent.getHiveEvent(), 0, amsEvent.getHiveEvent().length );

		    // flush the stream finally
		    outputStream.flush();

		} catch( NotSerializableException error ) {
		    Debug.warning( amsEvent.getDestinationAgents() + " is not serializable.  punting event." );

		} catch( IOException error ) {
		    Debug.error( "IOException on the event connection to " + destinationAddress + ".  terminating connection." );
		    IncomingRunningQueue incomingQueue = null;
		    synchronized( outgoingQueues ) {
			incomingQueue = (IncomingRunningQueue)incomingQueues.get( destinationAddress );
			outgoingQueues.remove( destinationAddress );
			incomingQueues.remove( destinationAddress );
			openSockets.remove( destinationAddress );

		    }
		    incomingQueue.stop();
		    return;

		}
	    }

	}

	/**
	 * queue an event to be sent out the output stream -- the
	 * sender will get no error saying whether or not the event
	 * went through
	 *
	 * @param receivingAgents the agents that are going to receiving this event
	 * @param he the serialized hive event to send to all these agents
	 */
	public void sendEvent( EventReceivingAgent[] receivingAgents, byte[] he ) {
	    queue.enQueue( new AgentMessagingShadowEvent( receivingAgents, he ) );
	    synchronized( this ) {
		notifyAll();
	    }

	}

    }
    
    public byte[] readStream( InputStream stream, int length ) 
    throws IOException {
	byte[] outputBytes = new byte[length];
	int readBytes = 0, readCurrent = 0;
	while( ( readBytes < length ) &&
	       ( ( readCurrent = stream.read( outputBytes, readBytes, length - readBytes ) ) > 0 ) )
	    readBytes += readCurrent;
	return outputBytes;

    }
    
    public byte[] intToByteArray( int src ) {
        byte[] newarray = new byte[4];
        for( int i=0;i<4;i++ ) 
            newarray[3-i] = (byte)( src >>> ( 8 * i ) );
        return newarray;
    }

    public int byteArrayToInt( byte[] src ) {
        int newint = 0;
        for( int i=0;i<4;i++ )
            if( src[3-i] < 0 ) 
                newint += ( ((int)src[3-i] ) + 256 ) << ( 8 * i );
            else 
                newint += (int)( src[3-i] << ( 8 * i ) );
        return newint;

    }

}
    
/**
 * use this just as a wrapper so we can temporarily hold the list of
 * agents this event is going to go, and the event itself for the
 * queue 
 */
class AgentMessagingShadowEvent 
implements Serializable {

    // the internal and unchangable variables

    private byte[] he; // the serialized hive event that we are wrapping
    private Agent[] destinationAgents; // the destination of the HiveEvent
    
    /**
     * get the hive event that we are wrapping
     * 
     * @return the hive event
     */
    public byte[] getHiveEvent() { return he; }

    /**
     * get the agent destination of this event
     *
     * @return the agent destinations
     */
    public Agent[] getDestinationAgents() { return destinationAgents; }


    /**
     * create a new agent messaging shadow event
     *
     * @param destinationAgents the agents that this hive event is going to be delivered to
     * @param hiveEvent the actual event (in serialized form) that we wish to deliver
     */
    public AgentMessagingShadowEvent( Agent[] destinationAgents, byte[] hiveEvent ) {
	this.destinationAgents = destinationAgents;
	this.he = hiveEvent;

    }

}

