// $Id: CellEventMonitorImpl.java,v 1.1 2000/05/18 22:01:15 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.shadow.cell;

import net.hivecell.hive.*;
import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.event.cell.*;
import net.hivecell.hive.cell.*;
import net.hivecell.hive.shadow.*;
import net.hivecell.hive.support.CellAddress;
import net.hivecell.hive.support.Debug;

/** Monitor class that tells listeners about server events.
 ** @version $Revision: 1.1 $
 **/
public class CellEventMonitorImpl
extends EventPublisherShadowImpl
implements CellEventMonitor {

    public CellEventMonitorImpl() { 
	super();

    }

    public void init() 
    throws ShadowInitializationException {
	setReady( true );

    }

    /** 
     * Tell subscribers that an agent was born. 
     *
     * @param sender the object that wishes to send this notification out
     * @param agent the agent which has just come alive
     */
    public void sendAgentBorn( Object sender, Agent agent ) {
	
	if( myCell.isAgentLocal( agent ) )
	    notifySubscribers( new AgentBornCellEvent( this, getCount(), agent, myCell.getAddress() ) );

    }

    /** 
     * Tell subscribers that an agent is dying. 
     * 
     * @param sender the object that wishes to send this notification out
     * @param the agent which has gone down
     */
    public void sendAgentDying( Object sender, Agent agent ) {
	
	if( myCell.isAgentLocal( agent ) )
	    notifySubscribers( new AgentDeathCellEvent( this, getCount(), agent, myCell.getAddress() ) );
    }

    /** 
     * Tell subscribers that an agent is dying. 
     * 
     * @param sender the object that wishes to send this notification out
     * @param the agent which has gone down
     */
    public void sendAgentMovingTo( Object sender, Agent agent, CellAddress destinationAddress ) {
	
	if( myCell.isAgentLocal( agent ) )
	    notifySubscribers( new AgentMovingToCellEvent( this, getCount(), agent, myCell.getAddress(), destinationAddress ) );
    }

    /** 
     * Tell subscribers that an agent is dying. 
     * 
     * @param sender the object that wishes to send this notification out
     * @param the agent which has gone down
     */
    public void sendAgentMovingFrom( Object sender, Agent agent, CellAddress sourceAddress ) {

	if( myCell.isAgentLocal( agent ) )
	    notifySubscribers( new AgentMovingFromCellEvent( this, getCount(), agent, sourceAddress, myCell.getAddress() ) );
    }

    public void sendAgentsConnecting( Object sender, Agent from, Agent to ) {
	sendAgentsConnecting( sender, from, null, to, null );
    }

    /** 
     * Tell subscribers that two agents are connecting. 
     *
     * @param sender the object that wishes to send this notification out
     * @param from the agent the connection is coming from
     * @param to the agent the connection is going to
     */
    public void sendAgentsConnecting( Object sender, Agent from, CellAddress fromAddress, Agent to, CellAddress toAddress ) {
	
	if( ( sender instanceof Agent ) &&
	    !myCell.isAgentLocal( (Agent)sender ) )
	    return;

	boolean fromLocal = myCell.isAgentLocal( from );
	boolean toLocal = myCell.isAgentLocal( to );
	if( ( fromLocal || toLocal ) &&
	    !( !fromLocal && !toLocal ) ) {
	    sendCellReport( sender, "Connection from " + Global.shortString( from ) + " to " + Global.shortString( to ) );
	    notifySubscribers( new AgentConnectCellEvent( this, getCount(), from, to ) );
	}      

    }

    /**
     *  Tell subscribers that two agents are disconnecting. 
     * 
     * @param sender the object that wishes to send this notification out
     * @param from the agent the connection is coming from
     * @param to the agent the connection is going to
     */
    public void sendAgentsDisconnecting( Object sender, Agent from, Agent to ) {

	if( ( sender instanceof Agent ) &&
	    !myCell.isAgentLocal( (Agent)sender ) )
	    return;

	boolean fromLocal = myCell.isAgentLocal( from );
	boolean toLocal = myCell.isAgentLocal( to );
	if( ( fromLocal || toLocal ) &&
	    !( !fromLocal && !toLocal ) ) {
	    sendCellReport( sender, "Disconnection from " + Global.shortString(from) + " to " + Global.shortString(to) );
	    notifySubscribers( new AgentDisconnectCellEvent( this, getCount(), from, to) );
	}
	
    }
    
    /** 
     * Tell subscribers that an agent is disconnecting from all. 
     * 
     * @param sender the object that wishes to send this notification out
     * @param from the agent which is disconnecting from all
     */
    public void sendAgentDisconnectingAll( Object sender, Agent from ) {

	sendCellReport( sender, "Disconnection from " + Global.shortString( from ) );

	// NOTE -- need to confirm that we only send out an agent
	// disconnect from all server event when the agent is local to
	// our cell
	notifySubscribers( new AgentDisconnectFromAllCellEvent( this, getCount(), from, myCell.getAddress() ) );

    }

    /** 
     * Tell subscribes that an agent is messaging another. 
     *
     * @param sender the object that wishes to send this information out
     * @param from the agent which the message is coming from
     * @param to the agent which the message is going to
     */
    public void sendAgentMessage( Object sender, String message, Agent from, Agent to ) {
	
	// if the destination of this message is not on our system,
	// then forget it being displayed 
	if( !myCell.isAgentLocal( to ) ) 
	    return;

	notifySubscribers( new AgentMessageCellEvent( this, getCount(), from, to, message ) );
	
    }

    /** 
     * Tell subscribers that we're about to go down. 
     *
     * @param sender the object that wishes to send this information out
     */
    public void sendCellGoingDown( Object sender ) {
	
	notifySubscribers( new CellGoingDownCellEvent( this, getCount(), myCell.getAddress() ) );
	
    }

    /** 
     * Tell subscribers that we're about to go down. 
     *
     * @param sender the object that wishes to send this information out
     */
    public void sendCellWithdrawing( Object sender ) {
	
	notifySubscribers( new CellWithdrawnCellEvent( this, getCount(), myCell.getAddress() ) );
	
    }

    /** 
     * Tell subscribers of a report. 
     *
     * @param sender the object that wishes to send this information out
     * @param mess the message to display
     */
    public void sendCellReport( Object sender, String message ) {
	
	notifySubscribers( new CellReportCellEvent( this, getCount(), myCell.getAddress(), message ) );
	Debug.output( "***Report: " + message );
	
    }

}

