// $Id: HTTPServerShadowImpl.java,v 1.1 2000/05/18 22:01:15 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.shadow.cell;

import java.io.*;
import java.net.InetAddress;
import java.net.ServerSocket;
import java.net.Socket;
import java.util.Properties;

import net.hivecell.hive.Global;
import net.hivecell.hive.cell.Cell;
import net.hivecell.hive.shadow.EventPublisherShadowImpl;
import net.hivecell.hive.shadow.ShadowInitializationException;
import net.hivecell.hive.support.Debug;

public class HTTPServerShadowImpl
extends EventPublisherShadowImpl
implements Runnable, HTTPServerShadow {

    private String classpath; // the classpath where we will find all of our classes
    
    protected ServerSocket ss; // the socket which we will deal with the incoming HTTP connections

    private String codebase;

    
    public HTTPServerShadowImpl() 
    throws IOException {
	this( System.getProperty( Global.AgentClassdir, "." ), 0 );

    }

    public HTTPServerShadowImpl( String classPath, int port ) 
    throws IOException {
	super();
	this.classpath = classPath; // setup the classpath
	this.ss = new ServerSocket( port ); // setup our listening port
	this.codebase = setupCodebase(); // setup ourselves for the RMI codebase
	
	Debug.notice( "Starting the HTTP server on port " + getPort());
	(new Thread( this )).start();

    }

    public void init() 
    throws ShadowInitializationException {
	setReady( true );

    }

    /**
     * Find out the port the http server is on
     *
     * @return the port number
     */
    public int getPort() {

	return this.ss.getLocalPort();

    }

    /**
     * Find out what is our codebase
     */
    public String getCodebase() {
	return this.codebase;
    }

    /** 
     * Set the codebase property to point to the correct port.  This has
     * to be called before any RMI magic happens, so that the
     * java.rmi.server.codebase property gets set before RMI reads it.
     * This code is not technically portable, doing <tt>java
     * -Djava.rmi.server.codebase=foobar Cell</tt> is more correct but
     * is a real pain.
     *
     * @param port the port for our codebase 
     */
    protected String setupCodebase() {
	String host = "??unknown??";
	try {
	    // There is a wierd caching bug such that host lookups do not 
	    // return fully qualified addresses unless we lookup by host address.
	    host = InetAddress.getByName( InetAddress.getLocalHost().getHostAddress() ).getHostName();
	}catch( Exception e ) {
	    Debug.error( "Couldn't find hostname!\n", e );
	}
	String ourCodebase = "http://" + host + ":" + getPort() + "/";

	// Set up the codebase to point to where our server will be.
	Properties p = System.getProperties();
	// In Java 1.2, should be setProperty()
	p.put( "java.rmi.server.codebase", ourCodebase );
    
	// Print out the codebase, just to verify it.
	p = System.getProperties();
	String s = (String)p.get( "java.rmi.server.codebase" );
	Debug.notice("Codebase = " + s);
	return ourCodebase;
    }


    // the runnable method which will run to listen for incoming
    // connections and fork off a thread to deal with them
    public void run() { 

	while( true ) {

	    Socket incoming; // the handle to the socket we will need for incomings
	    try {
		Debug.notice( "Blocking for HTTP connection.");
		incoming = this.ss.accept(); // block for an incoming connection
		Debug.notice( "Received an HTTP connection" );
	    } catch( IOException error ) {
		Debug.error( "HTTPServerShadow died while getting a connection", error );
		return; // something more intelligent has to be done here
	    }

	    class requestProcessingRunnable 
	    implements Runnable {
		Socket incoming;
		
		public requestProcessingRunnable( Socket s ) {
		    this.incoming = s;
		}

		public void run() {
		    DataOutputStream out = null;
		    try {
			out = new DataOutputStream( incoming.getOutputStream() );
		    } catch( IOException error ) {
			Debug.error( "Problem while getting the handle to the output stream", error );
			return;
		    }
		    
		    DataInputStream in = null;
		    try {
			in = new DataInputStream( incoming.getInputStream() );
		    } catch( IOException error ) {
			Debug.error( "Problem while getting the handle to the input stream", error );
			return;
		    }
		    
		    try {
			Thread.sleep( 1000 );
		    } catch( InterruptedException error ) { }

		    // take the input request and put it into
		    // a "static" format for us to deal with
		    // -- this might actually be a problem in
		    // that the whole request may not have
		    // arrived yet
		    String incomingRequest = null;
		    try {
			BufferedReader br = new BufferedReader( new InputStreamReader( in ) );
			incomingRequest = br.readLine();
		    } catch( IOException error ) {
			Debug.error( "Problem while trying to decode incoming HTTP request", error );
			return; // here it is ok to return -- we are just punting this connection
		    }
		    // need to determine what the requester is looking for
		    
		    // first check to see if we are requesting a .class file
                    Debug.notice("request: " + incomingRequest);
		    String classFilePath = null;
		    if( ( classFilePath = parseForClass( incomingRequest ) ) != null ) {
			try {
			    sendClassBytes( out, classFilePath );
			} catch( Exception error ) {
			    sendErrorMessage( out, error.getMessage() );
			}
			try {
			    out.close();
			    in.close();
			} catch( IOException error ) { }
			return;
		    }
		    sendErrorMessage( out, "Cannot process request for " + incomingRequest);
		    try {
			out.close();
			in.close();
		    } catch( IOException error ) { }
		    return;

		}

	    }
	    Thread incomingThread = new Thread( (Runnable)( new requestProcessingRunnable( incoming ) ) );
	    incomingThread.start(); // start the thread to deal with this connection

	    // now loop again so we can listen for another connection
	}

    }

    /**
     * Try to parse an incoming request for a class file.  If a class
     * file is being requested, try to get the path for the class file .
     * If the request is not for a class file, or the requested class
     * file is not valid, then we just return null 
     *
     * @param request the incoming HTTP request
     *
     * @return null if we can't find anything, or the file path if the request is for a valid class file
     */
    private String parseForClass( String request ) {

	String line = null;
	String path = "";

	// extract class from GET line
	if( request.startsWith( "GET /" ) ) {
	     line = request.substring( 5, request.length() - 1 ).trim();

	    int index = line.indexOf( ".class " );
	    if( index != -1 )
		path = line.substring( 0, index ).replace( '/', '.' );
	}

	if( path.length() != 0 )
	    return path;

	else
	    return null;

    }

    /**
     * Returns an array of bytes containing the bytecodes for
     * the class represented by the argument <b>path</b>.
     * The <b>path</b> is a dot separated class name with
     * the ".class" extension removed.
     *
     * @return the bytecodes for the class
     * @exception ClassNotFoundException if the class corresponding
     * to <b>path</b> could not be loaded.
     */
    protected byte[] getBytes( String path )
    throws IOException, ClassNotFoundException {
        String filename = classpath + File.separator + path.replace( '.', File.separatorChar ) + ".class";
	File f = new File(filename);
        if (!f.exists()) {
            Debug.error("Was asked to serve class " + path + " but could not find it at file " + filename);
            throw new ClassNotFoundException();
        }
            
	Debug.notice( "Serving class " + path );
	int length = (int)f.length();
	if( length == 0 ) {
	    throw new IOException( "File length is zero: " + path );
	} else {
	    FileInputStream fin = new FileInputStream( f );
	    DataInputStream in = new DataInputStream( fin );
	    
	    byte[] bytecodes = new byte[length];
	    in.readFully( bytecodes );
            Debug.notice("served up!");
	    return bytecodes;
	}
    }

    /**
     * Reply with the class's byte stream.  It will thrown an
     * IOException if it cannot
     *
     * @param out the data output stream to use to send the request
     * @param classFilePath the file path to the class file we need to send 
     *
     * @exception IOException if there is any problem in the transmission
     */
    protected void sendClassBytes( DataOutputStream out, String classFilePath ) 
    throws IOException, ClassNotFoundException {

	// retrieve bytecodes
	byte[] bytecodes = getBytes( classFilePath );
	// send bytecodes in response (assumes HTTP/1.0 or later)
	out.writeBytes( "HTTP/1.0 200 OK\r\n ");
	out.writeBytes( "Content-Length: " + bytecodes.length + "\r\n");
	out.writeBytes( "Content-Type: application/java\r\n\r\n" );
	out.write( bytecodes );
	out.flush();

    }

    /**
     * Send an HTTP 400 error response.  If an exception occurs, too
     * bad, this will just close the connection
     *
     * @param out the output stream to send the error over
     * @param e the error message (if any) to send
     */
    protected void sendErrorMessage( DataOutputStream out, String e ) {
        Debug.warning("Failed to send requested resource. " + e);
	try {
	    out.writeBytes("HTTP/1.0 400 " + e + "\r\n");
	    out.writeBytes("Content-Type: text/html\r\n\r\n");
	    out.flush();
	} catch( IOException error ) {
	    // too bad
	}
	
    }

}
