package net.hivecell.hive.shadow;

import java.io.InputStreamReader;
import java.io.BufferedReader;
import java.lang.Process;
import java.io.IOException;
import java.io.FileWriter;
import java.io.PrintWriter;
import java.io.File;
import java.lang.reflect.Method;
import java.util.Enumeration;
import java.util.Vector;

import net.hivecell.hive.support.Debug;

import gnu.getopt.*;


/**
 * this is the shadow compiler. 
 *
 * the shadow compiler takes a shadow's class file, then generates a
 * stub that can be used for that shadow.  it takes a class file,
 * first confirms that it is a Shadow, then it gets all the interfaces
 * and makes sure that somewhere this shadow implements Shadow.  then
 * it generates a list of the methods from the shadow's interfaces and
 * outputs some java code that when compiled will be the proxy to this
 * shadow 
 */
public class shadowc {

    /**
     * a flag that states whether we keep the generated java file or not
     */
    protected static boolean keepIntermediateCode = false;

    /**
     * a flag which states whether we should actually compile anything
     */
    protected static boolean compileIntermediateCode = true;

    /**
     * the directory that we are going to send the compiled code to --
     * by default it is this directory 
     */
    protected static File outputDirectory = new File( "." );


    /**
     * the usage statements
     */
    private static void usage() {
	System.err.println( "java net.hivecell.hive.shadow.shadowc <classname>" );
	System.err.println( " --keep       keep the intermediate generated files" );
	System.err.println( " --nocompile  do not compile" );
	System.err.println( " --directory  specifiy the output directory for the compiled files" );
	System.err.println( " --verbose    show all the debugging messages" );
	System.err.println( " --help       display this message" );

    }

    /** 
     * the main method -- this takes from the command line and then
     * starts the process of generating the shadow's stub 
     */
    public static void main( String[] args ) {
	Debug.setDefaultLevel( Debug.WARNING ); // default level.
	Debug.setFallthrough( true );

	if( args.length == 0 ) {
	    usage();
	    System.exit( 0 );

	}

	// first extract what class file we are compiling
	String toStub = args[args.length - 1];
	String[] newargs = new String[args.length - 1];
	System.arraycopy( args, 0, newargs, 0, args.length - 1 );

	// parse the command line options
	int c;
	String arg;
	LongOpt[] longopts = new LongOpt[5];
	longopts[0] = new LongOpt( "keep", LongOpt.NO_ARGUMENT, null, 'k' );
	longopts[1] = new LongOpt( "directory", LongOpt.REQUIRED_ARGUMENT, null, 'd' );
	longopts[2] = new LongOpt( "help", LongOpt.NO_ARGUMENT, null, 'h' );
	longopts[3] = new LongOpt( "verbose", LongOpt.NO_ARGUMENT, null, 'v' );
	longopts[4] = new LongOpt( "nocompile", LongOpt.NO_ARGUMENT, null, 'n' );
	Getopt g = new Getopt( "shadowc", newargs, "-h", longopts );
	g.setOpterr( false );
	while( ( c = g.getopt() ) != -1 ) {
	    switch( c ) {
	    case 'h':
		usage();
		System.exit( 0 );
		break;

	    case 'k':
		keepIntermediateCode = true;
		break;

	    case 'd':
		outputDirectory = new File( g.getOptarg() );
		break;

	    case 'v':
		Debug.setDefaultLevel( Debug.NOTICE );
		break;

	    case 'n':
		compileIntermediateCode = false;
		break;

	    default:
		Debug.error( "unrecognized command line flag: " + g.getOptopt() + "(" + (char)c + ")" );
		break;

	    }

	}

	Debug.notice( "attempting to create a stub for " + toStub );

	// ok - let's actually start the conversion from the class
	// file to the java source file for the stub.  first get a
	// handle to the class file that we are going to do the
	// conversion on
	Class toStubify = null;
	try {
	    Debug.notice( "attempting to load " + toStub );
	    toStubify = Class.forName( toStub );
	    if( !( ShadowImpl.class ).isAssignableFrom( toStubify ) ) {
		Debug.error( toStub + " is not a derivative of net.hivecell.hive.shadow.ShadowImpl" );
		System.exit( 1 );

	    }

	} catch( ClassNotFoundException error ) {
	    Debug.error( "cannot find class " + toStub );

	}

	// let's figure out what we are going to call the stub's .java
	// file
	String javafileName = toStubify.getName();
	javafileName = javafileName.substring( javafileName.lastIndexOf( "." ) + 1 ) + "_Stub.java";
	String className = javafileName.substring( 0, javafileName.lastIndexOf( "." ) );
	String shadowName = className.substring( 0, className.lastIndexOf( "_" ) );
	Debug.notice( "outputting the source code as " + javafileName );

	// let's figure out what package this class is in
	String packageName = toStubify.getName();
	packageName = packageName.substring( 0, packageName.lastIndexOf( "." ) );
	Debug.notice( "package name is " + packageName );

	// figure out exactly what we need to extend -- if the shadow
	// class extends ShadowImpl, then we extend ShadowStub -- if
	// it extends anything else, then we extend that_Stub
	Vector toImport = new Vector();
	String superclassName = "";
	if( toStubify.equals( ShadowImpl.class ) || toStubify.getSuperclass().equals( ShadowImpl.class ) )
	    superclassName = "ShadowStub";
	
	else {
	    superclassName = toStubify.getSuperclass().getName();
	    superclassName += "_Stub";
	    try {
		toImport.addElement( Class.forName( superclassName ) );

	    } catch( ClassNotFoundException error ) {
		Debug.error( superclassName + " is not found to extend " + className );
		System.exit( 1 );

	    }
	    superclassName = superclassName.substring( superclassName.lastIndexOf( "." ) + 1 );

	}
	Debug.notice( "extending " + superclassName );

	// figure out which interfaces we need to extends in this class
	Debug.notice( "computing all the interfaces the stub needs to implement" );
	Vector interfaces = ShadowImpl.getShadowInterfaces( toStubify );
	
	// figure out all the methods we need
	Debug.notice( "computing all the methods that we need in the stub" );
	Vector methods = ShadowImpl.getAllShadowMethods( toStubify );

	// compute what classes we need to import
	Debug.notice( "extracting the appropriate information from the stub-destined methods" );
	Enumeration e = methods.elements();
	while( e.hasMoreElements() ) {
	    Method method = (Method)e.nextElement();
	    Class returnType = method.getReturnType();
	    Class[] parameterTypes = method.getParameterTypes();
	    Class[] exceptionTypes = method.getExceptionTypes();
	    if( returnType.isArray() )
		returnType = returnType.getComponentType();
	    if( !returnType.isPrimitive() && !toImport.contains( returnType ) ) {
		Debug.notice( "need to import " + returnType.getName() );
		toImport.addElement( returnType );

	    }
	    for( int count=0;count<parameterTypes.length;count++ ) {
		Class testType = parameterTypes[count];
		if( testType.isArray() )
		    testType = testType.getComponentType();
		if( !testType.isPrimitive() && !toImport.contains( testType ) ) {
		    Debug.notice( "need to import " + testType.getName() );
		    toImport.addElement( testType );

		}

	    }
	    for( int count=0;count<exceptionTypes.length;count++ ) {
		if( !exceptionTypes[count].isPrimitive() && !toImport.contains( exceptionTypes[count] ) ) {
		    Debug.notice( "need to import " + exceptionTypes[count].getName() );
		    toImport.addElement( exceptionTypes[count] );

		}

	    }

	}

	Debug.notice( "making sure we import net.hivecell.hive.shadow.Shadow" );
	if( !toImport.contains( Shadow.class ) )
	    toImport.addElement( Shadow.class );
	Debug.notice( "making sure we import net.hivecell.hive.shadow.ShadowStub" );
	if( !toImport.contains( ShadowStub.class ) )
	    toImport.addElement( ShadowStub.class );
	Debug.notice( "making sure we import net.hivecell.hive.shadow.ShadowImpl" );
	if( !toImport.contains( ShadowImpl.class ) )
	    toImport.addElement( ShadowImpl.class );

	Debug.notice( "opening " + javafileName + " for writing" );
	try {
	    PrintWriter stubCodeWriter = new PrintWriter( new FileWriter( javafileName ) );

	    Debug.notice( "writing the package statement" );
	    stubCodeWriter.println( "package " + packageName + ";\n" );

	    Debug.notice( "writing the import lines" );
	    e = toImport.elements();
	    while( e.hasMoreElements() )
		stubCodeWriter.println( "import " + ((Class)e.nextElement()).getName() + ";" );
	    stubCodeWriter.println();
	    
	    Debug.notice( "writing the class definition" );
	    stubCodeWriter.println( "public class " + className );
	    stubCodeWriter.print( "extends " + superclassName );
	    e = interfaces.elements();
	    if( e.hasMoreElements() )
		stubCodeWriter.print( "\nimplements " );
	    while( e.hasMoreElements() ) {
		Class interfaceClass = (Class)e.nextElement();
		Debug.notice( "implementing " + interfaceClass.getName() );
		stubCodeWriter.print( interfaceClass.getName() );
		if( e.hasMoreElements() )
		    stubCodeWriter.print( ", " );

	    }
	    stubCodeWriter.println( " {" );

	    Debug.notice( "writing the constructor" );
	    stubCodeWriter.println( "\tpublic " + className + "( ShadowImpl implementation ) {" );
	    stubCodeWriter.println( "\t\tsuper( implementation );\n" );
	    stubCodeWriter.println( "\t}" );
	    stubCodeWriter.println();

	    Debug.notice( "writing the methods" );
	    e = methods.elements();
	    int count = 0;
	    while( e.hasMoreElements() ) {
		Method method = (Method)e.nextElement();
		Debug.notice( "creating the method signature for " + method );
		stubCodeWriter.print( "\t" );
		stubCodeWriter.print( "public " );
		if( method.getReturnType().isArray() )
		    stubCodeWriter.print( method.getReturnType().getComponentType().getName() + "[]" );
		else
		    stubCodeWriter.print( method.getReturnType().getName() );
		stubCodeWriter.print( " " + method.getName() + "( " );
		Class[] parameterTypes = method.getParameterTypes();
		int variableCount = 0;
		for( count=0;count<parameterTypes.length;count++ ) {
		    if( parameterTypes[count].isArray() )
			stubCodeWriter.print( parameterTypes[count].getComponentType().getName() + "[]" );
		    else
			stubCodeWriter.print( parameterTypes[count].getName() );
		    stubCodeWriter.print( " " + makeUniqueVariableName( variableCount++ ) );
		    if( count < parameterTypes.length - 1 )
			stubCodeWriter.print( ", " );
		    
		}
		stubCodeWriter.print( " )" );
		Class[] exceptionTypes = method.getExceptionTypes();
		if( exceptionTypes.length != 0 )
		    stubCodeWriter.print( "\n\tthrows " );
		for( count=0;count<exceptionTypes.length;count++ ) {
		    stubCodeWriter.print( exceptionTypes[count].getName() );
		    if( count < exceptionTypes.length - 1 )
			stubCodeWriter.print( ", " );

		}
		stubCodeWriter.print( " {\n\t\t" );
		if( !method.getReturnType().equals( void.class ) )
		    stubCodeWriter.print( "return " );
		stubCodeWriter.print( "((" + shadowName + ")this.implementation)." );

		Debug.notice( "creating the passthrough call to the implementation" );
		stubCodeWriter.print( method.getName() + "( " );
		variableCount = 0;
		for( count=0;count<parameterTypes.length;count++ ) {
		    stubCodeWriter.print( makeUniqueVariableName( variableCount++ ) );
		    if( count < parameterTypes.length - 1 )
			stubCodeWriter.print( ", " );
		    
		}
		stubCodeWriter.print( " );\n\n\t}" );
		stubCodeWriter.println();

		stubCodeWriter.println();
		
	    }

	    stubCodeWriter.println( "}" );
	    stubCodeWriter.flush();
	    stubCodeWriter.close();

	} catch( IOException error ) {
	    Debug.error( "problem while writing to " + javafileName, error );
	    System.exit( 1 );

	}

	if( compileIntermediateCode )
	    try {
		Debug.notice( "compiling the stub and outputting to " + outputDirectory.getAbsolutePath() );
		Process compiling = Runtime.getRuntime().exec( "javac -d " + outputDirectory.getAbsolutePath() + " " + javafileName );
		BufferedReader reader = new BufferedReader( new InputStreamReader( compiling.getErrorStream() ) );
		String incoming = null;
		while( ( incoming = reader.readLine() ) != null )
		    Debug.error( incoming );
		reader.close();

	    } catch( IOException error ) {
		Debug.error( "problem running the java compiler", error );
		System.exit( 1 );

	    }

	if( !keepIntermediateCode ) {
	    Debug.notice( "removing the intermediate code files" );
	    (new File( javafileName )).delete();

	}

    }    

    /**
     * this makes a unique variable name -- you give it an integer,
     * and it produces a variable that will correspond with that
     * interger
     *
     * @param count the integer that we want the variable of
     * @return the string of a variable that corresponds to count 
     */
    private static String makeUniqueVariableName( int count ) {
	int workingNumber = Math.abs( count );
	String variable = "variable_";
	while( workingNumber > 0 ) {
	    variable += (char)( (int)'a' + workingNumber % 26 ) + "";
	    workingNumber /= 26;
	    
	}
	return variable;

    }

}
