// $Id: AgentGraph.java,v 1.1 2000/05/18 22:01:17 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.support;

import net.hivecell.hive.agent.Agent;
import net.hivecell.hive.support.graph.*;

import java.io.Serializable;
import java.util.Enumeration;


public class AgentGraph 
implements Serializable {

    protected DenseGraph dg; // the actual graph we use for maintaining the agent information

    public AgentGraph() { dg = new DenseGraph(); }

    /**
     * Add an agent to the graph
     *
     * @param a agent that we want to add to the graph 
     */
    public void addAgent( Agent a ) { dg.addNode( a ); }

    /**
     * Add an agent graph to our agent graph -- this is sort of like a merge
     *
     * @param ag the agent graph to add to our own
     */
    public void addAgentGraph( AgentGraph ag ) {

	Enumeration e = ag.listAllAgents();
	while( e.hasMoreElements() )
	    dg.addNode( e.nextElement() );
	e = ag.listAllConnections();
	try {
	    while( e.hasMoreElements() ) {
		AgentGraphEdge age = (AgentGraphEdge)e.nextElement();
		dg.connectNodes( age.getSourceAgent(), age.getDestinationAgent(), 1 );
	    } 
	} catch( NotInGraphException error ) {
	    Debug.error( "Huh?  Raffi, what went wrong?" );
	}

    }

    /**
     * Test to see if we have this agent in our graph 
     *
     * @param a agent to test
     * @return true if we have it, false otherwise
     */
    public boolean containsAgent( Agent a ) { return dg.containsNode( a ); }

    /** 
     * Remove an agent from the graph 
     * 
     * @param a agent to remove
     * @return true if the object was removed (it was there to be removed in the first place), false otherwise
     */
    public boolean removeAgent( Agent a ) { return dg.removeNode( a ); }

    /**
     * Test to see if two agents are connected
     *
     * @param source the start of the connection
     * @param destination the end of the connection to test
     * @return true if the agents are connected, false otherwise
     */
    public boolean isConnected( Agent source, Agent destination ) {

	try {
	    if( dg.isConnected( source, destination ) )
		return true;
	    return false;
	} catch( NotInGraphException error ) {
	    return false;
	}

    }

    /** 
     * Connect two agents.  If they are not in the graph, then they
     * will be added to the graph
     *
     * @param source the agent to start the connection from
     * @param destination the agent to end the connection to 
     */
    public void connectAgents( Agent source, Agent destination ) { 

	if( !dg.containsNode( source ) ) {
	    Debug.warning( "An agent (" + source + ") was not in our graph, but it has initiated a connection that has come to our attention -- adding it" ); 
	    dg.addNode( source );
	}

	if( !dg.containsNode( destination ) ) {
	    Debug.warning( "An agent (" + destination + ") was not in our graph, but it has received a connection that has come to our attention -- adding it" ); 
	    dg.addNode( destination );
	}

	try {
	    dg.connectNodes( source, destination, 1 ); 
	} catch( NotInGraphException error ) {
	    Debug.error( "Huh?  This should never happen!", error );
	}

    }

    /** 
     * Disconnect two agents.  If they are not in the graph, this will return false
     *
     * @param source the source agent to start the disconnection from
     * @param destination the destination agent to end the disconnection from
     * @return true if the disconnection worked, false otherwise 
     */
    public boolean disconnectAgents( Agent source, Agent destination ) {

	try {
	    dg.disconnectNodes( source, destination );
	    return true;
	} catch( NotInGraphException error ) {
	    return false;
	}

    }
    
    /**
     * Disconnect an agent from all its outgoing connections
     *
     * @param agent the agent to disconnect from all the outgoing connections
     * @return true if anything was removed, false otherwise
     */
    public boolean disconnectFromAll( Agent agent ) {
	
	Agent[] outgoing = null;
	try {
	    outgoing = listOutgoingConnections( agent );
	} catch( NotInGraphException error ) {
	    return false;
	}
	boolean toRemove = false;
	for( int count=0;count != outgoing.length;count++ ) 
	    toRemove = toRemove || disconnectAgents( agent, outgoing[count] );

	return toRemove;
	
    }
    
    /** 
     * List all the agents on connection edges that point away from this agent.
     *
     * @return an array of agents that are being pointed to by this agent 
     * @exception NotInGraphException if the agent is not in the graph
     */
    public Agent[] listOutgoingConnections( Agent a ) 
    throws NotInGraphException {
	Object[] outgoing = dg.listOutgoingNodes( a );
	Agent[] toReturn = new Agent[outgoing.length];
	for( int count=0;count < outgoing.length;count++ )
	    toReturn[count] = (Agent)outgoing[count];
	return toReturn;
    }

    /**
     * List all the agents on connection edges that point to this agent
     *
     * @return an array of agents that point to this agent
     * @exception NotInGraphException if the agent is not in this graph
     */
    public Agent[] listIncomingConnections( Agent a )
    throws NotInGraphException {
	Object[] outgoing = dg.listIncomingNodes( a );
	Agent[] toReturn = new Agent[outgoing.length];
	for( int count=0;count < outgoing.length;count++ )
	    toReturn[count] = (Agent)outgoing[count];
	return toReturn;
    }

    /**
     * Enumerate through all the agents in this graph 
     *
     * @return the enumeration of all the agents
     */
    public Enumeration listAllAgents() { return dg.nodes(); }

    public int size() { return dg.numberNodes(); }

    /**
     * Enumerate through all the edges in the agent graph
     *
     * @return an enumeration of AgentGraphEdge classes
     */
    public Enumeration listAllConnections() {
	
	final Enumeration e = dg.edges();

	return new Enumeration() {
	    public boolean hasMoreElements() { return e.hasMoreElements(); }
	    public Object nextElement() { 
		Edge edge = (Edge)e.nextElement();
		return new AgentGraphEdge( (Agent)edge.getSource(), (Agent)edge.getDestination() );
	    }
	};
	
    }

    /**
     * List all the agents which are in the new graph that are not in
     * this graph
     *
     * @param graph the graph to check against
     * @return an array of Agents which are in the other graph, but
     * not in this one 
     */
    public Agent[] getNewAgents( AgentGraph graph ) {

	Object[] objects = GraphDifferenceHelper.getAddedObjects( dg, graph.dg );
	Agent[] agents = new Agent[ objects.length ];
	for( int count=0;count < objects.length;count++ )
	    agents[count] = (Agent)objects[count];
	return agents;

    }

    /**
     * List all the agents which have disappeared between this graph
     * and the passed in graph
     *
     * @param graph the graph to check against
     * @return an array of Agent swhich are in our graph, but not the other graph 
     */
    public Agent[] getRemovedAgents( AgentGraph graph ) {

	Object[] objects = GraphDifferenceHelper.getRemovedObjects( dg, graph.dg );
	Agent[] agents = new Agent[ objects.length ];
	for( int count=0;count < objects.length;count++ )
	    agents[count] = (Agent)objects[count];
	return agents;

    }

    /**
     * List all the connections which are in the new graph, but are
     * not in this graph
     *
     * @param graph the graph to check against
     * @return an array of Edges which are in the other graph, but not this one 
     */
    public Edge[] getNewEdges( AgentGraph graph ) {

	return GraphDifferenceHelper.getAddedConnections( dg, graph.dg );

    }

    /**
     * List all the connections which are not in the new graph, but are
     * not in this graph
     *
     * @param graph the graph to check against
     * @return an array of Edges which are in the other graph, but not this one 
     */
    public Edge[] getRemovedEdges( AgentGraph graph ) {

	return GraphDifferenceHelper.getRemovedConnections( dg, graph.dg );

    }
    
}


