// $Id: BaseCanvas.java,v 1.1 2000/05/18 22:01:17 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

// Basic superclass for canvases for agents.

package net.hivecell.hive.support;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import net.hivecell.hive.support.Debug;

public abstract class BaseCanvas 
extends Canvas {

  protected Image imageBuffer;
  protected Dimension imageBufferSize;
  protected boolean doubleBuffering = false;
  
  public BaseCanvas() {
    // handle resize events
    addComponentListener(new ComponentAdapter() {
      public void componentResized(ComponentEvent event) {
        adjustDisplayToNewSize();
      }
    });
  }

  protected void adjustDisplayToNewSize() {
    render();
  }

  protected void setDoubleBuffering(boolean flag) {
    doubleBuffering = flag;
  }

  // Call this to actually redraw the whole display - this will
  // call render(Graphics imageBuffer.getGraphics()), which the subclass
  // of this library should override to do the real work of painting.
  protected void render() {
    Graphics g = getGraphics();
    if( g != null ) {
	paint( g );
	g.dispose();
    }
  }

  // Paint just calls render(Graphics), then update
  public void paint(Graphics g) {
    if (doubleBuffering) {
      if (imageBuffer == null) {
        buildNewImageBuffer(this.getSize());
        if (imageBuffer == null) {                // STILL null, AWT weirdness
          Debug.println(Debug.NOTICE, "No image buffer, not drawing. Hopefully the window is new.");
          return;
        }
      }
      render(imageBuffer.getGraphics());
      update(g);
    } else
      render(g);
  }

  // Erase the backing buffer of the image
  protected void clearImageBuffer() {
    Graphics ig = imageBuffer.getGraphics();
    Color oldColor = ig.getColor();
    ig.setColor(getBackground());
    ig.fillRect(0, 0, imageBufferSize.width, imageBufferSize.height);
    ig.setColor(oldColor);
  }

  private void buildNewImageBuffer(Dimension size) {
    if (size.width > 0 && size.height > 0) {      // weird AWT race condition
      Debug.println(Debug.NOTICE, "Rebuilding image buffer to size " + size);
      imageBuffer = createImage(size.width, size.height);
      imageBufferSize = size;
      clearImageBuffer();
    } else {
      imageBuffer = null;
      imageBufferSize = size;
    }
  }
  
  // Convenience method - call this to flush the image buffer
  // to the actual screen.
  protected void update() {
    if (doubleBuffering) {
      Graphics g = this.getGraphics();
      update(g);
      g.dispose();
    }
  }

  // update the actual display - render the buffer on screen.
  //   this also manages creating the image buffer
  // If we're not double buffering, then do nothing.
  public void update(Graphics g) {
    if (doubleBuffering) {
      Dimension size = this.getSize();

      // check that the buffer is valid - if not, build one
      if (imageBuffer == null ||
          size.width != imageBufferSize.width ||
          size.height != imageBufferSize.height) {
        buildNewImageBuffer(size);
      }
      
      g.drawImage(imageBuffer, 0, 0, this);         // draw the actual buffer
    }
  }
  
  // Override this to do the real drawing in your application.
  // If you need to do more drawing, you can call update() to force
  // a buffer swap.
  public abstract void render(Graphics g);
}
