// $Id: Debug.java,v 1.3 2000/05/20 20:53:36 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.support;

import java.io.PrintStream;
import java.io.PrintWriter;
import java.io.StringWriter;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.StringTokenizer;
import java.util.Vector;
import java.util.NoSuchElementException;


/**
 * The <code>Debug</code> class implements a simple debugging facility.
 * Debugging messages are sent using one of the <code>println</code>
 * methods, which are then sent out the debugging stream (by default,
 * <code>System.err</code>). Messages are tagged with a debugging level;
 * if the message is less serious than the current debugging level,
 * then the message is not printed.<br>
 * Hive code should use this facility for all debugging messages.<br>
 *
 * <p>The default debug level is <code>NOTICE</code>.<br>
 * Note: the class is entirely static, you should not instantiate it.
 */
public class Debug {
    /** 
     * A debug level indicating complete failure, the program is about
     * to crash.  
     */
    public static final int FAILURE = 0;
 
    /** 
     * A debug level indicating a critical problem, the program will 
     * very likely stop working.  
     */
    public static final int CRITICAL = 1;

    /** 
     * A debug level indicating an important error, the program will
     * probably recover but some things are now wrong.
     */
    public static final int ERROR = 2;

    /** 
     * A debug level warning of an undesireable occurrence, but the
     * program has recovered and is continuing OK. 
     */
    public static final int WARNING = 3;
    
    /** 
     * A debug level used for notification messages, does not
     * represent any serious problem. 
     */
    public static final int NOTICE = 4;


    /**
     * there are two ways to specify the interactions with the Debug
     * class -- either we have a list where we have a list of things
     * that we should be listening to, or we have a list of things
     * that we shouldn't be listening to.  this flag sets it.  if it
     * is set to true, than we fall through if we are not caught by
     * the interactions table.  if it is false, then we stop 
     */
    private static boolean fallthrough = true;

    /**
     * the mapping between teh classname and teh hashtable containing the
     * streams taht we are goning to send this output to.  if we have
     * fallthrough on, and the file is not in this hashtable, then we
     * just send it to the default stream.  however, if the
     * fallthrough is on, and the object is in here, then it will
     * -only- be sent to the streams noted.  
     */
    private static Hashtable streams = new Hashtable();
    
    private static int level = NOTICE; // the current level we are working at

    /**
     * the output stream that we are going to be using
     */
    private static PrintStream out;
    static {
	out = System.err;

    }

    /** 
     * You may not construct instances; it is static only. 
     */
    private Debug() { }

    /**
     * this is just a method to use so we can synchronize any output
     * so we don't clobber any debugging output
     *
     * @param string the string that we want to print 
     */
    public synchronized static void output( String string ) {
	out.println( string );

    }
    

    /**
     * print the current stack trace to the debugging output.  mainly
     * a convienience method so we don't have to throw the exception
     * and cathc it, etc, etc, etc
     *
     * @param message the message that will go along with the stack trace 
     */
    public synchronized static void printStackTrace( ) {
	printStackTrace("");
    }
    public synchronized static void printStackTrace( String message ) {
	// This is mostly cut and pasted from fileAndLine();

	// Lines that match this are ignored
	final String ignoreLine = "at net.hivecell.hive.support.Debug";
	// If we find this as the file and line, then give up.
	final String compiledCode = "Compiled Code";
	
	// Generate a stack trace and capture it into a string
	StringWriter stringWriter = new StringWriter();
	PrintWriter writer = new PrintWriter( stringWriter );
	(new Throwable()).printStackTrace( writer );
	String stackOutput = stringWriter.toString();
	
	// Split the stack trace up by lines, look for the non-Debug class
	StringTokenizer st = new StringTokenizer( stackOutput, "\n" );
	st.nextToken();       // skip the classname

	String stackTrace = "\n<STACK TRACE ";
	String fileAndLine = null;
	
	while( st.hasMoreTokens() ) {
	    String line = st.nextToken();
	    // Check if the line is not in the Debug package.
	    if( line.indexOf( ignoreLine ) < 0 ) {
		// OK, we're out of Debug in the stack, and if this is the first line past it, this is where we want te get fileAndLine.
		if( fileAndLine == null ) {
		    int openParen = line.lastIndexOf( '(' );
		    int closeParen = line.lastIndexOf( ')' );
		    try {
			fileAndLine = line.substring( openParen + 1, closeParen );
		    } catch( StringIndexOutOfBoundsException ex ) {
			fileAndLine = "???";
		    }
		    // If we're running under JIT, there's no hope.
		    if( fileAndLine.equals( compiledCode ) )
			fileAndLine = "(In JIT)";

		    stackTrace += fileAndLine + "> " + message + "\n" + line + "\n";

		} else
		    stackTrace += line + "\n";

	    }

	}

	output( stackTrace );

    }

    /** 
     * This is the Debug method inserted by the pre-processor,
     * debugPreProcessor.pl.  It is given the name of the debug level,
     * line number, and file name before all other options to avoid
     * the speed hit of throwing an Exception and then catching it.  
     *
     * @param code the code level to report this at
     * @param line the number number of the file that we are working with
     * @param filen the name of the file that we shold display
     * @param string the message that we shoudl display
     */
    public synchronized static void explicit( int code, int line, String filen, String string ) {
	// now get the actual output message to a string so we can
	// decide whether we are going to send it or not
	String outputDebug = "<";
	switch( code ) {
	case FAILURE:
	    outputDebug += "FAILURE"; break;
	case CRITICAL:
	    outputDebug += "CRITICAL"; break;
	case ERROR:
	    outputDebug += "ERROR"; break;
	case WARNING:
	    outputDebug += "WARNING"; break;
	case NOTICE:
	    outputDebug += "NOTICE"; break;
	default:
	    outputDebug += "UNKNOWN"; break;
	    
	}
	String fileAndLine = filen + ":" + line;
	if( line <= 0 )
	    fileAndLine = filen;
	if( fileAndLine.length() > 1 )
	    outputDebug += " " + fileAndLine;
	outputDebug += "> " + string;

	// now let's get the streams that we are going to set this to
	Hashtable streamsHash = (Hashtable)streams.get( filen );
	if( ( streamsHash == null ) && !fallthrough )
	    return;

	else if( ( streamsHash == null ) && fallthrough && ( code <= level ) ) {
	    output( outputDebug ); // this is the default -- we fell through -- make sure our level is right
	    return;

	}

	else if ( streamsHash == null )  // NOTE: We need to check this.  It should never be null, but it is.
	    return;

	// and here, we shoudl go thorugh all the streams and output
	// them -- this might be a problem if it takes "time" to send
	// the debug message to one of the streams
	Enumeration e = streamsHash.keys();
	while( e.hasMoreElements() ) {
	    PrintStream stream = (PrintStream)e.nextElement();
	    int level = ((Integer)streamsHash.get( stream )).intValue();
	    if( level >= code )
		stream.println( outputDebug );

	}

    }

    /**
     * pass the preproccessed information over to the other method,
     * just figure out what to display for the object
     *
     * @param code the code level to report this at
     * @param line the number number of the file that we are working with
     * @param filen the name of the file that we shold display
     * @param string the message that we shoudl display
     * @param obj the object that we want to print information for 
     */
    public synchronized static void explicit( int code, int line, String filen, String string, Object obj ) {
	String debugLine = "";
	if( string != null )
	    debugLine = string + "\n";

	debugLine += translateObject( obj );

	explicit( code, line, filen, debugLine );

    }

    /**
     * do preprocssed stuff
     *
     * @param code the code level to report this at
     * @param line the number number of the file that we are working with
     * @param filen the name of the file that we shold display
     * @param obj the object that we want to print information for 
     */
    public synchronized static void explicit( int code, int line, String filen, Object obj ) {
	explicit( code, line, filen, null, obj );

    }
  
    /** 
     * Prints an arbitrary debug message with the given code.  This is
     * the base method that prints things out; all other methods call
     * this to do their printing, with the exception of println(int,
     * Object) (which may call Throwable.printStackTrace() and
     * assert(boolean).  
     *
     * @param code the code level
     * @param string the message that we are doing this at
     */
    public synchronized static void println( int code, String string ) {
	if( code <= level ) {
	    String fileAndLine = findFileAndLine();

	    int colonIndex = fileAndLine.indexOf( ":" );
	    String file = "";
	    String number = "-1";
	    if( colonIndex != -1 ) {
		file = fileAndLine.substring( 0, fileAndLine.indexOf( ":" ) );
		number = fileAndLine.substring( fileAndLine.indexOf( ":" ) + 1 );

	    }
	    
	    explicit( code, Integer.parseInt( number ), file, string );

	}

    }

    /** 
     * Prints out the given object; if the object is an exception,
     * then its stacktrace is also printed.  
     * 
     * @param code the code level
     * @param obj the object that we want to print
     */
    public synchronized static void println( int code, Object obj ) {
	println( code, translateObject( obj ) );

    }

    /** 
     * Prints a debug message as well as the given object. 
     *
     * @param code the code level we want to report at
     * @param string the string message
     * @param obj the object that we want to display also
     */
    public static void println( int code, String string, Object obj ) {
	println( code, string + translateObject( obj ) );

    }

    /**
     * translate an object to the appropriate string.  exceptions and all
     *
     * @param object the object that we want to display
     * @return a string that is the representation
     */
    protected static String translateObject( Object obj ) {
	String debugLine = "";
	if( obj == null )
	    debugLine += "(null)";

	else if( obj instanceof Throwable ) {
	    Throwable ex = (Throwable)obj;
	    StringWriter stringWriter = new StringWriter();
	    ex.printStackTrace( new PrintWriter( stringWriter ) );
	    debugLine += stringWriter.toString();

	} else
	    debugLine += obj.toString();

	return debugLine;

    }

    public synchronized static void assert( boolean exp ) {
	assert( exp, "unknown assertion" );

    }

    /** 
     * Tests a condition; if it is false, print out an assertion
     * failure and throw an error.  
     *
     * @param exp the expression that we are going for
     * @param msg the messge that we want to display
     */
    public synchronized static void assert( boolean exp, String msg ) {
        if( !exp ) {
	    output( "<ASSERTION FAILURE" );
	    String fileAndLine = findFileAndLine();
	    if( fileAndLine.length() > 0 )
		output( " " + findFileAndLine() );

	    output ("> " + msg );
	    throw new Error();

        }

    }

    /**
     * Calculate what file and line number the Debug package was
     * called from.  This is analagous to the use of __LINE__ and
     * __FILE__ macros in popular C debugging packages, although
     * trickier because we have to get the information from the Java
     * runtime.
     *
     * This code is a big ugly hack - the idea is to generate a
     * Throwable, print a stack trace, then walk up the stack trace to
     * find the first line there that does not belong to the Debug
     * package itself.  That line should be of the form at
     * net.hivecell.hive.cell.Cell.createAgent(Cell.java:201) The
     * thing in parentheses is what we want!
     *
     * Obviously, this code relies on the string form of stack traces.
     * It seems to work fine in Sun JDK 1.1.7, other versions may
     * vary.  It won't work with a JIT, because this info is not
     * available.  
     */
    private static String findFileAndLine() {
	// Lines that match this are ignored
	final String ignoreLine = "at net.hivecell.hive.support.Debug";
	// If we find this as the file and line, then give up.
	final String compiledCode = "Compiled Code";
	
	// Generate a stack trace and capture it into a string
	StringWriter stringWriter = new StringWriter();
	PrintWriter writer = new PrintWriter( stringWriter );
	( new Throwable() ).printStackTrace(writer);
	String stackOutput = stringWriter.toString();
	
	
	// Split the stack trace up by lines, look for the non-Debug class
	StringTokenizer st = new StringTokenizer( stackOutput, "\n" );
	st.nextToken();       // skip the classname
	while( st.hasMoreTokens() ) {
	    String line = st.nextToken();
	    // Check if the line is not in the Debug package.
	    if( line.indexOf( ignoreLine ) < 0 ) {
		// OK, we're out of Debug in the stack, grab the thing inside parens
		int openParen = line.lastIndexOf( '(' );
		int closeParen = line.lastIndexOf( ')' );
		String fileAndLine;
		try {
		    fileAndLine = line.substring( openParen + 1, closeParen );
		    
		} catch( StringIndexOutOfBoundsException ex ) {
		    return "???";                         // parse error.
		    
		}
		// If we're running under JIT, there's no hope.
		if( fileAndLine.equals( compiledCode ) )
		    return "";
		
		else
		    return fileAndLine;
		
	    }
	}

	// Hmm, we never found a suitable line, give up.
	return "???";
    }

    /**
     * print out an error message
     *
     * @param string the message 
     */
    public static void error( String string ) {
	println( ERROR, string );

    }

    /**
     * print out an error message
     *
     * @param string the message
     * @param obj the object
     */
    public static void error( String string, Object obj ) {
	println( ERROR, string, obj );

    }

    /**
     * print out an error message
     *
     * @param obj the object that we want to print out
     */
    public static void error( Object obj ){
	println(ERROR, obj);

    }

    /**
     * print out a failure message
     *
     * @param string the message
     */
    public static void failure( String string ) {
	println( FAILURE, string );

    }

    /**
     * print out a failure message
     *
     * @param string the message
     * @param obj the object that we want to print out
     */
    public static void failure( String string, Object obj ) {
	println( FAILURE, string, obj );

    }

    /**
     * print out a failure message
     *
     * @param obj the object that we want to print out
     */
    public static void failure( Object obj ) {
	println( FAILURE, obj );

    }

    /**
     * print out a critical message
     *
     * @param string the message
     */
    public static void critical( String string ) {
	println( CRITICAL, string );

    }

    /**
     * print out a critical message
     *
     * @param string the message
     * @param obj the object that we want printed out also
     */
    public static void critical( String string, Object obj ) {
	println( CRITICAL, string, obj );

    }

    /**
     * print out a critical message
     *
     * @param obj the object that we want to print out
     */
    public static void critical( Object obj ) {
	println( CRITICAL, obj );

    }

    /**
     * print out a warning
     *
     * @param string the string that we want to print out
     */
    public static void warning( String string ) {
	println( WARNING, string );

    }

    /**
     * print out a warning
     *
     * @param string the string that we want to print out
     * @param obj the object that we want to print out
     */
    public static void warning( String string, Object obj ) {
	println( WARNING, string, obj );

    }

    /**
     * print out a warning
     * 
     * @param obj the object that we want to print out
     */
    public static void warning( Object obj ) {
	println( WARNING, obj );

    }

    /**
     * print out a notice
     *
     * @param string the string that we want to print out
     */
    public static void notice( String string ) {
	println( NOTICE, string );

    }
    
    /**
     * print out a notice
     *
     * @param string the string that we want to print out
     * @param obj the object that we want associated with it
     */
    public static void notice( String string, Object obj ) {
	println( NOTICE, string, obj );

    }

    /**
     * print out a notice
     *
     * @param obj the object that we want associated with it
     */
    public static void notice( Object obj ) {
	println( NOTICE, obj );

    }


    /** 
     * Sets the stream used for error messages 
     *
     * @param out_arg the stream that we want to send all the outputs to
     */
    public synchronized static void setStream( PrintStream out_arg ) {
	out = out_arg;

    }

    /** 
     * Sets the current debug level; messages this serious or more will
     * be printed out. 
     *
     * @param level_arg the level that we want to set the minimum to
     */
    public static void setDefaultLevel( int level_arg ) {
	level = level_arg;

    }

    /**
     * Returns the current debug level. 
     *
     * @return the level at which we are reporting debug calls
     */
    public static int getDefaultLevel() {
	return level;

    }

    /**
     * set whether we have fall through on or not
     *
     * @param fallthrough the flag to set whether we have fallthrough on or not for teh debug messages
     */
    public static void setFallthrough( boolean fallthrough_arg ) {
	fallthrough = fallthrough_arg;

    }

    /**
     * tell us whether or not we have fallthrough on
     *
     * @return whether or not we have fallthrough on
     */
    public static boolean getFallthrough() {
	return fallthrough;

    }

    /**
     * convert the classname into its java file name
     *
     * @param classname the class object
     * @return the name of the java file
     */
    protected static String convertClass( Class classname ) {
	String classfile = classname.getName();
	classfile = classfile.substring( classfile.lastIndexOf( "." ) + 1 ) + ".java";
	return classfile;

    }

    /**
     * set the flag on a certain class to know what to do with it
     *
     * @param classname the class that we want to note the level with
     * @param level the level at which we shoudl be dealing with this class
     */
    public synchronized static void setClassLevel( Class classname, int level ) {
	setClassStream( classname, out, level );

    }

    /**
     * add an output stream for a specific class
     *
     * @param classname the class that we want to add astream for
     * @param out the PrintStream that we want to send the data to
     * @param level the level that we want to associate with this stream
     */
    public synchronized static void setClassStream( Class classname, PrintStream out, int level ) {
	String classFileName = convertClass( classname );
	Hashtable streamHash = (Hashtable)streams.get( classFileName );
	if( streamHash == null ) {
	    streamHash = new Hashtable();
	    streams.put( classFileName, streamHash );

	}
	streamHash.put( out, new Integer( level ) );

    }

    /**
     * add an output stream for a specific class at the default level
     *
     * @param classname the class that we want to add a stream for
     * @param out the PrintStream that we want to send the data to
     */
    public synchronized static void setClassStream( Class classname, PrintStream out ) {
	setClassStream( classname, out, getDefaultLevel() );

    }

    /**
     * get the list of output streams for a specific class
     *
     * @param classname the class that we want to get the streams for
     * @return an array of StreamLevelPairs
     */
    public synchronized static PrintStream[] getClassStreams( Class classname ) {
	String classFileName = convertClass( classname );
	Hashtable streamHash = (Hashtable)streams.get( classFileName );
	if( streamHash == null )
	    return null;
	PrintStream[] streamArray = new PrintStream[streamHash.size()];
	Enumeration e = streamHash.keys();
	int count = 0;
	while( e.hasMoreElements() )
	    streamArray[count++] = (PrintStream)e.nextElement();
	return streamArray;

    }

    /**
     * get the level a certain stream is registered at
     *
     * @param out the print stream we are querying
     * @return the level the stream is at
     * @exception NoSuchElementException if the stream is not here
     */
    public synchronized static int getClassStreamLevel( Class classname, PrintStream out ) 
    throws NoSuchElementException {
	String classFileName = convertClass( classname );
	Hashtable streamHash = (Hashtable)streams.get( classFileName );
	if( streamHash == null )
	    throw new NoSuchElementException( classname.getClass() + " is not registered" );
	if( !streamHash.containsKey( out ) )
	    throw new NoSuchElementException( classname.getClass() + " does not have " + out + " as a stream" );
	return ((Integer)streamHash.get( out )).intValue();

    } 

    /**
     * remove a stream from a specific class
     *
     * @param classname the class that we want to remove the streams for
     * @param out the PrintStream we want to remove
     * @return true if we removed it, false otherwise
     */
    public synchronized static boolean removeClassStream( Class classname, PrintStream out ) {
	String classFileName = convertClass( classname );
	Hashtable streamHash = (Hashtable)streams.get( classFileName );
	if( ( streamHash == null ) ||
	    ( !streamHash.containsKey( out ) ) )
	    return false;
	return( streamHash.remove( out ) != null );

    }

    /**
     * remove all streams from a specific class 
     *
     * @param classname the class that we want to remove all the streams for
     * @return true if we removed all the streams
     */
    public synchronized static boolean removeClassAllStreams( Class classname ) {
	String classFileName = convertClass( classname );
	return( streams.remove( classFileName ) != null );

    }
    
}
