// $Id: PPM.java,v 1.1 2000/05/18 22:01:17 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.support; 

import net.hivecell.hive.support.Debug;
import java.io.*;
import java.net.*;
import java.awt.*;
import java.awt.image.*;
import net.hivecell.hive.support.*;

/** Each instance of this class represents an image stored as an array of
 * ints in ARGB format. PPMs can be constructed from images, or more often
 * from files or inputstreams in PPM format.
 * Although this library can decode any bit depth from 1 - 24 per pixel,
 * it always writes in 24 bits per pixel.
 * 
 */
public class PPM extends SerializableImage {
    
    
    /** When writing PPM files they can be either BINARY or ASCII. Binary
     * is recommended because it is smaller and faster to process.
     */
    public static final int BINARY = 0;
    /** Write PPM in ASCII format
     */
    public static final int ASCII = 1;

    /** when asked to get a PPM from an image, it asks its superclass
     *  SerializableImage
     * @deprecated
     */
    public PPM(Image img) {
        super(img);
    }
    
    /** Create a new PPM from an input stream.
     */

    public PPM(InputStream in) throws IOException, UnsupportedEncodingException {
	StreamTokenizer st = new StreamTokenizer(in);
	st.commentChar('#');
	st.nextToken();
	int mode;
	if (st.sval.equals("P3")) {
	    mode = ASCII;
	}
	else if (st.sval.equals("P6")) {
	    mode = BINARY;
	}
	else {
	    throw new UnsupportedEncodingException();
	}
	dim = new Dimension();
	st.nextToken();
	dim.width = (int) Math.round(st.nval);
	st.nextToken();
	dim.height = (int) Math.round(st.nval);
	st.nextToken();
	data = new int[dim.width * dim.height];
	int maxVal = (int) Math.round(st.nval);
	switch(mode) {
	case ASCII: 
	    // This is an optimization so that 24 bit files need not go to the
	    // trouble of rescaling to 24 bit values.
	    if (maxVal != 255) {
		for (int i = 0; i < dim.width * dim.height; i++) {
		    st.nextToken();
		    data[i] =
			(255 << 24) + 
			(rescale((int) Math.round(st.nval), maxVal, 255) << 16);
		    st.nextToken();
		    data[i] +=
			(rescale((int) Math.round(st.nval), maxVal, 255) <<  8);
		    st.nextToken();
		    data[i] +=
			(rescale((int) Math.round(st.nval), maxVal, 255));
		}
	    }
	    else {
		// This is an optimization so that 24 bit files need not go
		// to the trouble of rescaling to 24 bit values.
		for (int i = 0; i < dim.width * dim.height; i++) {
		    st.nextToken();
		    data[i] = (255 << 24) +
			((int) st.nval) << 16;
		    st.nextToken();
		    data[i] += ((int) st.nval) << 8;
		    st.nextToken();
		    data[i] += (int) st.nval;
		}
	    }
	    in.close();
	    break;
	case BINARY:
	    if (maxVal != 255) {
		for (int i = 0; i < dim.width * dim.height; i++) {
		    data[i] =
			(255 << 24) + 
			(rescale(in.read(), maxVal, 255) << 16) +
			(rescale(in.read(), maxVal, 255) <<  8) +
			(rescale(in.read(), maxVal, 255)      );
		}
	    }
	    else {
		for (int i = 0; i < dim.width * dim.height; i++) {
		    data[i] =
			(255 << 24) +
			(in.read() << 16) +
			(in.read() <<  8) +
			in.read();
		}
	    }
	    in.close();
	    break;
	}
    }

    /** Create a new PPM from a file.
     */
    public PPM(String filename)
	throws IOException, UnsupportedEncodingException {
	this(new BufferedInputStream (new FileInputStream (filename)));
    }

    /** Create a new PPM from an URL.
     */
    public PPM(URL url) throws IOException, UnsupportedEncodingException {
	this(url.openStream());
    }

    /** Write the PPM out to a file in binary format.
     */
    public void write(String filename) throws IOException {
	write(filename, BINARY);
    }

    /** Write the PPM out to a file in the given mode.
     */
    public void write(String filename, int mode) throws IOException {
	FileOutputStream fos = new FileOutputStream(filename);
	write(fos, mode);
	fos.flush();
	fos.close();
    }

    /** Write the PPM out to an arbitrary output stream in binary format.
     */
    public void write(OutputStream out) throws IOException {
	write(out, BINARY);
    }
    
    
    /** Write the PPM out to an arbitrary output stream in the given mode
     */
    public void write(OutputStream out, int mode) throws IOException {
	switch(mode) {
	case ASCII:
	    PrintStream ps = new PrintStream(out);
	    ps.println("P3");
	    ps.println(dim.width + " " + dim.height);
	    ps.println(255);
	    for (int a = 0; a < dim.height; a++) {
		for (int b = 0; b < dim.width; b++) {
		    int offset = (dim.width * a) + b;
		    ps.print(((data[offset] >> 16) & 255) + " " +
			     ((data[offset] >>  8) & 255) + " " +
			     ((data[offset]      ) & 255) + " ");
		}
		ps.println();
	    }
	    ps.close();
	    break;
	case BINARY:
	    DataOutputStream dos = new DataOutputStream(out);
	    dos.writeBytes("P6\n");
	    dos.writeBytes(dim.width + " " + dim.height + "\n");
	    dos.writeBytes("255\n");
	    for(int offset = 0; offset < dim.width * dim.height; offset++) {
		out.write((data[offset] >> 16) & 255);
		out.write((data[offset] >>  8) & 255);
		out.write((data[offset]      ) & 255);
	    }
	    dos.close();
	    out.close();
	    break;
	default:
	    throw new IOException("Unknown Mode Exception.");
	}
    }
    
    /**      * 
     * @param maxValue 
     * @param desiredMax 
     * @param value 
     */
    protected static int rescale(int value, int maxValue, int desiredMax) {
	double frac = (double) value / (double) maxValue;
	int val = (int) Math.round(frac * (double) desiredMax);
	return val;
    }

    /**      * 
     * @param maxValue 
     * @param desiredMax 
     * @param value 
     */
    protected static int rescale(byte value, int maxValue, int desiredMax) {
	return rescale((int) value, maxValue, desiredMax);
    }

    
    /**       
     * @param args 
     */
    public static void main(String[] args) {
	try {
	    PPM ppm = new PPM(args[0]);
	    ppm.write("out.ppm");
	}
	catch(Exception e) {
	    e.printStackTrace();
	}
    }

}



