// $Id: SerializableImage.java,v 1.1 2000/05/18 22:01:17 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.support; 

import net.hivecell.hive.support.Debug;
import java.io.*;
import java.net.*;
import java.awt.*;
import java.awt.image.*;
import java.util.zip.*;

/** Each instance of this class represents an image stored as an array of
 * ints. These images are not AWT images - they are device
 * independent images that can be serialized. They can be created from
 * AWT Images, or from subclasses such as PPM that read from files.
 * The format is an array of integers - each integer is a pixel in ARGB
 * format (the same that MemoryImageSource uses in the default color model).
 * 
 */
public class SerializableImage implements Serializable, Cloneable {


    /** The dimension of the image being represented.
     */
    protected Dimension dim;
    
    /** The pixels of the image in ARGB format (default java color model)
     */
    protected int[] data;

    /** Clone a serializable image
     */
    public Object clone() {
	Dimension newDim = new Dimension(dim.width, dim.height);
	int[] newData = new int[data.length];
	System.arraycopy(data, 0, newData, 0, data.length);
	return new SerializableImage(newData, newDim);
    }

    /** Reads the compressed image from the input stream
     */
    private void readObject(ObjectInputStream in) throws IOException, ClassNotFoundException {

        GZIPInputStream gis = new GZIPInputStream(in);
        ObjectInputStream is = new ObjectInputStream(gis);
        
        dim = (Dimension)is.readObject();
        data = (int[])is.readObject();
    }        

    /** Writes the dimensions and the data compressed to the outputstream
     *  For now, the image will always be compressed.
     */
    private void writeObject(ObjectOutputStream out) throws IOException {

        GZIPOutputStream gos = new GZIPOutputStream (out);
        ObjectOutputStream os = new ObjectOutputStream(gos);
        
        os.writeObject(dim);
        os.writeObject(data);
        
        os.flush();
        gos.finish();        
    }    
    
    public SerializableImage() {
        super();
    }
    
    /** Create a new image from a pixel array and a Dimension.
     */
    public SerializableImage(int[] data, Dimension dim) {
	this.data = data;
	this.dim = dim;
    }
    
    /** Create a new serializable image from an AWT image.
     */
    public SerializableImage(Image image) {
	dim = new Dimension(-1, -1);
	while (dim.width == -1) {
	    dim.width = image.getWidth(null);
	}
	while (dim.height == -1) {
	    dim.height = image.getHeight(null);
	}
	data = new int[dim.width*dim.height];
	PixelGrabber pg = new PixelGrabber(image, 0, 0,
					   dim.width, dim.height,
					   data,
					   0, dim.width);
	try { pg.grabPixels(); } catch (Exception e) { Debug.println(Debug.ERROR, "Pixelgrabber failed"); }
    }

    /** Create an AWT image from our internal image data.
     */
    public Image getImage() {
	return Toolkit.getDefaultToolkit().createImage(getImageProducer());
    }    
    
    /** Return an ImageProducer from our internal image data.
     */
    public ImageProducer getImageProducer() {
        return new MemoryImageSource(dim.width, dim.height, data, 0,dim.width);
    }

    /** returns the dimension of the image.
     */
    public Dimension getSize() {
	return dim;
    }

    /** This method returns the raw pixels of the underlying image. It
     * kind of breaks the abstraction barrier, but it comes in handy to
     * implement filters without needing a PixelGrabber
     */
    public int[] getRep() {
	return data;
    }
    
}



