// $Id: VariablePropertyManipulator.java,v 1.1 2000/05/18 22:01:17 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.support;

import net.hivecell.hive.description.Describable;
import net.hivecell.hive.description.Description;

import java.beans.BeanInfo;
import java.beans.IntrospectionException;
import java.beans.Introspector;
import java.beans.PropertyDescriptor;
import java.beans.PropertyEditor;
import java.beans.PropertyEditorManager;
import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.rmi.RemoteException; 
import java.util.Enumeration;
import java.util.Hashtable;

/**
 * this class is meant to give access to the variables that can be set
 * and unset in the bean style method calls on a certain class (mostly
 * agents, but whatever) 
 */
public class VariablePropertyManipulator {

    /**
     * the object that we want to manipulate
     */
    private Object o; 

    /**
     * the methods that we use to set the beans (variable names->set methods)
     */
    private Hashtable setMethods = new Hashtable();

    /**
     * the methods that we use to get the beans (variable names->get methods)
     */
    private Hashtable getMethods = new Hashtable();

    /**
     * the types of classes the parameters take
     */
    private Hashtable getParamClasses = new Hashtable();

    /**
     * the property descriptors -- these are left as last minute stuff
     * so just in case the set method is not cooperating with us, we
     * can fall back on this 
     */
    private Hashtable propertyDescriptors = new Hashtable();


    /**
     * create a new manipulator class for this object
     *
     * @param o the object that we want the manipulation for
     * @param stopClass the class to stop the analysis at
     * @exception VariablePropertyManipulationException is thrown if something went wrong introspecting 
     */
    public VariablePropertyManipulator( Object o, Class stopClass ) 
    throws VariablePropertyManipulationException {

	this.o = o;
	if( stopClass == null )
	    stopClass = java.lang.Object.class;
	getAccessingInfo( stopClass ); // just set up our hashtables

    }

    /**
     * create a new manipulator class for this object
     *
     * @param o the object that we want the manipulation for
     * @exception VariablePropertyManipulationException is thrown if something went wrong introspecting 
     */
    public VariablePropertyManipulator( Object o ) 
    throws VariablePropertyManipulationException {
	this( o, null );

    }

    /**
     * get the (pseudo)names of the variables that we can set via this
     * bean manipulator class.  these names are the pseudo names,
     * because there is no guarantee that these are the real names of
     * the variablbes, but these are the names that we are extracting
     * and using through this class.  this helper function fills the
     * hashtable mapping the variable names to the methods needed to
     * get the information from the o
     *
     * @param stopClass the class to stop the analysis at -- set at null if we want to see everything
     * @exception VariablePropertyManipulationException is thrown if there is something wrong getting the information 
     */
    private void getAccessingInfo( Class stopClass ) 
    throws VariablePropertyManipulationException {
	
	// first get all the bean information
	BeanInfo bi = null;
	try {
	    bi = Introspector.getBeanInfo( this.o.getClass(), stopClass );
	} catch( IntrospectionException error ) { 
	    throw new VariablePropertyManipulationException( "cannot introspect " + this.o.getClass().getName(), error );
	}

	// get all the information on the beans such as the methods to
	// get and set them and then we can compute the name of the
	// variable from all this
	PropertyDescriptor[] pd = bi.getPropertyDescriptors();
	for( int count=0;count < pd.length;count++ ) {
	    Method getMethod = pd[count].getReadMethod();
	    Method setMethod = pd[count].getWriteMethod();
	    Class paramClass = pd[count].getPropertyType();
	    if( ( getMethod != null ) &&
		( setMethod != null ) &&
		( paramClass != null ) ) {
		
		// this should get the name of the variable and pull
		// off the "set" in the front
		String variableName = setMethod.getName().substring( 3 );

		// now store the information away
		getMethods.put( variableName, getMethod );
		setMethods.put( variableName, setMethod );
		getParamClasses.put( variableName, paramClass );
		propertyDescriptors.put( variableName, pd[count] );
		
	    }

	}

    }

    /**
     * get an array of the variables that we can manipulate on our object
     *
     * @return an array of string naming the variables
     */
    public String[] getAccessibleVariableNames() {

	String[] varNames = new String[ getMethods.size() ];
	int position = 0;
	Enumeration e = getMethods.keys();
	while( e.hasMoreElements() )
	    varNames[position++] = (String)e.nextElement();
	return varNames;

    }

    /**
     * get what java class a certain variable is
     *
     * @param variableName the name of the variable that we want the information for
     * @return the class of the parameter
     * @exception UnknownVariablePropertyName if the variable is not named in the object
     */
    public Class getAccessibleVariableType( String variableName ) 
    throws UnknownVariablePropertyName {

	if( !getParamClasses.containsKey( variableName ) )
	    throw new UnknownVariablePropertyName();

	return (Class)getParamClasses.get( variableName );

    }

    /**
     * get what a certain variable is already set at
     *
     * @param variableName the name of the variable we want information for
     * @return an object with the data in it
     * @exception UnknownVariablePropertyName if the variable is not named in the object, VariablePropertyManipulationException if something went wrong getting the information
     */
    public Object getAccessibleVariable( String variableName ) 
    throws UnknownVariablePropertyName, VariablePropertyManipulationException {

	if( !getMethods.containsKey( variableName ) )
	    throw new UnknownVariablePropertyName();

	Method getMethod = (Method)getMethods.get( variableName );
	try {
	    Object[] vars = { };
	    return getMethod.invoke( this.o, vars );
	} catch( IllegalAccessException error ) {
	    // this one should not happen

	} catch( IllegalArgumentException error ) {
	    // this one does not catch either

	} catch( InvocationTargetException error ) {
	    // this one might occur depending on remote stuff and all that fun crap
	    throw new VariablePropertyManipulationException( "problem invoking " + getMethod.getName() + " on " + this.o, error );

	}

	Debug.error( "problem getting the value of " + variableName );
	return null;
	    
    }

    /**
     * set a certain bean variable
     *
     * @param variableName the name of the variable we want to set
     * @param value the object which is the value we want to set the object at
     * @return the old value of the object
     * @exception UnknownVariablePropertyName if the variable is not named in the object, VariablePropertyManipulationException if something went wrong getting the information
     */
    public Object setAccessibleVariable( String variableName, Object value ) 
    throws UnknownVariablePropertyName, VariablePropertyManipulationException {

	Object oldValue = getAccessibleVariable( variableName );
	
	Method setMethod = (Method)setMethods.get( variableName );
	try {
	    Object[] vars = { value };
	    setMethod.invoke( this.o, vars );
	} catch( IllegalAccessException error ) {
	    // this one should not happen

	} catch( IllegalArgumentException error ) {
	    // this one does not catch either

	} catch( InvocationTargetException error ) {
	    // hmm -- something went wrong when we are setting this --
	    // if value is a string, then let's try setting this as
	    // text, if that works, then good!  else, forget it --
	    // just throw the exception
	    if( value instanceof String ) {
		// use the property descriptor to do the conversion for ys
		PropertyDescriptor pd = (PropertyDescriptor)propertyDescriptors.get( variableName );
		PropertyEditor editor = PropertyEditorManager.findEditor( pd.getPropertyType() );
		editor.setAsText( (String)value );

		// now let's set this the hard way
		Object[] vars = { editor.getValue() };
		try {
		    setMethod.invoke( this.o, vars );
		} catch( Exception internalError ) {
		    throw new VariablePropertyManipulationException( "problem invoking " + setMethod.getName() + " on " + this.o, internalError );
		}
	    } else 
		// this one might occur depending on remote stuff and all that fun crap
		throw new VariablePropertyManipulationException( "problem invoking " + setMethod.getName() + " on " + this.o, error );

	} finally {
	    return oldValue;

	}

    }


    /**
     * use this to manipulate an object so we set its variables to
     * what this object's description has it in
     *
     * @param o the object we want configured 
     */
    public static void configure( Object o ) { configure( o, null ); }

    /**
     * use this to manipulate an object so we set its variables to
     * what this object's description has it in
     *
     * @param o the object we want configured 
     * @param prefix the prefix of the configuration stuff
     */
    public static void configure( Object o, String prefix ) { if( o instanceof Describable ) configure( (Describable)o, prefix ); }

    /**
     * use this to manipulate an object so we set its variables to
     * what this object's description has it in
     *
     * @param d the object we want configured 
     */
    public static void configure( Describable d ) { configure( d, null ); }

    /**
     * use this to manipulate an object so we set its variables to
     * what this object's description has it in
     *
     * @param o the object we want configured 
     * @param prefix the prefix of the configuration stuff
     */
    public static void configure( Describable d, String prefix ) {

	// get the description of the object as this is how we are setting its variables on startup
	Description description = null;
	description = d.getDescription();
	if( description == null ) return;

	// take the desc set and narrow it down so we only have the pertinant information
	Description config = description.configContext();
	if (config != null)
	    if( prefix != null && config.select( prefix ) > 0 )
		description = config.getMatch();

	VariablePropertyManipulator vpm = null;
	try {
	    vpm = new VariablePropertyManipulator( d );
	} catch( VariablePropertyManipulationException error ) {
	    Debug.warning( "cannot create a variable property manipulator for " + d );
	    return;
	}
	String[] variableNames = vpm.getAccessibleVariableNames();
	for( int count=0;count<variableNames.length;count++ ) {

	    // figure out the property as defined by the description,
	    // and then attempt to set the value 
	    String variableName = variableNames[count];
	    Description propertyDescription;
	    String value = null;
	    
	    if ( description.select(variableName) > 0 ) {
		propertyDescription = description.getMatch();
		value = propertyDescription.getText();
	    }

	    if( value != null )
		try {
		    Debug.notice( "setting " + variableName + " to " + value + " in " + d );
		    vpm.setAccessibleVariable( variableName, value );
		    
		} catch( UnknownVariablePropertyName error ) {
		    Debug.warning( variableName + " is not a valid method to set on this object" );
		    
		} catch( VariablePropertyManipulationException error ) {
		    Debug.warning( "problem setting the variable on " + variableName );
		    
		}
	}
	

    }


}



