// $Id: AsyncCall.java,v 1.1 2000/05/18 22:01:20 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.support.async;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.io.Serializable;

/**
 * The class to allow an asynchronous call to anything -- this caller
 * creates a thread every time a call is made to wait for the result
 * of the method call 
 */
public class AsyncCall {

    /** the class used to make the call with -- this class -has- to
        implement AsyncCallerInterface **/
    private static AsyncCallerClass asyncObject = AsyncCaller.getInstance();


    /** The name of the method that we are going to call **/
    private Method method;
    /** The objects which will be passed as the parameters to the call **/
    private Object[] parameters;


    protected AsyncCall() { }

    /**
     * Create the object that can be used to make the asynchronous
     * call.  Take the method we wish to call on which class with the
     * parameters to use.
     *
     * @param methodName the name of the method to call
     * @param receiverType the class to maek the call onto
     * @param params the parameters to use 
     * @exception NoSuchMethodException if a matching method is not found
     * @exception SecurityException if access to the method is denied
     */
    public AsyncCall( String methodName, Class receiverType, Object[] params ) 
    throws NoSuchMethodException, SecurityException {

	Class[] paramClass = new Class[params.length];
	for( int count=0;count != params.length;count++ )
	    paramClass[count] = params[count].getClass();

	this.method = receiverType.getMethod( methodName, paramClass );
	this.parameters = params;

    }

    /**
     * Create the object to make the asynchronous call
     *
     * @param toCall the method object of the method we are going to call
     * @param params the parameters to use
     */
    public AsyncCall( Method toCall, Object[] params ) {
	
	this.method = toCall;
	this.parameters = params;

    }
    

    /**
     * Get the method object for what we are going to call
     * asynchronously
     * 
     * @return the method object 
     */
    public Method getMethod() {

	return this.method;

    }

    /**
     * Get the parameters we are going to use on this method call
     *
     * @return the array of objects we are going to use on the method
     * call 
     */
    public Object[] getParams() {

	return this.parameters;

    }

    /**
     * Change the implementation of the AsyncCaller.  It only accepts
     * classes which implement AsyncCallerInterface.
     *
     * @param asyncClass the new asyncCaller class
     * @exception NotAsyncCallerInterfaceException is thrown if the
     * class we are using to set does not implement
     * AsyncCallerInterface 
     *
     * @see AsyncCallerInterface
     */
    public static void setAsyncCallerClass( AsyncCallerClass asyncObject ) {

	AsyncCall.asyncObject = asyncObject;

    }

    /**
     * Get the objecy that we are currently using for the AsyncDelivery
     *
     * @return the class we are using
     */
    public Object getAsyncCallerObject() {

	return AsyncCall.asyncObject;

    }
    
    /**
     * Make the asynchronous call to the method.  This particular
     * method simply calls the other callOn method with a -1 time to
     * pass to the watching thread.
     *
     * @param callee the object to make the call on 
     */
    public AsyncResult callOn( Object callee ) {
	
	return callOn( callee, (long)-1 );

    }

    /**
     * Make the asynchronous call to the method with a "watching"
     * thread.  This delegates the call to the AsyncCaller
     * implementation that this object has it set to.
     *
     * @param callee the object to make the call on
     * @param millis the number of milliseconds before to give up 
     *
     * @see async.AsyncCallerInterface
     * @see async.AsyncCaller
     */
    public AsyncResult callOn( Object callee, long millis ) {

	return AsyncCall.asyncObject.makeCall( callee, this, millis );

    }
    
    /**
     * This is used by the async objects when they are actually making
     * the call asynchronously
     *
     * @param callee the object to call our method on
     * @return the result of this invocation 
     */
    public Object invoke( Object callee ) 
    throws IllegalAccessException, IllegalArgumentException, InvocationTargetException {
	
	return getMethod().invoke( callee, getParams() );

    }

    /**
     * This will make sure the types of the params class match those
     * which the Method object says.  This is a relatively slow
     * procedure, so that is why it is optional.  However, it is not
     * necessary if you use the first constructor
     *
     * @exception IllegalArgumentException if the types of the
     * parameters we are trying to pass in do not match those of the
     * method we created */
    public void validate() 
    throws IllegalArgumentException {

	Class[] paramTypes = this.method.getParameterTypes();
	if( paramTypes.length != this.parameters.length ) 
	    throw new IllegalArgumentException( "Incorrect amount of elements in method call" );
	for( int count=0;count != this.parameters.length;count++ ) 
	    if( !paramTypes[count].equals( this.parameters[count] ) )
		throw new IllegalArgumentException( "Incorrect parameter types in method call" );

    }

    /** Print out the method name this asynchronous call is for. **/
    public String toString() {
        /** Could also turn all the parameters into strings... */
        return method.toString();
    }
}
