// $Id: AsyncCaller.java,v 1.1 2000/05/18 22:01:20 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.support.async;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.rmi.ServerRuntimeException;
import java.util.Vector;

/** 
 * This is the default implementation of the asynchronous caller.  It
 * first creates the AsyncResult object, then spawns a thread which
 * will do the asynchronous call.  If the delay time passed to the
 * watching thread is non-negative, it will spin off another thread
 * which will sleep for a specified amount of time.  If the calling
 * thread does not complete before the watching thread, then it is
 * killed and the AsyncResult will contain an
 * AsyncCallNotCompleteException.
 *
 * This class is implemented as a singleton because there is no need
 * to have more than one object at a time running.  
 */
public class AsyncCaller 
extends AsyncCallerClass {

    /** The single instance of the caller that is allowed to run **/
    private static final AsyncCaller singleInstance = new AsyncCaller();

    /** A vector of all the async calls that are pending completion **/
    private static final ThreadGroup tg = new ThreadGroup( "AsyncCaller Threads" );
    

    /** 
     * Get a reference to the singleton instance of AsyncCaller. 
     */
    public static AsyncCallerClass getInstance() { 

	return (AsyncCallerClass)singleInstance; 

    }

    /**
     * Make the asynchronous call.  All this method call does is call
     * the other implementation of makeCall but with a -1 time passed
     * to the watching thread
     *
     * @param receiver the object to do the call on
     * @param call the specifics of the call
     * @return the async result of this call 
     */
    public AsyncResult makeCall( Object receiver, AsyncCall call ) {

	return makeCall( receiver, call, (long)-1 );

    }

    /**
     * Make the asynchronous call with an optional watching thread.
     * First the thread to do the asynchronous call is spawned off,
     * and then if the watching thread delay time is non-negative then
     * the watching thread is spawned off too.  If the watching thread
     * awakens before the calling thread terminates, then it kills the
     * watching thread and returns an AsyncCallNotCompleteException
     * throough the AsyncResult
     *
     * @param receiver the object to do the call on
     * @param call the specifics of the call
     * @param millis the number of milliseconds until we should just give up and return a AsyncCallNotCompleteException
     * @return the async result of this call 
     *
     * @see async.AsyncCallNotCompleteException
     */
    public AsyncResult makeCall( final Object receiver, final AsyncCall call, final long millis ) {

	final AsyncResultImpl result = new AsyncResultImpl( receiver, call );

	class CTRunnable
	implements Runnable {
	    private Thread wt = null;

	    public void setWatchingThread( Thread watchingThread ) {
		wt = watchingThread;
	    }

	    public void run() {
		
		Object o = null;
		boolean isException = false;
		try {
		    o = call.invoke( receiver );

		} catch( InvocationTargetException error ) {
		    // if the callee throws an exception, just pull it
		    // out and return that -- the
		    // InvocationTargetException will just be
		    // encapsulating an exception in the
		    // getMethod().invoke, and i think its an
		    // unnecessary wrapper.  the server runtime
		    // exception is similar, but means it happened
		    // over RMI
		    o = error.getTargetException();
		    if( o instanceof ServerRuntimeException )
			o = ((ServerRuntimeException)o).detail;

		    isException = true;
		    
		} catch( Exception error ) {
		    // the only other exceptions that this will throws
		    // is the IllegalArgumentException which we will
		    // just want to return and the same goes with the
		    // IllegalAccessException -- perhaps these should
		    // be encapsulated in a AsyncCallException?
		    o = error;
		    isException = true;

		}
		
		result.putResult( o, isException );

		// over here we want to tell if the watchingThread has
		// finished, and if not it should kill it
		
		if( wt != null )
		    wt.interrupt();

		// remove this from the pending calls list
		synchronized( AsyncCaller.this ) {
		    AsyncCaller.this.notifyAll();
		}

	    }
	};
	final CTRunnable callingThreadRunnable = new CTRunnable();


	final Thread callingThread = new Thread( tg, callingThreadRunnable );

	final Thread watchingThread = new Thread( new Runnable() {
		public void run() {

		    // this might not be the best way to do it.  what
		    // happens if it is interrupted
		    try {
			Thread.sleep( millis );
		    } catch( InterruptedException error ) {
			// somebody told us to stop doing what we are
			// doing.  this probably means the thread that
			// made the call is done
			return;
		    }
		    
		    result.putResult( new AsyncCallNotCompleteException(), true );

		    // will this cause it to be garbage collected? --
		    // do we need to garbage collect it?  this needs
		    // to be redone so it will be done through an
		    // interrupt
		    callingThread.stop();

		    // remove this from our list of pending calls
		    synchronized( AsyncCaller.this ) {
			AsyncCaller.this.notifyAll();
		    }

		}
	});

	
	if( millis >= 0 )
	    callingThreadRunnable.setWatchingThread( watchingThread );
	callingThread.start();
	if( millis >= 0 ) {
	    watchingThread.setPriority( callingThread.getPriority() + 1 );
	    watchingThread.start();
	}
	
	return result;
	
    }

    /**
     * This method should return true if this instance of an async
     * messenger has finished sending all the messages it has been
     * asked to
     *
     * @return true if we have finished sending all the messages we
     * have been asked to 
     */
    public boolean doneWithCalls() {

	if( tg.activeCount() == 0 )
	    return true;
	return false;

    }

    /**
     * This method should block until all the messages in this caller
     * have been finished
     */
    public void blockUntilDone() {

	while( !doneWithCalls() ) 
	    try {
		synchronized( this ) {
		    wait();
		}
	    } catch( InterruptedException error ) { }

    }


}




