// $Id: AsyncOneMethodCaller.java,v 1.1 2000/05/18 22:01:20 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.support.async;

import java.lang.reflect.Method;
import java.lang.reflect.InvocationTargetException;
import java.util.Hashtable;
import java.util.Vector;

import java.rmi.ServerRuntimeException;

public class AsyncOneMethodCaller 
extends AsyncCallerClass {

    /** The single instance of the caller that is allowed to run **/
    private static final AsyncOneMethodCaller singleInstance = new AsyncOneMethodCaller();

    /** 
     * The queues for each object.  This data structure hashes the
     * object to make the asynch call too with a Vector of AsyncCalls
     * to make on the object.  If an object is not in the data
     * structure, then it does not have a queue.  An object is in the
     * queue until the call has finished -- this is used so we can
     * tell if a remote object is being called on 
     */
    private static final Hashtable queues = new Hashtable();

    /** 
     * All the threads that are currently running -- useful to tell if
     * we have completed our calls 
     */
    private static final ThreadGroup tg = new ThreadGroup( "AsyncOneMethodCaller Threads" );


    /** 
     * Get a reference to the singleton instance of AsyncCaller.  */
    public static AsyncCallerClass getInstance() { 

	return (AsyncCallerClass)singleInstance; 

    }

    /**
     * Make the asynchronous call.  All this method call does is call
     * the other implementation of makeCall but with a -1 time passed
     * to the watching thread
     *
     * @param receiver the object to do the call on
     * @param call the specifics of the call
     * @return the async result of this call 
     */
    public AsyncResult makeCall( Object receiver, AsyncCall call ) {

	return makeCall( receiver, call, (long)-1 );

    }

    /**
     * Make the asynchronous call with an optional watching thread.
     *
     * First, check to see if there is a queue for the object.  If
     * there is a queue, then append ourselves to the vector for the
     * object.  If there isn't, then make a queue for our AsyncCall,
     * then spawn off a thread (with an optional watching thread -- if
     * the time passed is non negative).  At the termination of the
     * thread, it then checks the queue -- if there is something, it
     * acts on it.  If there isn't, then just remove the stuff from
     * the queue
     *
     * @param receiver the object to do the call on
     * @param call the specifics of the call
     * @param millis the number of milliseconds until we should just give up and return a AsyncCallNotCompleteException
     * @return the async result of this call 
     *
     * @see async.AsyncCallNotCompleteException */
    public synchronized AsyncResult makeCall( final Object receiver, final AsyncCall call, final long millis ) {

	final AsyncResultImpl result = new AsyncResultImpl( receiver, call );
	final AsyncOneMethodCallQueueElement queueElement = new AsyncOneMethodCallQueueElement( call, millis, result );
	
	// the runnable we will be using to monitor each individual
	// queue item.  this runnable will sleep for the time
	// specified in the milliseconds.  then after that, it will
	// awaken (it will be killed if our element has been dealt
	// with), and check on its element.  if our element has a
	// thread running with it then it kills the thread, it sets
	// the result to the proper exception, then removes the
	// element from the queue
	Runnable monitoringRunnable = new Runnable() {
	    public void run() {

		try {
		    Thread.sleep( millis ); // sleep for the proper amount
		} catch( InterruptedException error ) {
		    // the interruption has to come from the
		    // CTRunnable (see down below) interrupting us
		    // because the call is odne.  we need to return so
		    // it won't effect us
		    return;
		}

		if( queueElement.executingThread != null ) {
		    queueElement.executingThread.stop(); // kill the thread if there is one
		    result.putResult( new AsyncCallNotCompleteException( "Call taking too long" ), true ); // put the problem
		} else
		    result.putResult( new AsyncCallNotCompleteException( "Call never processed" ), true ); // put the problem

		// if our element is not the first item on the list,
		// then pull it from the queue.  otherwise, the other
		// thread will be dealing with it
		synchronized( queues ) {
		    if( ((Vector)queues.get( receiver )).indexOf( queueElement ) != 0 )
		        ((Vector)queues.get( receiver )).removeElement( queueElement );
		}

	    }
	};
    
	synchronized( queues ) {
	    if( queues.containsKey( receiver ) ) {
		// there is a queue for this reciever already!
		
		// we first need to spawn off a watching thread, and then
		// we need to add the element to the queue
		((Vector)queues.get( receiver )).addElement( queueElement );
		if( millis != -1 ) {
		    Thread monitoringThread = new Thread( monitoringRunnable );
		    queueElement.monitoringThread = monitoringThread;
		    monitoringThread.start();
		}
		
		return result;
		
	    }

	    // there is no queue for this element -- so create a queue
	    // and put the async call stuff into it so we can add on
	    // it
	
	    Vector queue = new Vector();
	    queues.put( receiver, queue );
	    
	    queue.addElement( queueElement ); // put the object into the queue

	    if( millis != -1 ) {
		Thread monitoringThread = new Thread( monitoringRunnable );
		queueElement.monitoringThread = monitoringThread;
		monitoringThread.start(); // start its montoring thread
	    }	
	}

	Runnable queueWatchingRunnable = new Runnable() {
	    public void run() {
		
		while( queues.containsKey( receiver ) ) {

		    // get the queue element
		    AsyncOneMethodCallQueueElement currentQueueElement = (AsyncOneMethodCallQueueElement)((Vector)queues.get( receiver )).elementAt( 0 );
		    
		    // the information out of this queue element
		    final AsyncCall currentCall = currentQueueElement.ca;
		    final long delayTime = currentQueueElement.waitingTime;
		    final AsyncResultImpl currentResult = currentQueueElement.ari;
		    
		    class CTRunnable 
		    implements Runnable {
			private Thread wt = null;
			
			public void setWatchingThread( Thread watchingThread ) {
			    wt = watchingThread;
			}
			
			public void run() {
				
			    Object o = null;
			    boolean isException = false;
			    try {
				o = currentCall.invoke( receiver );
				
			    } catch( InvocationTargetException error ) {
				// if the callee throws an exception, just pull it
				// out and return that -- the
				// InvocationTargetException will just be
				// encapsulating an exception in the
				// getMethod().invoke, and i think its an
				// unnecessary wrapper.  the server runtime
				// exception is similar, but means it happened
				// over RMI
				o = error.getTargetException();
				if( o instanceof ServerRuntimeException )
				   o = ((ServerRuntimeException)o).detail;

				isException = true;
				
			    } catch( Exception error ) {
				// the only other exceptions that this will throws
				// is the IllegalArgumentException which we will
				// just want to return and the same goes with the
				// IllegalAccessException -- perhaps these should
				// be encapsulated in a AsyncCallException?
				o = error;
				isException = true;
				
			    }
				
			    currentResult.putResult( o, isException );
			    
				// over here we want to tell if the watchingThread has
				// finished, and if not it should kill it
			    
			    if( wt != null )
			       wt.interrupt();
				
			}
		    };
 		    CTRunnable callingThreadRunnable = new CTRunnable();
		    final Thread callingThread = new Thread( callingThreadRunnable );
		    
		    if( millis >= 0 ) {
			currentQueueElement.executingThread = callingThread;
		        callingThreadRunnable.setWatchingThread( currentQueueElement.monitoringThread );
		    }
		    callingThread.start();
		    
		    // we need to spin here until this call is finished
		    while( callingThread.isAlive() )
		        Thread.yield();

		    // now we get to remove this element from the queue
		    synchronized( queues ) {
			Vector receiverQueue = (Vector)queues.get( receiver );
			receiverQueue.removeElementAt( 0 );
			if( receiverQueue.size() == 0 )
                            queues.remove( receiver );
		    }
		}

		// remove ourselves from the list of threads that are
		// still running because we are exiting now
		synchronized( AsyncOneMethodCaller.this ) {
		    AsyncOneMethodCaller.this.notifyAll();
		}
	    }
	};
	final Thread queueWatchingThread = new Thread( tg, queueWatchingRunnable );

	queueWatchingThread.start();

	
	return result;
	
    }

    /**
     * This method should return true if this instance of an async
     * messenger has finished sending all the messages it has been
     * asked to
     *
     * @return true if we have finished sending all the messages we
     * have been asked to 
     */
    public boolean doneWithCalls() {
	
	if( tg.activeCount() == 0 )
	    return true;
	return false;

    }

    /**
     * This method should block until all the messages in this caller
     * have been finished
     */
    public void blockUntilDone() {

	while( !doneWithCalls() )
	    try {
		synchronized( this ) {
		    wait();
		}
	    } catch( InterruptedException error ) { }

    }


}

// this object is just used to encode everything needed for the queue
class AsyncOneMethodCallQueueElement {

    AsyncCall ca; 
    long waitingTime;
    AsyncResultImpl ari;
    Thread monitoringThread;
    Thread executingThread;


    public AsyncOneMethodCallQueueElement( AsyncCall ca, long waitingTime, AsyncResultImpl ari ) {

	this.ca = ca;
	this.waitingTime = waitingTime;
	this.ari = ari;

    }

}

