// $Id: AsyncResultImpl.java,v 1.1 2000/05/18 22:01:20 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.support.async;

import java.lang.reflect.Method;
import java.io.Serializable;
import java.util.Enumeration;
import java.util.Vector;


/**
 * The result of the asynchronous call we make
 */
public class AsyncResultImpl
implements AsyncResult {
    
    /** The object which we called the method on **/
    final private Object receiver;

    /** The object which performed the call **/
    final private AsyncCall call;
    
    /** The result of this call **/
    private Object result = null;

    
    /** The flags to test for finish state **/
    private boolean isFinished = false;
    private boolean isException;

    
    /** A list of the listeners **/
    private Vector listeners = new Vector();
    private Vector exceptionOnlyListeners = new Vector();
    

    /**
     * Create the beginnings of the AsyncResultImpl object, it just
     * simply records the receiving object and the call used to do it.
     * eventually after this is constructed, the rest of the data will
     * be filled in
     *
     * @param receiver the object we are calling on
     * @param call the call used to generate this result 
     */
    public AsyncResultImpl( Object receiver, AsyncCall call ) {

	this.receiver = receiver;
	this.call = call;

    }

    /**
     * Put the data into this object.  It will update the flags in a
     * synchronized block so this class will signal that it has data.
     * This can only be called once on this object -- any other calls
     * to this method will be ignored,
     *
     * @param result the result of the data
     * @param isException if this result is actually an exception 
     */
    public synchronized void putResult( Object result, boolean isException ) {
	
	if( !this.isFinished ) {

	    this.result = result;
	    
	    this.isFinished = true;
	    this.isException = isException;
	    
	    notify();
	    
	    notifyListeners();

	}

    }

    /** 
     * Get the object we called on 
     *
     * @return the reference to the object we made the call on
     */
    public Object getReceiver() {

	return this.receiver;

    }

    /** 
     * Get the AsyncCall object that we used to call the receiver 
     *
     * @return the AsyncCall object that did the call to get this
     */
    public AsyncCall getCall() {

	return this.call;

    }
    

    /**
     * Return whether or not we have the result -- this does not block
     *
     * @return true if we have a result, false otherwise
     */
    public boolean isFinished() {

	return this.isFinished;

    }
    
    /**
     * Just hold the thread until its finished 
     */
    private synchronized void waitUntilFinished() {

	while( !isFinished() ) {

	    // this will do the blocking
	    try {
		wait();
	    } catch( InterruptedException error ) {
	    }

	}
	
    }

    /**
     * Return whether or not the result of the call was an exception
     * or not -- this method blocks.  This blocks because there may be
     * a case in which the method call returns void, and we just need
     * to capture any exceptions.  To get around this first call
     * isFinished() and if that returns true, then you can test for an
     * exception
     *
     * @return true if this is an exception 
     */
    public boolean wasException() {
	
	waitUntilFinished();

	return this.isException;

    }

    /**
     * The result of the method call.  Could be a Void object, it
     * could be an exception, or anything else this could possibly
     * return.  This method will block, so call isFinished() first.
     *
     * @return the result of running the method 
     */
    public Object getResult() {

	waitUntilFinished();

	return result;

    }
    
    /**
     * Tell all the listeners we have an object for them.
     *
     * @param ar the object for the result, namely us 
     * @param isException if the object is not an exception then the
     * exception only listeners couldn't care less 
     */
    private synchronized void notifyListeners() {
	
	if( this.isException ) {
	    Enumeration e = this.exceptionOnlyListeners.elements();
	    while( e.hasMoreElements() ) 
		((AsyncResultListener)e.nextElement()).asyncFinished( this );
	}

	Enumeration e = this.listeners.elements();
	while( e.hasMoreElements() )
	    ((AsyncResultListener)e.nextElement()).asyncFinished( this );

    }

    /**
     * Add a listener which will recieve notices when isFinished()
     * will be returning true.  if the result is already here, it just
     * simply calls back immediately
     *
     * @param arl the AsyncResultListener which will recieve the notices 
     */
    public void addListener( AsyncResultListener arl ) {

	if( !isFinished() )
	    listeners.addElement( arl );

	else
	    arl.asyncFinished( this );

    }

    /**
     * Similar to addListener, except these AsyncResultListeners will
     * not get all the events, but they will get the notice if the
     * AsyncResult is an exception.
     *
     * @param arl the AsyncResultListener which will recieve only
     * exception notices 
     */
    public void addExceptionOnlyListener( AsyncResultListener arl ) {

	if( !isFinished() )
	    exceptionOnlyListeners.addElement( arl );

	else if( wasException() )
	    arl.asyncFinished( this );

    }

}
