// $Id: DenseGraph.java,v 1.1 2000/05/18 22:01:21 tucker Exp $
// Hive. Copyright (c) 1998-1999, The Massachusetts Institute of Technology.
// All rights reserved. Distributed with no warranty, without even the
// implied warranty of merchantability or fitness for a particular purpose.
// For more details see COPYING.GPL, the GNU General Public License.

package net.hivecell.hive.support.graph;

import java.io.Serializable;
import java.util.Enumeration;
import java.util.Hashtable;
import java.util.NoSuchElementException;

/**
 * This data structure provides a directed and weighted graph.  This
 * has the ability to add objects and place them at nodes.  
 */
public class DenseGraph 
extends Object
implements Graph, Serializable {
    
    /**
     * The rate at which to increase the size of the matrix 
     */
    private double increasingFactor = 0.5;

    /**
     * The initial amount of objects for the adjacency matrix
     */
    private int initialSize = 16;


    /**
     * How many nodes do we have in our matrix
     */
    private int numberNodes = 0;

    /**
     * How many connections do we have in our matrix
     */
    private int numberConnections = 0;

    
    /**
     * This is useful to allow a programmer to get the node for a
     * particular object in case he didn't keep the node object
     * returned off addNode 
     */
    private Hashtable objectToNode = new Hashtable();

    /**
     * The values for the nodes in the adjacency matrix
     */
    private Object[] adjacencyMatrixValues;
    
    /**
     * The actual adjacency matrix.  The initial size is
     * initializeSize + 1 -- we use the +1 space to hold the objects
     * in the node 
     */
    private int[][] adjacencyMatrix;


    public DenseGraph() { this( 16 ); }
    public DenseGraph( int capacity ) { 
	
	this.adjacencyMatrixValues = new Object[capacity];
	this.adjacencyMatrix = new int[capacity][capacity];

    }

    /** 
     * Return how many nodes there are in this graph
     *
     * @return a count of the number of verbs
     */
    public int numberNodes() { return this.numberNodes; }

    /**
     * return how many connections there ar ein this graph
     *
     * @return a count of the number of edges
     */
    public int numberEdges() { return this.numberConnections; }
    
    /** 
     * Return an enumeration of the objects
     * 
     * @return the node enumeration
     */
    public Enumeration nodes() { return this.objectToNode.keys(); }

    /**
     * Return an enumeration of all the edges
     *
     * @return the edge enumeration
     */
    public Enumeration edges() {

	return new Enumeration() {
	    int count = 0;
	    int startx = 0, starty = 0;
	    
	    public boolean hasMoreElements() { return( count < numberConnections ); }

      	    public Object nextElement() {

		// no more nodes
		if( !hasMoreElements() )
		    throw new NoSuchElementException();

		// this is the current edge that we return
		if( ( startx == 0 ) &&
		    ( starty == 0 ) ) {
		    while( (( adjacencyMatrix[startx][starty] == 0 ) ||
			    ( adjacencyMatrixValues[startx] == null ) ||
			    ( adjacencyMatrixValues[starty] == null ) )) {
			startx++;
			if( startx >= adjacencyMatrixValues.length ) {
			    startx = 0;
			    starty++;
			}
		    } 
		}


		Edge toReturn = new Edge( adjacencyMatrixValues[startx], adjacencyMatrixValues[starty], adjacencyMatrix[startx][starty] );
		if( !( (++count) < numberConnections ) )
		    // there are no more nodes, so we cannot go on a search
		    return toReturn;
		
		// search for the next node in the system
		do {
		    startx++;
		    if( startx >= adjacencyMatrixValues.length ) {
			startx = 0;
			starty++;
		    }
		} while( ( adjacencyMatrix[startx][starty] == 0 ) ||
			 ( adjacencyMatrixValues[startx] == null ) ||
			 ( adjacencyMatrixValues[starty] == null ) );
		
		// return our last edge even though we are pointing to the next one
		return toReturn;

	    }

	};

    }

    /**
     * Return an array of all the objects which are nodes in this graph
     *
     * @return an array of all the nodes in this graph
     */
    protected Node[] listAllNodes() {

	Node[] toReturn = new Node[this.numberNodes];
	Enumeration e = this.objectToNode.elements();
	for( int count=0;count < numberNodes;count++ )
	    toReturn[count] = (Node)e.nextElement();

	return toReturn;

    }

    /**
     * Add an object to the graph and make a node in the graph.  If
     * this object is already in the graph, then its Node object is
     * just simply returned
     *
     * @param o the object to add to the graph
     * @return the node in the graph 
     */
    public void addNode( Object o ) {

	// hopefully the implementation of the hashtable in java is
	// pretty fast for lookups and insertions
	
	// first make sure that we don't already have this object
	// inside the graph
	if( this.objectToNode.containsKey( o ) )
	    return;

	// find the first free indexing slot
	int freeIndex = this.findFreeIndex();

	// put the object to the adjacency matrix
	this.adjacencyMatrixValues[freeIndex] = o;

	// create the node
	Node n = new Node( o, this, freeIndex );
	this.objectToNode.put( o, n );

	numberNodes++;
	
    }

    /**
     * Find a free index inside our adjacency table.  If there are no
     * free indexes, then this will call another function to increase
     * the size of our adjacency tabled
     *
     * @return the first free index that we can find in the table 
     */
    private int findFreeIndex() {

	// unfortunately this is a linear search :(  sigh O(n)
	int foundSpace = -1;

	if( numberNodes == this.adjacencyMatrixValues.length ) {
	    increaseMatrixSize();
	    return numberNodes;
	}
	    
	for( int count = 0;count != this.adjacencyMatrixValues.length;count++ )
	    if( this.adjacencyMatrixValues[count] == null ) {
		foundSpace = count;
		break;
	    }
	return foundSpace; // return the index for our new matrix

    }
    
    /** 
     * Increase the size of the adjacency matrix by the increasing factor
     */
    private void increaseMatrixSize() {
	// oh no -- time to expand our matrix -- first calculate
	// its new size, allocate the size, then copy our old
	// matrix there....
	int newSize = (int)( (double)this.adjacencyMatrixValues.length * increasingFactor ) + this.adjacencyMatrixValues.length;
	int[][] newMatrix = new int[newSize][newSize];
	for( int count = 0;count != this.adjacencyMatrix[0].length;count++ ) 
	    System.arraycopy( this.adjacencyMatrix[count], 0, newMatrix[count], 0, this.adjacencyMatrix[count].length );
	Object[] newMatrixValues = new Object[newSize];
	System.arraycopy( this.adjacencyMatrixValues, 0, newMatrixValues, 0, this.adjacencyMatrixValues.length );
	this.adjacencyMatrix = newMatrix; // replace our matrix
	this.adjacencyMatrixValues = newMatrixValues;
    }
    
    /**
     * Remove this node from the graph.  First it removes the mapping
     * from object to node mapping, then it nulls out its columns in
     * the adjacency matrix.
     *
     * @param o the object to remove from the matrix 
     * @return true if the object was removed, false if we don't have it
     */
    public boolean removeNode( Object o ) {

	Node n = null;
	if( ( n = (Node)this.objectToNode.get( o ) ) == null )
	    return false;
	return removeNode( n );

    }

    protected boolean removeNode( Node n ) {

	if( this.objectToNode.remove( n.getObject() ) == null )
	    return false;

	this.adjacencyMatrixValues[n.getIndexNumber()] = null;
	for( int count=0;count != this.adjacencyMatrix[0].length;count++ ) {
	    this.adjacencyMatrix[n.getIndexNumber()][count] = 0;
	    this.adjacencyMatrix[count][n.getIndexNumber()] = 0;
	}

	numberNodes--;

	return true;

    }

    /** 
     * Return true if the object is in the graph
     *
     * @return true if the object is in the graph, false else
     */
    public boolean containsNode( Object o ) {

	return this.objectToNode.containsKey( o );

    }

    /**
     * A helper function to get the node that has this object -- it
     * will throw an exception if it doesn't have it
     *
     * @param o the object to look for
     * @return Node the node the object is in
     * @exception NotInGraphException if it is not in the graph 
     */
    protected Node getNode( Object o ) 
    throws NotInGraphException {

	Node n = null;
	if(( n = (Node)this.objectToNode.get( o ) ) == null )
	    throw new NotInGraphException( o );
	return n;

    }
    
    /** 
     * A helper function to test to see if a passed Node object is
     * even valid.  It will just return if the node is valid --
     * otherwise it will throw an exception
     *
     * @param n the node object to test
     * @exception NotInGraphException if the node object is invalid
     */
    private void testNode( Node n ) 
    throws NotInGraphException {
	
	if( ( Math.abs( n.getIndexNumber() ) >= this.adjacencyMatrixValues.length ) || ( n.getObject() != this.adjacencyMatrixValues[n.getIndexNumber()] ) || ( n.getGraph() != this ) )
	    throw new NotInGraphException( n.getObject() );

    }

    /**
     * Connect two nodes in a specific direction with a specific weight.
     *
     * @param o1 the source node
     * @param o2 the destination node
     * @param weight the weight to assign this direction
     * @exception NotInGraphException if either object is not in the graph */
    public void connectNodes( Object o1, Object o2, int weight ) 
    throws NotInGraphException {

	Node n1 = this.getNode( o1 );
	Node n2 = this.getNode( o2 );
	connectNodesHelper( n1, n2, weight );

    }

    protected void connectNodes( Node n1, Node n2, int weight )
    throws NotInGraphException {

	this.testNode( n1 );
	this.testNode( n2 );
	connectNodesHelper( n1, n2, weight );

    }

    private void connectNodesHelper( Node n1, Node n2, int weight ) {

	if( weight < 0 ) {
	    if( this.adjacencyMatrix[n2.getIndexNumber()][n1.getIndexNumber()] == 0 ) numberConnections++;
	    this.adjacencyMatrix[n2.getIndexNumber()][n1.getIndexNumber()] = -1 * weight;

	} else {
	    if( this.adjacencyMatrix[n1.getIndexNumber()][n2.getIndexNumber()] == 0 ) numberConnections++;
	    this.adjacencyMatrix[n1.getIndexNumber()][n2.getIndexNumber()] = weight;

	}

    }

    /**
     * Returns true if these objects (if they are nodes in the graph)
     * are connected in this direction
     *
     * @param o1 the origin object
     * @param o2 the destination object
     * @return true if there is an edge in this direction
     * @exception NotInGraphException if either of these objects are not in the graph
     */
    public boolean isConnected( Object o1, Object o2 )
    throws NotInGraphException {

	Node n1 = this.getNode( o1 );
	Node n2 = this.getNode( o2 );
	return areConnectedHelper( n1, n2 );

    }

    /**
     * Test to see if these two objects are connected in the graph
     *
     * @param n1 the origin node for the testing edge
     * @param n2 the destination node for the testing edge
     * @return true if there is an edge between these nodes in this direction
     * @exception NotInGraphException if these nodes aren't in the graphs 
     */
    protected boolean areConnected( Node n1, Node n2 )
    throws NotInGraphException {

	this.testNode( n1 );
	this.testNode( n2 );
	return areConnectedHelper( n1, n2 );

    }

    private boolean areConnectedHelper( Node n1, Node n2 ) {
	
	return( this.adjacencyMatrix[n1.getIndexNumber()][n2.getIndexNumber()] > 0 );

    }

    /**
     * Get the weight of the edge between these two nodes - a weight
     * of 0 means the objects are not connected 
     * 
     * @param o1 the origin object we are looking for
     * @param o2 the destination object
     * @return the weight of the edge -- 0 means that there is no edge
     * @exception NotInGraphException is thrown if these objects are not in the graph
     */
    public int getWeight( Object o1, Object o2 )
    throws NotInGraphException {

	Node n1 = this.getNode( o1 );
	Node n2 = this.getNode( o2 );
	return this.getWeightHelper( n1, n2 );

    }

    protected int getWeight( Node n1, Node n2 )
    throws NotInGraphException {

	this.testNode( n1 );
	this.testNode( n2 );
	return this.getWeightHelper( n1, n2 );

    }

    private int getWeightHelper( Node n1, Node n2 ) {

	return this.adjacencyMatrix[n1.getIndexNumber()][n2.getIndexNumber()];

    }

    /**
     * Disconnect two nodes in a specific direction
     *
     * @param o1 the source node
     * @param o2 the destination node
     * @return the weight of the connection that used to be between these nodes
     * @exception NotInGraphException if either object is not in the graph */
    public int disconnectNodes( Object o1, Object o2 )
    throws NotInGraphException {

	Node n1 = this.getNode( o1 );
	Node n2 = this.getNode( o2 );
	return disconnectNodesHelper( n1, n2 );

    }

    protected int disconnectNodes( Node n1, Node n2 )
    throws NotInGraphException {

	this.testNode( n1 );
	this.testNode( n2 );
	return disconnectNodesHelper( n1, n2 );

    }

    private int disconnectNodesHelper( Node n1, Node n2 ) {

	int weight = this.adjacencyMatrix[n1.getIndexNumber()][n2.getIndexNumber()];
	this.adjacencyMatrix[n1.getIndexNumber()][n2.getIndexNumber()] = 0;
	if( weight != 0 ) numberConnections--;
	return weight;
    }

    /** 
     * Get a list of all the nodes which have outgoing connections from this object
     *
     * @param o the object we want the list for
     * @return an array of the nodes that this node points to
     */
    public Object[] listOutgoingNodes( Object o ) 
    throws NotInGraphException {

	Node n = this.getNode( o );
	return listOutgoingNodesHelper( n );

    }

    public Object[] listOutgoingNodes( Node n ) 
    throws NotInGraphException {

	this.testNode( n );
	return listOutgoingNodesHelper( n );

    }

    private Object[] listOutgoingNodesHelper( Node n ) {

	Object[] holder = new Object[this.adjacencyMatrixValues.length];
	int position = 0;
	for( int count=0;count < this.adjacencyMatrixValues.length;count++ )
	    if( this.adjacencyMatrix[n.getIndexNumber()][count] != 0 ) {
		holder[position] = this.adjacencyMatrixValues[count];
		position++;
	    }
	Object[] toReturn = new Object[position]; 
	System.arraycopy( holder, 0, toReturn, 0, position );
	return toReturn;
	
    }

    /** 
     * Get a list of all the nodes which have incoming connections from this object
     *
     * @param o the object we want the list for
     * @return an array of the nodes that point to this node
     */
    public Object[] listIncomingNodes( Object o ) 
    throws NotInGraphException {

	Node n = this.getNode( o );
	return listIncomingNodesHelper( n );

    }

    public Object[] listIncomingNodes( Node n ) 
    throws NotInGraphException {

	this.testNode( n );
	return listIncomingNodesHelper( n );

    }

    private Object[] listIncomingNodesHelper( Node n ) {

	Object[] holder = new Object[this.adjacencyMatrixValues.length];
	int position = 0;
	for( int count=0;count < this.adjacencyMatrixValues.length;count++ )
	    if( this.adjacencyMatrix[count][n.getIndexNumber()] != 0 ) {
		holder[position] = this.adjacencyMatrixValues[count];
		position++;
	    }
	Object[] toReturn = new Object[position]; 
	System.arraycopy( holder, 0, toReturn, 0, position );
	return toReturn;
	
    }

    /**
     * Return how many nodes we can hold at oyr current size of our adjacency matrix
     *
     * @return the capacity of our current matrix
     */
    public int capacity() { return this.adjacencyMatrixValues.length; }
    
    /**
     * Find out by how much the adjacency matrix grows every time it needs to expand 
     *
     * @return the increasing factor of our matrix
     */
    public double getGrowthFactor() { return this.increasingFactor; }

    /**
     * Set the growth factor of our abstraction
     *
     * @param newGrowthFactor the new growth factor
     */
    public void setGrowthFactor( double newGrowthFactor ) { this.increasingFactor = newGrowthFactor; }


//      public static void main( String[] args ) {

//  	System.out.println( "Creating the graph..." );
//  	DenseGraph dg = new DenseGraph();
//  	System.out.println( "...created" );

//  	System.out.println( "Adding three nodes..." );
//  	Node n1 = dg.addNode( new Integer( 1 ) );
//  	Node n2 = dg.addNode( new Integer( 2 ) );
//  	Node n3 = dg.addNode( new Integer( 3 ) );
//  	System.out.println( "...added" );

//  	System.out.println( "Testing to make sure those nodes are still there..." ); 
//  	Node[] nodes = dg.listAllNodes();
//  	System.out.println( " Searching for the first one..." );
//  	int position = -1;
//  	for( int count=0;count < nodes.length;count++ )
//  	    if( nodes[count].equals( n1 ) ) {
//  		position = count;
//  		break;
//  	    }
//  	if( position == -1 ) {
//  	    System.out.println( " ...cannot find node 1 in the graph!" );
//  	    System.exit( 1 );
//  	} else 
//  	    System.out.println( " ...found node one." );
//  	System.out.println( " Searching for the second one..." );
//  	position = -1;
//  	for( int count=0;count < nodes.length;count++ )
//  	    if( nodes[count].equals( n2 ) ) {
//  		position = count;
//  		break;
//  	    }
//  	if( position == -1 ) {
//  	    System.out.println( " ...cannot find node 2 in the graph!" );
//  	    System.exit( 1 );
//  	} else 
//  	    System.out.println( " ...found node two." );
//  	System.out.println( " Searching for the third one..." );
//  	position = -1;
//  	for( int count=0;count < nodes.length;count++ )
//  	    if( nodes[count].equals( n3 ) ) {
//  		position = count;
//  		break;
//  	    }
//  	if( position == -1 ) {
//  	    System.out.println( " ...cannot find node 3 in the graph!" );
//  	    System.exit( 1 );
//  	} else 
//  	    System.out.println( " ...found node three." );
//  	System.out.println( "...found all nodes." );

//  	try {
//  	    System.out.println( "Testing to make sure there is not already a node between node 1 and node 2" );
//  	    if( !dg.areConnected( n1, n2 ) )
//  		System.out.println( "...the nodes are not connected." );
//  	    else {
//  		System.out.println( "...there is apparently an edge of weight " + dg.getWeight( n1, n2 ) + " connecting them?" );
//  		System.exit( 1 );
//  	    }

//  	    System.out.println( "Testing to make sure there is not already a node between node 2 and node 1" );
//  	    if( !dg.areConnected( n2, n1 ) )
//  		System.out.println( "...the nodes are not connected." );
//  	    else {
//  		System.out.println( "...there is apparently an edge of weight " + dg.getWeight( n2, n1 ) + " connecting them?" );
//  		System.exit( 1 );
//  	    }

//  	    System.out.println( "Connecting node 1 to node 2 with an edge of 10..." );
//  	    dg.connectNodes( n1, n2, 10 );
//  	    System.out.println( "...connected" );

//  	    System.out.println( "Testing to see if node 1 and node 2 have an edge between them,,," );
//  	    if( dg.areConnected( n1, n2 ) )
//  		System.out.println( "...they are connected." );
//  	    else {
//  		System.out.println( "...they are not connected!" );
//  		System.exit( 1 );
//  	    }
	    
//  	    System.out.println( "Testing the weight on the edge between node 1 and node 2..." );
//  	    if( dg.getWeight( n1, n2 ) == 10 )
//  		System.out.println( "...there is an edge of weight 10." );
//  	    else {
//  		System.out.println( "...there is apparently an edge of weight " + dg.getWeight( n1, n2 ) + "?" );
//  		System.exit( 1 );
//  	    }
	    
//  	    System.out.println( "Testing to make sure there is not an edge from node 2 to node 1..." );
//  	    if( !dg.areConnected( n2, n1 ) )
//  		System.out.println( "...they are not connected." );
//  	    else {
//  		System.out.println( "...the are apparently connected with an edge of weight " + dg.getWeight( n2, n1 ) + "?" );
//  		System.exit( 1 );
//  	    }

//  	    System.out.println( "Removing the edge between node 1 and node 2..." );
//  	    dg.disconnectNodes( n1, n2 );
//  	    System.out.println( "...disconnected." );

//  	    System.out.println( "Testing to see if there is an edge between node 1 and node 2..." );
//  	    if( !dg.areConnected( n1, n2 ) )
//  		System.out.println( "...there is not an edge between node 1 and node 2." );
//  	    else {
//  		System.out.println( "...there is apparently an edge between node 1 and node 2 of weight " + dg.getWeight( n1, n2 ) + "?" );
//  		System.exit( 1 );
//  	    }
	    
//  	    System.out.println( "Testing to see if there is an edge between node 2 and node 1..." );
//  	    if( !dg.areConnected( n2, n1 ) )
//  		System.out.println( "...there is not an edge between node 2 and node 1." );
//  	    else {
//  		System.out.println( "...there is apparently an edge between node 2 and node 1 of weight " + dg.getWeight( n2, n1 ) + "?" );
//  		System.exit( 1 );
//  	    }
	    
//  	} catch( NotInGraphException error ) {
//  	    System.out.println( "What?  All these operations should of been safe!\n" + error.getMessage() );
//  	    System.exit( 1 );
//  	}

//  	System.out.println();
//  	System.out.println( "Testing the growth of this graph..." );
//  	System.out.println( "Creating a graph of size 2..." );
//  	DenseGraph dg1 = new DenseGraph( 2 );
//  	System.out.println( "...created." );

//  	System.out.println( "Testing the capacity of the graph..." );
//  	if( dg1.capacity() == 2 )
//  	    System.out.println( "...the graph has a capacity of two." );
//  	else {
//  	    System.out.println( "...the graph apparently has a capacity of " + dg1.capacity() + "?" );
//  	    System.exit( 1 );
//  	}

//  	System.out.println( "Adding two nodes to the graph..." );
//  	Node n4 = dg1.addNode( new Integer( 4 ) );
//  	Node n5 = dg1.addNode( new Integer( 5 ) );
//  	System.out.println( "...added." );

//  	System.out.println( "Testing the capacity of the graph..." );
//  	if( dg1.capacity() == 2 )
//  	    System.out.println( "...the graph has a capacity of two." );
//  	else {
//  	    System.out.println( "...the graph apparently has a capacity of " + dg1.capacity() + "?" );
//  	    System.exit( 1 );
//  	}

//  	try {
//  	    System.out.println( "Making a connection of weight 5 between node 5 and node 4..." );
//  	    dg1.connectNodes( n5, n4, 5 );
//  	    System.out.println( "...connected." );
//  	} catch( NotInGraphException error ) {
//  	    System.out.println( "What?  All these operations should of been safe!\n" + error.getMessage() );
//  	    System.exit( 1 );
//  	}

//  	System.out.println( "Adding a third node (this should cause growth to a capacity of three)..." );
//  	Node n6 = dg1.addNode( new Integer( 6 ) );
//  	System.out.println( "...added." );
	
//  	System.out.println( "Testing the capacity of the graph..." );
//  	if( dg1.capacity() == 3 )
//  	    System.out.println( "...the graph has a capacity of three." );
//  	else {
//  	    System.out.println( "...the graph apparently has a capacity of " + dg1.capacity() + "?" );
//  	    System.exit( 1 );
//  	}

//  	System.out.println( "Testing to see all three nodes are still there..." );
//  	System.out.println( "Testing to make sure those nodes are still there..." ); 
//  	nodes = dg1.listAllNodes();
//  	System.out.println( " Searching for the first one..." );
//  	position = -1;
//  	for( int count=0;count < nodes.length;count++ )
//  	    if( nodes[count].equals( n4 ) ) {
//  		position = count;
//  		break;
//  	    }
//  	if( position == -1 ) {
//  	    System.out.println( " ...cannot find node 4 in the graph!" );
//  	    System.exit( 1 );
//  	} else 
//  	    System.out.println( " ...found node four." );
//  	System.out.println( " Searching for the second one..." );
//  	position = -1;
//  	for( int count=0;count < nodes.length;count++ )
//  	    if( nodes[count].equals( n5 ) ) {
//  		position = count;
//  		break;
//  	    }
//  	if( position == -1 ) {
//  	    System.out.println( " ...cannot find node 5 in the graph!" );
//  	    System.exit( 1 );
//  	} else 
//  	    System.out.println( " ...found node five." );
//  	System.out.println( " Searching for the third one..." );
//  	position = -1;
//  	for( int count=0;count < nodes.length;count++ )
//  	    if( nodes[count].equals( n6 ) ) {
//  		position = count;
//  		break;
//  	    }
//  	if( position == -1 ) {
//  	    System.out.println( " ...cannot find node 6 in the graph!" );
//  	    System.exit( 1 );
//  	} else 
//  	    System.out.println( " ...found node six." );
//  	System.out.println( "...found all nodes." );

//  	try {
//  	    System.out.println( "Testing to see if node 5 and node 4 have an edge between them,,," );
//  	    if( dg1.areConnected( n5, n4 ) )
//  		System.out.println( "...they are connected." );
//  	    else {
//  		System.out.println( "...they are not connected!" );
//  		System.exit( 1 );
//  	    }
	    
//  	    System.out.println( "Testing the weight on the edge between node 5 and node 4..." );
//  	    if( dg1.getWeight( n5, n4 ) == 5 )
//  		System.out.println( "...there is an edge of weight 5." );
//  	    else {
//  		System.out.println( "...there is apparently an edge of weight " + dg1.getWeight( n5, n4 ) + "?" );
//  		System.exit( 1 );
//  	    }
	    
//  	    System.out.println( "Testing to make sure there is not an edge from node 4 to node 5..." );
//  	    if( !dg1.areConnected( n4, n5 ) )
//  		System.out.println( "...they are not connected." );
//  	    else {
//  		System.out.println( "...the are apparently connected with an edge of weight " + dg.getWeight( n4, n5 ) + "?" );
//  		System.exit( 1 );
//  	    }
	
//  	} catch( NotInGraphException error ) {
//  	    System.out.println( "What?  All these operations should of been safe!\n" + error.getMessage() );
//  	    System.exit( 1 );
//  	}

//  	System.out.println();
//  	System.out.println( "Testing connecting two nodes together which are in different graphs..." );
//  	try {
//  	    dg1.connectNodes( n1, n6, 10 );
//  	    System.out.println( "...the exception didn't catch this case" );
//  	    System.exit( 1 );
//  	} catch( NotInGraphException error ) {
//  	    System.out.println( "...exception throw -- good." );
//  	}

//  	System.out.println();
//  	System.out.println( "Testing the connection stuff..." );
//  	try {
//  	    System.out.println( "Making a connection between node 5 and node 6..." );
//  	    dg1.connectNodes( n5, n6, 10 );
//  	    System.out.println( "...connected." );

//  	    System.out.println( "Asking for a list of all the outgoing edges from node 5..." );
//  	    nodes = dg1.listOutgoingNodes( n5 );
//  	    System.out.println( "...got the list." );

//  	    System.out.println( "Making sure there are only two entries in this list..." );
//  	    if( nodes.length == 2 ) 
//  		System.out.println( "...only two connections in the outgoing edges list." );
//  	    else {
//  		System.out.println( "...there are apparently " + nodes.length + " connections in the outgoing edges list!" );
//  		System.exit( 1 );
//  	    }

//  	    System.out.println( "Making sure they are the correct entries..." );
//  	    System.out.println( " Searching for the first one..." );
//  	    position = -1;
//  	    for( int count=0;count < nodes.length;count++ ) {
//  		if( nodes[count].equals( n4 ) ) {
//  		    position = count;
//  		    break;
//  		}
//  	    }
//  	    if( position == -1 ) {
//  		System.out.println( " ...cannot find node 4 in the list!" );
//  		System.exit( 1 );
//  	    } else 
//  		System.out.println( " ...found node four." );
//  	    System.out.println( " Searching for the second one..." );
//  	    position = -1;
//  	    for( int count=0;count < nodes.length;count++ )
//  		if( nodes[count].equals( n6 ) ) {
//  		    position = count;
//  		    break;
//  		}
//  	    if( position == -1 ) {
//  		System.out.println( " ...cannot find node 5 in the graph!" );
//  		System.exit( 1 );
//  	    } else 
//  		System.out.println( " ...found node five." );
//  	    System.out.println( "...found both." );

//  	    System.out.println( "Asking for a list of all the incoming edges to node 6..." );
//  	    nodes = dg1.listIncomingNodes( n6 );
//  	    System.out.println( "...got the list." );

//  	    System.out.println( "Making sure there are only one entry in this list..." );
//  	    if( nodes.length == 1 ) 
//  		System.out.println( "...only one connections in the outgoing edges list." );
//  	    else {
//  		System.out.println( "...there are apparently " + nodes.length + " connections in the outgoing edges list!" );
//  		System.exit( 1 );
//  	    }

//  	    System.out.println( "Making sure it is the correct entries..." );
//  	    System.out.println( " Searching for the first one..." );
//  	    if( !nodes[0].equals( n5 ) ) {
//  		System.out.println( " ...node 5 is not here!" );
//  		System.exit( 0 );
//  	    } else 
//  		System.out.println( " ...found node 5." );
//  	    System.out.println( "...found it." );
	
//  	} catch( NotInGraphException error ) {
//  	    System.out.println( "What?  All these operations should of been safe!\n" + error.getMessage() );
//  	    System.exit( 1 );
//  	}

//      }

}

















